#include <tree_sitter/parser.h>

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 13
#define STATE_COUNT 1967
#define LARGE_STATE_COUNT 14
#define SYMBOL_COUNT 274
#define ALIAS_COUNT 1
#define TOKEN_COUNT 96
#define EXTERNAL_TOKEN_COUNT 1
#define FIELD_COUNT 32
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define PRODUCTION_ID_COUNT 78

enum {
  sym_module_comment = 1,
  sym_statement_comment = 2,
  sym_comment = 3,
  anon_sym_if = 4,
  anon_sym_LBRACE = 5,
  anon_sym_RBRACE = 6,
  anon_sym_erlang = 7,
  anon_sym_javascript = 8,
  anon_sym_AT = 9,
  anon_sym_LPAREN = 10,
  anon_sym_COMMA = 11,
  anon_sym_RPAREN = 12,
  anon_sym_COLON = 13,
  anon_sym_import = 14,
  anon_sym_DOT = 15,
  anon_sym_as = 16,
  anon_sym_SLASH = 17,
  anon_sym_type = 18,
  anon_sym_const = 19,
  anon_sym_EQ = 20,
  anon_sym_POUND = 21,
  anon_sym_LBRACK = 22,
  anon_sym_RBRACK = 23,
  anon_sym_LT_LT = 24,
  anon_sym_GT_GT = 25,
  anon_sym_DASH = 26,
  anon_sym_size = 27,
  anon_sym_fn = 28,
  anon_sym_DASH_GT = 29,
  anon_sym_external = 30,
  anon_sym_PIPE_PIPE = 31,
  anon_sym_AMP_AMP = 32,
  anon_sym_EQ_EQ = 33,
  anon_sym_BANG_EQ = 34,
  anon_sym_LT = 35,
  anon_sym_LT_EQ = 36,
  anon_sym_LT_DOT = 37,
  anon_sym_LT_EQ_DOT = 38,
  anon_sym_GT = 39,
  anon_sym_GT_EQ = 40,
  anon_sym_GT_DOT = 41,
  anon_sym_GT_EQ_DOT = 42,
  anon_sym_PIPE_GT = 43,
  anon_sym_PLUS = 44,
  anon_sym_PLUS_DOT = 45,
  anon_sym_DASH_DOT = 46,
  anon_sym_STAR = 47,
  anon_sym_STAR_DOT = 48,
  anon_sym_SLASH_DOT = 49,
  anon_sym_PERCENT = 50,
  anon_sym_LT_GT = 51,
  anon_sym_todo = 52,
  anon_sym_panic = 53,
  anon_sym_echo = 54,
  anon_sym_DOT_DOT = 55,
  anon_sym_case = 56,
  anon_sym_PIPE = 57,
  anon_sym_let = 58,
  anon_sym_assert = 59,
  anon_sym_use = 60,
  anon_sym_LT_DASH = 61,
  anon_sym_BANG = 62,
  sym_visibility_modifier = 63,
  sym_opacity_modifier = 64,
  anon_sym_DQUOTE = 65,
  anon_sym_DQUOTE2 = 66,
  aux_sym_escape_sequence_token1 = 67,
  aux_sym_escape_sequence_token2 = 68,
  sym_float = 69,
  sym__hex = 70,
  sym__decimal = 71,
  sym__octal = 72,
  sym__binary = 73,
  anon_sym_binary = 74,
  anon_sym_bytes = 75,
  anon_sym_int = 76,
  anon_sym_float = 77,
  anon_sym_bit_string = 78,
  anon_sym_bits = 79,
  anon_sym_utf8 = 80,
  anon_sym_utf16 = 81,
  anon_sym_utf32 = 82,
  anon_sym_utf8_codepoint = 83,
  anon_sym_utf16_codepoint = 84,
  anon_sym_utf32_codepoint = 85,
  anon_sym_signed = 86,
  anon_sym_unsigned = 87,
  anon_sym_big = 88,
  anon_sym_little = 89,
  anon_sym_native = 90,
  anon_sym_unit = 91,
  sym__discard_name = 92,
  sym__name = 93,
  sym__upname = 94,
  sym_quoted_content = 95,
  sym_source_file = 96,
  sym__module_statement = 97,
  sym_target_group = 98,
  sym_target = 99,
  sym_attribute = 100,
  sym__attribute_arguments = 101,
  sym_attribute_value = 102,
  sym_import = 103,
  sym_module = 104,
  sym_unqualified_imports = 105,
  sym_unqualified_import = 106,
  sym_constant = 107,
  sym__constant_value = 108,
  sym_constant_tuple = 109,
  sym_constant_list = 110,
  sym__constant_bit_string = 111,
  sym_constant_bit_string_segment = 112,
  sym_constant_bit_string_segment_options = 113,
  sym__constant_bit_string_segment_option = 114,
  sym__constant_bit_string_named_segment_option = 115,
  sym__constant_bit_string_segment_option_size = 116,
  sym_constant_record = 117,
  sym_constant_record_arguments = 118,
  sym_constant_record_argument = 119,
  sym_constant_field_access = 120,
  sym__constant_type = 121,
  sym__constant_type_annotation = 122,
  sym_constant_tuple_type = 123,
  sym_constant_function_type = 124,
  sym_constant_function_parameter_types = 125,
  sym_constant_type = 126,
  sym_constant_type_arguments = 127,
  sym_constant_type_argument = 128,
  sym_external_type = 129,
  sym_external_function = 130,
  sym_external_function_parameters = 131,
  sym_external_function_parameter = 132,
  sym_external_function_body = 133,
  sym_function = 134,
  sym_function_parameters = 135,
  sym_function_parameter = 136,
  sym__labeled_discard_param = 137,
  sym__discard_param = 138,
  sym__labeled_name_param = 139,
  sym__name_param = 140,
  aux_sym__statement_seq = 141,
  sym__statement = 142,
  sym__expression = 143,
  sym_binary_expression = 144,
  sym__expression_unit = 145,
  sym_record = 146,
  sym_todo = 147,
  sym_panic = 148,
  sym_pipeline_echo = 149,
  sym_echo = 150,
  sym_tuple = 151,
  sym_list = 152,
  sym__expression_bit_string = 153,
  sym_expression_bit_string_segment = 154,
  sym_expression_bit_string_segment_options = 155,
  sym__expression_bit_string_segment_option = 156,
  sym__expression_bit_string_named_segment_option = 157,
  sym__expression_bit_string_segment_option_size = 158,
  sym_anonymous_function = 159,
  sym_anonymous_function_parameters = 160,
  sym_anonymous_function_parameter = 161,
  sym_block = 162,
  sym_case = 163,
  sym_case_subjects = 164,
  sym_case_clauses = 165,
  sym_case_clause = 166,
  sym_case_clause_patterns = 167,
  sym_case_clause_pattern = 168,
  sym_case_clause_guard = 169,
  sym__case_clause_guard_expression = 170,
  sym__case_clause_guard_binary_expression = 171,
  sym__case_clause_guard_unit = 172,
  sym__case_clause_tuple_access = 173,
  sym_let_assert = 174,
  sym_let = 175,
  sym_use = 176,
  sym_use_assignments = 177,
  sym_use_assignment = 178,
  sym_assert = 179,
  sym_boolean_negation = 180,
  sym_integer_negation = 181,
  sym__assignment = 182,
  sym_record_update = 183,
  sym_record_update_arguments = 184,
  sym_record_update_argument = 185,
  sym__maybe_tuple_expression = 186,
  sym_tuple_access = 187,
  sym__maybe_record_expression = 188,
  sym_field_access = 189,
  sym__maybe_function_expression = 190,
  sym_arguments = 191,
  sym_argument = 192,
  sym_hole = 193,
  sym_function_call = 194,
  sym__pattern_expression = 195,
  sym__pattern_binary_expression = 196,
  sym__pattern = 197,
  sym_record_pattern = 198,
  sym_record_pattern_arguments = 199,
  sym_record_pattern_argument = 200,
  sym_pattern_spread = 201,
  sym_tuple_pattern = 202,
  sym__pattern_bit_string = 203,
  sym_pattern_bit_string_segment = 204,
  sym_pattern_bit_string_segment_options = 205,
  sym__pattern_bit_string_segment_option = 206,
  sym__pattern_bit_string_named_segment_option = 207,
  sym__pattern_bit_string_segment_option_size = 208,
  sym__pattern_bit_string_segment_argument = 209,
  sym_list_pattern = 210,
  sym_list_pattern_tail = 211,
  sym_type_definition = 212,
  sym_data_constructors = 213,
  sym_data_constructor = 214,
  sym_data_constructor_arguments = 215,
  sym_data_constructor_argument = 216,
  sym_type_alias = 217,
  sym_string = 218,
  sym_escape_sequence = 219,
  sym_integer = 220,
  sym__bit_string_segment_option = 221,
  sym__type = 222,
  sym__type_annotation = 223,
  sym_type_hole = 224,
  sym_tuple_type = 225,
  sym_function_type = 226,
  sym_function_parameter_types = 227,
  sym_type = 228,
  sym_type_arguments = 229,
  sym_type_argument = 230,
  sym_type_var = 231,
  sym_type_name = 232,
  sym_type_parameters = 233,
  sym_type_parameter = 234,
  sym_identifier = 235,
  sym_label = 236,
  sym_discard = 237,
  sym_type_identifier = 238,
  sym_remote_type_identifier = 239,
  sym_constructor_name = 240,
  sym_remote_constructor_name = 241,
  aux_sym_source_file_repeat1 = 242,
  aux_sym_target_group_repeat1 = 243,
  aux_sym__attribute_arguments_repeat1 = 244,
  aux_sym_module_repeat1 = 245,
  aux_sym_unqualified_imports_repeat1 = 246,
  aux_sym_constant_tuple_repeat1 = 247,
  aux_sym__constant_bit_string_repeat1 = 248,
  aux_sym_constant_bit_string_segment_options_repeat1 = 249,
  aux_sym_constant_record_arguments_repeat1 = 250,
  aux_sym_constant_tuple_type_repeat1 = 251,
  aux_sym_constant_type_arguments_repeat1 = 252,
  aux_sym_external_function_parameters_repeat1 = 253,
  aux_sym_function_parameters_repeat1 = 254,
  aux_sym_tuple_repeat1 = 255,
  aux_sym__expression_bit_string_repeat1 = 256,
  aux_sym_expression_bit_string_segment_options_repeat1 = 257,
  aux_sym_anonymous_function_parameters_repeat1 = 258,
  aux_sym_case_clauses_repeat1 = 259,
  aux_sym_case_clause_patterns_repeat1 = 260,
  aux_sym_case_clause_pattern_repeat1 = 261,
  aux_sym_use_assignments_repeat1 = 262,
  aux_sym_record_update_arguments_repeat1 = 263,
  aux_sym_arguments_repeat1 = 264,
  aux_sym_record_pattern_arguments_repeat1 = 265,
  aux_sym__pattern_bit_string_repeat1 = 266,
  aux_sym_pattern_bit_string_segment_options_repeat1 = 267,
  aux_sym_data_constructors_repeat1 = 268,
  aux_sym_data_constructor_arguments_repeat1 = 269,
  aux_sym_string_repeat1 = 270,
  aux_sym_tuple_type_repeat1 = 271,
  aux_sym_type_arguments_repeat1 = 272,
  aux_sym_type_parameters_repeat1 = 273,
  alias_sym_function_body = 274,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_module_comment] = "module_comment",
  [sym_statement_comment] = "statement_comment",
  [sym_comment] = "comment",
  [anon_sym_if] = "if",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_erlang] = "erlang",
  [anon_sym_javascript] = "javascript",
  [anon_sym_AT] = "@",
  [anon_sym_LPAREN] = "(",
  [anon_sym_COMMA] = ",",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COLON] = ":",
  [anon_sym_import] = "import",
  [anon_sym_DOT] = ".",
  [anon_sym_as] = "as",
  [anon_sym_SLASH] = "/",
  [anon_sym_type] = "type",
  [anon_sym_const] = "const",
  [anon_sym_EQ] = "=",
  [anon_sym_POUND] = "#",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_DASH] = "-",
  [anon_sym_size] = "size",
  [anon_sym_fn] = "fn",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_external] = "external",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT] = "<",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_LT_DOT] = "<.",
  [anon_sym_LT_EQ_DOT] = "<=.",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_GT_DOT] = ">.",
  [anon_sym_GT_EQ_DOT] = ">=.",
  [anon_sym_PIPE_GT] = "|>",
  [anon_sym_PLUS] = "+",
  [anon_sym_PLUS_DOT] = "+.",
  [anon_sym_DASH_DOT] = "-.",
  [anon_sym_STAR] = "*",
  [anon_sym_STAR_DOT] = "*.",
  [anon_sym_SLASH_DOT] = "/.",
  [anon_sym_PERCENT] = "%",
  [anon_sym_LT_GT] = "<>",
  [anon_sym_todo] = "todo",
  [anon_sym_panic] = "panic",
  [anon_sym_echo] = "echo",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_case] = "case",
  [anon_sym_PIPE] = "|",
  [anon_sym_let] = "let",
  [anon_sym_assert] = "assert",
  [anon_sym_use] = "use",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_BANG] = "!",
  [sym_visibility_modifier] = "visibility_modifier",
  [sym_opacity_modifier] = "opacity_modifier",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_DQUOTE2] = "\"",
  [aux_sym_escape_sequence_token1] = "escape_sequence_token1",
  [aux_sym_escape_sequence_token2] = "escape_sequence_token2",
  [sym_float] = "float",
  [sym__hex] = "_hex",
  [sym__decimal] = "_decimal",
  [sym__octal] = "_octal",
  [sym__binary] = "_binary",
  [anon_sym_binary] = "binary",
  [anon_sym_bytes] = "bytes",
  [anon_sym_int] = "int",
  [anon_sym_float] = "float",
  [anon_sym_bit_string] = "bit_string",
  [anon_sym_bits] = "bits",
  [anon_sym_utf8] = "utf8",
  [anon_sym_utf16] = "utf16",
  [anon_sym_utf32] = "utf32",
  [anon_sym_utf8_codepoint] = "utf8_codepoint",
  [anon_sym_utf16_codepoint] = "utf16_codepoint",
  [anon_sym_utf32_codepoint] = "utf32_codepoint",
  [anon_sym_signed] = "signed",
  [anon_sym_unsigned] = "unsigned",
  [anon_sym_big] = "big",
  [anon_sym_little] = "little",
  [anon_sym_native] = "native",
  [anon_sym_unit] = "unit",
  [sym__discard_name] = "_discard_name",
  [sym__name] = "_name",
  [sym__upname] = "_upname",
  [sym_quoted_content] = "quoted_content",
  [sym_source_file] = "source_file",
  [sym__module_statement] = "_module_statement",
  [sym_target_group] = "target_group",
  [sym_target] = "target",
  [sym_attribute] = "attribute",
  [sym__attribute_arguments] = "arguments",
  [sym_attribute_value] = "attribute_value",
  [sym_import] = "import",
  [sym_module] = "module",
  [sym_unqualified_imports] = "unqualified_imports",
  [sym_unqualified_import] = "unqualified_import",
  [sym_constant] = "constant",
  [sym__constant_value] = "_constant_value",
  [sym_constant_tuple] = "tuple",
  [sym_constant_list] = "list",
  [sym__constant_bit_string] = "bit_string",
  [sym_constant_bit_string_segment] = "bit_string_segment",
  [sym_constant_bit_string_segment_options] = "bit_string_segment_options",
  [sym__constant_bit_string_segment_option] = "_constant_bit_string_segment_option",
  [sym__constant_bit_string_named_segment_option] = "_constant_bit_string_named_segment_option",
  [sym__constant_bit_string_segment_option_size] = "bit_string_segment_option",
  [sym_constant_record] = "record",
  [sym_constant_record_arguments] = "arguments",
  [sym_constant_record_argument] = "argument",
  [sym_constant_field_access] = "field_access",
  [sym__constant_type] = "_constant_type",
  [sym__constant_type_annotation] = "_constant_type_annotation",
  [sym_constant_tuple_type] = "tuple_type",
  [sym_constant_function_type] = "function_type",
  [sym_constant_function_parameter_types] = "function_parameter_types",
  [sym_constant_type] = "type",
  [sym_constant_type_arguments] = "type_arguments",
  [sym_constant_type_argument] = "type_argument",
  [sym_external_type] = "external_type",
  [sym_external_function] = "external_function",
  [sym_external_function_parameters] = "function_parameters",
  [sym_external_function_parameter] = "function_parameter",
  [sym_external_function_body] = "external_function_body",
  [sym_function] = "function",
  [sym_function_parameters] = "function_parameters",
  [sym_function_parameter] = "function_parameter",
  [sym__labeled_discard_param] = "_labeled_discard_param",
  [sym__discard_param] = "_discard_param",
  [sym__labeled_name_param] = "_labeled_name_param",
  [sym__name_param] = "_name_param",
  [aux_sym__statement_seq] = "_statement_seq",
  [sym__statement] = "_statement",
  [sym__expression] = "_expression",
  [sym_binary_expression] = "binary_expression",
  [sym__expression_unit] = "_expression_unit",
  [sym_record] = "record",
  [sym_todo] = "todo",
  [sym_panic] = "panic",
  [sym_pipeline_echo] = "pipeline_echo",
  [sym_echo] = "echo",
  [sym_tuple] = "tuple",
  [sym_list] = "list",
  [sym__expression_bit_string] = "bit_string",
  [sym_expression_bit_string_segment] = "bit_string_segment",
  [sym_expression_bit_string_segment_options] = "bit_string_segment_options",
  [sym__expression_bit_string_segment_option] = "_expression_bit_string_segment_option",
  [sym__expression_bit_string_named_segment_option] = "_expression_bit_string_named_segment_option",
  [sym__expression_bit_string_segment_option_size] = "bit_string_segment_option",
  [sym_anonymous_function] = "anonymous_function",
  [sym_anonymous_function_parameters] = "function_parameters",
  [sym_anonymous_function_parameter] = "function_parameter",
  [sym_block] = "block",
  [sym_case] = "case",
  [sym_case_subjects] = "case_subjects",
  [sym_case_clauses] = "case_clauses",
  [sym_case_clause] = "case_clause",
  [sym_case_clause_patterns] = "case_clause_patterns",
  [sym_case_clause_pattern] = "case_clause_pattern",
  [sym_case_clause_guard] = "case_clause_guard",
  [sym__case_clause_guard_expression] = "_case_clause_guard_expression",
  [sym__case_clause_guard_binary_expression] = "binary_expression",
  [sym__case_clause_guard_unit] = "_case_clause_guard_unit",
  [sym__case_clause_tuple_access] = "tuple_access",
  [sym_let_assert] = "let_assert",
  [sym_let] = "let",
  [sym_use] = "use",
  [sym_use_assignments] = "use_assignments",
  [sym_use_assignment] = "use_assignment",
  [sym_assert] = "assert",
  [sym_boolean_negation] = "boolean_negation",
  [sym_integer_negation] = "integer_negation",
  [sym__assignment] = "_assignment",
  [sym_record_update] = "record_update",
  [sym_record_update_arguments] = "record_update_arguments",
  [sym_record_update_argument] = "record_update_argument",
  [sym__maybe_tuple_expression] = "_maybe_tuple_expression",
  [sym_tuple_access] = "tuple_access",
  [sym__maybe_record_expression] = "_maybe_record_expression",
  [sym_field_access] = "field_access",
  [sym__maybe_function_expression] = "_maybe_function_expression",
  [sym_arguments] = "arguments",
  [sym_argument] = "argument",
  [sym_hole] = "hole",
  [sym_function_call] = "function_call",
  [sym__pattern_expression] = "_pattern_expression",
  [sym__pattern_binary_expression] = "binary_expression",
  [sym__pattern] = "_pattern",
  [sym_record_pattern] = "record_pattern",
  [sym_record_pattern_arguments] = "record_pattern_arguments",
  [sym_record_pattern_argument] = "record_pattern_argument",
  [sym_pattern_spread] = "pattern_spread",
  [sym_tuple_pattern] = "tuple_pattern",
  [sym__pattern_bit_string] = "bit_string_pattern",
  [sym_pattern_bit_string_segment] = "bit_string_segment",
  [sym_pattern_bit_string_segment_options] = "bit_string_segment_options",
  [sym__pattern_bit_string_segment_option] = "_pattern_bit_string_segment_option",
  [sym__pattern_bit_string_named_segment_option] = "_pattern_bit_string_named_segment_option",
  [sym__pattern_bit_string_segment_option_size] = "bit_string_segment_option",
  [sym__pattern_bit_string_segment_argument] = "_pattern_bit_string_segment_argument",
  [sym_list_pattern] = "list_pattern",
  [sym_list_pattern_tail] = "list_pattern_tail",
  [sym_type_definition] = "type_definition",
  [sym_data_constructors] = "data_constructors",
  [sym_data_constructor] = "data_constructor",
  [sym_data_constructor_arguments] = "data_constructor_arguments",
  [sym_data_constructor_argument] = "data_constructor_argument",
  [sym_type_alias] = "type_alias",
  [sym_string] = "string",
  [sym_escape_sequence] = "escape_sequence",
  [sym_integer] = "integer",
  [sym__bit_string_segment_option] = "bit_string_segment_option",
  [sym__type] = "_type",
  [sym__type_annotation] = "_type_annotation",
  [sym_type_hole] = "type_hole",
  [sym_tuple_type] = "tuple_type",
  [sym_function_type] = "function_type",
  [sym_function_parameter_types] = "function_parameter_types",
  [sym_type] = "type",
  [sym_type_arguments] = "type_arguments",
  [sym_type_argument] = "type_argument",
  [sym_type_var] = "type_var",
  [sym_type_name] = "type_name",
  [sym_type_parameters] = "type_parameters",
  [sym_type_parameter] = "type_parameter",
  [sym_identifier] = "identifier",
  [sym_label] = "label",
  [sym_discard] = "discard",
  [sym_type_identifier] = "type_identifier",
  [sym_remote_type_identifier] = "remote_type_identifier",
  [sym_constructor_name] = "constructor_name",
  [sym_remote_constructor_name] = "remote_constructor_name",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_target_group_repeat1] = "target_group_repeat1",
  [aux_sym__attribute_arguments_repeat1] = "_attribute_arguments_repeat1",
  [aux_sym_module_repeat1] = "module_repeat1",
  [aux_sym_unqualified_imports_repeat1] = "unqualified_imports_repeat1",
  [aux_sym_constant_tuple_repeat1] = "constant_tuple_repeat1",
  [aux_sym__constant_bit_string_repeat1] = "_constant_bit_string_repeat1",
  [aux_sym_constant_bit_string_segment_options_repeat1] = "constant_bit_string_segment_options_repeat1",
  [aux_sym_constant_record_arguments_repeat1] = "constant_record_arguments_repeat1",
  [aux_sym_constant_tuple_type_repeat1] = "constant_tuple_type_repeat1",
  [aux_sym_constant_type_arguments_repeat1] = "constant_type_arguments_repeat1",
  [aux_sym_external_function_parameters_repeat1] = "external_function_parameters_repeat1",
  [aux_sym_function_parameters_repeat1] = "function_parameters_repeat1",
  [aux_sym_tuple_repeat1] = "tuple_repeat1",
  [aux_sym__expression_bit_string_repeat1] = "_expression_bit_string_repeat1",
  [aux_sym_expression_bit_string_segment_options_repeat1] = "expression_bit_string_segment_options_repeat1",
  [aux_sym_anonymous_function_parameters_repeat1] = "anonymous_function_parameters_repeat1",
  [aux_sym_case_clauses_repeat1] = "case_clauses_repeat1",
  [aux_sym_case_clause_patterns_repeat1] = "case_clause_patterns_repeat1",
  [aux_sym_case_clause_pattern_repeat1] = "case_clause_pattern_repeat1",
  [aux_sym_use_assignments_repeat1] = "use_assignments_repeat1",
  [aux_sym_record_update_arguments_repeat1] = "record_update_arguments_repeat1",
  [aux_sym_arguments_repeat1] = "arguments_repeat1",
  [aux_sym_record_pattern_arguments_repeat1] = "record_pattern_arguments_repeat1",
  [aux_sym__pattern_bit_string_repeat1] = "_pattern_bit_string_repeat1",
  [aux_sym_pattern_bit_string_segment_options_repeat1] = "pattern_bit_string_segment_options_repeat1",
  [aux_sym_data_constructors_repeat1] = "data_constructors_repeat1",
  [aux_sym_data_constructor_arguments_repeat1] = "data_constructor_arguments_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [aux_sym_tuple_type_repeat1] = "tuple_type_repeat1",
  [aux_sym_type_arguments_repeat1] = "type_arguments_repeat1",
  [aux_sym_type_parameters_repeat1] = "type_parameters_repeat1",
  [alias_sym_function_body] = "function_body",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_module_comment] = sym_module_comment,
  [sym_statement_comment] = sym_statement_comment,
  [sym_comment] = sym_comment,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_erlang] = anon_sym_erlang,
  [anon_sym_javascript] = anon_sym_javascript,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_size] = anon_sym_size,
  [anon_sym_fn] = anon_sym_fn,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_external] = anon_sym_external,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_LT_DOT] = anon_sym_LT_DOT,
  [anon_sym_LT_EQ_DOT] = anon_sym_LT_EQ_DOT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_GT_DOT] = anon_sym_GT_DOT,
  [anon_sym_GT_EQ_DOT] = anon_sym_GT_EQ_DOT,
  [anon_sym_PIPE_GT] = anon_sym_PIPE_GT,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_PLUS_DOT] = anon_sym_PLUS_DOT,
  [anon_sym_DASH_DOT] = anon_sym_DASH_DOT,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_STAR_DOT] = anon_sym_STAR_DOT,
  [anon_sym_SLASH_DOT] = anon_sym_SLASH_DOT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_LT_GT] = anon_sym_LT_GT,
  [anon_sym_todo] = anon_sym_todo,
  [anon_sym_panic] = anon_sym_panic,
  [anon_sym_echo] = anon_sym_echo,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_use] = anon_sym_use,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [sym_visibility_modifier] = sym_visibility_modifier,
  [sym_opacity_modifier] = sym_opacity_modifier,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_DQUOTE2] = anon_sym_DQUOTE,
  [aux_sym_escape_sequence_token1] = aux_sym_escape_sequence_token1,
  [aux_sym_escape_sequence_token2] = aux_sym_escape_sequence_token2,
  [sym_float] = sym_float,
  [sym__hex] = sym__hex,
  [sym__decimal] = sym__decimal,
  [sym__octal] = sym__octal,
  [sym__binary] = sym__binary,
  [anon_sym_binary] = anon_sym_binary,
  [anon_sym_bytes] = anon_sym_bytes,
  [anon_sym_int] = anon_sym_int,
  [anon_sym_float] = anon_sym_float,
  [anon_sym_bit_string] = anon_sym_bit_string,
  [anon_sym_bits] = anon_sym_bits,
  [anon_sym_utf8] = anon_sym_utf8,
  [anon_sym_utf16] = anon_sym_utf16,
  [anon_sym_utf32] = anon_sym_utf32,
  [anon_sym_utf8_codepoint] = anon_sym_utf8_codepoint,
  [anon_sym_utf16_codepoint] = anon_sym_utf16_codepoint,
  [anon_sym_utf32_codepoint] = anon_sym_utf32_codepoint,
  [anon_sym_signed] = anon_sym_signed,
  [anon_sym_unsigned] = anon_sym_unsigned,
  [anon_sym_big] = anon_sym_big,
  [anon_sym_little] = anon_sym_little,
  [anon_sym_native] = anon_sym_native,
  [anon_sym_unit] = anon_sym_unit,
  [sym__discard_name] = sym__discard_name,
  [sym__name] = sym__name,
  [sym__upname] = sym__upname,
  [sym_quoted_content] = sym_quoted_content,
  [sym_source_file] = sym_source_file,
  [sym__module_statement] = sym__module_statement,
  [sym_target_group] = sym_target_group,
  [sym_target] = sym_target,
  [sym_attribute] = sym_attribute,
  [sym__attribute_arguments] = sym_arguments,
  [sym_attribute_value] = sym_attribute_value,
  [sym_import] = sym_import,
  [sym_module] = sym_module,
  [sym_unqualified_imports] = sym_unqualified_imports,
  [sym_unqualified_import] = sym_unqualified_import,
  [sym_constant] = sym_constant,
  [sym__constant_value] = sym__constant_value,
  [sym_constant_tuple] = sym_tuple,
  [sym_constant_list] = sym_list,
  [sym__constant_bit_string] = sym__constant_bit_string,
  [sym_constant_bit_string_segment] = sym_constant_bit_string_segment,
  [sym_constant_bit_string_segment_options] = sym_constant_bit_string_segment_options,
  [sym__constant_bit_string_segment_option] = sym__constant_bit_string_segment_option,
  [sym__constant_bit_string_named_segment_option] = sym__constant_bit_string_named_segment_option,
  [sym__constant_bit_string_segment_option_size] = sym__constant_bit_string_segment_option_size,
  [sym_constant_record] = sym_record,
  [sym_constant_record_arguments] = sym_arguments,
  [sym_constant_record_argument] = sym_argument,
  [sym_constant_field_access] = sym_field_access,
  [sym__constant_type] = sym__constant_type,
  [sym__constant_type_annotation] = sym__constant_type_annotation,
  [sym_constant_tuple_type] = sym_tuple_type,
  [sym_constant_function_type] = sym_function_type,
  [sym_constant_function_parameter_types] = sym_function_parameter_types,
  [sym_constant_type] = sym_type,
  [sym_constant_type_arguments] = sym_type_arguments,
  [sym_constant_type_argument] = sym_type_argument,
  [sym_external_type] = sym_external_type,
  [sym_external_function] = sym_external_function,
  [sym_external_function_parameters] = sym_function_parameters,
  [sym_external_function_parameter] = sym_function_parameter,
  [sym_external_function_body] = sym_external_function_body,
  [sym_function] = sym_function,
  [sym_function_parameters] = sym_function_parameters,
  [sym_function_parameter] = sym_function_parameter,
  [sym__labeled_discard_param] = sym__labeled_discard_param,
  [sym__discard_param] = sym__discard_param,
  [sym__labeled_name_param] = sym__labeled_name_param,
  [sym__name_param] = sym__name_param,
  [aux_sym__statement_seq] = aux_sym__statement_seq,
  [sym__statement] = sym__statement,
  [sym__expression] = sym__expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym__expression_unit] = sym__expression_unit,
  [sym_record] = sym_record,
  [sym_todo] = sym_todo,
  [sym_panic] = sym_panic,
  [sym_pipeline_echo] = sym_pipeline_echo,
  [sym_echo] = sym_echo,
  [sym_tuple] = sym_tuple,
  [sym_list] = sym_list,
  [sym__expression_bit_string] = sym__constant_bit_string,
  [sym_expression_bit_string_segment] = sym_constant_bit_string_segment,
  [sym_expression_bit_string_segment_options] = sym_constant_bit_string_segment_options,
  [sym__expression_bit_string_segment_option] = sym__expression_bit_string_segment_option,
  [sym__expression_bit_string_named_segment_option] = sym__expression_bit_string_named_segment_option,
  [sym__expression_bit_string_segment_option_size] = sym__constant_bit_string_segment_option_size,
  [sym_anonymous_function] = sym_anonymous_function,
  [sym_anonymous_function_parameters] = sym_function_parameters,
  [sym_anonymous_function_parameter] = sym_function_parameter,
  [sym_block] = sym_block,
  [sym_case] = sym_case,
  [sym_case_subjects] = sym_case_subjects,
  [sym_case_clauses] = sym_case_clauses,
  [sym_case_clause] = sym_case_clause,
  [sym_case_clause_patterns] = sym_case_clause_patterns,
  [sym_case_clause_pattern] = sym_case_clause_pattern,
  [sym_case_clause_guard] = sym_case_clause_guard,
  [sym__case_clause_guard_expression] = sym__case_clause_guard_expression,
  [sym__case_clause_guard_binary_expression] = sym_binary_expression,
  [sym__case_clause_guard_unit] = sym__case_clause_guard_unit,
  [sym__case_clause_tuple_access] = sym_tuple_access,
  [sym_let_assert] = sym_let_assert,
  [sym_let] = sym_let,
  [sym_use] = sym_use,
  [sym_use_assignments] = sym_use_assignments,
  [sym_use_assignment] = sym_use_assignment,
  [sym_assert] = sym_assert,
  [sym_boolean_negation] = sym_boolean_negation,
  [sym_integer_negation] = sym_integer_negation,
  [sym__assignment] = sym__assignment,
  [sym_record_update] = sym_record_update,
  [sym_record_update_arguments] = sym_record_update_arguments,
  [sym_record_update_argument] = sym_record_update_argument,
  [sym__maybe_tuple_expression] = sym__maybe_tuple_expression,
  [sym_tuple_access] = sym_tuple_access,
  [sym__maybe_record_expression] = sym__maybe_record_expression,
  [sym_field_access] = sym_field_access,
  [sym__maybe_function_expression] = sym__maybe_function_expression,
  [sym_arguments] = sym_arguments,
  [sym_argument] = sym_argument,
  [sym_hole] = sym_hole,
  [sym_function_call] = sym_function_call,
  [sym__pattern_expression] = sym__pattern_expression,
  [sym__pattern_binary_expression] = sym_binary_expression,
  [sym__pattern] = sym__pattern,
  [sym_record_pattern] = sym_record_pattern,
  [sym_record_pattern_arguments] = sym_record_pattern_arguments,
  [sym_record_pattern_argument] = sym_record_pattern_argument,
  [sym_pattern_spread] = sym_pattern_spread,
  [sym_tuple_pattern] = sym_tuple_pattern,
  [sym__pattern_bit_string] = sym__pattern_bit_string,
  [sym_pattern_bit_string_segment] = sym_constant_bit_string_segment,
  [sym_pattern_bit_string_segment_options] = sym_constant_bit_string_segment_options,
  [sym__pattern_bit_string_segment_option] = sym__pattern_bit_string_segment_option,
  [sym__pattern_bit_string_named_segment_option] = sym__pattern_bit_string_named_segment_option,
  [sym__pattern_bit_string_segment_option_size] = sym__constant_bit_string_segment_option_size,
  [sym__pattern_bit_string_segment_argument] = sym__pattern_bit_string_segment_argument,
  [sym_list_pattern] = sym_list_pattern,
  [sym_list_pattern_tail] = sym_list_pattern_tail,
  [sym_type_definition] = sym_type_definition,
  [sym_data_constructors] = sym_data_constructors,
  [sym_data_constructor] = sym_data_constructor,
  [sym_data_constructor_arguments] = sym_data_constructor_arguments,
  [sym_data_constructor_argument] = sym_data_constructor_argument,
  [sym_type_alias] = sym_type_alias,
  [sym_string] = sym_string,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_integer] = sym_integer,
  [sym__bit_string_segment_option] = sym__constant_bit_string_segment_option_size,
  [sym__type] = sym__type,
  [sym__type_annotation] = sym__type_annotation,
  [sym_type_hole] = sym_type_hole,
  [sym_tuple_type] = sym_tuple_type,
  [sym_function_type] = sym_function_type,
  [sym_function_parameter_types] = sym_function_parameter_types,
  [sym_type] = sym_type,
  [sym_type_arguments] = sym_type_arguments,
  [sym_type_argument] = sym_type_argument,
  [sym_type_var] = sym_type_var,
  [sym_type_name] = sym_type_name,
  [sym_type_parameters] = sym_type_parameters,
  [sym_type_parameter] = sym_type_parameter,
  [sym_identifier] = sym_identifier,
  [sym_label] = sym_label,
  [sym_discard] = sym_discard,
  [sym_type_identifier] = sym_type_identifier,
  [sym_remote_type_identifier] = sym_remote_type_identifier,
  [sym_constructor_name] = sym_constructor_name,
  [sym_remote_constructor_name] = sym_remote_constructor_name,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_target_group_repeat1] = aux_sym_target_group_repeat1,
  [aux_sym__attribute_arguments_repeat1] = aux_sym__attribute_arguments_repeat1,
  [aux_sym_module_repeat1] = aux_sym_module_repeat1,
  [aux_sym_unqualified_imports_repeat1] = aux_sym_unqualified_imports_repeat1,
  [aux_sym_constant_tuple_repeat1] = aux_sym_constant_tuple_repeat1,
  [aux_sym__constant_bit_string_repeat1] = aux_sym__constant_bit_string_repeat1,
  [aux_sym_constant_bit_string_segment_options_repeat1] = aux_sym_constant_bit_string_segment_options_repeat1,
  [aux_sym_constant_record_arguments_repeat1] = aux_sym_constant_record_arguments_repeat1,
  [aux_sym_constant_tuple_type_repeat1] = aux_sym_constant_tuple_type_repeat1,
  [aux_sym_constant_type_arguments_repeat1] = aux_sym_constant_type_arguments_repeat1,
  [aux_sym_external_function_parameters_repeat1] = aux_sym_external_function_parameters_repeat1,
  [aux_sym_function_parameters_repeat1] = aux_sym_function_parameters_repeat1,
  [aux_sym_tuple_repeat1] = aux_sym_tuple_repeat1,
  [aux_sym__expression_bit_string_repeat1] = aux_sym__expression_bit_string_repeat1,
  [aux_sym_expression_bit_string_segment_options_repeat1] = aux_sym_expression_bit_string_segment_options_repeat1,
  [aux_sym_anonymous_function_parameters_repeat1] = aux_sym_anonymous_function_parameters_repeat1,
  [aux_sym_case_clauses_repeat1] = aux_sym_case_clauses_repeat1,
  [aux_sym_case_clause_patterns_repeat1] = aux_sym_case_clause_patterns_repeat1,
  [aux_sym_case_clause_pattern_repeat1] = aux_sym_case_clause_pattern_repeat1,
  [aux_sym_use_assignments_repeat1] = aux_sym_use_assignments_repeat1,
  [aux_sym_record_update_arguments_repeat1] = aux_sym_record_update_arguments_repeat1,
  [aux_sym_arguments_repeat1] = aux_sym_arguments_repeat1,
  [aux_sym_record_pattern_arguments_repeat1] = aux_sym_record_pattern_arguments_repeat1,
  [aux_sym__pattern_bit_string_repeat1] = aux_sym__pattern_bit_string_repeat1,
  [aux_sym_pattern_bit_string_segment_options_repeat1] = aux_sym_pattern_bit_string_segment_options_repeat1,
  [aux_sym_data_constructors_repeat1] = aux_sym_data_constructors_repeat1,
  [aux_sym_data_constructor_arguments_repeat1] = aux_sym_data_constructor_arguments_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [aux_sym_tuple_type_repeat1] = aux_sym_tuple_type_repeat1,
  [aux_sym_type_arguments_repeat1] = aux_sym_type_arguments_repeat1,
  [aux_sym_type_parameters_repeat1] = aux_sym_type_parameters_repeat1,
  [alias_sym_function_body] = alias_sym_function_body,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_module_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_statement_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_erlang] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_javascript] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_size] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_external] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_todo] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_panic] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_echo] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_use] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [sym_visibility_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_opacity_modifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_escape_sequence_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token2] = {
    .visible = false,
    .named = false,
  },
  [sym_float] = {
    .visible = true,
    .named = true,
  },
  [sym__hex] = {
    .visible = false,
    .named = true,
  },
  [sym__decimal] = {
    .visible = false,
    .named = true,
  },
  [sym__octal] = {
    .visible = false,
    .named = true,
  },
  [sym__binary] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_binary] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bytes] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_int] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_float] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bit_string] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_bits] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf8] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf16] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf32] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf8_codepoint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf16_codepoint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_utf32_codepoint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_signed] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsigned] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_big] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_little] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_native] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unit] = {
    .visible = true,
    .named = false,
  },
  [sym__discard_name] = {
    .visible = false,
    .named = true,
  },
  [sym__name] = {
    .visible = false,
    .named = true,
  },
  [sym__upname] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_content] = {
    .visible = true,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__module_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_target_group] = {
    .visible = true,
    .named = true,
  },
  [sym_target] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_value] = {
    .visible = true,
    .named = true,
  },
  [sym_import] = {
    .visible = true,
    .named = true,
  },
  [sym_module] = {
    .visible = true,
    .named = true,
  },
  [sym_unqualified_imports] = {
    .visible = true,
    .named = true,
  },
  [sym_unqualified_import] = {
    .visible = true,
    .named = true,
  },
  [sym_constant] = {
    .visible = true,
    .named = true,
  },
  [sym__constant_value] = {
    .visible = false,
    .named = true,
  },
  [sym_constant_tuple] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_list] = {
    .visible = true,
    .named = true,
  },
  [sym__constant_bit_string] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_bit_string_segment] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_bit_string_segment_options] = {
    .visible = true,
    .named = true,
  },
  [sym__constant_bit_string_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__constant_bit_string_named_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__constant_bit_string_segment_option_size] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_record] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_record_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_record_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_field_access] = {
    .visible = true,
    .named = true,
  },
  [sym__constant_type] = {
    .visible = false,
    .named = true,
  },
  [sym__constant_type_annotation] = {
    .visible = false,
    .named = true,
  },
  [sym_constant_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_function_parameter_types] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_type_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_external_type] = {
    .visible = true,
    .named = true,
  },
  [sym_external_function] = {
    .visible = true,
    .named = true,
  },
  [sym_external_function_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_external_function_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_external_function_body] = {
    .visible = true,
    .named = true,
  },
  [sym_function] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym__labeled_discard_param] = {
    .visible = false,
    .named = true,
  },
  [sym__discard_param] = {
    .visible = false,
    .named = true,
  },
  [sym__labeled_name_param] = {
    .visible = false,
    .named = true,
  },
  [sym__name_param] = {
    .visible = false,
    .named = true,
  },
  [aux_sym__statement_seq] = {
    .visible = false,
    .named = false,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_unit] = {
    .visible = false,
    .named = true,
  },
  [sym_record] = {
    .visible = true,
    .named = true,
  },
  [sym_todo] = {
    .visible = true,
    .named = true,
  },
  [sym_panic] = {
    .visible = true,
    .named = true,
  },
  [sym_pipeline_echo] = {
    .visible = true,
    .named = true,
  },
  [sym_echo] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple] = {
    .visible = true,
    .named = true,
  },
  [sym_list] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_bit_string] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_bit_string_segment] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_bit_string_segment_options] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_bit_string_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__expression_bit_string_named_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__expression_bit_string_segment_option_size] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_anonymous_function_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym_case] = {
    .visible = true,
    .named = true,
  },
  [sym_case_subjects] = {
    .visible = true,
    .named = true,
  },
  [sym_case_clauses] = {
    .visible = true,
    .named = true,
  },
  [sym_case_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_case_clause_patterns] = {
    .visible = true,
    .named = true,
  },
  [sym_case_clause_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_case_clause_guard] = {
    .visible = true,
    .named = true,
  },
  [sym__case_clause_guard_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__case_clause_guard_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__case_clause_guard_unit] = {
    .visible = false,
    .named = true,
  },
  [sym__case_clause_tuple_access] = {
    .visible = true,
    .named = true,
  },
  [sym_let_assert] = {
    .visible = true,
    .named = true,
  },
  [sym_let] = {
    .visible = true,
    .named = true,
  },
  [sym_use] = {
    .visible = true,
    .named = true,
  },
  [sym_use_assignments] = {
    .visible = true,
    .named = true,
  },
  [sym_use_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_assert] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_negation] = {
    .visible = true,
    .named = true,
  },
  [sym_integer_negation] = {
    .visible = true,
    .named = true,
  },
  [sym__assignment] = {
    .visible = false,
    .named = true,
  },
  [sym_record_update] = {
    .visible = true,
    .named = true,
  },
  [sym_record_update_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_record_update_argument] = {
    .visible = true,
    .named = true,
  },
  [sym__maybe_tuple_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_tuple_access] = {
    .visible = true,
    .named = true,
  },
  [sym__maybe_record_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_field_access] = {
    .visible = true,
    .named = true,
  },
  [sym__maybe_function_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_hole] = {
    .visible = true,
    .named = true,
  },
  [sym_function_call] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern_expression] = {
    .visible = false,
    .named = true,
  },
  [sym__pattern_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_record_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_record_pattern_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_record_pattern_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern_spread] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern_bit_string] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern_bit_string_segment] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern_bit_string_segment_options] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern_bit_string_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__pattern_bit_string_named_segment_option] = {
    .visible = false,
    .named = true,
  },
  [sym__pattern_bit_string_segment_option_size] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern_bit_string_segment_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_pattern_tail] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_data_constructors] = {
    .visible = true,
    .named = true,
  },
  [sym_data_constructor] = {
    .visible = true,
    .named = true,
  },
  [sym_data_constructor_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_data_constructor_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_type_alias] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_integer] = {
    .visible = true,
    .named = true,
  },
  [sym__bit_string_segment_option] = {
    .visible = true,
    .named = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
  },
  [sym__type_annotation] = {
    .visible = false,
    .named = true,
  },
  [sym_type_hole] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameter_types] = {
    .visible = true,
    .named = true,
  },
  [sym_type] = {
    .visible = true,
    .named = true,
  },
  [sym_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_type_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_type_var] = {
    .visible = true,
    .named = true,
  },
  [sym_type_name] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_label] = {
    .visible = true,
    .named = true,
  },
  [sym_discard] = {
    .visible = true,
    .named = true,
  },
  [sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_remote_type_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_name] = {
    .visible = true,
    .named = true,
  },
  [sym_remote_constructor_name] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_target_group_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__attribute_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_unqualified_imports_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_tuple_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__constant_bit_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_bit_string_segment_options_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_record_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_tuple_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_external_function_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tuple_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__expression_bit_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_bit_string_segment_options_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_anonymous_function_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_clauses_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_clause_patterns_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_clause_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_use_assignments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_update_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_pattern_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__pattern_bit_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_pattern_bit_string_segment_options_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_data_constructors_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_data_constructor_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tuple_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_function_body] = {
    .visible = true,
    .named = true,
  },
};

enum {
  field_alias = 1,
  field_arguments = 2,
  field_assign = 3,
  field_assignments = 4,
  field_body = 5,
  field_clauses = 6,
  field_constructor = 7,
  field_field = 8,
  field_function = 9,
  field_guard = 10,
  field_imports = 11,
  field_index = 12,
  field_label = 13,
  field_left = 14,
  field_message = 15,
  field_module = 16,
  field_name = 17,
  field_operator = 18,
  field_options = 19,
  field_parameter_types = 20,
  field_parameters = 21,
  field_pattern = 22,
  field_patterns = 23,
  field_record = 24,
  field_return_type = 25,
  field_right = 26,
  field_spread = 27,
  field_subjects = 28,
  field_target = 29,
  field_tuple = 30,
  field_type = 31,
  field_value = 32,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_arguments] = "arguments",
  [field_assign] = "assign",
  [field_assignments] = "assignments",
  [field_body] = "body",
  [field_clauses] = "clauses",
  [field_constructor] = "constructor",
  [field_field] = "field",
  [field_function] = "function",
  [field_guard] = "guard",
  [field_imports] = "imports",
  [field_index] = "index",
  [field_label] = "label",
  [field_left] = "left",
  [field_message] = "message",
  [field_module] = "module",
  [field_name] = "name",
  [field_operator] = "operator",
  [field_options] = "options",
  [field_parameter_types] = "parameter_types",
  [field_parameters] = "parameters",
  [field_pattern] = "pattern",
  [field_patterns] = "patterns",
  [field_record] = "record",
  [field_return_type] = "return_type",
  [field_right] = "right",
  [field_spread] = "spread",
  [field_subjects] = "subjects",
  [field_target] = "target",
  [field_tuple] = "tuple",
  [field_type] = "type",
  [field_value] = "value",
};

static const TSFieldMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 1},
  [2] = {.index = 1, .length = 1},
  [3] = {.index = 2, .length = 1},
  [4] = {.index = 3, .length = 1},
  [5] = {.index = 4, .length = 4},
  [6] = {.index = 8, .length = 3},
  [7] = {.index = 11, .length = 1},
  [8] = {.index = 12, .length = 2},
  [9] = {.index = 14, .length = 2},
  [10] = {.index = 16, .length = 2},
  [11] = {.index = 18, .length = 2},
  [12] = {.index = 20, .length = 1},
  [13] = {.index = 21, .length = 2},
  [14] = {.index = 23, .length = 1},
  [15] = {.index = 24, .length = 2},
  [16] = {.index = 26, .length = 4},
  [17] = {.index = 30, .length = 1},
  [18] = {.index = 31, .length = 2},
  [19] = {.index = 33, .length = 3},
  [20] = {.index = 36, .length = 2},
  [21] = {.index = 38, .length = 2},
  [22] = {.index = 40, .length = 2},
  [23] = {.index = 42, .length = 1},
  [24] = {.index = 43, .length = 2},
  [25] = {.index = 45, .length = 2},
  [26] = {.index = 47, .length = 1},
  [27] = {.index = 48, .length = 2},
  [28] = {.index = 50, .length = 3},
  [29] = {.index = 53, .length = 2},
  [30] = {.index = 55, .length = 2},
  [31] = {.index = 57, .length = 1},
  [32] = {.index = 58, .length = 1},
  [33] = {.index = 59, .length = 2},
  [34] = {.index = 61, .length = 3},
  [35] = {.index = 64, .length = 2},
  [36] = {.index = 66, .length = 2},
  [37] = {.index = 68, .length = 2},
  [38] = {.index = 70, .length = 1},
  [39] = {.index = 71, .length = 3},
  [40] = {.index = 74, .length = 1},
  [41] = {.index = 75, .length = 2},
  [42] = {.index = 77, .length = 3},
  [43] = {.index = 80, .length = 2},
  [44] = {.index = 82, .length = 3},
  [45] = {.index = 85, .length = 1},
  [46] = {.index = 86, .length = 2},
  [47] = {.index = 88, .length = 2},
  [48] = {.index = 90, .length = 1},
  [49] = {.index = 91, .length = 2},
  [50] = {.index = 93, .length = 4},
  [51] = {.index = 97, .length = 5},
  [52] = {.index = 102, .length = 4},
  [53] = {.index = 106, .length = 2},
  [54] = {.index = 108, .length = 2},
  [55] = {.index = 110, .length = 3},
  [56] = {.index = 113, .length = 1},
  [57] = {.index = 114, .length = 1},
  [58] = {.index = 115, .length = 3},
  [59] = {.index = 118, .length = 2},
  [60] = {.index = 120, .length = 2},
  [61] = {.index = 122, .length = 2},
  [62] = {.index = 124, .length = 3},
  [63] = {.index = 127, .length = 3},
  [64] = {.index = 130, .length = 3},
  [65] = {.index = 133, .length = 2},
  [66] = {.index = 135, .length = 2},
  [67] = {.index = 137, .length = 1},
  [68] = {.index = 138, .length = 3},
  [69] = {.index = 141, .length = 2},
  [70] = {.index = 143, .length = 3},
  [71] = {.index = 146, .length = 3},
  [72] = {.index = 149, .length = 3},
  [73] = {.index = 152, .length = 2},
  [74] = {.index = 154, .length = 4},
  [75] = {.index = 158, .length = 4},
  [76] = {.index = 158, .length = 4},
  [77] = {.index = 162, .length = 4},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_name, 0},
  [1] =
    {field_name, 1},
  [2] =
    {field_module, 1},
  [3] =
    {field_value, 0},
  [4] =
    {field_assign, 1, .inherited = true},
    {field_pattern, 1, .inherited = true},
    {field_type, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [8] =
    {field_left, 0, .inherited = true},
    {field_operator, 0, .inherited = true},
    {field_right, 0, .inherited = true},
  [11] =
    {field_assign, 0, .inherited = true},
  [12] =
    {field_arguments, 1},
    {field_function, 0},
  [14] =
    {field_arguments, 1},
    {field_name, 0},
  [16] =
    {field_arguments, 2},
    {field_name, 1},
  [18] =
    {field_name, 0},
    {field_parameters, 1},
  [20] =
    {field_name, 0, .inherited = true},
  [21] =
    {field_name, 1},
    {field_parameters, 2},
  [23] =
    {field_message, 2},
  [24] =
    {field_assign, 0, .inherited = true},
    {field_value, 0},
  [26] =
    {field_assign, 2, .inherited = true},
    {field_pattern, 2, .inherited = true},
    {field_type, 2, .inherited = true},
    {field_value, 2, .inherited = true},
  [30] =
    {field_value, 2},
  [31] =
    {field_assign, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [33] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [36] =
    {field_index, 2},
    {field_tuple, 0},
  [38] =
    {field_field, 2},
    {field_record, 0},
  [40] =
    {field_module, 0},
    {field_name, 2},
  [42] =
    {field_target, 1},
  [43] =
    {field_imports, 3},
    {field_module, 1},
  [45] =
    {field_alias, 3},
    {field_module, 1},
  [47] =
    {field_type, 1},
  [48] =
    {field_name, 1},
    {field_value, 3},
  [50] =
    {field_options, 1},
    {field_options, 2},
    {field_value, 0},
  [53] =
    {field_name, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [55] =
    {field_label, 0, .inherited = true},
    {field_name, 0, .inherited = true},
  [57] =
    {field_subjects, 1},
  [58] =
    {field_assign, 1, .inherited = true},
  [59] =
    {field_assign, 1},
    {field_assign, 2},
  [61] =
    {field_assign, 0, .inherited = true},
    {field_pattern, 0},
    {field_value, 2},
  [64] =
    {field_assign, 0, .inherited = true},
    {field_pattern, 0},
  [66] =
    {field_assignments, 1},
    {field_value, 3},
  [68] =
    {field_name, 2},
    {field_parameters, 3},
  [70] =
    {field_label, 0},
  [71] =
    {field_name, 1},
    {field_type, 2, .inherited = true},
    {field_value, 4},
  [74] =
    {field_spread, 3},
  [75] =
    {field_body, 3},
    {field_parameters, 1},
  [77] =
    {field_label, 0, .inherited = true},
    {field_name, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [80] =
    {field_label, 0},
    {field_name, 1},
  [82] =
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [85] =
    {field_type, 0},
  [86] =
    {field_clauses, 3},
    {field_subjects, 1},
  [88] =
    {field_assign, 0, .inherited = true},
    {field_assign, 1, .inherited = true},
  [90] =
    {field_assign, 2, .inherited = true},
  [91] =
    {field_assign, 1, .inherited = true},
    {field_assign, 2, .inherited = true},
  [93] =
    {field_assign, 0, .inherited = true},
    {field_options, 1},
    {field_options, 2},
    {field_value, 0},
  [97] =
    {field_assign, 2, .inherited = true},
    {field_message, 4},
    {field_pattern, 2, .inherited = true},
    {field_type, 2, .inherited = true},
    {field_value, 2, .inherited = true},
  [102] =
    {field_assign, 0, .inherited = true},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
    {field_value, 3},
  [106] =
    {field_name, 2},
    {field_value, 4},
  [108] =
    {field_label, 0},
    {field_value, 2},
  [110] =
    {field_alias, 5},
    {field_imports, 3},
    {field_module, 1},
  [113] =
    {field_return_type, 2},
  [114] =
    {field_spread, 4},
  [115] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 2},
  [118] =
    {field_index, 0, .inherited = true},
    {field_tuple, 0, .inherited = true},
  [120] =
    {field_patterns, 0},
    {field_value, 2},
  [122] =
    {field_assign, 2, .inherited = true},
    {field_assign, 3, .inherited = true},
  [124] =
    {field_assign, 2, .inherited = true},
    {field_label, 0},
    {field_pattern, 2},
  [127] =
    {field_name, 2},
    {field_type, 3, .inherited = true},
    {field_value, 5},
  [130] =
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [133] =
    {field_alias, 2},
    {field_name, 0},
  [135] =
    {field_parameter_types, 1},
    {field_return_type, 3},
  [137] =
    {field_spread, 5},
  [138] =
    {field_body, 5},
    {field_parameters, 1},
    {field_return_type, 3},
  [141] =
    {field_name, 0},
    {field_type, 2},
  [143] =
    {field_guard, 1},
    {field_patterns, 0},
    {field_value, 3},
  [146] =
    {field_body, 5},
    {field_name, 2},
    {field_parameters, 3},
  [149] =
    {field_arguments, 5},
    {field_constructor, 0},
    {field_spread, 3},
  [152] =
    {field_alias, 3},
    {field_name, 1},
  [154] =
    {field_body, 6},
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [158] =
    {field_body, 7},
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [162] =
    {field_body, 8},
    {field_name, 3},
    {field_parameters, 4},
    {field_return_type, 6},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [41] = {
    [3] = alias_sym_function_body,
  },
  [58] = {
    [4] = alias_sym_function_body,
  },
  [68] = {
    [5] = alias_sym_function_body,
  },
  [71] = {
    [5] = alias_sym_function_body,
  },
  [74] = {
    [6] = alias_sym_function_body,
  },
  [76] = {
    [7] = alias_sym_function_body,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  aux_sym__statement_seq, 2,
    aux_sym__statement_seq,
    alias_sym_function_body,
  0,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(170)
      if (lookahead == '\r') SKIP(170)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(268);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(216);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(207);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '\\') ADVANCE(127);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'a') ADVANCE(430);
      if (lookahead == 'b') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(336);
      if (lookahead == 'f') ADVANCE(388);
      if (lookahead == 'i') ADVANCE(367);
      if (lookahead == 'j') ADVANCE(323);
      if (lookahead == 'l') ADVANCE(359);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 's') ADVANCE(378);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(393);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(256);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('d' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 1:
      if (lookahead == '&') ADVANCE(228);
      END_STATE();
    case 2:
      if (lookahead == '-') ADVANCE(167);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(272);
      END_STATE();
    case 3:
      if (lookahead == '-') ADVANCE(260);
      if (lookahead == '<') ADVANCE(211);
      END_STATE();
    case 4:
      if (lookahead == '-') ADVANCE(260);
      if (lookahead == '>') ADVANCE(249);
      END_STATE();
    case 5:
      if (lookahead == '.') ADVANCE(253);
      END_STATE();
    case 6:
      if (lookahead == '.') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(6);
      END_STATE();
    case 7:
      if (lookahead == '/') ADVANCE(179);
      END_STATE();
    case 8:
      if (lookahead == '1') ADVANCE(10);
      if (lookahead == '3') ADVANCE(9);
      if (lookahead == '8') ADVANCE(292);
      END_STATE();
    case 9:
      if (lookahead == '2') ADVANCE(296);
      END_STATE();
    case 10:
      if (lookahead == '6') ADVANCE(294);
      END_STATE();
    case 11:
      if (lookahead == '<') ADVANCE(211);
      END_STATE();
    case 12:
      if (lookahead == '=') ADVANCE(230);
      END_STATE();
    case 13:
      if (lookahead == '=') ADVANCE(229);
      END_STATE();
    case 14:
      if (lookahead == '>') ADVANCE(224);
      END_STATE();
    case 15:
      if (lookahead == '>') ADVANCE(212);
      END_STATE();
    case 16:
      if (lookahead == '>') ADVANCE(241);
      if (lookahead == '|') ADVANCE(227);
      END_STATE();
    case 17:
      if (lookahead == '_') ADVANCE(111);
      if (lookahead == 's') ADVANCE(290);
      END_STATE();
    case 18:
      if (lookahead == 'a') ADVANCE(130);
      END_STATE();
    case 19:
      if (lookahead == 'a') ADVANCE(100);
      END_STATE();
    case 20:
      if (lookahead == 'a') ADVANCE(102);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(76);
      END_STATE();
    case 22:
      if (lookahead == 'a') ADVANCE(67);
      END_STATE();
    case 23:
      if (lookahead == 'a') ADVANCE(109);
      END_STATE();
    case 24:
      if (lookahead == 'a') ADVANCE(124);
      END_STATE();
    case 25:
      if (lookahead == 'a') ADVANCE(116);
      END_STATE();
    case 26:
      if (lookahead == 'b') ADVANCE(263);
      END_STATE();
    case 27:
      if (lookahead == 'c') ADVANCE(86);
      END_STATE();
    case 28:
      if (lookahead == 'c') ADVANCE(105);
      END_STATE();
    case 29:
      if (lookahead == 'c') ADVANCE(91);
      END_STATE();
    case 30:
      if (lookahead == 'c') ADVANCE(92);
      END_STATE();
    case 31:
      if (lookahead == 'd') ADVANCE(304);
      END_STATE();
    case 32:
      if (lookahead == 'd') ADVANCE(306);
      END_STATE();
    case 33:
      if (lookahead == 'd') ADVANCE(45);
      END_STATE();
    case 34:
      if (lookahead == 'd') ADVANCE(46);
      END_STATE();
    case 35:
      if (lookahead == 'd') ADVANCE(47);
      END_STATE();
    case 36:
      if (lookahead == 'e') ADVANCE(220);
      END_STATE();
    case 37:
      if (lookahead == 'e') ADVANCE(202);
      END_STATE();
    case 38:
      if (lookahead == 'e') ADVANCE(31);
      END_STATE();
    case 39:
      if (lookahead == 'e') ADVANCE(310);
      END_STATE();
    case 40:
      if (lookahead == 'e') ADVANCE(312);
      END_STATE();
    case 41:
      if (lookahead == 'e') ADVANCE(265);
      END_STATE();
    case 42:
      if (lookahead == 'e') ADVANCE(103);
      END_STATE();
    case 43:
      if (lookahead == 'e') ADVANCE(32);
      END_STATE();
    case 44:
      if (lookahead == 'e') ADVANCE(108);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(96);
      END_STATE();
    case 46:
      if (lookahead == 'e') ADVANCE(98);
      END_STATE();
    case 47:
      if (lookahead == 'e') ADVANCE(99);
      END_STATE();
    case 48:
      if (lookahead == 'f') ADVANCE(8);
      END_STATE();
    case 49:
      if (lookahead == 'f') ADVANCE(181);
      END_STATE();
    case 50:
      if (lookahead == 'f') ADVANCE(181);
      if (lookahead == 'm') ADVANCE(93);
      END_STATE();
    case 51:
      if (lookahead == 'g') ADVANCE(308);
      if (lookahead == 'n') ADVANCE(20);
      if (lookahead == 't') ADVANCE(17);
      END_STATE();
    case 52:
      if (lookahead == 'g') ADVANCE(185);
      END_STATE();
    case 53:
      if (lookahead == 'g') ADVANCE(288);
      END_STATE();
    case 54:
      if (lookahead == 'g') ADVANCE(78);
      if (lookahead == 'z') ADVANCE(36);
      END_STATE();
    case 55:
      if (lookahead == 'g') ADVANCE(80);
      END_STATE();
    case 56:
      if (lookahead == 'i') ADVANCE(51);
      if (lookahead == 'y') ADVANCE(112);
      END_STATE();
    case 57:
      if (lookahead == 'i') ADVANCE(54);
      END_STATE();
    case 58:
      if (lookahead == 'i') ADVANCE(55);
      END_STATE();
    case 59:
      if (lookahead == 'i') ADVANCE(131);
      END_STATE();
    case 60:
      if (lookahead == 'i') ADVANCE(126);
      END_STATE();
    case 61:
      if (lookahead == 'i') ADVANCE(97);
      END_STATE();
    case 62:
      if (lookahead == 'i') ADVANCE(77);
      END_STATE();
    case 63:
      if (lookahead == 'i') ADVANCE(114);
      if (lookahead == 's') ADVANCE(58);
      END_STATE();
    case 64:
      if (lookahead == 'i') ADVANCE(81);
      END_STATE();
    case 65:
      if (lookahead == 'i') ADVANCE(82);
      END_STATE();
    case 66:
      if (lookahead == 'i') ADVANCE(83);
      END_STATE();
    case 67:
      if (lookahead == 'l') ADVANCE(225);
      END_STATE();
    case 68:
      if (lookahead == 'l') ADVANCE(87);
      if (lookahead == 'n') ADVANCE(222);
      END_STATE();
    case 69:
      if (lookahead == 'l') ADVANCE(21);
      END_STATE();
    case 70:
      if (lookahead == 'l') ADVANCE(39);
      END_STATE();
    case 71:
      if (lookahead == 'm') ADVANCE(93);
      END_STATE();
    case 72:
      if (lookahead == 'm') ADVANCE(93);
      if (lookahead == 'n') ADVANCE(113);
      END_STATE();
    case 73:
      if (lookahead == 'n') ADVANCE(222);
      END_STATE();
    case 74:
      if (lookahead == 'n') ADVANCE(63);
      if (lookahead == 't') ADVANCE(48);
      END_STATE();
    case 75:
      if (lookahead == 'n') ADVANCE(110);
      END_STATE();
    case 76:
      if (lookahead == 'n') ADVANCE(52);
      END_STATE();
    case 77:
      if (lookahead == 'n') ADVANCE(53);
      END_STATE();
    case 78:
      if (lookahead == 'n') ADVANCE(38);
      END_STATE();
    case 79:
      if (lookahead == 'n') ADVANCE(22);
      END_STATE();
    case 80:
      if (lookahead == 'n') ADVANCE(43);
      END_STATE();
    case 81:
      if (lookahead == 'n') ADVANCE(119);
      END_STATE();
    case 82:
      if (lookahead == 'n') ADVANCE(120);
      END_STATE();
    case 83:
      if (lookahead == 'n') ADVANCE(121);
      END_STATE();
    case 84:
      if (lookahead == 'o') ADVANCE(75);
      END_STATE();
    case 85:
      if (lookahead == 'o') ADVANCE(106);
      END_STATE();
    case 86:
      if (lookahead == 'o') ADVANCE(33);
      END_STATE();
    case 87:
      if (lookahead == 'o') ADVANCE(25);
      END_STATE();
    case 88:
      if (lookahead == 'o') ADVANCE(64);
      END_STATE();
    case 89:
      if (lookahead == 'o') ADVANCE(65);
      END_STATE();
    case 90:
      if (lookahead == 'o') ADVANCE(66);
      END_STATE();
    case 91:
      if (lookahead == 'o') ADVANCE(34);
      END_STATE();
    case 92:
      if (lookahead == 'o') ADVANCE(35);
      END_STATE();
    case 93:
      if (lookahead == 'p') ADVANCE(85);
      END_STATE();
    case 94:
      if (lookahead == 'p') ADVANCE(19);
      END_STATE();
    case 95:
      if (lookahead == 'p') ADVANCE(37);
      END_STATE();
    case 96:
      if (lookahead == 'p') ADVANCE(88);
      END_STATE();
    case 97:
      if (lookahead == 'p') ADVANCE(118);
      END_STATE();
    case 98:
      if (lookahead == 'p') ADVANCE(89);
      END_STATE();
    case 99:
      if (lookahead == 'p') ADVANCE(90);
      END_STATE();
    case 100:
      if (lookahead == 'q') ADVANCE(129);
      END_STATE();
    case 101:
      if (lookahead == 'r') ADVANCE(69);
      if (lookahead == 'x') ADVANCE(123);
      END_STATE();
    case 102:
      if (lookahead == 'r') ADVANCE(133);
      END_STATE();
    case 103:
      if (lookahead == 'r') ADVANCE(79);
      END_STATE();
    case 104:
      if (lookahead == 'r') ADVANCE(62);
      END_STATE();
    case 105:
      if (lookahead == 'r') ADVANCE(61);
      END_STATE();
    case 106:
      if (lookahead == 'r') ADVANCE(117);
      END_STATE();
    case 107:
      if (lookahead == 's') ADVANCE(198);
      END_STATE();
    case 108:
      if (lookahead == 's') ADVANCE(282);
      END_STATE();
    case 109:
      if (lookahead == 's') ADVANCE(28);
      END_STATE();
    case 110:
      if (lookahead == 's') ADVANCE(115);
      END_STATE();
    case 111:
      if (lookahead == 's') ADVANCE(125);
      END_STATE();
    case 112:
      if (lookahead == 't') ADVANCE(44);
      END_STATE();
    case 113:
      if (lookahead == 't') ADVANCE(284);
      END_STATE();
    case 114:
      if (lookahead == 't') ADVANCE(314);
      END_STATE();
    case 115:
      if (lookahead == 't') ADVANCE(204);
      END_STATE();
    case 116:
      if (lookahead == 't') ADVANCE(286);
      END_STATE();
    case 117:
      if (lookahead == 't') ADVANCE(194);
      END_STATE();
    case 118:
      if (lookahead == 't') ADVANCE(187);
      END_STATE();
    case 119:
      if (lookahead == 't') ADVANCE(298);
      END_STATE();
    case 120:
      if (lookahead == 't') ADVANCE(300);
      END_STATE();
    case 121:
      if (lookahead == 't') ADVANCE(302);
      END_STATE();
    case 122:
      if (lookahead == 't') ADVANCE(70);
      END_STATE();
    case 123:
      if (lookahead == 't') ADVANCE(42);
      END_STATE();
    case 124:
      if (lookahead == 't') ADVANCE(59);
      END_STATE();
    case 125:
      if (lookahead == 't') ADVANCE(104);
      END_STATE();
    case 126:
      if (lookahead == 't') ADVANCE(122);
      END_STATE();
    case 127:
      if (lookahead == 'u') ADVANCE(135);
      if (lookahead == '"' ||
          lookahead == '\\' ||
          lookahead == 'e' ||
          lookahead == 'f' ||
          lookahead == 'n' ||
          lookahead == 'r' ||
          lookahead == 't') ADVANCE(269);
      END_STATE();
    case 128:
      if (lookahead == 'u') ADVANCE(26);
      END_STATE();
    case 129:
      if (lookahead == 'u') ADVANCE(41);
      END_STATE();
    case 130:
      if (lookahead == 'v') ADVANCE(23);
      END_STATE();
    case 131:
      if (lookahead == 'v') ADVANCE(40);
      END_STATE();
    case 132:
      if (lookahead == 'x') ADVANCE(123);
      END_STATE();
    case 133:
      if (lookahead == 'y') ADVANCE(280);
      END_STATE();
    case 134:
      if (lookahead == 'y') ADVANCE(95);
      END_STATE();
    case 135:
      if (lookahead == '{') ADVANCE(168);
      END_STATE();
    case 136:
      if (lookahead == '}') ADVANCE(270);
      END_STATE();
    case 137:
      if (lookahead == '}') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(136);
      END_STATE();
    case 138:
      if (lookahead == '}') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(137);
      END_STATE();
    case 139:
      if (lookahead == '}') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(138);
      END_STATE();
    case 140:
      if (lookahead == '}') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(139);
      END_STATE();
    case 141:
      if (lookahead == '}') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(140);
      END_STATE();
    case 142:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(142)
      if (lookahead == '\r') SKIP(142)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(5);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 143:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(143)
      if (lookahead == '\r') SKIP(143)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(430);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 144:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(144)
      if (lookahead == '\r') SKIP(144)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 145:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(145)
      if (lookahead == '\r') SKIP(145)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(5);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 146:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(146)
      if (lookahead == '\r') SKIP(146)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == '-') ADVANCE(216);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 147:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(148)
      if (lookahead == '\r') SKIP(148)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '"') ADVANCE(268);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(214);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(276);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(232);
      if (lookahead == '=') ADVANCE(207);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '\\') ADVANCE(127);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'b') ADVANCE(56);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'f') ADVANCE(68);
      if (lookahead == 'i') ADVANCE(72);
      if (lookahead == 'j') ADVANCE(18);
      if (lookahead == 'l') ADVANCE(60);
      if (lookahead == 'n') ADVANCE(24);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 's') ADVANCE(57);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == 'u') ADVANCE(74);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(277);
      END_STATE();
    case 148:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(148)
      if (lookahead == '\r') SKIP(148)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(214);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(276);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(232);
      if (lookahead == '=') ADVANCE(207);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'b') ADVANCE(56);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'f') ADVANCE(68);
      if (lookahead == 'i') ADVANCE(72);
      if (lookahead == 'j') ADVANCE(18);
      if (lookahead == 'l') ADVANCE(60);
      if (lookahead == 'n') ADVANCE(24);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 's') ADVANCE(57);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == 'u') ADVANCE(74);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(277);
      END_STATE();
    case 149:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(149)
      if (lookahead == '\r') SKIP(149)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(214);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(232);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(132);
      if (lookahead == 'f') ADVANCE(73);
      if (lookahead == 'i') ADVANCE(71);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      END_STATE();
    case 150:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(150)
      if (lookahead == '\r') SKIP(150)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 151:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(151)
      if (lookahead == '\r') SKIP(151)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'a') ADVANCE(433);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 152:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(152)
      if (lookahead == '\r') SKIP(152)
      if (lookahead == '!') ADVANCE(261);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(219);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(11);
      if (lookahead == '>') ADVANCE(15);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == '{') ADVANCE(183);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 153:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(153)
      if (lookahead == '\r') SKIP(153)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(215);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '<') ADVANCE(232);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'i') ADVANCE(49);
      if (lookahead == '|') ADVANCE(256);
      END_STATE();
    case 154:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(154)
      if (lookahead == '\r') SKIP(154)
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '-') ADVANCE(218);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(11);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'i') ADVANCE(369);
      if (lookahead == '|') ADVANCE(255);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 155:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(155)
      if (lookahead == '\r') SKIP(155)
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '-') ADVANCE(219);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(3);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(316);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 156:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(156)
      if (lookahead == '\r') SKIP(156)
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '-') ADVANCE(219);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(11);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'a') ADVANCE(439);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 157:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(157)
      if (lookahead == '\r') SKIP(157)
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(213);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(4);
      if (lookahead == '=') ADVANCE(206);
      if (lookahead == '>') ADVANCE(15);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(132);
      if (lookahead == 'f') ADVANCE(73);
      if (lookahead == 'i') ADVANCE(71);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '}') ADVANCE(184);
      END_STATE();
    case 158:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(158)
      if (lookahead == '\r') SKIP(158)
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(14);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(4);
      if (lookahead == '=') ADVANCE(206);
      if (lookahead == '>') ADVANCE(15);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(132);
      if (lookahead == 'f') ADVANCE(73);
      if (lookahead == 'i') ADVANCE(50);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == '|') ADVANCE(255);
      if (lookahead == '}') ADVANCE(184);
      END_STATE();
    case 159:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(159)
      if (lookahead == '\r') SKIP(159)
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '_') ADVANCE(317);
      if (lookahead == 'f') ADVANCE(392);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 160:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(160)
      if (lookahead == '\r') SKIP(160)
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(317);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 161:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(161)
      if (lookahead == '\r') SKIP(161)
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '-') ADVANCE(213);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(276);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(277);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      if (lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 162:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(162)
      if (lookahead == '\r') SKIP(162)
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == 't') ADVANCE(461);
      if (lookahead == '}') ADVANCE(184);
      if (lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 163:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(163)
      if (lookahead == '\r') SKIP(163)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(207);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(326);
      if (lookahead == 'e') ADVANCE(338);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'p') ADVANCE(328);
      if (lookahead == 't') ADVANCE(405);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 164:
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(164)
      if (lookahead == '\r') SKIP(164)
      if (lookahead == '!') ADVANCE(12);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(215);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(232);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == 'a') ADVANCE(107);
      if (lookahead == 'c') ADVANCE(84);
      if (lookahead == 'e') ADVANCE(132);
      if (lookahead == 'f') ADVANCE(73);
      if (lookahead == 'i') ADVANCE(50);
      if (lookahead == 'o') ADVANCE(94);
      if (lookahead == 'p') ADVANCE(128);
      if (lookahead == 't') ADVANCE(134);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (lookahead == '}') ADVANCE(184);
      END_STATE();
    case 165:
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(279);
      END_STATE();
    case 166:
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(278);
      END_STATE();
    case 167:
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(272);
      END_STATE();
    case 168:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(141);
      END_STATE();
    case 169:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(273);
      END_STATE();
    case 170:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(170)
      if (lookahead == '\r') SKIP(170)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(216);
      if (lookahead == '.') ADVANCE(197);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(207);
      if (lookahead == '>') ADVANCE(237);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == ']') ADVANCE(210);
      if (lookahead == '_') ADVANCE(316);
      if (lookahead == 'a') ADVANCE(430);
      if (lookahead == 'b') ADVANCE(377);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(336);
      if (lookahead == 'f') ADVANCE(388);
      if (lookahead == 'i') ADVANCE(367);
      if (lookahead == 'j') ADVANCE(323);
      if (lookahead == 'l') ADVANCE(359);
      if (lookahead == 'n') ADVANCE(333);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 's') ADVANCE(378);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(393);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(256);
      if (lookahead == '}') ADVANCE(184);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('d' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 171:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(171)
      if (lookahead == '\r') SKIP(171)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(337);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'i') ADVANCE(368);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 172:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(172)
      if (lookahead == '\r') SKIP(172)
      if (lookahead == '!') ADVANCE(261);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == ')') ADVANCE(192);
      if (lookahead == ',') ADVANCE(191);
      if (lookahead == '-') ADVANCE(219);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == ':') ADVANCE(193);
      if (lookahead == '<') ADVANCE(3);
      if (lookahead == '=') ADVANCE(206);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(337);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'i') ADVANCE(368);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 173:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(173)
      if (lookahead == '\r') SKIP(173)
      if (lookahead == '!') ADVANCE(261);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '-') ADVANCE(218);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(11);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(439);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(337);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'i') ADVANCE(368);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 174:
      if (eof) ADVANCE(175);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == ' ') SKIP(174)
      if (lookahead == '\r') SKIP(174)
      if (lookahead == '!') ADVANCE(262);
      if (lookahead == '"') ADVANCE(267);
      if (lookahead == '#') ADVANCE(208);
      if (lookahead == '%') ADVANCE(248);
      if (lookahead == '&') ADVANCE(1);
      if (lookahead == '(') ADVANCE(190);
      if (lookahead == '*') ADVANCE(245);
      if (lookahead == '+') ADVANCE(242);
      if (lookahead == '-') ADVANCE(217);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == '/') ADVANCE(201);
      if (lookahead == '0') ADVANCE(274);
      if (lookahead == '<') ADVANCE(231);
      if (lookahead == '=') ADVANCE(13);
      if (lookahead == '>') ADVANCE(236);
      if (lookahead == '@') ADVANCE(189);
      if (lookahead == '[') ADVANCE(209);
      if (lookahead == '_') ADVANCE(318);
      if (lookahead == 'a') ADVANCE(430);
      if (lookahead == 'c') ADVANCE(325);
      if (lookahead == 'e') ADVANCE(337);
      if (lookahead == 'f') ADVANCE(392);
      if (lookahead == 'i') ADVANCE(368);
      if (lookahead == 'l') ADVANCE(360);
      if (lookahead == 'o') ADVANCE(416);
      if (lookahead == 'p') ADVANCE(327);
      if (lookahead == 't') ADVANCE(404);
      if (lookahead == 'u') ADVANCE(431);
      if (lookahead == '{') ADVANCE(183);
      if (lookahead == '|') ADVANCE(16);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(275);
      if (('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      if (('A' <= lookahead && lookahead <= 'Z')) ADVANCE(463);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(sym_module_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(176);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(sym_statement_comment);
      if (lookahead == '/') ADVANCE(176);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(178);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(sym_statement_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(178);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(177);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(180);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(180);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_if);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_erlang);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_erlang);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_javascript);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_javascript);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_import);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(253);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_as);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 's') ADVANCE(351);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_as);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '.') ADVANCE(247);
      if (lookahead == '/') ADVANCE(179);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_type);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_const);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_const);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(229);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(244);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(244);
      if (lookahead == '>') ADVANCE(224);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(244);
      if (lookahead == '>') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(6);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(244);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(6);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(6);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(6);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_size);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_size);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_fn);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_fn);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_external);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_external);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '.') ADVANCE(234);
      if (lookahead == '<') ADVANCE(211);
      if (lookahead == '=') ADVANCE(233);
      if (lookahead == '>') ADVANCE(249);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '.') ADVANCE(234);
      if (lookahead == '=') ADVANCE(233);
      if (lookahead == '>') ADVANCE(249);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      if (lookahead == '.') ADVANCE(235);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_LT_DOT);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_LT_EQ_DOT);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '.') ADVANCE(239);
      if (lookahead == '=') ADVANCE(238);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '.') ADVANCE(239);
      if (lookahead == '=') ADVANCE(238);
      if (lookahead == '>') ADVANCE(212);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      if (lookahead == '.') ADVANCE(240);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_GT_DOT);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_GT_EQ_DOT);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(anon_sym_PIPE_GT);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(243);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(anon_sym_PLUS_DOT);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(anon_sym_DASH_DOT);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '.') ADVANCE(246);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_STAR_DOT);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_SLASH_DOT);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_LT_GT);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_todo);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_panic);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_echo);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_case);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '>') ADVANCE(241);
      if (lookahead == '|') ADVANCE(227);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_let);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(anon_sym_assert);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(anon_sym_use);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(230);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_visibility_modifier);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_visibility_modifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_opacity_modifier);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_opacity_modifier);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(anon_sym_DQUOTE2);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token1);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token2);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(sym_float);
      if (lookahead == 'e') ADVANCE(2);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(271);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(sym_float);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(272);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(sym__hex);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(273);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(sym__decimal);
      if (lookahead == '.') ADVANCE(271);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(165);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(166);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(169);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(275);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(sym__decimal);
      if (lookahead == '.') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(275);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(sym__decimal);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(165);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(166);
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(169);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(277);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(sym__decimal);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(277);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(sym__octal);
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(278);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(sym__binary);
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(279);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(anon_sym_binary);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(anon_sym_binary);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(anon_sym_bytes);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(anon_sym_bytes);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(anon_sym_int);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(anon_sym_int);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(anon_sym_float);
      END_STATE();
    case 287:
      ACCEPT_TOKEN(anon_sym_float);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(anon_sym_bit_string);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(anon_sym_bit_string);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(anon_sym_bits);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(anon_sym_bits);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(anon_sym_utf8);
      if (lookahead == '_') ADVANCE(27);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(anon_sym_utf8);
      if (lookahead == '_') ADVANCE(340);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 294:
      ACCEPT_TOKEN(anon_sym_utf16);
      if (lookahead == '_') ADVANCE(29);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(anon_sym_utf16);
      if (lookahead == '_') ADVANCE(342);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(anon_sym_utf32);
      if (lookahead == '_') ADVANCE(30);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(anon_sym_utf32);
      if (lookahead == '_') ADVANCE(343);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(anon_sym_utf8_codepoint);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(anon_sym_utf8_codepoint);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(anon_sym_utf16_codepoint);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(anon_sym_utf16_codepoint);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(anon_sym_utf32_codepoint);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(anon_sym_utf32_codepoint);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(anon_sym_signed);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(anon_sym_signed);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(anon_sym_unsigned);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(anon_sym_unsigned);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(anon_sym_big);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(anon_sym_big);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(anon_sym_little);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(anon_sym_little);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(anon_sym_native);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(anon_sym_native);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(anon_sym_unit);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(anon_sym_unit);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym__discard_name);
      if (lookahead == '.') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(316);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(317);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym__discard_name);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(317);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == '.') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(318);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == '1') ADVANCE(321);
      if (lookahead == '3') ADVANCE(320);
      if (lookahead == '8') ADVANCE(293);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == '2') ADVANCE(297);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == '6') ADVANCE(295);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == '_') ADVANCE(438);
      if (lookahead == 's') ADVANCE(291);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(458);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(423);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(436);
      if (lookahead == 'o') ADVANCE(395);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(436);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(396);
      if (lookahead == 'u') ADVANCE(335);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(396);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(424);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(394);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(435);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(451);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'a') ADVANCE(444);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'b') ADVANCE(264);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(376);
      if (lookahead == 'r') ADVANCE(390);
      if (lookahead == 'x') ADVANCE(455);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(376);
      if (lookahead == 'x') ADVANCE(455);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(376);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(251);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(408);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(427);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(414);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'c') ADVANCE(415);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(305);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(307);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(407);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(364);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(365);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'd') ADVANCE(366);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(259);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(428);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(254);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(221);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(203);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(311);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(313);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(266);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(344);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(441);
      if (lookahead == 'i') ADVANCE(456);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 360:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(441);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 361:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(345);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 362:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(426);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 363:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(432);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 364:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(418);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 365:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(421);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 366:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'e') ADVANCE(422);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 367:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'f') ADVANCE(182);
      if (lookahead == 'm') ADVANCE(417);
      if (lookahead == 'n') ADVANCE(440);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 368:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'f') ADVANCE(182);
      if (lookahead == 'm') ADVANCE(417);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 369:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'f') ADVANCE(182);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 370:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'f') ADVANCE(319);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 371:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'g') ADVANCE(309);
      if (lookahead == 'n') ADVANCE(329);
      if (lookahead == 't') ADVANCE(322);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 372:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'g') ADVANCE(186);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 373:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'g') ADVANCE(289);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 374:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'g') ADVANCE(399);
      if (lookahead == 'z') ADVANCE(353);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'y')) ADVANCE(462);
      END_STATE();
    case 375:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'g') ADVANCE(400);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 376:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'h') ADVANCE(406);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 377:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(371);
      if (lookahead == 'y') ADVANCE(453);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 378:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(374);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 379:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(339);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 380:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(375);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 381:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(459);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 382:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(420);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 383:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(397);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 384:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(442);
      if (lookahead == 's') ADVANCE(380);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 385:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(401);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 386:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(402);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 387:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'i') ADVANCE(403);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 388:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'l') ADVANCE(410);
      if (lookahead == 'n') ADVANCE(223);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 389:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'l') ADVANCE(226);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 390:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'l') ADVANCE(331);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 391:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'l') ADVANCE(355);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 392:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(223);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 393:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(384);
      if (lookahead == 's') ADVANCE(350);
      if (lookahead == 't') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 394:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 395:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(437);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 396:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(379);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 397:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(373);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 398:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 399:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(358);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 400:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(361);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 401:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(448);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 402:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(449);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 403:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'n') ADVANCE(450);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 404:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(346);
      if (lookahead == 'y') ADVANCE(419);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 405:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(346);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 406:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(252);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 407:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(250);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 408:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(347);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 409:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(429);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 410:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(334);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 411:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(385);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 412:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(386);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 413:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(387);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 414:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(348);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 415:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'o') ADVANCE(349);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 416:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(324);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 417:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(409);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 418:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(411);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 419:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(354);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 420:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(447);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 421:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(412);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 422:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'p') ADVANCE(413);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 423:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'q') ADVANCE(457);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 424:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(460);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 425:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(383);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 426:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(398);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 427:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(382);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 428:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(445);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 429:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'r') ADVANCE(446);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 430:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(199);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 431:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(350);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 432:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(283);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 433:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(200);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 434:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(351);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 435:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(341);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 436:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(352);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 437:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(443);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 438:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(454);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 439:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 's') ADVANCE(434);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 440:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(285);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 441:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 442:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(315);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 443:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(205);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 444:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(287);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 445:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(258);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 446:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(195);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 447:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(188);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 448:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(299);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 449:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(301);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 450:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(303);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 451:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(381);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 452:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(391);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 453:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(363);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 454:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(425);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 455:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(362);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 456:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 't') ADVANCE(452);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 457:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'u') ADVANCE(357);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 458:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'v') ADVANCE(332);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 459:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'v') ADVANCE(356);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 460:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'y') ADVANCE(281);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 461:
      ACCEPT_TOKEN(sym__name);
      if (lookahead == 'y') ADVANCE(419);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 462:
      ACCEPT_TOKEN(sym__name);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(462);
      END_STATE();
    case 463:
      ACCEPT_TOKEN(sym__upname);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(463);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 171},
  [2] = {.lex_state = 174},
  [3] = {.lex_state = 171},
  [4] = {.lex_state = 142},
  [5] = {.lex_state = 171},
  [6] = {.lex_state = 171},
  [7] = {.lex_state = 143},
  [8] = {.lex_state = 144},
  [9] = {.lex_state = 145},
  [10] = {.lex_state = 142},
  [11] = {.lex_state = 146},
  [12] = {.lex_state = 171},
  [13] = {.lex_state = 171},
  [14] = {.lex_state = 174},
  [15] = {.lex_state = 174},
  [16] = {.lex_state = 144},
  [17] = {.lex_state = 144},
  [18] = {.lex_state = 144},
  [19] = {.lex_state = 144},
  [20] = {.lex_state = 144},
  [21] = {.lex_state = 144},
  [22] = {.lex_state = 144},
  [23] = {.lex_state = 174},
  [24] = {.lex_state = 144},
  [25] = {.lex_state = 174},
  [26] = {.lex_state = 144},
  [27] = {.lex_state = 144},
  [28] = {.lex_state = 144},
  [29] = {.lex_state = 144},
  [30] = {.lex_state = 174},
  [31] = {.lex_state = 174},
  [32] = {.lex_state = 144},
  [33] = {.lex_state = 144},
  [34] = {.lex_state = 144},
  [35] = {.lex_state = 144},
  [36] = {.lex_state = 144},
  [37] = {.lex_state = 144},
  [38] = {.lex_state = 144},
  [39] = {.lex_state = 144},
  [40] = {.lex_state = 174},
  [41] = {.lex_state = 144},
  [42] = {.lex_state = 144},
  [43] = {.lex_state = 144},
  [44] = {.lex_state = 171},
  [45] = {.lex_state = 174},
  [46] = {.lex_state = 144},
  [47] = {.lex_state = 174},
  [48] = {.lex_state = 144},
  [49] = {.lex_state = 174},
  [50] = {.lex_state = 144},
  [51] = {.lex_state = 144},
  [52] = {.lex_state = 144},
  [53] = {.lex_state = 174},
  [54] = {.lex_state = 144},
  [55] = {.lex_state = 174},
  [56] = {.lex_state = 144},
  [57] = {.lex_state = 144},
  [58] = {.lex_state = 144},
  [59] = {.lex_state = 174},
  [60] = {.lex_state = 174},
  [61] = {.lex_state = 174},
  [62] = {.lex_state = 174},
  [63] = {.lex_state = 174},
  [64] = {.lex_state = 144},
  [65] = {.lex_state = 144},
  [66] = {.lex_state = 174},
  [67] = {.lex_state = 144},
  [68] = {.lex_state = 144},
  [69] = {.lex_state = 144},
  [70] = {.lex_state = 174},
  [71] = {.lex_state = 144},
  [72] = {.lex_state = 171},
  [73] = {.lex_state = 144},
  [74] = {.lex_state = 171},
  [75] = {.lex_state = 144},
  [76] = {.lex_state = 144},
  [77] = {.lex_state = 171},
  [78] = {.lex_state = 144},
  [79] = {.lex_state = 144},
  [80] = {.lex_state = 174},
  [81] = {.lex_state = 144},
  [82] = {.lex_state = 144},
  [83] = {.lex_state = 144},
  [84] = {.lex_state = 144},
  [85] = {.lex_state = 174},
  [86] = {.lex_state = 171},
  [87] = {.lex_state = 174},
  [88] = {.lex_state = 174},
  [89] = {.lex_state = 171},
  [90] = {.lex_state = 174},
  [91] = {.lex_state = 144},
  [92] = {.lex_state = 174},
  [93] = {.lex_state = 174},
  [94] = {.lex_state = 171},
  [95] = {.lex_state = 174},
  [96] = {.lex_state = 144},
  [97] = {.lex_state = 174},
  [98] = {.lex_state = 171},
  [99] = {.lex_state = 144},
  [100] = {.lex_state = 171},
  [101] = {.lex_state = 144},
  [102] = {.lex_state = 144},
  [103] = {.lex_state = 171},
  [104] = {.lex_state = 174},
  [105] = {.lex_state = 174},
  [106] = {.lex_state = 144},
  [107] = {.lex_state = 144},
  [108] = {.lex_state = 171},
  [109] = {.lex_state = 171},
  [110] = {.lex_state = 174},
  [111] = {.lex_state = 144},
  [112] = {.lex_state = 144},
  [113] = {.lex_state = 171},
  [114] = {.lex_state = 171},
  [115] = {.lex_state = 144},
  [116] = {.lex_state = 171},
  [117] = {.lex_state = 144},
  [118] = {.lex_state = 171},
  [119] = {.lex_state = 171},
  [120] = {.lex_state = 171},
  [121] = {.lex_state = 144},
  [122] = {.lex_state = 174},
  [123] = {.lex_state = 171},
  [124] = {.lex_state = 174},
  [125] = {.lex_state = 174},
  [126] = {.lex_state = 174},
  [127] = {.lex_state = 174},
  [128] = {.lex_state = 163},
  [129] = {.lex_state = 171},
  [130] = {.lex_state = 174},
  [131] = {.lex_state = 145},
  [132] = {.lex_state = 145},
  [133] = {.lex_state = 145},
  [134] = {.lex_state = 171},
  [135] = {.lex_state = 174},
  [136] = {.lex_state = 171},
  [137] = {.lex_state = 171},
  [138] = {.lex_state = 174},
  [139] = {.lex_state = 171},
  [140] = {.lex_state = 174},
  [141] = {.lex_state = 145},
  [142] = {.lex_state = 171},
  [143] = {.lex_state = 171},
  [144] = {.lex_state = 174},
  [145] = {.lex_state = 171},
  [146] = {.lex_state = 171},
  [147] = {.lex_state = 174},
  [148] = {.lex_state = 145},
  [149] = {.lex_state = 174},
  [150] = {.lex_state = 174},
  [151] = {.lex_state = 174},
  [152] = {.lex_state = 171},
  [153] = {.lex_state = 174},
  [154] = {.lex_state = 174},
  [155] = {.lex_state = 174},
  [156] = {.lex_state = 174},
  [157] = {.lex_state = 174},
  [158] = {.lex_state = 174},
  [159] = {.lex_state = 174},
  [160] = {.lex_state = 171},
  [161] = {.lex_state = 145},
  [162] = {.lex_state = 174},
  [163] = {.lex_state = 174},
  [164] = {.lex_state = 174},
  [165] = {.lex_state = 174},
  [166] = {.lex_state = 174},
  [167] = {.lex_state = 174},
  [168] = {.lex_state = 174},
  [169] = {.lex_state = 174},
  [170] = {.lex_state = 145},
  [171] = {.lex_state = 174},
  [172] = {.lex_state = 174},
  [173] = {.lex_state = 174},
  [174] = {.lex_state = 174},
  [175] = {.lex_state = 174},
  [176] = {.lex_state = 174},
  [177] = {.lex_state = 171},
  [178] = {.lex_state = 171},
  [179] = {.lex_state = 171},
  [180] = {.lex_state = 145},
  [181] = {.lex_state = 171},
  [182] = {.lex_state = 145},
  [183] = {.lex_state = 171},
  [184] = {.lex_state = 145},
  [185] = {.lex_state = 145},
  [186] = {.lex_state = 171},
  [187] = {.lex_state = 171},
  [188] = {.lex_state = 171},
  [189] = {.lex_state = 171},
  [190] = {.lex_state = 145},
  [191] = {.lex_state = 145},
  [192] = {.lex_state = 145},
  [193] = {.lex_state = 171},
  [194] = {.lex_state = 145},
  [195] = {.lex_state = 171},
  [196] = {.lex_state = 171},
  [197] = {.lex_state = 145},
  [198] = {.lex_state = 171},
  [199] = {.lex_state = 171},
  [200] = {.lex_state = 171},
  [201] = {.lex_state = 171},
  [202] = {.lex_state = 171},
  [203] = {.lex_state = 171},
  [204] = {.lex_state = 171},
  [205] = {.lex_state = 171},
  [206] = {.lex_state = 171},
  [207] = {.lex_state = 171},
  [208] = {.lex_state = 171},
  [209] = {.lex_state = 145},
  [210] = {.lex_state = 171},
  [211] = {.lex_state = 145},
  [212] = {.lex_state = 145},
  [213] = {.lex_state = 171},
  [214] = {.lex_state = 145},
  [215] = {.lex_state = 171},
  [216] = {.lex_state = 145},
  [217] = {.lex_state = 171},
  [218] = {.lex_state = 171},
  [219] = {.lex_state = 171},
  [220] = {.lex_state = 145},
  [221] = {.lex_state = 171},
  [222] = {.lex_state = 171},
  [223] = {.lex_state = 145},
  [224] = {.lex_state = 171},
  [225] = {.lex_state = 171},
  [226] = {.lex_state = 171},
  [227] = {.lex_state = 145},
  [228] = {.lex_state = 145},
  [229] = {.lex_state = 145},
  [230] = {.lex_state = 171},
  [231] = {.lex_state = 145},
  [232] = {.lex_state = 171},
  [233] = {.lex_state = 145},
  [234] = {.lex_state = 144},
  [235] = {.lex_state = 144},
  [236] = {.lex_state = 144},
  [237] = {.lex_state = 145},
  [238] = {.lex_state = 145},
  [239] = {.lex_state = 144},
  [240] = {.lex_state = 144},
  [241] = {.lex_state = 142},
  [242] = {.lex_state = 144},
  [243] = {.lex_state = 142},
  [244] = {.lex_state = 144},
  [245] = {.lex_state = 142},
  [246] = {.lex_state = 142},
  [247] = {.lex_state = 144},
  [248] = {.lex_state = 144},
  [249] = {.lex_state = 144},
  [250] = {.lex_state = 144},
  [251] = {.lex_state = 142},
  [252] = {.lex_state = 142},
  [253] = {.lex_state = 142},
  [254] = {.lex_state = 142},
  [255] = {.lex_state = 142},
  [256] = {.lex_state = 142},
  [257] = {.lex_state = 144},
  [258] = {.lex_state = 142},
  [259] = {.lex_state = 144},
  [260] = {.lex_state = 144},
  [261] = {.lex_state = 144},
  [262] = {.lex_state = 144},
  [263] = {.lex_state = 144},
  [264] = {.lex_state = 142},
  [265] = {.lex_state = 142},
  [266] = {.lex_state = 144},
  [267] = {.lex_state = 144},
  [268] = {.lex_state = 142},
  [269] = {.lex_state = 142},
  [270] = {.lex_state = 142},
  [271] = {.lex_state = 142},
  [272] = {.lex_state = 144},
  [273] = {.lex_state = 144},
  [274] = {.lex_state = 144},
  [275] = {.lex_state = 142},
  [276] = {.lex_state = 142},
  [277] = {.lex_state = 142},
  [278] = {.lex_state = 142},
  [279] = {.lex_state = 142},
  [280] = {.lex_state = 142},
  [281] = {.lex_state = 142},
  [282] = {.lex_state = 142},
  [283] = {.lex_state = 144},
  [284] = {.lex_state = 142},
  [285] = {.lex_state = 144},
  [286] = {.lex_state = 142},
  [287] = {.lex_state = 142},
  [288] = {.lex_state = 142},
  [289] = {.lex_state = 142},
  [290] = {.lex_state = 142},
  [291] = {.lex_state = 144},
  [292] = {.lex_state = 142},
  [293] = {.lex_state = 142},
  [294] = {.lex_state = 144},
  [295] = {.lex_state = 144},
  [296] = {.lex_state = 144},
  [297] = {.lex_state = 144},
  [298] = {.lex_state = 144},
  [299] = {.lex_state = 142},
  [300] = {.lex_state = 142},
  [301] = {.lex_state = 142},
  [302] = {.lex_state = 142},
  [303] = {.lex_state = 142},
  [304] = {.lex_state = 142},
  [305] = {.lex_state = 142},
  [306] = {.lex_state = 142},
  [307] = {.lex_state = 142},
  [308] = {.lex_state = 142},
  [309] = {.lex_state = 142},
  [310] = {.lex_state = 142},
  [311] = {.lex_state = 142},
  [312] = {.lex_state = 142},
  [313] = {.lex_state = 142},
  [314] = {.lex_state = 142},
  [315] = {.lex_state = 143},
  [316] = {.lex_state = 142},
  [317] = {.lex_state = 142},
  [318] = {.lex_state = 142},
  [319] = {.lex_state = 142},
  [320] = {.lex_state = 142},
  [321] = {.lex_state = 142},
  [322] = {.lex_state = 142},
  [323] = {.lex_state = 142},
  [324] = {.lex_state = 142},
  [325] = {.lex_state = 142},
  [326] = {.lex_state = 142},
  [327] = {.lex_state = 142},
  [328] = {.lex_state = 142},
  [329] = {.lex_state = 142},
  [330] = {.lex_state = 142},
  [331] = {.lex_state = 142},
  [332] = {.lex_state = 142},
  [333] = {.lex_state = 142},
  [334] = {.lex_state = 142},
  [335] = {.lex_state = 142},
  [336] = {.lex_state = 142},
  [337] = {.lex_state = 142},
  [338] = {.lex_state = 142},
  [339] = {.lex_state = 142},
  [340] = {.lex_state = 142},
  [341] = {.lex_state = 142},
  [342] = {.lex_state = 144},
  [343] = {.lex_state = 142},
  [344] = {.lex_state = 144},
  [345] = {.lex_state = 142},
  [346] = {.lex_state = 144},
  [347] = {.lex_state = 144},
  [348] = {.lex_state = 142},
  [349] = {.lex_state = 142},
  [350] = {.lex_state = 142},
  [351] = {.lex_state = 142},
  [352] = {.lex_state = 144},
  [353] = {.lex_state = 142},
  [354] = {.lex_state = 143},
  [355] = {.lex_state = 144},
  [356] = {.lex_state = 142},
  [357] = {.lex_state = 142},
  [358] = {.lex_state = 142},
  [359] = {.lex_state = 144},
  [360] = {.lex_state = 142},
  [361] = {.lex_state = 142},
  [362] = {.lex_state = 144},
  [363] = {.lex_state = 142},
  [364] = {.lex_state = 142},
  [365] = {.lex_state = 142},
  [366] = {.lex_state = 142},
  [367] = {.lex_state = 142},
  [368] = {.lex_state = 142},
  [369] = {.lex_state = 142},
  [370] = {.lex_state = 142},
  [371] = {.lex_state = 142},
  [372] = {.lex_state = 142},
  [373] = {.lex_state = 142},
  [374] = {.lex_state = 142},
  [375] = {.lex_state = 142},
  [376] = {.lex_state = 144},
  [377] = {.lex_state = 144},
  [378] = {.lex_state = 142},
  [379] = {.lex_state = 142},
  [380] = {.lex_state = 142},
  [381] = {.lex_state = 142},
  [382] = {.lex_state = 142},
  [383] = {.lex_state = 142},
  [384] = {.lex_state = 142},
  [385] = {.lex_state = 142},
  [386] = {.lex_state = 142},
  [387] = {.lex_state = 142},
  [388] = {.lex_state = 142},
  [389] = {.lex_state = 142},
  [390] = {.lex_state = 142},
  [391] = {.lex_state = 142},
  [392] = {.lex_state = 142},
  [393] = {.lex_state = 142},
  [394] = {.lex_state = 142},
  [395] = {.lex_state = 142},
  [396] = {.lex_state = 142},
  [397] = {.lex_state = 142},
  [398] = {.lex_state = 142},
  [399] = {.lex_state = 142},
  [400] = {.lex_state = 142},
  [401] = {.lex_state = 142},
  [402] = {.lex_state = 142},
  [403] = {.lex_state = 142},
  [404] = {.lex_state = 142},
  [405] = {.lex_state = 142},
  [406] = {.lex_state = 142},
  [407] = {.lex_state = 142},
  [408] = {.lex_state = 142},
  [409] = {.lex_state = 142},
  [410] = {.lex_state = 142},
  [411] = {.lex_state = 142},
  [412] = {.lex_state = 142},
  [413] = {.lex_state = 143},
  [414] = {.lex_state = 143},
  [415] = {.lex_state = 142},
  [416] = {.lex_state = 143},
  [417] = {.lex_state = 142},
  [418] = {.lex_state = 142},
  [419] = {.lex_state = 142},
  [420] = {.lex_state = 142},
  [421] = {.lex_state = 142},
  [422] = {.lex_state = 142},
  [423] = {.lex_state = 142},
  [424] = {.lex_state = 142},
  [425] = {.lex_state = 143},
  [426] = {.lex_state = 142},
  [427] = {.lex_state = 142},
  [428] = {.lex_state = 143},
  [429] = {.lex_state = 142},
  [430] = {.lex_state = 143},
  [431] = {.lex_state = 142},
  [432] = {.lex_state = 142},
  [433] = {.lex_state = 142},
  [434] = {.lex_state = 142},
  [435] = {.lex_state = 142},
  [436] = {.lex_state = 143},
  [437] = {.lex_state = 142},
  [438] = {.lex_state = 142},
  [439] = {.lex_state = 143},
  [440] = {.lex_state = 142},
  [441] = {.lex_state = 142},
  [442] = {.lex_state = 142},
  [443] = {.lex_state = 142},
  [444] = {.lex_state = 144},
  [445] = {.lex_state = 142},
  [446] = {.lex_state = 142},
  [447] = {.lex_state = 142},
  [448] = {.lex_state = 143},
  [449] = {.lex_state = 143},
  [450] = {.lex_state = 143},
  [451] = {.lex_state = 143},
  [452] = {.lex_state = 142},
  [453] = {.lex_state = 144},
  [454] = {.lex_state = 142},
  [455] = {.lex_state = 144},
  [456] = {.lex_state = 142},
  [457] = {.lex_state = 142},
  [458] = {.lex_state = 142},
  [459] = {.lex_state = 142},
  [460] = {.lex_state = 142},
  [461] = {.lex_state = 142},
  [462] = {.lex_state = 142},
  [463] = {.lex_state = 142},
  [464] = {.lex_state = 142},
  [465] = {.lex_state = 142},
  [466] = {.lex_state = 143},
  [467] = {.lex_state = 142},
  [468] = {.lex_state = 142},
  [469] = {.lex_state = 142},
  [470] = {.lex_state = 142},
  [471] = {.lex_state = 142},
  [472] = {.lex_state = 142},
  [473] = {.lex_state = 142},
  [474] = {.lex_state = 142},
  [475] = {.lex_state = 143},
  [476] = {.lex_state = 142},
  [477] = {.lex_state = 143},
  [478] = {.lex_state = 142},
  [479] = {.lex_state = 142},
  [480] = {.lex_state = 142},
  [481] = {.lex_state = 142},
  [482] = {.lex_state = 142},
  [483] = {.lex_state = 142},
  [484] = {.lex_state = 142},
  [485] = {.lex_state = 142},
  [486] = {.lex_state = 142},
  [487] = {.lex_state = 142},
  [488] = {.lex_state = 142},
  [489] = {.lex_state = 142},
  [490] = {.lex_state = 144},
  [491] = {.lex_state = 142},
  [492] = {.lex_state = 142},
  [493] = {.lex_state = 142},
  [494] = {.lex_state = 144},
  [495] = {.lex_state = 142},
  [496] = {.lex_state = 142},
  [497] = {.lex_state = 143},
  [498] = {.lex_state = 143},
  [499] = {.lex_state = 142},
  [500] = {.lex_state = 144},
  [501] = {.lex_state = 142},
  [502] = {.lex_state = 144},
  [503] = {.lex_state = 142},
  [504] = {.lex_state = 142},
  [505] = {.lex_state = 142},
  [506] = {.lex_state = 143},
  [507] = {.lex_state = 143},
  [508] = {.lex_state = 143},
  [509] = {.lex_state = 143},
  [510] = {.lex_state = 143},
  [511] = {.lex_state = 143},
  [512] = {.lex_state = 143},
  [513] = {.lex_state = 144},
  [514] = {.lex_state = 143},
  [515] = {.lex_state = 142},
  [516] = {.lex_state = 143},
  [517] = {.lex_state = 143},
  [518] = {.lex_state = 143},
  [519] = {.lex_state = 143},
  [520] = {.lex_state = 143},
  [521] = {.lex_state = 143},
  [522] = {.lex_state = 143},
  [523] = {.lex_state = 143},
  [524] = {.lex_state = 143},
  [525] = {.lex_state = 143},
  [526] = {.lex_state = 142},
  [527] = {.lex_state = 143},
  [528] = {.lex_state = 143},
  [529] = {.lex_state = 143},
  [530] = {.lex_state = 142},
  [531] = {.lex_state = 142},
  [532] = {.lex_state = 143},
  [533] = {.lex_state = 143},
  [534] = {.lex_state = 142},
  [535] = {.lex_state = 143},
  [536] = {.lex_state = 143},
  [537] = {.lex_state = 142},
  [538] = {.lex_state = 142},
  [539] = {.lex_state = 142},
  [540] = {.lex_state = 142},
  [541] = {.lex_state = 143},
  [542] = {.lex_state = 143},
  [543] = {.lex_state = 142},
  [544] = {.lex_state = 143},
  [545] = {.lex_state = 143},
  [546] = {.lex_state = 143},
  [547] = {.lex_state = 142},
  [548] = {.lex_state = 143},
  [549] = {.lex_state = 142},
  [550] = {.lex_state = 143},
  [551] = {.lex_state = 143},
  [552] = {.lex_state = 143},
  [553] = {.lex_state = 142},
  [554] = {.lex_state = 143},
  [555] = {.lex_state = 143},
  [556] = {.lex_state = 143},
  [557] = {.lex_state = 143},
  [558] = {.lex_state = 143},
  [559] = {.lex_state = 143},
  [560] = {.lex_state = 143},
  [561] = {.lex_state = 143},
  [562] = {.lex_state = 143},
  [563] = {.lex_state = 142},
  [564] = {.lex_state = 143},
  [565] = {.lex_state = 142},
  [566] = {.lex_state = 143},
  [567] = {.lex_state = 142},
  [568] = {.lex_state = 142},
  [569] = {.lex_state = 143},
  [570] = {.lex_state = 142},
  [571] = {.lex_state = 143},
  [572] = {.lex_state = 142},
  [573] = {.lex_state = 144},
  [574] = {.lex_state = 144},
  [575] = {.lex_state = 144},
  [576] = {.lex_state = 144},
  [577] = {.lex_state = 144},
  [578] = {.lex_state = 144},
  [579] = {.lex_state = 144},
  [580] = {.lex_state = 144},
  [581] = {.lex_state = 144},
  [582] = {.lex_state = 144},
  [583] = {.lex_state = 144},
  [584] = {.lex_state = 144},
  [585] = {.lex_state = 144},
  [586] = {.lex_state = 144},
  [587] = {.lex_state = 144},
  [588] = {.lex_state = 144},
  [589] = {.lex_state = 144},
  [590] = {.lex_state = 147},
  [591] = {.lex_state = 147},
  [592] = {.lex_state = 147},
  [593] = {.lex_state = 147},
  [594] = {.lex_state = 147},
  [595] = {.lex_state = 147},
  [596] = {.lex_state = 149},
  [597] = {.lex_state = 150},
  [598] = {.lex_state = 150},
  [599] = {.lex_state = 150},
  [600] = {.lex_state = 150},
  [601] = {.lex_state = 150},
  [602] = {.lex_state = 150},
  [603] = {.lex_state = 150},
  [604] = {.lex_state = 164},
  [605] = {.lex_state = 150},
  [606] = {.lex_state = 150},
  [607] = {.lex_state = 150},
  [608] = {.lex_state = 150},
  [609] = {.lex_state = 150},
  [610] = {.lex_state = 150},
  [611] = {.lex_state = 150},
  [612] = {.lex_state = 150},
  [613] = {.lex_state = 150},
  [614] = {.lex_state = 150},
  [615] = {.lex_state = 150},
  [616] = {.lex_state = 150},
  [617] = {.lex_state = 151},
  [618] = {.lex_state = 151},
  [619] = {.lex_state = 150},
  [620] = {.lex_state = 150},
  [621] = {.lex_state = 150},
  [622] = {.lex_state = 150},
  [623] = {.lex_state = 150},
  [624] = {.lex_state = 150},
  [625] = {.lex_state = 150},
  [626] = {.lex_state = 150},
  [627] = {.lex_state = 150},
  [628] = {.lex_state = 150},
  [629] = {.lex_state = 150},
  [630] = {.lex_state = 150},
  [631] = {.lex_state = 150},
  [632] = {.lex_state = 164},
  [633] = {.lex_state = 150},
  [634] = {.lex_state = 164},
  [635] = {.lex_state = 150},
  [636] = {.lex_state = 150},
  [637] = {.lex_state = 164},
  [638] = {.lex_state = 164},
  [639] = {.lex_state = 150},
  [640] = {.lex_state = 150},
  [641] = {.lex_state = 150},
  [642] = {.lex_state = 164},
  [643] = {.lex_state = 164},
  [644] = {.lex_state = 150},
  [645] = {.lex_state = 150},
  [646] = {.lex_state = 150},
  [647] = {.lex_state = 150},
  [648] = {.lex_state = 150},
  [649] = {.lex_state = 164},
  [650] = {.lex_state = 150},
  [651] = {.lex_state = 164},
  [652] = {.lex_state = 150},
  [653] = {.lex_state = 150},
  [654] = {.lex_state = 150},
  [655] = {.lex_state = 150},
  [656] = {.lex_state = 150},
  [657] = {.lex_state = 164},
  [658] = {.lex_state = 164},
  [659] = {.lex_state = 150},
  [660] = {.lex_state = 150},
  [661] = {.lex_state = 150},
  [662] = {.lex_state = 150},
  [663] = {.lex_state = 150},
  [664] = {.lex_state = 150},
  [665] = {.lex_state = 150},
  [666] = {.lex_state = 150},
  [667] = {.lex_state = 150},
  [668] = {.lex_state = 150},
  [669] = {.lex_state = 150},
  [670] = {.lex_state = 150},
  [671] = {.lex_state = 150},
  [672] = {.lex_state = 150},
  [673] = {.lex_state = 150},
  [674] = {.lex_state = 164},
  [675] = {.lex_state = 164},
  [676] = {.lex_state = 164},
  [677] = {.lex_state = 164},
  [678] = {.lex_state = 164},
  [679] = {.lex_state = 164},
  [680] = {.lex_state = 164},
  [681] = {.lex_state = 164},
  [682] = {.lex_state = 174},
  [683] = {.lex_state = 174},
  [684] = {.lex_state = 174},
  [685] = {.lex_state = 174},
  [686] = {.lex_state = 149},
  [687] = {.lex_state = 172},
  [688] = {.lex_state = 171},
  [689] = {.lex_state = 149},
  [690] = {.lex_state = 149},
  [691] = {.lex_state = 172},
  [692] = {.lex_state = 174},
  [693] = {.lex_state = 149},
  [694] = {.lex_state = 149},
  [695] = {.lex_state = 149},
  [696] = {.lex_state = 149},
  [697] = {.lex_state = 171},
  [698] = {.lex_state = 172},
  [699] = {.lex_state = 149},
  [700] = {.lex_state = 149},
  [701] = {.lex_state = 149},
  [702] = {.lex_state = 149},
  [703] = {.lex_state = 149},
  [704] = {.lex_state = 149},
  [705] = {.lex_state = 149},
  [706] = {.lex_state = 171},
  [707] = {.lex_state = 174},
  [708] = {.lex_state = 172},
  [709] = {.lex_state = 171},
  [710] = {.lex_state = 150},
  [711] = {.lex_state = 150},
  [712] = {.lex_state = 172},
  [713] = {.lex_state = 171},
  [714] = {.lex_state = 149},
  [715] = {.lex_state = 174},
  [716] = {.lex_state = 173},
  [717] = {.lex_state = 149},
  [718] = {.lex_state = 173},
  [719] = {.lex_state = 150},
  [720] = {.lex_state = 150},
  [721] = {.lex_state = 149},
  [722] = {.lex_state = 174},
  [723] = {.lex_state = 174},
  [724] = {.lex_state = 149},
  [725] = {.lex_state = 174},
  [726] = {.lex_state = 149},
  [727] = {.lex_state = 149},
  [728] = {.lex_state = 173},
  [729] = {.lex_state = 150},
  [730] = {.lex_state = 149},
  [731] = {.lex_state = 173},
  [732] = {.lex_state = 172},
  [733] = {.lex_state = 150},
  [734] = {.lex_state = 172},
  [735] = {.lex_state = 173},
  [736] = {.lex_state = 174},
  [737] = {.lex_state = 149},
  [738] = {.lex_state = 173},
  [739] = {.lex_state = 149},
  [740] = {.lex_state = 149},
  [741] = {.lex_state = 150},
  [742] = {.lex_state = 172},
  [743] = {.lex_state = 149},
  [744] = {.lex_state = 149},
  [745] = {.lex_state = 172},
  [746] = {.lex_state = 149},
  [747] = {.lex_state = 172},
  [748] = {.lex_state = 171},
  [749] = {.lex_state = 149},
  [750] = {.lex_state = 171},
  [751] = {.lex_state = 149},
  [752] = {.lex_state = 171},
  [753] = {.lex_state = 171},
  [754] = {.lex_state = 171},
  [755] = {.lex_state = 171},
  [756] = {.lex_state = 171},
  [757] = {.lex_state = 171},
  [758] = {.lex_state = 171},
  [759] = {.lex_state = 149},
  [760] = {.lex_state = 171},
  [761] = {.lex_state = 171},
  [762] = {.lex_state = 150},
  [763] = {.lex_state = 149},
  [764] = {.lex_state = 171},
  [765] = {.lex_state = 171},
  [766] = {.lex_state = 171},
  [767] = {.lex_state = 171},
  [768] = {.lex_state = 171},
  [769] = {.lex_state = 150},
  [770] = {.lex_state = 171},
  [771] = {.lex_state = 171},
  [772] = {.lex_state = 171},
  [773] = {.lex_state = 149},
  [774] = {.lex_state = 149},
  [775] = {.lex_state = 171},
  [776] = {.lex_state = 149},
  [777] = {.lex_state = 171},
  [778] = {.lex_state = 149},
  [779] = {.lex_state = 171},
  [780] = {.lex_state = 171},
  [781] = {.lex_state = 171},
  [782] = {.lex_state = 149},
  [783] = {.lex_state = 171},
  [784] = {.lex_state = 147},
  [785] = {.lex_state = 171},
  [786] = {.lex_state = 171},
  [787] = {.lex_state = 171},
  [788] = {.lex_state = 171},
  [789] = {.lex_state = 171},
  [790] = {.lex_state = 171},
  [791] = {.lex_state = 171},
  [792] = {.lex_state = 171},
  [793] = {.lex_state = 150},
  [794] = {.lex_state = 171},
  [795] = {.lex_state = 149},
  [796] = {.lex_state = 171},
  [797] = {.lex_state = 171},
  [798] = {.lex_state = 171},
  [799] = {.lex_state = 147},
  [800] = {.lex_state = 171},
  [801] = {.lex_state = 171},
  [802] = {.lex_state = 171},
  [803] = {.lex_state = 171},
  [804] = {.lex_state = 171},
  [805] = {.lex_state = 171},
  [806] = {.lex_state = 171},
  [807] = {.lex_state = 171},
  [808] = {.lex_state = 171},
  [809] = {.lex_state = 171},
  [810] = {.lex_state = 147},
  [811] = {.lex_state = 171},
  [812] = {.lex_state = 171},
  [813] = {.lex_state = 147},
  [814] = {.lex_state = 171},
  [815] = {.lex_state = 171},
  [816] = {.lex_state = 171},
  [817] = {.lex_state = 150},
  [818] = {.lex_state = 149},
  [819] = {.lex_state = 149},
  [820] = {.lex_state = 171},
  [821] = {.lex_state = 171},
  [822] = {.lex_state = 147},
  [823] = {.lex_state = 150},
  [824] = {.lex_state = 149},
  [825] = {.lex_state = 171},
  [826] = {.lex_state = 149},
  [827] = {.lex_state = 149},
  [828] = {.lex_state = 171},
  [829] = {.lex_state = 149},
  [830] = {.lex_state = 171},
  [831] = {.lex_state = 149},
  [832] = {.lex_state = 149},
  [833] = {.lex_state = 171},
  [834] = {.lex_state = 171},
  [835] = {.lex_state = 149},
  [836] = {.lex_state = 147},
  [837] = {.lex_state = 171},
  [838] = {.lex_state = 171},
  [839] = {.lex_state = 171},
  [840] = {.lex_state = 171},
  [841] = {.lex_state = 171},
  [842] = {.lex_state = 171},
  [843] = {.lex_state = 171},
  [844] = {.lex_state = 147},
  [845] = {.lex_state = 152},
  [846] = {.lex_state = 152},
  [847] = {.lex_state = 152},
  [848] = {.lex_state = 152},
  [849] = {.lex_state = 149},
  [850] = {.lex_state = 152},
  [851] = {.lex_state = 152},
  [852] = {.lex_state = 153},
  [853] = {.lex_state = 149},
  [854] = {.lex_state = 152},
  [855] = {.lex_state = 152},
  [856] = {.lex_state = 152},
  [857] = {.lex_state = 152},
  [858] = {.lex_state = 152},
  [859] = {.lex_state = 152},
  [860] = {.lex_state = 152},
  [861] = {.lex_state = 149},
  [862] = {.lex_state = 149},
  [863] = {.lex_state = 149},
  [864] = {.lex_state = 149},
  [865] = {.lex_state = 152},
  [866] = {.lex_state = 152},
  [867] = {.lex_state = 147},
  [868] = {.lex_state = 147},
  [869] = {.lex_state = 149},
  [870] = {.lex_state = 149},
  [871] = {.lex_state = 149},
  [872] = {.lex_state = 149},
  [873] = {.lex_state = 149},
  [874] = {.lex_state = 149},
  [875] = {.lex_state = 147},
  [876] = {.lex_state = 150},
  [877] = {.lex_state = 147},
  [878] = {.lex_state = 147},
  [879] = {.lex_state = 150},
  [880] = {.lex_state = 154},
  [881] = {.lex_state = 154},
  [882] = {.lex_state = 154},
  [883] = {.lex_state = 150},
  [884] = {.lex_state = 150},
  [885] = {.lex_state = 155},
  [886] = {.lex_state = 154},
  [887] = {.lex_state = 155},
  [888] = {.lex_state = 155},
  [889] = {.lex_state = 150},
  [890] = {.lex_state = 150},
  [891] = {.lex_state = 155},
  [892] = {.lex_state = 150},
  [893] = {.lex_state = 150},
  [894] = {.lex_state = 156},
  [895] = {.lex_state = 156},
  [896] = {.lex_state = 150},
  [897] = {.lex_state = 150},
  [898] = {.lex_state = 149},
  [899] = {.lex_state = 149},
  [900] = {.lex_state = 150},
  [901] = {.lex_state = 150},
  [902] = {.lex_state = 150},
  [903] = {.lex_state = 150},
  [904] = {.lex_state = 150},
  [905] = {.lex_state = 150},
  [906] = {.lex_state = 149},
  [907] = {.lex_state = 150},
  [908] = {.lex_state = 150},
  [909] = {.lex_state = 164},
  [910] = {.lex_state = 149},
  [911] = {.lex_state = 149},
  [912] = {.lex_state = 150},
  [913] = {.lex_state = 150},
  [914] = {.lex_state = 150},
  [915] = {.lex_state = 149},
  [916] = {.lex_state = 150},
  [917] = {.lex_state = 150},
  [918] = {.lex_state = 149},
  [919] = {.lex_state = 149},
  [920] = {.lex_state = 149},
  [921] = {.lex_state = 150},
  [922] = {.lex_state = 150},
  [923] = {.lex_state = 149},
  [924] = {.lex_state = 152},
  [925] = {.lex_state = 152},
  [926] = {.lex_state = 152},
  [927] = {.lex_state = 164},
  [928] = {.lex_state = 164},
  [929] = {.lex_state = 149},
  [930] = {.lex_state = 164},
  [931] = {.lex_state = 149},
  [932] = {.lex_state = 164},
  [933] = {.lex_state = 149},
  [934] = {.lex_state = 164},
  [935] = {.lex_state = 152},
  [936] = {.lex_state = 164},
  [937] = {.lex_state = 164},
  [938] = {.lex_state = 164},
  [939] = {.lex_state = 149},
  [940] = {.lex_state = 152},
  [941] = {.lex_state = 164},
  [942] = {.lex_state = 149},
  [943] = {.lex_state = 149},
  [944] = {.lex_state = 164},
  [945] = {.lex_state = 152},
  [946] = {.lex_state = 164},
  [947] = {.lex_state = 164},
  [948] = {.lex_state = 164},
  [949] = {.lex_state = 164},
  [950] = {.lex_state = 150},
  [951] = {.lex_state = 152},
  [952] = {.lex_state = 164},
  [953] = {.lex_state = 164},
  [954] = {.lex_state = 149},
  [955] = {.lex_state = 152},
  [956] = {.lex_state = 164},
  [957] = {.lex_state = 164},
  [958] = {.lex_state = 152},
  [959] = {.lex_state = 164},
  [960] = {.lex_state = 152},
  [961] = {.lex_state = 150},
  [962] = {.lex_state = 152},
  [963] = {.lex_state = 149},
  [964] = {.lex_state = 152},
  [965] = {.lex_state = 152},
  [966] = {.lex_state = 152},
  [967] = {.lex_state = 164},
  [968] = {.lex_state = 152},
  [969] = {.lex_state = 150},
  [970] = {.lex_state = 164},
  [971] = {.lex_state = 152},
  [972] = {.lex_state = 164},
  [973] = {.lex_state = 152},
  [974] = {.lex_state = 164},
  [975] = {.lex_state = 149},
  [976] = {.lex_state = 149},
  [977] = {.lex_state = 164},
  [978] = {.lex_state = 152},
  [979] = {.lex_state = 149},
  [980] = {.lex_state = 149},
  [981] = {.lex_state = 149},
  [982] = {.lex_state = 149},
  [983] = {.lex_state = 149},
  [984] = {.lex_state = 149},
  [985] = {.lex_state = 164},
  [986] = {.lex_state = 150},
  [987] = {.lex_state = 152},
  [988] = {.lex_state = 152},
  [989] = {.lex_state = 152},
  [990] = {.lex_state = 164},
  [991] = {.lex_state = 164},
  [992] = {.lex_state = 152},
  [993] = {.lex_state = 152},
  [994] = {.lex_state = 164},
  [995] = {.lex_state = 149},
  [996] = {.lex_state = 152},
  [997] = {.lex_state = 150},
  [998] = {.lex_state = 164},
  [999] = {.lex_state = 152},
  [1000] = {.lex_state = 149},
  [1001] = {.lex_state = 152},
  [1002] = {.lex_state = 152},
  [1003] = {.lex_state = 149},
  [1004] = {.lex_state = 149},
  [1005] = {.lex_state = 149},
  [1006] = {.lex_state = 164},
  [1007] = {.lex_state = 149},
  [1008] = {.lex_state = 149},
  [1009] = {.lex_state = 164},
  [1010] = {.lex_state = 164},
  [1011] = {.lex_state = 164},
  [1012] = {.lex_state = 149},
  [1013] = {.lex_state = 149},
  [1014] = {.lex_state = 164},
  [1015] = {.lex_state = 149},
  [1016] = {.lex_state = 152},
  [1017] = {.lex_state = 149},
  [1018] = {.lex_state = 149},
  [1019] = {.lex_state = 149},
  [1020] = {.lex_state = 149},
  [1021] = {.lex_state = 149},
  [1022] = {.lex_state = 164},
  [1023] = {.lex_state = 152},
  [1024] = {.lex_state = 152},
  [1025] = {.lex_state = 149},
  [1026] = {.lex_state = 149},
  [1027] = {.lex_state = 149},
  [1028] = {.lex_state = 149},
  [1029] = {.lex_state = 149},
  [1030] = {.lex_state = 164},
  [1031] = {.lex_state = 152},
  [1032] = {.lex_state = 149},
  [1033] = {.lex_state = 152},
  [1034] = {.lex_state = 149},
  [1035] = {.lex_state = 149},
  [1036] = {.lex_state = 149},
  [1037] = {.lex_state = 149},
  [1038] = {.lex_state = 152},
  [1039] = {.lex_state = 152},
  [1040] = {.lex_state = 152},
  [1041] = {.lex_state = 152},
  [1042] = {.lex_state = 164},
  [1043] = {.lex_state = 152},
  [1044] = {.lex_state = 152},
  [1045] = {.lex_state = 164},
  [1046] = {.lex_state = 149},
  [1047] = {.lex_state = 164},
  [1048] = {.lex_state = 164},
  [1049] = {.lex_state = 164},
  [1050] = {.lex_state = 164},
  [1051] = {.lex_state = 152},
  [1052] = {.lex_state = 164},
  [1053] = {.lex_state = 164},
  [1054] = {.lex_state = 152},
  [1055] = {.lex_state = 149},
  [1056] = {.lex_state = 152},
  [1057] = {.lex_state = 164},
  [1058] = {.lex_state = 149},
  [1059] = {.lex_state = 149},
  [1060] = {.lex_state = 149},
  [1061] = {.lex_state = 149},
  [1062] = {.lex_state = 152},
  [1063] = {.lex_state = 149},
  [1064] = {.lex_state = 164},
  [1065] = {.lex_state = 164},
  [1066] = {.lex_state = 152},
  [1067] = {.lex_state = 164},
  [1068] = {.lex_state = 164},
  [1069] = {.lex_state = 164},
  [1070] = {.lex_state = 164},
  [1071] = {.lex_state = 164},
  [1072] = {.lex_state = 149},
  [1073] = {.lex_state = 149},
  [1074] = {.lex_state = 164},
  [1075] = {.lex_state = 164},
  [1076] = {.lex_state = 164},
  [1077] = {.lex_state = 164},
  [1078] = {.lex_state = 164},
  [1079] = {.lex_state = 152},
  [1080] = {.lex_state = 149},
  [1081] = {.lex_state = 149},
  [1082] = {.lex_state = 164},
  [1083] = {.lex_state = 164},
  [1084] = {.lex_state = 164},
  [1085] = {.lex_state = 164},
  [1086] = {.lex_state = 152},
  [1087] = {.lex_state = 149},
  [1088] = {.lex_state = 143},
  [1089] = {.lex_state = 164},
  [1090] = {.lex_state = 152},
  [1091] = {.lex_state = 144},
  [1092] = {.lex_state = 152},
  [1093] = {.lex_state = 152},
  [1094] = {.lex_state = 152},
  [1095] = {.lex_state = 152},
  [1096] = {.lex_state = 152},
  [1097] = {.lex_state = 152},
  [1098] = {.lex_state = 149},
  [1099] = {.lex_state = 152},
  [1100] = {.lex_state = 152},
  [1101] = {.lex_state = 152},
  [1102] = {.lex_state = 164},
  [1103] = {.lex_state = 144},
  [1104] = {.lex_state = 152},
  [1105] = {.lex_state = 164},
  [1106] = {.lex_state = 149},
  [1107] = {.lex_state = 164},
  [1108] = {.lex_state = 164},
  [1109] = {.lex_state = 164},
  [1110] = {.lex_state = 164},
  [1111] = {.lex_state = 157},
  [1112] = {.lex_state = 158},
  [1113] = {.lex_state = 164},
  [1114] = {.lex_state = 164},
  [1115] = {.lex_state = 164},
  [1116] = {.lex_state = 149},
  [1117] = {.lex_state = 149},
  [1118] = {.lex_state = 149},
  [1119] = {.lex_state = 149},
  [1120] = {.lex_state = 149},
  [1121] = {.lex_state = 164},
  [1122] = {.lex_state = 149},
  [1123] = {.lex_state = 149},
  [1124] = {.lex_state = 149},
  [1125] = {.lex_state = 149},
  [1126] = {.lex_state = 149},
  [1127] = {.lex_state = 149},
  [1128] = {.lex_state = 149},
  [1129] = {.lex_state = 149},
  [1130] = {.lex_state = 149},
  [1131] = {.lex_state = 149},
  [1132] = {.lex_state = 149},
  [1133] = {.lex_state = 149},
  [1134] = {.lex_state = 149},
  [1135] = {.lex_state = 149},
  [1136] = {.lex_state = 164},
  [1137] = {.lex_state = 164},
  [1138] = {.lex_state = 149},
  [1139] = {.lex_state = 149},
  [1140] = {.lex_state = 149},
  [1141] = {.lex_state = 149},
  [1142] = {.lex_state = 149},
  [1143] = {.lex_state = 149},
  [1144] = {.lex_state = 149},
  [1145] = {.lex_state = 149},
  [1146] = {.lex_state = 149},
  [1147] = {.lex_state = 149},
  [1148] = {.lex_state = 147},
  [1149] = {.lex_state = 147},
  [1150] = {.lex_state = 147},
  [1151] = {.lex_state = 159},
  [1152] = {.lex_state = 159},
  [1153] = {.lex_state = 159},
  [1154] = {.lex_state = 159},
  [1155] = {.lex_state = 159},
  [1156] = {.lex_state = 159},
  [1157] = {.lex_state = 159},
  [1158] = {.lex_state = 159},
  [1159] = {.lex_state = 159},
  [1160] = {.lex_state = 159},
  [1161] = {.lex_state = 157},
  [1162] = {.lex_state = 157},
  [1163] = {.lex_state = 157},
  [1164] = {.lex_state = 159},
  [1165] = {.lex_state = 159},
  [1166] = {.lex_state = 159},
  [1167] = {.lex_state = 157},
  [1168] = {.lex_state = 159},
  [1169] = {.lex_state = 159},
  [1170] = {.lex_state = 159},
  [1171] = {.lex_state = 159},
  [1172] = {.lex_state = 159},
  [1173] = {.lex_state = 159},
  [1174] = {.lex_state = 159},
  [1175] = {.lex_state = 159},
  [1176] = {.lex_state = 159},
  [1177] = {.lex_state = 159},
  [1178] = {.lex_state = 159},
  [1179] = {.lex_state = 159},
  [1180] = {.lex_state = 158},
  [1181] = {.lex_state = 159},
  [1182] = {.lex_state = 159},
  [1183] = {.lex_state = 159},
  [1184] = {.lex_state = 159},
  [1185] = {.lex_state = 157},
  [1186] = {.lex_state = 159},
  [1187] = {.lex_state = 159},
  [1188] = {.lex_state = 159},
  [1189] = {.lex_state = 157},
  [1190] = {.lex_state = 159},
  [1191] = {.lex_state = 159},
  [1192] = {.lex_state = 159},
  [1193] = {.lex_state = 159},
  [1194] = {.lex_state = 159},
  [1195] = {.lex_state = 159},
  [1196] = {.lex_state = 157},
  [1197] = {.lex_state = 159},
  [1198] = {.lex_state = 157},
  [1199] = {.lex_state = 157},
  [1200] = {.lex_state = 159},
  [1201] = {.lex_state = 159},
  [1202] = {.lex_state = 159},
  [1203] = {.lex_state = 159},
  [1204] = {.lex_state = 159},
  [1205] = {.lex_state = 157},
  [1206] = {.lex_state = 159},
  [1207] = {.lex_state = 159},
  [1208] = {.lex_state = 159},
  [1209] = {.lex_state = 159},
  [1210] = {.lex_state = 159},
  [1211] = {.lex_state = 159},
  [1212] = {.lex_state = 159},
  [1213] = {.lex_state = 159},
  [1214] = {.lex_state = 159},
  [1215] = {.lex_state = 157},
  [1216] = {.lex_state = 159},
  [1217] = {.lex_state = 159},
  [1218] = {.lex_state = 157},
  [1219] = {.lex_state = 159},
  [1220] = {.lex_state = 159},
  [1221] = {.lex_state = 157},
  [1222] = {.lex_state = 159},
  [1223] = {.lex_state = 159},
  [1224] = {.lex_state = 159},
  [1225] = {.lex_state = 157},
  [1226] = {.lex_state = 157},
  [1227] = {.lex_state = 159},
  [1228] = {.lex_state = 159},
  [1229] = {.lex_state = 159},
  [1230] = {.lex_state = 157},
  [1231] = {.lex_state = 159},
  [1232] = {.lex_state = 158},
  [1233] = {.lex_state = 158},
  [1234] = {.lex_state = 158},
  [1235] = {.lex_state = 158},
  [1236] = {.lex_state = 158},
  [1237] = {.lex_state = 158},
  [1238] = {.lex_state = 158},
  [1239] = {.lex_state = 158},
  [1240] = {.lex_state = 158},
  [1241] = {.lex_state = 158},
  [1242] = {.lex_state = 158},
  [1243] = {.lex_state = 158},
  [1244] = {.lex_state = 158},
  [1245] = {.lex_state = 159},
  [1246] = {.lex_state = 158},
  [1247] = {.lex_state = 157},
  [1248] = {.lex_state = 158},
  [1249] = {.lex_state = 158},
  [1250] = {.lex_state = 147},
  [1251] = {.lex_state = 158},
  [1252] = {.lex_state = 159},
  [1253] = {.lex_state = 158},
  [1254] = {.lex_state = 158},
  [1255] = {.lex_state = 158},
  [1256] = {.lex_state = 158},
  [1257] = {.lex_state = 158},
  [1258] = {.lex_state = 147},
  [1259] = {.lex_state = 158},
  [1260] = {.lex_state = 158},
  [1261] = {.lex_state = 147},
  [1262] = {.lex_state = 158},
  [1263] = {.lex_state = 159},
  [1264] = {.lex_state = 158},
  [1265] = {.lex_state = 158},
  [1266] = {.lex_state = 158},
  [1267] = {.lex_state = 147},
  [1268] = {.lex_state = 157},
  [1269] = {.lex_state = 157},
  [1270] = {.lex_state = 164},
  [1271] = {.lex_state = 158},
  [1272] = {.lex_state = 160},
  [1273] = {.lex_state = 157},
  [1274] = {.lex_state = 157},
  [1275] = {.lex_state = 160},
  [1276] = {.lex_state = 147},
  [1277] = {.lex_state = 157},
  [1278] = {.lex_state = 160},
  [1279] = {.lex_state = 160},
  [1280] = {.lex_state = 164},
  [1281] = {.lex_state = 147},
  [1282] = {.lex_state = 157},
  [1283] = {.lex_state = 164},
  [1284] = {.lex_state = 164},
  [1285] = {.lex_state = 164},
  [1286] = {.lex_state = 157},
  [1287] = {.lex_state = 160},
  [1288] = {.lex_state = 164},
  [1289] = {.lex_state = 160},
  [1290] = {.lex_state = 157},
  [1291] = {.lex_state = 147},
  [1292] = {.lex_state = 147},
  [1293] = {.lex_state = 147},
  [1294] = {.lex_state = 147},
  [1295] = {.lex_state = 147},
  [1296] = {.lex_state = 161},
  [1297] = {.lex_state = 147},
  [1298] = {.lex_state = 160},
  [1299] = {.lex_state = 147},
  [1300] = {.lex_state = 161},
  [1301] = {.lex_state = 147},
  [1302] = {.lex_state = 147},
  [1303] = {.lex_state = 147},
  [1304] = {.lex_state = 158},
  [1305] = {.lex_state = 147},
  [1306] = {.lex_state = 158},
  [1307] = {.lex_state = 147},
  [1308] = {.lex_state = 158},
  [1309] = {.lex_state = 157},
  [1310] = {.lex_state = 147},
  [1311] = {.lex_state = 147},
  [1312] = {.lex_state = 147},
  [1313] = {.lex_state = 147},
  [1314] = {.lex_state = 147},
  [1315] = {.lex_state = 147},
  [1316] = {.lex_state = 157},
  [1317] = {.lex_state = 147},
  [1318] = {.lex_state = 147},
  [1319] = {.lex_state = 147},
  [1320] = {.lex_state = 147},
  [1321] = {.lex_state = 158},
  [1322] = {.lex_state = 147},
  [1323] = {.lex_state = 147},
  [1324] = {.lex_state = 147},
  [1325] = {.lex_state = 147},
  [1326] = {.lex_state = 147},
  [1327] = {.lex_state = 147},
  [1328] = {.lex_state = 147},
  [1329] = {.lex_state = 161},
  [1330] = {.lex_state = 147},
  [1331] = {.lex_state = 147},
  [1332] = {.lex_state = 147},
  [1333] = {.lex_state = 147},
  [1334] = {.lex_state = 147},
  [1335] = {.lex_state = 147},
  [1336] = {.lex_state = 147},
  [1337] = {.lex_state = 147},
  [1338] = {.lex_state = 160},
  [1339] = {.lex_state = 160},
  [1340] = {.lex_state = 158},
  [1341] = {.lex_state = 160},
  [1342] = {.lex_state = 158},
  [1343] = {.lex_state = 158},
  [1344] = {.lex_state = 158},
  [1345] = {.lex_state = 162},
  [1346] = {.lex_state = 160},
  [1347] = {.lex_state = 162},
  [1348] = {.lex_state = 162},
  [1349] = {.lex_state = 162},
  [1350] = {.lex_state = 162},
  [1351] = {.lex_state = 158},
  [1352] = {.lex_state = 162},
  [1353] = {.lex_state = 147, .external_lex_state = 1},
  [1354] = {.lex_state = 147},
  [1355] = {.lex_state = 161},
  [1356] = {.lex_state = 147, .external_lex_state = 1},
  [1357] = {.lex_state = 147, .external_lex_state = 1},
  [1358] = {.lex_state = 147},
  [1359] = {.lex_state = 161},
  [1360] = {.lex_state = 147, .external_lex_state = 1},
  [1361] = {.lex_state = 147, .external_lex_state = 1},
  [1362] = {.lex_state = 147, .external_lex_state = 1},
  [1363] = {.lex_state = 147},
  [1364] = {.lex_state = 161},
  [1365] = {.lex_state = 147, .external_lex_state = 1},
  [1366] = {.lex_state = 147},
  [1367] = {.lex_state = 147},
  [1368] = {.lex_state = 161},
  [1369] = {.lex_state = 147, .external_lex_state = 1},
  [1370] = {.lex_state = 161},
  [1371] = {.lex_state = 161},
  [1372] = {.lex_state = 161},
  [1373] = {.lex_state = 147, .external_lex_state = 1},
  [1374] = {.lex_state = 147, .external_lex_state = 1},
  [1375] = {.lex_state = 147},
  [1376] = {.lex_state = 161},
  [1377] = {.lex_state = 147, .external_lex_state = 1},
  [1378] = {.lex_state = 147, .external_lex_state = 1},
  [1379] = {.lex_state = 147, .external_lex_state = 1},
  [1380] = {.lex_state = 0},
  [1381] = {.lex_state = 147},
  [1382] = {.lex_state = 147, .external_lex_state = 1},
  [1383] = {.lex_state = 161},
  [1384] = {.lex_state = 161},
  [1385] = {.lex_state = 147, .external_lex_state = 1},
  [1386] = {.lex_state = 158},
  [1387] = {.lex_state = 147, .external_lex_state = 1},
  [1388] = {.lex_state = 147, .external_lex_state = 1},
  [1389] = {.lex_state = 147},
  [1390] = {.lex_state = 161},
  [1391] = {.lex_state = 147},
  [1392] = {.lex_state = 147},
  [1393] = {.lex_state = 161},
  [1394] = {.lex_state = 162},
  [1395] = {.lex_state = 160},
  [1396] = {.lex_state = 0},
  [1397] = {.lex_state = 0},
  [1398] = {.lex_state = 0},
  [1399] = {.lex_state = 160},
  [1400] = {.lex_state = 0},
  [1401] = {.lex_state = 147},
  [1402] = {.lex_state = 0},
  [1403] = {.lex_state = 0},
  [1404] = {.lex_state = 0},
  [1405] = {.lex_state = 0},
  [1406] = {.lex_state = 158},
  [1407] = {.lex_state = 0},
  [1408] = {.lex_state = 158},
  [1409] = {.lex_state = 0},
  [1410] = {.lex_state = 158},
  [1411] = {.lex_state = 147},
  [1412] = {.lex_state = 0},
  [1413] = {.lex_state = 0},
  [1414] = {.lex_state = 0},
  [1415] = {.lex_state = 157},
  [1416] = {.lex_state = 147},
  [1417] = {.lex_state = 147},
  [1418] = {.lex_state = 0},
  [1419] = {.lex_state = 172},
  [1420] = {.lex_state = 0},
  [1421] = {.lex_state = 160},
  [1422] = {.lex_state = 0},
  [1423] = {.lex_state = 147},
  [1424] = {.lex_state = 147},
  [1425] = {.lex_state = 147},
  [1426] = {.lex_state = 147},
  [1427] = {.lex_state = 161},
  [1428] = {.lex_state = 161},
  [1429] = {.lex_state = 147},
  [1430] = {.lex_state = 0},
  [1431] = {.lex_state = 0},
  [1432] = {.lex_state = 147},
  [1433] = {.lex_state = 158},
  [1434] = {.lex_state = 147},
  [1435] = {.lex_state = 161},
  [1436] = {.lex_state = 147},
  [1437] = {.lex_state = 0},
  [1438] = {.lex_state = 0},
  [1439] = {.lex_state = 161},
  [1440] = {.lex_state = 161},
  [1441] = {.lex_state = 160},
  [1442] = {.lex_state = 0},
  [1443] = {.lex_state = 147},
  [1444] = {.lex_state = 0},
  [1445] = {.lex_state = 161},
  [1446] = {.lex_state = 160},
  [1447] = {.lex_state = 0},
  [1448] = {.lex_state = 0},
  [1449] = {.lex_state = 0},
  [1450] = {.lex_state = 161},
  [1451] = {.lex_state = 147},
  [1452] = {.lex_state = 161},
  [1453] = {.lex_state = 160},
  [1454] = {.lex_state = 172},
  [1455] = {.lex_state = 0},
  [1456] = {.lex_state = 158},
  [1457] = {.lex_state = 161},
  [1458] = {.lex_state = 161},
  [1459] = {.lex_state = 0},
  [1460] = {.lex_state = 158},
  [1461] = {.lex_state = 147},
  [1462] = {.lex_state = 147},
  [1463] = {.lex_state = 147, .external_lex_state = 1},
  [1464] = {.lex_state = 0},
  [1465] = {.lex_state = 147},
  [1466] = {.lex_state = 161},
  [1467] = {.lex_state = 147},
  [1468] = {.lex_state = 0},
  [1469] = {.lex_state = 147},
  [1470] = {.lex_state = 0},
  [1471] = {.lex_state = 161},
  [1472] = {.lex_state = 160},
  [1473] = {.lex_state = 161},
  [1474] = {.lex_state = 147},
  [1475] = {.lex_state = 147},
  [1476] = {.lex_state = 0},
  [1477] = {.lex_state = 172},
  [1478] = {.lex_state = 0},
  [1479] = {.lex_state = 171},
  [1480] = {.lex_state = 172},
  [1481] = {.lex_state = 0},
  [1482] = {.lex_state = 0},
  [1483] = {.lex_state = 0},
  [1484] = {.lex_state = 0},
  [1485] = {.lex_state = 0},
  [1486] = {.lex_state = 0},
  [1487] = {.lex_state = 0},
  [1488] = {.lex_state = 0},
  [1489] = {.lex_state = 0},
  [1490] = {.lex_state = 0},
  [1491] = {.lex_state = 0},
  [1492] = {.lex_state = 0},
  [1493] = {.lex_state = 172},
  [1494] = {.lex_state = 172},
  [1495] = {.lex_state = 0},
  [1496] = {.lex_state = 0},
  [1497] = {.lex_state = 0},
  [1498] = {.lex_state = 0},
  [1499] = {.lex_state = 147},
  [1500] = {.lex_state = 0},
  [1501] = {.lex_state = 0},
  [1502] = {.lex_state = 147},
  [1503] = {.lex_state = 161},
  [1504] = {.lex_state = 0},
  [1505] = {.lex_state = 0},
  [1506] = {.lex_state = 0},
  [1507] = {.lex_state = 172},
  [1508] = {.lex_state = 0},
  [1509] = {.lex_state = 0},
  [1510] = {.lex_state = 172},
  [1511] = {.lex_state = 0},
  [1512] = {.lex_state = 0},
  [1513] = {.lex_state = 0},
  [1514] = {.lex_state = 0},
  [1515] = {.lex_state = 0},
  [1516] = {.lex_state = 0},
  [1517] = {.lex_state = 0},
  [1518] = {.lex_state = 0},
  [1519] = {.lex_state = 161},
  [1520] = {.lex_state = 0},
  [1521] = {.lex_state = 0},
  [1522] = {.lex_state = 0},
  [1523] = {.lex_state = 0},
  [1524] = {.lex_state = 164},
  [1525] = {.lex_state = 0},
  [1526] = {.lex_state = 172},
  [1527] = {.lex_state = 172},
  [1528] = {.lex_state = 0},
  [1529] = {.lex_state = 0},
  [1530] = {.lex_state = 0},
  [1531] = {.lex_state = 0},
  [1532] = {.lex_state = 0},
  [1533] = {.lex_state = 0},
  [1534] = {.lex_state = 0},
  [1535] = {.lex_state = 147},
  [1536] = {.lex_state = 171},
  [1537] = {.lex_state = 161},
  [1538] = {.lex_state = 147},
  [1539] = {.lex_state = 0},
  [1540] = {.lex_state = 0},
  [1541] = {.lex_state = 0},
  [1542] = {.lex_state = 147},
  [1543] = {.lex_state = 161},
  [1544] = {.lex_state = 0},
  [1545] = {.lex_state = 172},
  [1546] = {.lex_state = 171},
  [1547] = {.lex_state = 0},
  [1548] = {.lex_state = 172},
  [1549] = {.lex_state = 0},
  [1550] = {.lex_state = 0},
  [1551] = {.lex_state = 147},
  [1552] = {.lex_state = 172},
  [1553] = {.lex_state = 172},
  [1554] = {.lex_state = 158},
  [1555] = {.lex_state = 171},
  [1556] = {.lex_state = 0},
  [1557] = {.lex_state = 0},
  [1558] = {.lex_state = 0},
  [1559] = {.lex_state = 0},
  [1560] = {.lex_state = 172},
  [1561] = {.lex_state = 0},
  [1562] = {.lex_state = 147},
  [1563] = {.lex_state = 0},
  [1564] = {.lex_state = 172},
  [1565] = {.lex_state = 172},
  [1566] = {.lex_state = 0},
  [1567] = {.lex_state = 172},
  [1568] = {.lex_state = 0},
  [1569] = {.lex_state = 0},
  [1570] = {.lex_state = 172},
  [1571] = {.lex_state = 0},
  [1572] = {.lex_state = 161},
  [1573] = {.lex_state = 0},
  [1574] = {.lex_state = 0},
  [1575] = {.lex_state = 0},
  [1576] = {.lex_state = 147},
  [1577] = {.lex_state = 0},
  [1578] = {.lex_state = 0},
  [1579] = {.lex_state = 0},
  [1580] = {.lex_state = 0},
  [1581] = {.lex_state = 0},
  [1582] = {.lex_state = 0},
  [1583] = {.lex_state = 0},
  [1584] = {.lex_state = 0},
  [1585] = {.lex_state = 0},
  [1586] = {.lex_state = 161},
  [1587] = {.lex_state = 0},
  [1588] = {.lex_state = 172},
  [1589] = {.lex_state = 0},
  [1590] = {.lex_state = 0},
  [1591] = {.lex_state = 147},
  [1592] = {.lex_state = 0},
  [1593] = {.lex_state = 172},
  [1594] = {.lex_state = 0},
  [1595] = {.lex_state = 0},
  [1596] = {.lex_state = 0},
  [1597] = {.lex_state = 0},
  [1598] = {.lex_state = 0},
  [1599] = {.lex_state = 0},
  [1600] = {.lex_state = 0},
  [1601] = {.lex_state = 0},
  [1602] = {.lex_state = 0},
  [1603] = {.lex_state = 0},
  [1604] = {.lex_state = 0},
  [1605] = {.lex_state = 0},
  [1606] = {.lex_state = 0},
  [1607] = {.lex_state = 0},
  [1608] = {.lex_state = 0},
  [1609] = {.lex_state = 0},
  [1610] = {.lex_state = 0},
  [1611] = {.lex_state = 0},
  [1612] = {.lex_state = 0},
  [1613] = {.lex_state = 0},
  [1614] = {.lex_state = 0},
  [1615] = {.lex_state = 0},
  [1616] = {.lex_state = 0},
  [1617] = {.lex_state = 0},
  [1618] = {.lex_state = 0},
  [1619] = {.lex_state = 0},
  [1620] = {.lex_state = 172},
  [1621] = {.lex_state = 172},
  [1622] = {.lex_state = 0},
  [1623] = {.lex_state = 0},
  [1624] = {.lex_state = 0},
  [1625] = {.lex_state = 0},
  [1626] = {.lex_state = 0},
  [1627] = {.lex_state = 172},
  [1628] = {.lex_state = 0},
  [1629] = {.lex_state = 0},
  [1630] = {.lex_state = 0},
  [1631] = {.lex_state = 0},
  [1632] = {.lex_state = 0},
  [1633] = {.lex_state = 147},
  [1634] = {.lex_state = 172},
  [1635] = {.lex_state = 0},
  [1636] = {.lex_state = 0},
  [1637] = {.lex_state = 0},
  [1638] = {.lex_state = 0},
  [1639] = {.lex_state = 172},
  [1640] = {.lex_state = 0},
  [1641] = {.lex_state = 0},
  [1642] = {.lex_state = 161},
  [1643] = {.lex_state = 0},
  [1644] = {.lex_state = 0},
  [1645] = {.lex_state = 0},
  [1646] = {.lex_state = 0},
  [1647] = {.lex_state = 0},
  [1648] = {.lex_state = 0},
  [1649] = {.lex_state = 0},
  [1650] = {.lex_state = 0},
  [1651] = {.lex_state = 0},
  [1652] = {.lex_state = 172},
  [1653] = {.lex_state = 0},
  [1654] = {.lex_state = 0},
  [1655] = {.lex_state = 0},
  [1656] = {.lex_state = 172},
  [1657] = {.lex_state = 0},
  [1658] = {.lex_state = 0},
  [1659] = {.lex_state = 0},
  [1660] = {.lex_state = 0},
  [1661] = {.lex_state = 147},
  [1662] = {.lex_state = 147},
  [1663] = {.lex_state = 172},
  [1664] = {.lex_state = 0},
  [1665] = {.lex_state = 0},
  [1666] = {.lex_state = 0},
  [1667] = {.lex_state = 161},
  [1668] = {.lex_state = 161},
  [1669] = {.lex_state = 0},
  [1670] = {.lex_state = 0},
  [1671] = {.lex_state = 161},
  [1672] = {.lex_state = 161},
  [1673] = {.lex_state = 0},
  [1674] = {.lex_state = 0},
  [1675] = {.lex_state = 0},
  [1676] = {.lex_state = 161},
  [1677] = {.lex_state = 0},
  [1678] = {.lex_state = 171},
  [1679] = {.lex_state = 0},
  [1680] = {.lex_state = 161},
  [1681] = {.lex_state = 0},
  [1682] = {.lex_state = 172},
  [1683] = {.lex_state = 0},
  [1684] = {.lex_state = 161},
  [1685] = {.lex_state = 0},
  [1686] = {.lex_state = 0},
  [1687] = {.lex_state = 161},
  [1688] = {.lex_state = 0},
  [1689] = {.lex_state = 0},
  [1690] = {.lex_state = 172},
  [1691] = {.lex_state = 161},
  [1692] = {.lex_state = 172},
  [1693] = {.lex_state = 0},
  [1694] = {.lex_state = 0},
  [1695] = {.lex_state = 0},
  [1696] = {.lex_state = 0},
  [1697] = {.lex_state = 0},
  [1698] = {.lex_state = 0},
  [1699] = {.lex_state = 161},
  [1700] = {.lex_state = 0},
  [1701] = {.lex_state = 0},
  [1702] = {.lex_state = 0},
  [1703] = {.lex_state = 0},
  [1704] = {.lex_state = 161},
  [1705] = {.lex_state = 0},
  [1706] = {.lex_state = 147},
  [1707] = {.lex_state = 172},
  [1708] = {.lex_state = 0},
  [1709] = {.lex_state = 0},
  [1710] = {.lex_state = 0},
  [1711] = {.lex_state = 0},
  [1712] = {.lex_state = 0},
  [1713] = {.lex_state = 0},
  [1714] = {.lex_state = 0},
  [1715] = {.lex_state = 161},
  [1716] = {.lex_state = 0},
  [1717] = {.lex_state = 0},
  [1718] = {.lex_state = 0},
  [1719] = {.lex_state = 0},
  [1720] = {.lex_state = 0},
  [1721] = {.lex_state = 161},
  [1722] = {.lex_state = 0},
  [1723] = {.lex_state = 0},
  [1724] = {.lex_state = 0},
  [1725] = {.lex_state = 172},
  [1726] = {.lex_state = 0},
  [1727] = {.lex_state = 0},
  [1728] = {.lex_state = 0},
  [1729] = {.lex_state = 0},
  [1730] = {.lex_state = 0},
  [1731] = {.lex_state = 0},
  [1732] = {.lex_state = 0},
  [1733] = {.lex_state = 0},
  [1734] = {.lex_state = 0},
  [1735] = {.lex_state = 0},
  [1736] = {.lex_state = 0},
  [1737] = {.lex_state = 0},
  [1738] = {.lex_state = 0},
  [1739] = {.lex_state = 0},
  [1740] = {.lex_state = 0},
  [1741] = {.lex_state = 0},
  [1742] = {.lex_state = 171},
  [1743] = {.lex_state = 171},
  [1744] = {.lex_state = 0},
  [1745] = {.lex_state = 161},
  [1746] = {.lex_state = 0},
  [1747] = {.lex_state = 0},
  [1748] = {.lex_state = 0},
  [1749] = {.lex_state = 0},
  [1750] = {.lex_state = 0},
  [1751] = {.lex_state = 0},
  [1752] = {.lex_state = 0},
  [1753] = {.lex_state = 0},
  [1754] = {.lex_state = 0},
  [1755] = {.lex_state = 0},
  [1756] = {.lex_state = 0},
  [1757] = {.lex_state = 0},
  [1758] = {.lex_state = 0},
  [1759] = {.lex_state = 0},
  [1760] = {.lex_state = 171},
  [1761] = {.lex_state = 171},
  [1762] = {.lex_state = 0},
  [1763] = {.lex_state = 161},
  [1764] = {.lex_state = 161},
  [1765] = {.lex_state = 161},
  [1766] = {.lex_state = 0},
  [1767] = {.lex_state = 161},
  [1768] = {.lex_state = 0},
  [1769] = {.lex_state = 171},
  [1770] = {.lex_state = 0},
  [1771] = {.lex_state = 171},
  [1772] = {.lex_state = 171},
  [1773] = {.lex_state = 161},
  [1774] = {.lex_state = 171},
  [1775] = {.lex_state = 161},
  [1776] = {.lex_state = 0},
  [1777] = {.lex_state = 171},
  [1778] = {.lex_state = 0},
  [1779] = {.lex_state = 171},
  [1780] = {.lex_state = 0},
  [1781] = {.lex_state = 171},
  [1782] = {.lex_state = 0},
  [1783] = {.lex_state = 0},
  [1784] = {.lex_state = 0},
  [1785] = {.lex_state = 0},
  [1786] = {.lex_state = 171},
  [1787] = {.lex_state = 171},
  [1788] = {.lex_state = 0},
  [1789] = {.lex_state = 161},
  [1790] = {.lex_state = 147},
  [1791] = {.lex_state = 0},
  [1792] = {.lex_state = 147},
  [1793] = {.lex_state = 172},
  [1794] = {.lex_state = 0},
  [1795] = {.lex_state = 0},
  [1796] = {.lex_state = 161},
  [1797] = {.lex_state = 171},
  [1798] = {.lex_state = 171},
  [1799] = {.lex_state = 161},
  [1800] = {.lex_state = 0},
  [1801] = {.lex_state = 161},
  [1802] = {.lex_state = 161},
  [1803] = {.lex_state = 0},
  [1804] = {.lex_state = 161},
  [1805] = {.lex_state = 0},
  [1806] = {.lex_state = 0},
  [1807] = {.lex_state = 0},
  [1808] = {.lex_state = 0},
  [1809] = {.lex_state = 0},
  [1810] = {.lex_state = 0},
  [1811] = {.lex_state = 0},
  [1812] = {.lex_state = 147},
  [1813] = {.lex_state = 161},
  [1814] = {.lex_state = 0},
  [1815] = {.lex_state = 161},
  [1816] = {.lex_state = 172},
  [1817] = {.lex_state = 0},
  [1818] = {.lex_state = 0},
  [1819] = {.lex_state = 0},
  [1820] = {.lex_state = 172},
  [1821] = {.lex_state = 147},
  [1822] = {.lex_state = 0},
  [1823] = {.lex_state = 0},
  [1824] = {.lex_state = 0},
  [1825] = {.lex_state = 172},
  [1826] = {.lex_state = 0},
  [1827] = {.lex_state = 172},
  [1828] = {.lex_state = 0},
  [1829] = {.lex_state = 0},
  [1830] = {.lex_state = 0},
  [1831] = {.lex_state = 0},
  [1832] = {.lex_state = 0},
  [1833] = {.lex_state = 0},
  [1834] = {.lex_state = 0},
  [1835] = {.lex_state = 172},
  [1836] = {.lex_state = 0},
  [1837] = {.lex_state = 0},
  [1838] = {.lex_state = 0},
  [1839] = {.lex_state = 172},
  [1840] = {.lex_state = 0},
  [1841] = {.lex_state = 0},
  [1842] = {.lex_state = 0},
  [1843] = {.lex_state = 0},
  [1844] = {.lex_state = 0},
  [1845] = {.lex_state = 0},
  [1846] = {.lex_state = 161},
  [1847] = {.lex_state = 0},
  [1848] = {.lex_state = 0},
  [1849] = {.lex_state = 0},
  [1850] = {.lex_state = 161},
  [1851] = {.lex_state = 0},
  [1852] = {.lex_state = 0},
  [1853] = {.lex_state = 0},
  [1854] = {.lex_state = 0},
  [1855] = {.lex_state = 0},
  [1856] = {.lex_state = 0},
  [1857] = {.lex_state = 0},
  [1858] = {.lex_state = 0},
  [1859] = {.lex_state = 0},
  [1860] = {.lex_state = 0},
  [1861] = {.lex_state = 172},
  [1862] = {.lex_state = 0},
  [1863] = {.lex_state = 0},
  [1864] = {.lex_state = 0},
  [1865] = {.lex_state = 0},
  [1866] = {.lex_state = 0},
  [1867] = {.lex_state = 0},
  [1868] = {.lex_state = 0},
  [1869] = {.lex_state = 172},
  [1870] = {.lex_state = 172},
  [1871] = {.lex_state = 0},
  [1872] = {.lex_state = 0},
  [1873] = {.lex_state = 0},
  [1874] = {.lex_state = 0},
  [1875] = {.lex_state = 172},
  [1876] = {.lex_state = 0},
  [1877] = {.lex_state = 172},
  [1878] = {.lex_state = 0},
  [1879] = {.lex_state = 0},
  [1880] = {.lex_state = 0},
  [1881] = {.lex_state = 0},
  [1882] = {.lex_state = 0},
  [1883] = {.lex_state = 0},
  [1884] = {.lex_state = 0},
  [1885] = {.lex_state = 0},
  [1886] = {.lex_state = 0},
  [1887] = {.lex_state = 172},
  [1888] = {.lex_state = 0},
  [1889] = {.lex_state = 172},
  [1890] = {.lex_state = 0},
  [1891] = {.lex_state = 0},
  [1892] = {.lex_state = 0},
  [1893] = {.lex_state = 0},
  [1894] = {.lex_state = 0},
  [1895] = {.lex_state = 0},
  [1896] = {.lex_state = 0},
  [1897] = {.lex_state = 0},
  [1898] = {.lex_state = 0},
  [1899] = {.lex_state = 0},
  [1900] = {.lex_state = 0},
  [1901] = {.lex_state = 0},
  [1902] = {.lex_state = 0},
  [1903] = {.lex_state = 147},
  [1904] = {.lex_state = 0},
  [1905] = {.lex_state = 147},
  [1906] = {.lex_state = 0},
  [1907] = {.lex_state = 0},
  [1908] = {.lex_state = 0},
  [1909] = {.lex_state = 0},
  [1910] = {.lex_state = 0},
  [1911] = {.lex_state = 0},
  [1912] = {.lex_state = 0},
  [1913] = {.lex_state = 0},
  [1914] = {.lex_state = 172},
  [1915] = {.lex_state = 0},
  [1916] = {.lex_state = 0},
  [1917] = {.lex_state = 0},
  [1918] = {.lex_state = 0},
  [1919] = {.lex_state = 147},
  [1920] = {.lex_state = 0},
  [1921] = {.lex_state = 0},
  [1922] = {.lex_state = 0},
  [1923] = {.lex_state = 0},
  [1924] = {.lex_state = 0},
  [1925] = {.lex_state = 0},
  [1926] = {.lex_state = 0},
  [1927] = {.lex_state = 0},
  [1928] = {.lex_state = 0},
  [1929] = {.lex_state = 0},
  [1930] = {.lex_state = 0},
  [1931] = {.lex_state = 0},
  [1932] = {.lex_state = 0},
  [1933] = {.lex_state = 0},
  [1934] = {.lex_state = 0},
  [1935] = {.lex_state = 0},
  [1936] = {.lex_state = 0},
  [1937] = {.lex_state = 0},
  [1938] = {.lex_state = 0},
  [1939] = {.lex_state = 172},
  [1940] = {.lex_state = 0},
  [1941] = {.lex_state = 0},
  [1942] = {.lex_state = 172},
  [1943] = {.lex_state = 0},
  [1944] = {.lex_state = 0},
  [1945] = {.lex_state = 172},
  [1946] = {.lex_state = 0},
  [1947] = {.lex_state = 0},
  [1948] = {.lex_state = 172},
  [1949] = {.lex_state = 0},
  [1950] = {.lex_state = 0},
  [1951] = {.lex_state = 0},
  [1952] = {.lex_state = 0},
  [1953] = {.lex_state = 0},
  [1954] = {.lex_state = 0},
  [1955] = {.lex_state = 0},
  [1956] = {.lex_state = 0},
  [1957] = {.lex_state = 0},
  [1958] = {.lex_state = 0},
  [1959] = {.lex_state = 0},
  [1960] = {.lex_state = 0},
  [1961] = {.lex_state = 0},
  [1962] = {.lex_state = 0},
  [1963] = {.lex_state = 172},
  [1964] = {.lex_state = 0},
  [1965] = {.lex_state = 0},
  [1966] = {.lex_state = 0},
};

enum {
  ts_external_token_quoted_content = 0,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_quoted_content] = sym_quoted_content,
};

static const bool ts_external_scanner_states[2][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_quoted_content] = true,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [0] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_erlang] = ACTIONS(1),
    [anon_sym_javascript] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_size] = ACTIONS(1),
    [anon_sym_fn] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_external] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_LT_DOT] = ACTIONS(1),
    [anon_sym_LT_EQ_DOT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_GT_DOT] = ACTIONS(1),
    [anon_sym_GT_EQ_DOT] = ACTIONS(1),
    [anon_sym_PIPE_GT] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_PLUS_DOT] = ACTIONS(1),
    [anon_sym_DASH_DOT] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_STAR_DOT] = ACTIONS(1),
    [anon_sym_SLASH_DOT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_LT_GT] = ACTIONS(1),
    [anon_sym_todo] = ACTIONS(1),
    [anon_sym_panic] = ACTIONS(1),
    [anon_sym_echo] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_use] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [sym_visibility_modifier] = ACTIONS(1),
    [sym_opacity_modifier] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE2] = ACTIONS(1),
    [aux_sym_escape_sequence_token1] = ACTIONS(1),
    [aux_sym_escape_sequence_token2] = ACTIONS(1),
    [sym_float] = ACTIONS(1),
    [sym__hex] = ACTIONS(1),
    [sym__decimal] = ACTIONS(1),
    [sym__octal] = ACTIONS(1),
    [sym__binary] = ACTIONS(1),
    [anon_sym_binary] = ACTIONS(1),
    [anon_sym_bytes] = ACTIONS(1),
    [anon_sym_int] = ACTIONS(1),
    [anon_sym_float] = ACTIONS(1),
    [anon_sym_bit_string] = ACTIONS(1),
    [anon_sym_bits] = ACTIONS(1),
    [anon_sym_utf8] = ACTIONS(1),
    [anon_sym_utf16] = ACTIONS(1),
    [anon_sym_utf32] = ACTIONS(1),
    [anon_sym_utf8_codepoint] = ACTIONS(1),
    [anon_sym_utf16_codepoint] = ACTIONS(1),
    [anon_sym_utf32_codepoint] = ACTIONS(1),
    [anon_sym_signed] = ACTIONS(1),
    [anon_sym_unsigned] = ACTIONS(1),
    [anon_sym_big] = ACTIONS(1),
    [anon_sym_little] = ACTIONS(1),
    [anon_sym_native] = ACTIONS(1),
    [anon_sym_unit] = ACTIONS(1),
    [sym__discard_name] = ACTIONS(1),
    [sym__name] = ACTIONS(1),
    [sym__upname] = ACTIONS(1),
    [sym_quoted_content] = ACTIONS(1),
  },
  [1] = {
    [sym_source_file] = STATE(1915),
    [sym__module_statement] = STATE(6),
    [sym_target_group] = STATE(6),
    [sym_attribute] = STATE(6),
    [sym_import] = STATE(6),
    [sym_constant] = STATE(6),
    [sym_external_type] = STATE(6),
    [sym_external_function] = STATE(6),
    [sym_function] = STATE(6),
    [aux_sym__statement_seq] = STATE(13),
    [sym__statement] = STATE(761),
    [sym__expression] = STATE(230),
    [sym_binary_expression] = STATE(230),
    [sym__expression_unit] = STATE(230),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(230),
    [sym_panic] = STATE(230),
    [sym_echo] = STATE(230),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(230),
    [sym__expression_bit_string] = STATE(230),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_let_assert] = STATE(761),
    [sym_let] = STATE(761),
    [sym_use] = STATE(761),
    [sym_assert] = STATE(230),
    [sym_boolean_negation] = STATE(230),
    [sym_integer_negation] = STATE(230),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_type_definition] = STATE(6),
    [sym_type_alias] = STATE(6),
    [sym_string] = STATE(230),
    [sym_integer] = STATE(230),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [aux_sym_source_file_repeat1] = STATE(6),
    [ts_builtin_sym_end] = ACTIONS(7),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(9),
    [anon_sym_LBRACE] = ACTIONS(11),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_import] = ACTIONS(15),
    [anon_sym_type] = ACTIONS(17),
    [anon_sym_const] = ACTIONS(19),
    [anon_sym_POUND] = ACTIONS(21),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_LT_LT] = ACTIONS(25),
    [anon_sym_DASH] = ACTIONS(27),
    [anon_sym_fn] = ACTIONS(29),
    [anon_sym_external] = ACTIONS(31),
    [anon_sym_todo] = ACTIONS(33),
    [anon_sym_panic] = ACTIONS(35),
    [anon_sym_echo] = ACTIONS(37),
    [anon_sym_case] = ACTIONS(39),
    [anon_sym_let] = ACTIONS(41),
    [anon_sym_assert] = ACTIONS(43),
    [anon_sym_use] = ACTIONS(45),
    [anon_sym_BANG] = ACTIONS(47),
    [sym_visibility_modifier] = ACTIONS(49),
    [sym_opacity_modifier] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_float] = ACTIONS(55),
    [sym__hex] = ACTIONS(57),
    [sym__decimal] = ACTIONS(59),
    [sym__octal] = ACTIONS(57),
    [sym__binary] = ACTIONS(57),
    [sym__name] = ACTIONS(61),
    [sym__upname] = ACTIONS(63),
  },
  [2] = {
    [sym__expression] = STATE(163),
    [sym_binary_expression] = STATE(163),
    [sym__expression_unit] = STATE(163),
    [sym_record] = STATE(92),
    [sym_todo] = STATE(163),
    [sym_panic] = STATE(163),
    [sym_echo] = STATE(163),
    [sym_tuple] = STATE(85),
    [sym_list] = STATE(163),
    [sym__expression_bit_string] = STATE(163),
    [sym_anonymous_function] = STATE(80),
    [sym_block] = STATE(66),
    [sym_case] = STATE(66),
    [sym_assert] = STATE(163),
    [sym_boolean_negation] = STATE(163),
    [sym_integer_negation] = STATE(163),
    [sym_record_update] = STATE(92),
    [sym__maybe_tuple_expression] = STATE(1957),
    [sym_tuple_access] = STATE(66),
    [sym__maybe_record_expression] = STATE(1856),
    [sym_field_access] = STATE(66),
    [sym__maybe_function_expression] = STATE(1719),
    [sym_function_call] = STATE(66),
    [sym_string] = STATE(163),
    [sym_integer] = STATE(163),
    [sym_identifier] = STATE(70),
    [sym_constructor_name] = STATE(14),
    [sym_remote_constructor_name] = STATE(14),
    [ts_builtin_sym_end] = ACTIONS(65),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(67),
    [anon_sym_LBRACE] = ACTIONS(65),
    [anon_sym_AT] = ACTIONS(65),
    [anon_sym_import] = ACTIONS(67),
    [anon_sym_as] = ACTIONS(67),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_type] = ACTIONS(67),
    [anon_sym_const] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(65),
    [anon_sym_LBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(67),
    [anon_sym_external] = ACTIONS(67),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(67),
    [anon_sym_panic] = ACTIONS(67),
    [anon_sym_echo] = ACTIONS(67),
    [anon_sym_case] = ACTIONS(67),
    [anon_sym_let] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [sym_visibility_modifier] = ACTIONS(67),
    [sym_opacity_modifier] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(65),
    [sym_float] = ACTIONS(65),
    [sym__hex] = ACTIONS(65),
    [sym__decimal] = ACTIONS(67),
    [sym__octal] = ACTIONS(65),
    [sym__binary] = ACTIONS(65),
    [sym__name] = ACTIONS(67),
    [sym__upname] = ACTIONS(65),
  },
  [3] = {
    [sym__expression] = STATE(201),
    [sym_binary_expression] = STATE(201),
    [sym__expression_unit] = STATE(201),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(201),
    [sym_panic] = STATE(201),
    [sym_echo] = STATE(201),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(201),
    [sym__expression_bit_string] = STATE(201),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_assert] = STATE(201),
    [sym_boolean_negation] = STATE(201),
    [sym_integer_negation] = STATE(201),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_string] = STATE(201),
    [sym_integer] = STATE(201),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [ts_builtin_sym_end] = ACTIONS(65),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(67),
    [anon_sym_LBRACE] = ACTIONS(65),
    [anon_sym_AT] = ACTIONS(65),
    [anon_sym_import] = ACTIONS(67),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_type] = ACTIONS(67),
    [anon_sym_const] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(65),
    [anon_sym_LBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(67),
    [anon_sym_external] = ACTIONS(67),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(67),
    [anon_sym_panic] = ACTIONS(67),
    [anon_sym_echo] = ACTIONS(67),
    [anon_sym_case] = ACTIONS(67),
    [anon_sym_let] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [sym_visibility_modifier] = ACTIONS(67),
    [sym_opacity_modifier] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(65),
    [sym_float] = ACTIONS(65),
    [sym__hex] = ACTIONS(65),
    [sym__decimal] = ACTIONS(67),
    [sym__octal] = ACTIONS(65),
    [sym__binary] = ACTIONS(65),
    [sym__name] = ACTIONS(67),
    [sym__upname] = ACTIONS(65),
  },
  [4] = {
    [sym__expression] = STATE(861),
    [sym_binary_expression] = STATE(861),
    [sym__expression_unit] = STATE(861),
    [sym_record] = STATE(740),
    [sym_todo] = STATE(861),
    [sym_panic] = STATE(861),
    [sym_echo] = STATE(861),
    [sym_tuple] = STATE(743),
    [sym_list] = STATE(861),
    [sym__expression_bit_string] = STATE(861),
    [sym_anonymous_function] = STATE(744),
    [sym_block] = STATE(705),
    [sym_case] = STATE(705),
    [sym_assert] = STATE(861),
    [sym_boolean_negation] = STATE(861),
    [sym_integer_negation] = STATE(861),
    [sym_record_update] = STATE(740),
    [sym__maybe_tuple_expression] = STATE(1819),
    [sym_tuple_access] = STATE(705),
    [sym__maybe_record_expression] = STATE(1842),
    [sym_field_access] = STATE(705),
    [sym__maybe_function_expression] = STATE(1808),
    [sym_function_call] = STATE(705),
    [sym_string] = STATE(861),
    [sym_integer] = STATE(861),
    [sym_identifier] = STATE(696),
    [sym_constructor_name] = STATE(686),
    [sym_remote_constructor_name] = STATE(686),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(69),
    [anon_sym_RBRACE] = ACTIONS(65),
    [anon_sym_COMMA] = ACTIONS(65),
    [anon_sym_RPAREN] = ACTIONS(65),
    [anon_sym_COLON] = ACTIONS(65),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(71),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_RBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(75),
    [anon_sym_GT_GT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(77),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(79),
    [anon_sym_panic] = ACTIONS(81),
    [anon_sym_echo] = ACTIONS(83),
    [anon_sym_DOT_DOT] = ACTIONS(65),
    [anon_sym_case] = ACTIONS(85),
    [anon_sym_assert] = ACTIONS(87),
    [anon_sym_BANG] = ACTIONS(89),
    [anon_sym_DQUOTE] = ACTIONS(91),
    [sym_float] = ACTIONS(93),
    [sym__hex] = ACTIONS(95),
    [sym__decimal] = ACTIONS(97),
    [sym__octal] = ACTIONS(95),
    [sym__binary] = ACTIONS(95),
    [sym__name] = ACTIONS(99),
    [sym__upname] = ACTIONS(101),
  },
  [5] = {
    [sym__module_statement] = STATE(5),
    [sym_target_group] = STATE(5),
    [sym_attribute] = STATE(5),
    [sym_import] = STATE(5),
    [sym_constant] = STATE(5),
    [sym_external_type] = STATE(5),
    [sym_external_function] = STATE(5),
    [sym_function] = STATE(5),
    [aux_sym__statement_seq] = STATE(13),
    [sym__statement] = STATE(761),
    [sym__expression] = STATE(230),
    [sym_binary_expression] = STATE(230),
    [sym__expression_unit] = STATE(230),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(230),
    [sym_panic] = STATE(230),
    [sym_echo] = STATE(230),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(230),
    [sym__expression_bit_string] = STATE(230),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_let_assert] = STATE(761),
    [sym_let] = STATE(761),
    [sym_use] = STATE(761),
    [sym_assert] = STATE(230),
    [sym_boolean_negation] = STATE(230),
    [sym_integer_negation] = STATE(230),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_type_definition] = STATE(5),
    [sym_type_alias] = STATE(5),
    [sym_string] = STATE(230),
    [sym_integer] = STATE(230),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [aux_sym_source_file_repeat1] = STATE(5),
    [ts_builtin_sym_end] = ACTIONS(103),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(105),
    [anon_sym_LBRACE] = ACTIONS(108),
    [anon_sym_AT] = ACTIONS(111),
    [anon_sym_import] = ACTIONS(114),
    [anon_sym_type] = ACTIONS(117),
    [anon_sym_const] = ACTIONS(120),
    [anon_sym_POUND] = ACTIONS(123),
    [anon_sym_LBRACK] = ACTIONS(126),
    [anon_sym_LT_LT] = ACTIONS(129),
    [anon_sym_DASH] = ACTIONS(132),
    [anon_sym_fn] = ACTIONS(135),
    [anon_sym_external] = ACTIONS(138),
    [anon_sym_todo] = ACTIONS(141),
    [anon_sym_panic] = ACTIONS(144),
    [anon_sym_echo] = ACTIONS(147),
    [anon_sym_case] = ACTIONS(150),
    [anon_sym_let] = ACTIONS(153),
    [anon_sym_assert] = ACTIONS(156),
    [anon_sym_use] = ACTIONS(159),
    [anon_sym_BANG] = ACTIONS(162),
    [sym_visibility_modifier] = ACTIONS(165),
    [sym_opacity_modifier] = ACTIONS(168),
    [anon_sym_DQUOTE] = ACTIONS(171),
    [sym_float] = ACTIONS(174),
    [sym__hex] = ACTIONS(177),
    [sym__decimal] = ACTIONS(180),
    [sym__octal] = ACTIONS(177),
    [sym__binary] = ACTIONS(177),
    [sym__name] = ACTIONS(183),
    [sym__upname] = ACTIONS(186),
  },
  [6] = {
    [sym__module_statement] = STATE(5),
    [sym_target_group] = STATE(5),
    [sym_attribute] = STATE(5),
    [sym_import] = STATE(5),
    [sym_constant] = STATE(5),
    [sym_external_type] = STATE(5),
    [sym_external_function] = STATE(5),
    [sym_function] = STATE(5),
    [aux_sym__statement_seq] = STATE(13),
    [sym__statement] = STATE(761),
    [sym__expression] = STATE(230),
    [sym_binary_expression] = STATE(230),
    [sym__expression_unit] = STATE(230),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(230),
    [sym_panic] = STATE(230),
    [sym_echo] = STATE(230),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(230),
    [sym__expression_bit_string] = STATE(230),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_let_assert] = STATE(761),
    [sym_let] = STATE(761),
    [sym_use] = STATE(761),
    [sym_assert] = STATE(230),
    [sym_boolean_negation] = STATE(230),
    [sym_integer_negation] = STATE(230),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_type_definition] = STATE(5),
    [sym_type_alias] = STATE(5),
    [sym_string] = STATE(230),
    [sym_integer] = STATE(230),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [aux_sym_source_file_repeat1] = STATE(5),
    [ts_builtin_sym_end] = ACTIONS(189),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(9),
    [anon_sym_LBRACE] = ACTIONS(11),
    [anon_sym_AT] = ACTIONS(13),
    [anon_sym_import] = ACTIONS(15),
    [anon_sym_type] = ACTIONS(17),
    [anon_sym_const] = ACTIONS(19),
    [anon_sym_POUND] = ACTIONS(21),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_LT_LT] = ACTIONS(25),
    [anon_sym_DASH] = ACTIONS(27),
    [anon_sym_fn] = ACTIONS(29),
    [anon_sym_external] = ACTIONS(31),
    [anon_sym_todo] = ACTIONS(33),
    [anon_sym_panic] = ACTIONS(35),
    [anon_sym_echo] = ACTIONS(37),
    [anon_sym_case] = ACTIONS(39),
    [anon_sym_let] = ACTIONS(41),
    [anon_sym_assert] = ACTIONS(43),
    [anon_sym_use] = ACTIONS(45),
    [anon_sym_BANG] = ACTIONS(47),
    [sym_visibility_modifier] = ACTIONS(49),
    [sym_opacity_modifier] = ACTIONS(51),
    [anon_sym_DQUOTE] = ACTIONS(53),
    [sym_float] = ACTIONS(55),
    [sym__hex] = ACTIONS(57),
    [sym__decimal] = ACTIONS(59),
    [sym__octal] = ACTIONS(57),
    [sym__binary] = ACTIONS(57),
    [sym__name] = ACTIONS(61),
    [sym__upname] = ACTIONS(63),
  },
  [7] = {
    [sym__expression] = STATE(525),
    [sym_binary_expression] = STATE(525),
    [sym__expression_unit] = STATE(525),
    [sym_record] = STATE(507),
    [sym_todo] = STATE(525),
    [sym_panic] = STATE(525),
    [sym_echo] = STATE(525),
    [sym_tuple] = STATE(510),
    [sym_list] = STATE(525),
    [sym__expression_bit_string] = STATE(525),
    [sym_anonymous_function] = STATE(511),
    [sym_block] = STATE(498),
    [sym_case] = STATE(498),
    [sym_assert] = STATE(525),
    [sym_boolean_negation] = STATE(525),
    [sym_integer_negation] = STATE(525),
    [sym_record_update] = STATE(507),
    [sym__maybe_tuple_expression] = STATE(1936),
    [sym_tuple_access] = STATE(498),
    [sym__maybe_record_expression] = STATE(1868),
    [sym_field_access] = STATE(498),
    [sym__maybe_function_expression] = STATE(1686),
    [sym_function_call] = STATE(498),
    [sym_string] = STATE(525),
    [sym_integer] = STATE(525),
    [sym_identifier] = STATE(497),
    [sym_constructor_name] = STATE(315),
    [sym_remote_constructor_name] = STATE(315),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(65),
    [anon_sym_RBRACE] = ACTIONS(65),
    [anon_sym_as] = ACTIONS(67),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(65),
    [anon_sym_LBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(67),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(67),
    [anon_sym_panic] = ACTIONS(67),
    [anon_sym_echo] = ACTIONS(67),
    [anon_sym_case] = ACTIONS(67),
    [anon_sym_let] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(65),
    [sym_float] = ACTIONS(65),
    [sym__hex] = ACTIONS(65),
    [sym__decimal] = ACTIONS(67),
    [sym__octal] = ACTIONS(65),
    [sym__binary] = ACTIONS(65),
    [sym__name] = ACTIONS(67),
    [sym__upname] = ACTIONS(65),
  },
  [8] = {
    [sym__expression] = STATE(585),
    [sym_binary_expression] = STATE(585),
    [sym__expression_unit] = STATE(585),
    [sym_record] = STATE(272),
    [sym_todo] = STATE(585),
    [sym_panic] = STATE(585),
    [sym_echo] = STATE(585),
    [sym_tuple] = STATE(273),
    [sym_list] = STATE(585),
    [sym__expression_bit_string] = STATE(585),
    [sym_anonymous_function] = STATE(274),
    [sym_block] = STATE(266),
    [sym_case] = STATE(266),
    [sym_assert] = STATE(585),
    [sym_boolean_negation] = STATE(585),
    [sym_integer_negation] = STATE(585),
    [sym_record_update] = STATE(272),
    [sym__maybe_tuple_expression] = STATE(1830),
    [sym_tuple_access] = STATE(266),
    [sym__maybe_record_expression] = STATE(1817),
    [sym_field_access] = STATE(266),
    [sym__maybe_function_expression] = STATE(1805),
    [sym_function_call] = STATE(266),
    [sym_string] = STATE(585),
    [sym_integer] = STATE(585),
    [sym_identifier] = STATE(263),
    [sym_constructor_name] = STATE(240),
    [sym_remote_constructor_name] = STATE(240),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(65),
    [anon_sym_RBRACE] = ACTIONS(65),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(65),
    [anon_sym_LBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(67),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(67),
    [anon_sym_panic] = ACTIONS(67),
    [anon_sym_echo] = ACTIONS(67),
    [anon_sym_case] = ACTIONS(67),
    [anon_sym_let] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(67),
    [anon_sym_BANG] = ACTIONS(67),
    [anon_sym_DQUOTE] = ACTIONS(65),
    [sym_float] = ACTIONS(65),
    [sym__hex] = ACTIONS(65),
    [sym__decimal] = ACTIONS(67),
    [sym__octal] = ACTIONS(65),
    [sym__binary] = ACTIONS(65),
    [sym__name] = ACTIONS(67),
    [sym__upname] = ACTIONS(65),
  },
  [9] = {
    [sym__expression] = STATE(668),
    [sym_binary_expression] = STATE(668),
    [sym__expression_unit] = STATE(668),
    [sym_record] = STATE(620),
    [sym_todo] = STATE(668),
    [sym_panic] = STATE(668),
    [sym_echo] = STATE(668),
    [sym_tuple] = STATE(629),
    [sym_list] = STATE(668),
    [sym__expression_bit_string] = STATE(668),
    [sym_anonymous_function] = STATE(628),
    [sym_block] = STATE(598),
    [sym_case] = STATE(598),
    [sym_assert] = STATE(668),
    [sym_boolean_negation] = STATE(668),
    [sym_integer_negation] = STATE(668),
    [sym_record_update] = STATE(620),
    [sym__maybe_tuple_expression] = STATE(1886),
    [sym_tuple_access] = STATE(598),
    [sym__maybe_record_expression] = STATE(1880),
    [sym_field_access] = STATE(598),
    [sym__maybe_function_expression] = STATE(1713),
    [sym_function_call] = STATE(598),
    [sym_string] = STATE(668),
    [sym_integer] = STATE(668),
    [sym_identifier] = STATE(615),
    [sym_constructor_name] = STATE(597),
    [sym_remote_constructor_name] = STATE(597),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(191),
    [anon_sym_RBRACE] = ACTIONS(65),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(65),
    [anon_sym_LBRACK] = ACTIONS(65),
    [anon_sym_LT_LT] = ACTIONS(65),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(193),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(195),
    [anon_sym_panic] = ACTIONS(197),
    [anon_sym_echo] = ACTIONS(199),
    [anon_sym_case] = ACTIONS(201),
    [anon_sym_assert] = ACTIONS(203),
    [anon_sym_BANG] = ACTIONS(205),
    [anon_sym_DQUOTE] = ACTIONS(65),
    [sym_float] = ACTIONS(65),
    [sym__hex] = ACTIONS(65),
    [sym__decimal] = ACTIONS(67),
    [sym__octal] = ACTIONS(65),
    [sym__binary] = ACTIONS(65),
    [sym__discard_name] = ACTIONS(67),
    [sym__name] = ACTIONS(67),
    [sym__upname] = ACTIONS(65),
  },
  [10] = {
    [sym__expression] = STATE(1004),
    [sym_binary_expression] = STATE(1004),
    [sym__expression_unit] = STATE(1004),
    [sym_record] = STATE(740),
    [sym_todo] = STATE(1004),
    [sym_panic] = STATE(1004),
    [sym_echo] = STATE(1004),
    [sym_tuple] = STATE(743),
    [sym_list] = STATE(1004),
    [sym__expression_bit_string] = STATE(1004),
    [sym_anonymous_function] = STATE(744),
    [sym_block] = STATE(705),
    [sym_case] = STATE(705),
    [sym_assert] = STATE(1004),
    [sym_boolean_negation] = STATE(1004),
    [sym_integer_negation] = STATE(1004),
    [sym_record_update] = STATE(740),
    [sym__maybe_tuple_expression] = STATE(1819),
    [sym_tuple_access] = STATE(705),
    [sym__maybe_record_expression] = STATE(1842),
    [sym_field_access] = STATE(705),
    [sym__maybe_function_expression] = STATE(1808),
    [sym_function_call] = STATE(705),
    [sym_string] = STATE(1004),
    [sym_integer] = STATE(1004),
    [sym_identifier] = STATE(696),
    [sym_constructor_name] = STATE(686),
    [sym_remote_constructor_name] = STATE(686),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(65),
    [anon_sym_COMMA] = ACTIONS(65),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(71),
    [anon_sym_LBRACK] = ACTIONS(73),
    [anon_sym_LT_LT] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(77),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(207),
    [anon_sym_panic] = ACTIONS(209),
    [anon_sym_echo] = ACTIONS(211),
    [anon_sym_case] = ACTIONS(85),
    [anon_sym_assert] = ACTIONS(213),
    [anon_sym_BANG] = ACTIONS(215),
    [anon_sym_DQUOTE] = ACTIONS(91),
    [sym_float] = ACTIONS(217),
    [sym__hex] = ACTIONS(95),
    [sym__decimal] = ACTIONS(97),
    [sym__octal] = ACTIONS(95),
    [sym__binary] = ACTIONS(95),
    [sym__name] = ACTIONS(99),
    [sym__upname] = ACTIONS(101),
  },
  [11] = {
    [sym__expression] = STATE(1052),
    [sym_binary_expression] = STATE(1052),
    [sym__expression_unit] = STATE(1052),
    [sym_record] = STATE(967),
    [sym_todo] = STATE(1052),
    [sym_panic] = STATE(1052),
    [sym_echo] = STATE(1052),
    [sym_tuple] = STATE(1006),
    [sym_list] = STATE(1052),
    [sym__expression_bit_string] = STATE(1052),
    [sym_anonymous_function] = STATE(998),
    [sym_block] = STATE(938),
    [sym_case] = STATE(938),
    [sym_assert] = STATE(1052),
    [sym_boolean_negation] = STATE(1052),
    [sym_integer_negation] = STATE(1052),
    [sym_record_update] = STATE(967),
    [sym__maybe_tuple_expression] = STATE(1836),
    [sym_tuple_access] = STATE(938),
    [sym__maybe_record_expression] = STATE(1892),
    [sym_field_access] = STATE(938),
    [sym__maybe_function_expression] = STATE(1755),
    [sym_function_call] = STATE(938),
    [sym_string] = STATE(1052),
    [sym_integer] = STATE(1052),
    [sym_identifier] = STATE(937),
    [sym_constructor_name] = STATE(909),
    [sym_remote_constructor_name] = STATE(909),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_LBRACE] = ACTIONS(219),
    [anon_sym_SLASH] = ACTIONS(67),
    [anon_sym_POUND] = ACTIONS(221),
    [anon_sym_LBRACK] = ACTIONS(223),
    [anon_sym_LT_LT] = ACTIONS(225),
    [anon_sym_DASH] = ACTIONS(67),
    [anon_sym_fn] = ACTIONS(227),
    [anon_sym_DASH_GT] = ACTIONS(65),
    [anon_sym_PIPE_PIPE] = ACTIONS(65),
    [anon_sym_AMP_AMP] = ACTIONS(65),
    [anon_sym_EQ_EQ] = ACTIONS(65),
    [anon_sym_BANG_EQ] = ACTIONS(65),
    [anon_sym_LT] = ACTIONS(67),
    [anon_sym_LT_EQ] = ACTIONS(67),
    [anon_sym_LT_DOT] = ACTIONS(65),
    [anon_sym_LT_EQ_DOT] = ACTIONS(65),
    [anon_sym_GT] = ACTIONS(67),
    [anon_sym_GT_EQ] = ACTIONS(67),
    [anon_sym_GT_DOT] = ACTIONS(65),
    [anon_sym_GT_EQ_DOT] = ACTIONS(65),
    [anon_sym_PIPE_GT] = ACTIONS(65),
    [anon_sym_PLUS] = ACTIONS(67),
    [anon_sym_PLUS_DOT] = ACTIONS(65),
    [anon_sym_DASH_DOT] = ACTIONS(65),
    [anon_sym_STAR] = ACTIONS(67),
    [anon_sym_STAR_DOT] = ACTIONS(65),
    [anon_sym_SLASH_DOT] = ACTIONS(65),
    [anon_sym_PERCENT] = ACTIONS(65),
    [anon_sym_LT_GT] = ACTIONS(65),
    [anon_sym_todo] = ACTIONS(229),
    [anon_sym_panic] = ACTIONS(231),
    [anon_sym_echo] = ACTIONS(233),
    [anon_sym_case] = ACTIONS(235),
    [anon_sym_assert] = ACTIONS(237),
    [anon_sym_BANG] = ACTIONS(239),
    [anon_sym_DQUOTE] = ACTIONS(241),
    [sym_float] = ACTIONS(243),
    [sym__hex] = ACTIONS(245),
    [sym__decimal] = ACTIONS(247),
    [sym__octal] = ACTIONS(245),
    [sym__binary] = ACTIONS(245),
    [sym__name] = ACTIONS(249),
    [sym__upname] = ACTIONS(251),
  },
  [12] = {
    [aux_sym__statement_seq] = STATE(12),
    [sym__statement] = STATE(761),
    [sym__expression] = STATE(230),
    [sym_binary_expression] = STATE(230),
    [sym__expression_unit] = STATE(230),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(230),
    [sym_panic] = STATE(230),
    [sym_echo] = STATE(230),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(230),
    [sym__expression_bit_string] = STATE(230),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_let_assert] = STATE(761),
    [sym_let] = STATE(761),
    [sym_use] = STATE(761),
    [sym_assert] = STATE(230),
    [sym_boolean_negation] = STATE(230),
    [sym_integer_negation] = STATE(230),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_string] = STATE(230),
    [sym_integer] = STATE(230),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [ts_builtin_sym_end] = ACTIONS(253),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(255),
    [anon_sym_LBRACE] = ACTIONS(257),
    [anon_sym_AT] = ACTIONS(253),
    [anon_sym_import] = ACTIONS(255),
    [anon_sym_type] = ACTIONS(255),
    [anon_sym_const] = ACTIONS(255),
    [anon_sym_POUND] = ACTIONS(260),
    [anon_sym_LBRACK] = ACTIONS(263),
    [anon_sym_LT_LT] = ACTIONS(266),
    [anon_sym_DASH] = ACTIONS(269),
    [anon_sym_fn] = ACTIONS(272),
    [anon_sym_external] = ACTIONS(255),
    [anon_sym_todo] = ACTIONS(275),
    [anon_sym_panic] = ACTIONS(278),
    [anon_sym_echo] = ACTIONS(281),
    [anon_sym_case] = ACTIONS(284),
    [anon_sym_let] = ACTIONS(287),
    [anon_sym_assert] = ACTIONS(290),
    [anon_sym_use] = ACTIONS(293),
    [anon_sym_BANG] = ACTIONS(296),
    [sym_visibility_modifier] = ACTIONS(255),
    [sym_opacity_modifier] = ACTIONS(255),
    [anon_sym_DQUOTE] = ACTIONS(299),
    [sym_float] = ACTIONS(302),
    [sym__hex] = ACTIONS(305),
    [sym__decimal] = ACTIONS(308),
    [sym__octal] = ACTIONS(305),
    [sym__binary] = ACTIONS(305),
    [sym__name] = ACTIONS(311),
    [sym__upname] = ACTIONS(314),
  },
  [13] = {
    [aux_sym__statement_seq] = STATE(12),
    [sym__statement] = STATE(761),
    [sym__expression] = STATE(230),
    [sym_binary_expression] = STATE(230),
    [sym__expression_unit] = STATE(230),
    [sym_record] = STATE(137),
    [sym_todo] = STATE(230),
    [sym_panic] = STATE(230),
    [sym_echo] = STATE(230),
    [sym_tuple] = STATE(142),
    [sym_list] = STATE(230),
    [sym__expression_bit_string] = STATE(230),
    [sym_anonymous_function] = STATE(143),
    [sym_block] = STATE(114),
    [sym_case] = STATE(114),
    [sym_let_assert] = STATE(761),
    [sym_let] = STATE(761),
    [sym_use] = STATE(761),
    [sym_assert] = STATE(230),
    [sym_boolean_negation] = STATE(230),
    [sym_integer_negation] = STATE(230),
    [sym_record_update] = STATE(137),
    [sym__maybe_tuple_expression] = STATE(1913),
    [sym_tuple_access] = STATE(114),
    [sym__maybe_record_expression] = STATE(1902),
    [sym_field_access] = STATE(114),
    [sym__maybe_function_expression] = STATE(1666),
    [sym_function_call] = STATE(114),
    [sym_string] = STATE(230),
    [sym_integer] = STATE(230),
    [sym_identifier] = STATE(109),
    [sym_constructor_name] = STATE(44),
    [sym_remote_constructor_name] = STATE(44),
    [ts_builtin_sym_end] = ACTIONS(317),
    [sym_module_comment] = ACTIONS(3),
    [sym_statement_comment] = ACTIONS(5),
    [sym_comment] = ACTIONS(5),
    [anon_sym_if] = ACTIONS(319),
    [anon_sym_LBRACE] = ACTIONS(321),
    [anon_sym_AT] = ACTIONS(317),
    [anon_sym_import] = ACTIONS(319),
    [anon_sym_type] = ACTIONS(319),
    [anon_sym_const] = ACTIONS(319),
    [anon_sym_POUND] = ACTIONS(324),
    [anon_sym_LBRACK] = ACTIONS(327),
    [anon_sym_LT_LT] = ACTIONS(330),
    [anon_sym_DASH] = ACTIONS(333),
    [anon_sym_fn] = ACTIONS(336),
    [anon_sym_external] = ACTIONS(319),
    [anon_sym_todo] = ACTIONS(339),
    [anon_sym_panic] = ACTIONS(342),
    [anon_sym_echo] = ACTIONS(345),
    [anon_sym_case] = ACTIONS(348),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_assert] = ACTIONS(354),
    [anon_sym_use] = ACTIONS(357),
    [anon_sym_BANG] = ACTIONS(360),
    [sym_visibility_modifier] = ACTIONS(319),
    [sym_opacity_modifier] = ACTIONS(319),
    [anon_sym_DQUOTE] = ACTIONS(363),
    [sym_float] = ACTIONS(366),
    [sym__hex] = ACTIONS(369),
    [sym__decimal] = ACTIONS(372),
    [sym__octal] = ACTIONS(369),
    [sym__binary] = ACTIONS(369),
    [sym__name] = ACTIONS(375),
    [sym__upname] = ACTIONS(378),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(385), 1,
      anon_sym_LPAREN,
    STATE(90), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(381), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [73] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(387), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [141] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(393), 1,
      anon_sym_RBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [271] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(433), 1,
      anon_sym_RBRACE,
    STATE(19), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [401] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(435), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [531] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(437), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [661] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(253), 1,
      anon_sym_RBRACE,
    ACTIONS(439), 1,
      anon_sym_LBRACE,
    ACTIONS(442), 1,
      anon_sym_POUND,
    ACTIONS(445), 1,
      anon_sym_LBRACK,
    ACTIONS(448), 1,
      anon_sym_LT_LT,
    ACTIONS(451), 1,
      anon_sym_DASH,
    ACTIONS(454), 1,
      anon_sym_fn,
    ACTIONS(457), 1,
      anon_sym_todo,
    ACTIONS(460), 1,
      anon_sym_panic,
    ACTIONS(463), 1,
      anon_sym_echo,
    ACTIONS(466), 1,
      anon_sym_case,
    ACTIONS(469), 1,
      anon_sym_let,
    ACTIONS(472), 1,
      anon_sym_assert,
    ACTIONS(475), 1,
      anon_sym_use,
    ACTIONS(478), 1,
      anon_sym_BANG,
    ACTIONS(481), 1,
      anon_sym_DQUOTE,
    ACTIONS(484), 1,
      sym_float,
    ACTIONS(490), 1,
      sym__decimal,
    ACTIONS(493), 1,
      sym__name,
    ACTIONS(496), 1,
      sym__upname,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(487), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [791] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(499), 1,
      anon_sym_RBRACE,
    STATE(22), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [921] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(501), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1051] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(503), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [1119] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(507), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1249] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(509), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [1317] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(513), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1447] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(515), 1,
      anon_sym_RBRACE,
    STATE(42), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1577] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(517), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1707] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(519), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [1837] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(521), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [1905] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(525), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [1973] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(529), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2103] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(531), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2233] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(533), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2363] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(535), 1,
      anon_sym_RBRACE,
    STATE(69), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2493] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(537), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2623] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(539), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2753] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(541), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [2883] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(543), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3013] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(545), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [3081] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(549), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3211] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(551), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3341] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(553), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3471] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(555), 1,
      anon_sym_LPAREN,
    STATE(139), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(381), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [3543] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(557), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [3611] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(561), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3741] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(563), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [3809] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(567), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [3939] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(569), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [4007] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(573), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4137] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(575), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4267] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(577), 1,
      anon_sym_RBRACE,
    STATE(18), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4397] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(579), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [4465] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(583), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4595] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(585), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [4663] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(589), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4793] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(591), 1,
      anon_sym_RBRACE,
    STATE(50), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [4923] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(593), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [5053] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(595), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [5121] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(599), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [5189] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(603), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [5257] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(607), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [5325] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(611), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [5393] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(615), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [5523] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(617), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [5653] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(621), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [5725] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(628), 1,
      anon_sym_RBRACE,
    STATE(65), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [5855] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(630), 1,
      anon_sym_RBRACE,
    STATE(43), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [5985] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(632), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [6115] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(634), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(621), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [6187] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(638), 1,
      anon_sym_RBRACE,
    STATE(20), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [6317] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(525), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [6384] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(32), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [6511] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(387), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [6578] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(58), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [6705] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(51), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [6832] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(585), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [6899] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(38), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7026] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(37), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7153] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(621), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [7222] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(56), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7349] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(34), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7476] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(54), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7603] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(28), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [7730] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(621), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [7799] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(521), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [7866] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(642), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [7933] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(646), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8000] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(509), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8067] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(650), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8134] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(46), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [8261] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(621), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [8330] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(656), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8397] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(503), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8464] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(664), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(660), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(662), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [8533] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(71), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [8660] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(666), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8727] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(545), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8794] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(16), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [8921] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(557), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [8988] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(33), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [9115] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(36), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [9242] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(569), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [9309] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(670), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [9376] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(674), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [9443] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(26), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [9570] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(24), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [9697] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(563), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [9764] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(678), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [9835] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(686), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(682), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(684), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [9904] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(29), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [10031] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(39), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [10158] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(599), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10225] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10296] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(41), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [10423] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(611), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10490] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(48), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [10617] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(607), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10684] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(595), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10751] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(603), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [10818] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(413), 1,
      anon_sym_let,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(417), 1,
      anon_sym_use,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(423), 1,
      sym_float,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    STATE(64), 1,
      aux_sym__statement_seq,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1091), 4,
      sym__statement,
      sym_let_assert,
      sym_let,
      sym_use,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(582), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [10945] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(688), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11012] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(579), 29,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11079] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(692), 1,
      anon_sym_LPAREN,
    ACTIONS(694), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(660), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11150] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(696), 1,
      anon_sym_LPAREN,
    ACTIONS(698), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(682), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11221] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(700), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(702), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [11287] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(704), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(706), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [11353] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 21,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(607), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11419] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(688), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [11485] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(708), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(710), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [11551] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(712), 1,
      anon_sym_RPAREN,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(716), 1,
      anon_sym_DOT_DOT,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1481), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [11681] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(726), 1,
      anon_sym_RPAREN,
    ACTIONS(728), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1584), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [11811] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(730), 1,
      anon_sym_RPAREN,
    ACTIONS(732), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1585), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [11941] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(656), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12007] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(734), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [12073] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(642), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12139] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12207] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(738), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(740), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [12273] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(650), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12339] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(742), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(744), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [12405] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(746), 1,
      anon_sym_RPAREN,
    ACTIONS(748), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1654), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [12535] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12603] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(619), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12671] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(758), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(750), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(752), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [12757] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(674), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12823] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(670), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [12889] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 21,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 23,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [12963] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(774), 1,
      anon_sym_RPAREN,
    ACTIONS(776), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1648), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [13093] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 20,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 23,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13169] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 16,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13249] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13331] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(646), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [13397] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(778), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(780), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13463] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(782), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(784), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13529] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(786), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(788), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13595] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(790), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(792), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13661] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(794), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(796), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13727] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(798), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(800), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13793] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(802), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(804), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [13859] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(666), 28,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [13925] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(806), 1,
      anon_sym_RPAREN,
    ACTIONS(808), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1623), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [14055] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14139] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(758), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(810), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(812), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14225] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 23,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 25,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14295] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(734), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14361] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(758), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(814), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(816), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14447] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(818), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(821), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14513] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(758), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(824), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(826), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14599] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(758), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(760), 1,
      anon_sym_AMP_AMP,
    ACTIONS(768), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(754), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(756), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(762), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(770), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(764), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(766), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(772), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(828), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(830), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14685] = 36,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(832), 1,
      anon_sym_RPAREN,
    ACTIONS(834), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1660), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [14815] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(836), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(838), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14881] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(840), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(842), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [14947] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(844), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(846), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15013] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(848), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(850), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15079] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(852), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(854), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15145] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(856), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(858), 27,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15211] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15294] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(786), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [15359] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(828), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(830), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15444] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(880), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [15571] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(824), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(826), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [15656] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(806), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1623), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [15783] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(856), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [15848] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(882), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [15975] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(884), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [16102] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(790), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [16167] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(700), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [16232] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(798), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [16297] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(886), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(888), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [16382] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(890), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [16509] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(832), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1660), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [16636] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(892), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [16763] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(704), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [16828] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(894), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [16955] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(734), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17020] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 23,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 24,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17089] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(896), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [17216] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17297] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 16,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17376] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 20,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 22,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17451] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(810), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(812), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17536] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 21,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 22,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [17609] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(734), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17674] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(708), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17739] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(794), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17804] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(818), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17869] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(844), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17934] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(802), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [17999] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(898), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [18126] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(852), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [18191] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(900), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [18318] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(726), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1584), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [18445] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(814), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(816), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [18530] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(712), 1,
      anon_sym_RPAREN,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1481), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [18657] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(902), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(904), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [18742] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(730), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1585), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [18869] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(738), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [18934] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(742), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [18999] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(848), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [19064] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(774), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1648), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [19191] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(750), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(752), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [19276] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(906), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(908), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [19361] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(910), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [19488] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(836), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [19553] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(782), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [19618] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(778), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [19683] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(912), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [19810] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(746), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1654), 1,
      sym_argument,
    STATE(1675), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [19937] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(914), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [20064] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(864), 1,
      anon_sym_AMP_AMP,
    ACTIONS(872), 1,
      anon_sym_PIPE_GT,
    ACTIONS(878), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(860), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(862), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(866), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(874), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(868), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(870), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(876), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(916), 12,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(918), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [20149] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(920), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [20276] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 26,
      anon_sym_if,
      anon_sym_import,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
    ACTIONS(840), 27,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [20341] = 35,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(922), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [20468] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(569), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [20532] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(603), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [20596] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(611), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [20660] = 34,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(720), 1,
      sym_float,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(724), 1,
      sym__name,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1675), 1,
      sym_hole,
    STATE(1739), 1,
      sym_argument,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1961), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(995), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [20784] = 33,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(722), 1,
      sym__discard_name,
    ACTIONS(926), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1746), 1,
      sym_hole,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(924), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1000), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [20906] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(509), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [20970] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(928), 1,
      anon_sym_LPAREN,
    STATE(283), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(381), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21037] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(930), 1,
      anon_sym_RBRACK,
    ACTIONS(932), 1,
      anon_sym_DOT_DOT,
    ACTIONS(934), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [21155] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(521), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21217] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(936), 1,
      anon_sym_RBRACK,
    ACTIONS(938), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [21335] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(563), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21397] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(940), 1,
      anon_sym_RBRACK,
    ACTIONS(942), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [21515] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(944), 1,
      anon_sym_RBRACK,
    ACTIONS(946), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [21633] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(585), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21695] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(525), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21757] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(503), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21819] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(387), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [21881] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(948), 1,
      anon_sym_RBRACK,
    ACTIONS(950), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [21999] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(952), 1,
      anon_sym_RBRACK,
    ACTIONS(954), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22117] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(956), 1,
      anon_sym_RBRACK,
    ACTIONS(958), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22235] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(960), 1,
      anon_sym_RBRACK,
    ACTIONS(962), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22353] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(964), 1,
      anon_sym_RBRACK,
    ACTIONS(966), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22471] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(970), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(968), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1005), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22587] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(704), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [22649] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(972), 1,
      anon_sym_RBRACK,
    ACTIONS(974), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [22767] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(595), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [22829] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(599), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [22891] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(856), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [22953] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(557), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [23015] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(976), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [23081] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(980), 1,
      anon_sym_RBRACK,
    ACTIONS(982), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23199] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(984), 1,
      anon_sym_RBRACK,
    ACTIONS(986), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23317] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [23383] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(545), 31,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [23445] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(988), 1,
      anon_sym_RBRACK,
    ACTIONS(990), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23563] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(992), 1,
      anon_sym_RBRACK,
    ACTIONS(994), 1,
      anon_sym_DOT_DOT,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23681] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(996), 1,
      anon_sym_RPAREN,
    ACTIONS(998), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(942), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23796] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1008), 1,
      anon_sym_echo,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1014), 1,
      sym_float,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(650), 1,
      sym_pipeline_echo,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(652), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [23911] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [23974] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [24037] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [24100] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1026), 1,
      anon_sym_echo,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1030), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(818), 1,
      sym_pipeline_echo,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1008), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24215] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1032), 1,
      anon_sym_RBRACK,
    ACTIONS(1034), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(911), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24330] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1036), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24445] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1895), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24560] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1040), 1,
      anon_sym_echo,
    ACTIONS(1042), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(588), 1,
      sym_pipeline_echo,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(586), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24675] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1883), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24790] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1044), 1,
      anon_sym_RBRACK,
    ACTIONS(1046), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(915), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [24905] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1048), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25020] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(650), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [25081] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1050), 1,
      anon_sym_RPAREN,
    ACTIONS(1052), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(963), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25196] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(656), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [25257] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1866), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25372] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1054), 1,
      anon_sym_RPAREN,
    ACTIONS(1056), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(954), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25487] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1871), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25602] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1859), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25717] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    STATE(1845), 1,
      sym_case_subjects,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [25832] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(688), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [25893] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1058), 1,
      anon_sym_RBRACK,
    ACTIONS(1060), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(923), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26008] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1038), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1828), 1,
      sym_case_subjects,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(929), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26123] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(674), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [26184] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(670), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [26245] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(666), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [26306] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(646), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [26367] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(642), 30,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [26428] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1062), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26543] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1064), 1,
      anon_sym_echo,
    ACTIONS(1066), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(818), 1,
      sym_pipeline_echo,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(862), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26658] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1068), 1,
      anon_sym_RPAREN,
    ACTIONS(1070), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(933), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26773] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1088), 1,
      anon_sym_echo,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1098), 1,
      sym_float,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(135), 1,
      sym_pipeline_echo,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(147), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [26888] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1110), 1,
      anon_sym_echo,
    ACTIONS(1112), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(203), 1,
      sym_pipeline_echo,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(202), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27003] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1114), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27118] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1116), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27233] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1120), 1,
      anon_sym_echo,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1124), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1078), 1,
      sym_pipeline_echo,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1077), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27348] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1126), 1,
      anon_sym_RPAREN,
    ACTIONS(1128), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(931), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27463] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1130), 1,
      anon_sym_RBRACK,
    ACTIONS(1132), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(899), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27578] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1134), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27693] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1136), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27808] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1138), 1,
      anon_sym_RPAREN,
    ACTIONS(1140), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(939), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [27923] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1142), 1,
      anon_sym_RBRACK,
    ACTIONS(1144), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(898), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28038] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1146), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28153] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1148), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28268] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1150), 1,
      anon_sym_LPAREN,
    STATE(506), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(381), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [28333] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1168), 1,
      anon_sym_echo,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1178), 1,
      sym_float,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(550), 1,
      sym_pipeline_echo,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(548), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28448] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1188), 1,
      anon_sym_RBRACK,
    ACTIONS(1190), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(906), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28563] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1192), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28678] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28793] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1196), 1,
      anon_sym_RPAREN,
    ACTIONS(1198), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(943), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [28908] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1200), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29023] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1202), 1,
      anon_sym_RBRACK,
    ACTIONS(1204), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(910), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29138] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1206), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29253] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    ACTIONS(1208), 1,
      anon_sym_RPAREN,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29368] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1210), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(869), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29480] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1212), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1030), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29592] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1214), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(864), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29704] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1216), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(585), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29816] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1218), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(872), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [29928] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1220), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(215), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30040] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1222), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1026), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30152] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1224), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1028), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30264] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1226), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(849), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30376] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1228), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1032), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30488] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1230), 1,
      anon_sym_GT_GT,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30604] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1244), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(874), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30716] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1246), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1035), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30828] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1248), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [30944] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1250), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31060] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1252), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1055), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31172] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1254), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1530), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31288] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(802), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [31348] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1256), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(871), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31460] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(844), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [31520] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1258), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(636), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31632] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(818), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [31692] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(794), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [31752] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1260), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(795), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31864] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1262), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(580), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [31976] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1264), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(579), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32088] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1266), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(853), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32200] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(798), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [32260] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1268), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(576), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32372] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(569), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [32432] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(790), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [32492] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1270), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(863), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32604] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1272), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1025), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32716] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(93), 1,
      sym_float,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(861), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [32828] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(852), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [32888] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1274), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33004] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1276), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(575), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33116] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(836), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [33176] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1278), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(574), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33288] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1280), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(577), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33400] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1282), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(581), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33512] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1284), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(583), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33624] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1286), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(573), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33736] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1288), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1080), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33848] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1290), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1036), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [33960] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1292), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(644), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34072] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1294), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1034), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34184] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1296), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1568), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34300] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1300), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(163), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34412] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1302), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(201), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34524] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1304), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1027), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34636] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(848), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [34696] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(840), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [34756] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1306), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1015), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34868] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1308), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1013), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [34980] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1310), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(179), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35092] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1312), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1012), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35204] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1314), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(653), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35316] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1316), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35432] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1318), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(181), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35544] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1320), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1072), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35656] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(934), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(918), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35768] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1322), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1073), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35880] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1324), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(631), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [35992] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1326), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(654), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36104] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1328), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(630), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36216] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1330), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1017), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36328] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1332), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(655), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36440] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1334), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(656), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36552] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1336), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1020), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36664] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1338), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(578), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36776] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1340), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1632), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [36892] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1342), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1021), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37004] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1344), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1059), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37116] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1346), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(589), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37228] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1348), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(661), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37340] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1350), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(663), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37452] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1352), 1,
      anon_sym_LBRACE,
    ACTIONS(1355), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1003), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37564] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1357), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1063), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37676] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1359), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1087), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37788] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1361), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(984), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [37900] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1363), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(983), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38012] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1365), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(189), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38124] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1367), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(982), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38236] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1369), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(165), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38348] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1371), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(164), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38460] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1373), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(162), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38572] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1375), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(151), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38684] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(611), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [38744] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(607), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [38804] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1377), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [38920] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(603), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [38980] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1379), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(662), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39092] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1381), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(150), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39204] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1383), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(668), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39316] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1385), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(149), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39428] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1387), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1058), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39540] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1389), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(981), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39652] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1391), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(980), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39764] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1393), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(979), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [39876] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(387), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [39936] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1395), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(976), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40048] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1397), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(975), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40160] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(585), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [40220] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1355), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1003), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40332] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(563), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [40392] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1399), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1019), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40504] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1401), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1037), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40616] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1403), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(195), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40728] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1405), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(144), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40840] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1407), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(196), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [40952] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(557), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [41012] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1411), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(524), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41124] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1413), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(177), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41236] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(545), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [41296] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1415), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(198), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41408] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1417), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(199), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41520] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1419), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(660), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41632] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1421), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1029), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41744] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(786), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [41804] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1423), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(200), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [41916] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1425), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1014), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42028] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(243), 1,
      sym_float,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1052), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42140] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(525), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42200] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(503), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42260] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(509), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42320] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(521), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42380] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1427), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1018), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42492] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(782), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42552] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1429), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42668] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(778), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [42728] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1431), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42844] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1433), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1076), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [42956] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1435), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1060), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43068] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1437), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(584), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43180] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1439), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1065), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43292] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1441), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(873), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43404] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1443), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1653), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43520] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1445), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1081), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43632] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1447), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43748] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1449), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(556), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [43860] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(579), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [43920] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1451), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1558), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44036] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1453), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1597), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44152] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1455), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1075), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44264] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1457), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1074), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44376] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1459), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(166), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44488] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1461), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44604] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1463), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(587), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44716] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1465), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1071), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [44828] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(595), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [44888] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1467), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45004] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(599), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [45064] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1469), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1067), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45176] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1471), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(546), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45288] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1473), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1007), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45400] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1475), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(545), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45512] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1260), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(795), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45624] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1477), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(544), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45736] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1479), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(542), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45848] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1481), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(541), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [45960] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1483), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(523), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46072] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1485), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46188] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1355), 1,
      sym_float,
    ACTIONS(1487), 1,
      anon_sym_LBRACE,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1003), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46300] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1490), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1061), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46412] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(742), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [46472] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1492), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1628), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46588] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1494), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(213), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46700] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1496), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1070), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46812] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(738), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [46872] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1498), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [46988] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1500), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(221), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47100] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(1502), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [47164] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [47228] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1506), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1069), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47340] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(700), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [47400] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(217), 1,
      sym_float,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1004), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47512] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(708), 29,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [47572] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1508), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(222), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47684] = 32,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    ACTIONS(1510), 1,
      anon_sym_GT_GT,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47800] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1512), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(525), 13,
      sym__expression,
      sym_binary_expression,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [47912] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(650), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [47971] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48032] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1514), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(660), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48093] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1516), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(682), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48154] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48215] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(619), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48276] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1518), 1,
      anon_sym_LPAREN,
    ACTIONS(1520), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(660), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48339] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(579), 28,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48398] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1522), 1,
      anon_sym_LPAREN,
    ACTIONS(1524), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(682), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48461] = 31,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1242), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1730), 1,
      sym_expression_bit_string_segment,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [48574] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(666), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48633] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(656), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48692] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(646), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48751] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(688), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48810] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(670), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48869] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(674), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48928] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(642), 27,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [48987] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49045] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1530), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(814), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(816), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [49123] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1530), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(810), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(812), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [49201] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1546), 1,
      sym_float,
    ACTIONS(1550), 1,
      sym__decimal,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1548), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(569), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [49311] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(798), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49369] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(790), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49427] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(802), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49485] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1552), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(751), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [49595] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(207), 1,
      anon_sym_todo,
    ACTIONS(209), 1,
      anon_sym_panic,
    ACTIONS(211), 1,
      anon_sym_echo,
    ACTIONS(213), 1,
      anon_sym_assert,
    ACTIONS(1024), 1,
      anon_sym_DASH,
    ACTIONS(1028), 1,
      anon_sym_BANG,
    ACTIONS(1554), 1,
      sym_float,
    ACTIONS(1558), 1,
      sym__decimal,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1556), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(763), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [49705] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(848), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49763] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(840), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [49821] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1560), 1,
      sym_float,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(186), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [49931] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1530), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(828), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(830), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [50009] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1530), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(824), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(826), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [50087] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1562), 1,
      sym_float,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(673), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [50197] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1564), 1,
      sym_float,
    ACTIONS(1568), 1,
      sym__decimal,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1566), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(157), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [50307] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(191), 1,
      anon_sym_LBRACE,
    ACTIONS(193), 1,
      anon_sym_fn,
    ACTIONS(195), 1,
      anon_sym_todo,
    ACTIONS(197), 1,
      anon_sym_panic,
    ACTIONS(199), 1,
      anon_sym_echo,
    ACTIONS(201), 1,
      anon_sym_case,
    ACTIONS(203), 1,
      anon_sym_assert,
    ACTIONS(1000), 1,
      anon_sym_POUND,
    ACTIONS(1002), 1,
      anon_sym_LBRACK,
    ACTIONS(1004), 1,
      anon_sym_LT_LT,
    ACTIONS(1006), 1,
      anon_sym_DASH,
    ACTIONS(1010), 1,
      anon_sym_BANG,
    ACTIONS(1012), 1,
      anon_sym_DQUOTE,
    ACTIONS(1020), 1,
      sym__name,
    ACTIONS(1022), 1,
      sym__upname,
    ACTIONS(1570), 1,
      sym_float,
    ACTIONS(1574), 1,
      sym__decimal,
    STATE(615), 1,
      sym_identifier,
    STATE(628), 1,
      sym_anonymous_function,
    STATE(629), 1,
      sym_tuple,
    STATE(1713), 1,
      sym__maybe_function_expression,
    STATE(1880), 1,
      sym__maybe_record_expression,
    STATE(1886), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(597), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(620), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1572), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(598), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(669), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [50417] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1072), 1,
      anon_sym_LBRACE,
    ACTIONS(1074), 1,
      anon_sym_POUND,
    ACTIONS(1076), 1,
      anon_sym_LBRACK,
    ACTIONS(1078), 1,
      anon_sym_LT_LT,
    ACTIONS(1080), 1,
      anon_sym_DASH,
    ACTIONS(1082), 1,
      anon_sym_fn,
    ACTIONS(1084), 1,
      anon_sym_todo,
    ACTIONS(1086), 1,
      anon_sym_panic,
    ACTIONS(1090), 1,
      anon_sym_case,
    ACTIONS(1092), 1,
      anon_sym_assert,
    ACTIONS(1094), 1,
      anon_sym_BANG,
    ACTIONS(1096), 1,
      anon_sym_DQUOTE,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(1104), 1,
      sym__name,
    ACTIONS(1106), 1,
      sym__upname,
    ACTIONS(1298), 1,
      anon_sym_echo,
    ACTIONS(1576), 1,
      sym_float,
    STATE(70), 1,
      sym_identifier,
    STATE(80), 1,
      sym_anonymous_function,
    STATE(85), 1,
      sym_tuple,
    STATE(1719), 1,
      sym__maybe_function_expression,
    STATE(1856), 1,
      sym__maybe_record_expression,
    STATE(1957), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(14), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(92), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(66), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(156), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [50527] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 18,
      anon_sym_as,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 22,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [50589] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 12,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(736), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [50665] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(11), 1,
      anon_sym_LBRACE,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_LT_LT,
    ACTIONS(27), 1,
      anon_sym_DASH,
    ACTIONS(33), 1,
      anon_sym_todo,
    ACTIONS(35), 1,
      anon_sym_panic,
    ACTIONS(37), 1,
      anon_sym_echo,
    ACTIONS(39), 1,
      anon_sym_case,
    ACTIONS(43), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_BANG,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(1108), 1,
      anon_sym_fn,
    ACTIONS(1578), 1,
      sym_float,
    ACTIONS(1582), 1,
      sym__decimal,
    STATE(109), 1,
      sym_identifier,
    STATE(142), 1,
      sym_tuple,
    STATE(143), 1,
      sym_anonymous_function,
    STATE(1666), 1,
      sym__maybe_function_expression,
    STATE(1902), 1,
      sym__maybe_record_expression,
    STATE(1913), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(44), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(137), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1580), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(114), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(205), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [50775] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 13,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [50849] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 15,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [50921] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 16,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 19,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [50989] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1584), 1,
      sym_float,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1011), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [51099] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 16,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 20,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51165] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(219), 1,
      anon_sym_LBRACE,
    ACTIONS(221), 1,
      anon_sym_POUND,
    ACTIONS(223), 1,
      anon_sym_LBRACK,
    ACTIONS(225), 1,
      anon_sym_LT_LT,
    ACTIONS(227), 1,
      anon_sym_fn,
    ACTIONS(229), 1,
      anon_sym_todo,
    ACTIONS(231), 1,
      anon_sym_panic,
    ACTIONS(233), 1,
      anon_sym_echo,
    ACTIONS(235), 1,
      anon_sym_case,
    ACTIONS(237), 1,
      anon_sym_assert,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1118), 1,
      anon_sym_DASH,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1586), 1,
      sym_float,
    ACTIONS(1590), 1,
      sym__decimal,
    STATE(937), 1,
      sym_identifier,
    STATE(998), 1,
      sym_anonymous_function,
    STATE(1006), 1,
      sym_tuple,
    STATE(1755), 1,
      sym__maybe_function_expression,
    STATE(1836), 1,
      sym__maybe_tuple_expression,
    STATE(1892), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(909), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(967), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1588), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(938), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(1050), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [51275] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51333] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(708), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51391] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(700), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51449] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1552), 1,
      sym_float,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(751), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [51559] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(738), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51617] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(742), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51675] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1530), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1532), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1540), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1526), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1528), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1534), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1542), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1536), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1538), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1544), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(750), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(752), 12,
      anon_sym_as,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [51753] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(778), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51811] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(782), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51869] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(786), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51927] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(818), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [51985] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(836), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52043] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(852), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52101] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1592), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(355), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52211] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(844), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52269] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1152), 1,
      anon_sym_LBRACE,
    ACTIONS(1154), 1,
      anon_sym_POUND,
    ACTIONS(1156), 1,
      anon_sym_LBRACK,
    ACTIONS(1158), 1,
      anon_sym_LT_LT,
    ACTIONS(1160), 1,
      anon_sym_DASH,
    ACTIONS(1162), 1,
      anon_sym_fn,
    ACTIONS(1164), 1,
      anon_sym_todo,
    ACTIONS(1166), 1,
      anon_sym_panic,
    ACTIONS(1170), 1,
      anon_sym_case,
    ACTIONS(1172), 1,
      anon_sym_assert,
    ACTIONS(1174), 1,
      anon_sym_BANG,
    ACTIONS(1176), 1,
      anon_sym_DQUOTE,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(1184), 1,
      sym__name,
    ACTIONS(1186), 1,
      sym__upname,
    ACTIONS(1409), 1,
      anon_sym_echo,
    ACTIONS(1594), 1,
      sym_float,
    STATE(497), 1,
      sym_identifier,
    STATE(510), 1,
      sym_tuple,
    STATE(511), 1,
      sym_anonymous_function,
    STATE(1686), 1,
      sym__maybe_function_expression,
    STATE(1868), 1,
      sym__maybe_record_expression,
    STATE(1936), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(315), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(507), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(498), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(528), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52379] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(704), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52437] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(69), 1,
      anon_sym_LBRACE,
    ACTIONS(71), 1,
      anon_sym_POUND,
    ACTIONS(73), 1,
      anon_sym_LBRACK,
    ACTIONS(75), 1,
      anon_sym_LT_LT,
    ACTIONS(77), 1,
      anon_sym_fn,
    ACTIONS(79), 1,
      anon_sym_todo,
    ACTIONS(81), 1,
      anon_sym_panic,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(85), 1,
      anon_sym_case,
    ACTIONS(87), 1,
      anon_sym_assert,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(714), 1,
      anon_sym_DASH,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(1554), 1,
      sym_float,
    ACTIONS(1558), 1,
      sym__decimal,
    STATE(696), 1,
      sym_identifier,
    STATE(743), 1,
      sym_tuple,
    STATE(744), 1,
      sym_anonymous_function,
    STATE(1808), 1,
      sym__maybe_function_expression,
    STATE(1819), 1,
      sym__maybe_tuple_expression,
    STATE(1842), 1,
      sym__maybe_record_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(686), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(740), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1556), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(705), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(763), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52547] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1592), 1,
      sym_float,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(355), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52657] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(794), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52715] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(401), 1,
      anon_sym_DASH,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(405), 1,
      anon_sym_todo,
    ACTIONS(407), 1,
      anon_sym_panic,
    ACTIONS(409), 1,
      anon_sym_echo,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(415), 1,
      anon_sym_assert,
    ACTIONS(419), 1,
      anon_sym_BANG,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1596), 1,
      sym_float,
    ACTIONS(1600), 1,
      sym__decimal,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1830), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1598), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(347), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52825] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 20,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(856), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [52883] = 30,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(83), 1,
      anon_sym_echo,
    ACTIONS(391), 1,
      anon_sym_LBRACE,
    ACTIONS(395), 1,
      anon_sym_POUND,
    ACTIONS(397), 1,
      anon_sym_LBRACK,
    ACTIONS(399), 1,
      anon_sym_LT_LT,
    ACTIONS(403), 1,
      anon_sym_fn,
    ACTIONS(411), 1,
      anon_sym_case,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(1232), 1,
      anon_sym_DASH,
    ACTIONS(1234), 1,
      anon_sym_todo,
    ACTIONS(1236), 1,
      anon_sym_panic,
    ACTIONS(1238), 1,
      anon_sym_assert,
    ACTIONS(1240), 1,
      anon_sym_BANG,
    ACTIONS(1596), 1,
      sym_float,
    ACTIONS(1600), 1,
      sym__decimal,
    STATE(263), 1,
      sym_identifier,
    STATE(273), 1,
      sym_tuple,
    STATE(274), 1,
      sym_anonymous_function,
    STATE(1805), 1,
      sym__maybe_function_expression,
    STATE(1817), 1,
      sym__maybe_record_expression,
    STATE(1925), 1,
      sym__maybe_tuple_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(240), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    STATE(272), 2,
      sym_record,
      sym_record_update,
    ACTIONS(1598), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(266), 5,
      sym_block,
      sym_case,
      sym_tuple_access,
      sym_field_access,
      sym_function_call,
    STATE(347), 11,
      sym__expression_unit,
      sym_todo,
      sym_panic,
      sym_echo,
      sym_list,
      sym__expression_bit_string,
      sym_assert,
      sym_boolean_negation,
      sym_integer_negation,
      sym_string,
      sym_integer,
  [52993] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 15,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 19,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53060] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 17,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 22,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53121] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53178] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(886), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(888), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53255] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(736), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 12,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53330] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(750), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(752), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53407] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(824), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(826), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53484] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(828), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(830), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53561] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(736), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 13,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53634] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(916), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(918), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53711] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(736), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 15,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [53782] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(814), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(816), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53859] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(810), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(812), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [53936] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 15,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 20,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54001] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(902), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(904), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [54078] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 19,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
    ACTIONS(734), 26,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54135] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1606), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1612), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1614), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1602), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1604), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1608), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1616), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1610), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1618), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1620), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(906), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(908), 11,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      anon_sym_BANG,
      sym__decimal,
      sym__name,
  [54212] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 10,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(607), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54267] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 10,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(611), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54322] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 10,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(603), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54377] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 10,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(569), 33,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54432] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 9,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(704), 32,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54485] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 9,
      anon_sym_SLASH,
      anon_sym_EQ,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(856), 32,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54538] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(509), 32,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [54591] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1622), 1,
      anon_sym_LPAREN,
    STATE(627), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(381), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54646] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(619), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54700] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(509), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54750] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(579), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54800] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(599), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54850] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(525), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54900] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(611), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [54950] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1626), 1,
      anon_sym_LPAREN,
    STATE(649), 1,
      sym_constant_record_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1628), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1624), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [55004] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(521), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55054] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(607), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55104] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(603), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55154] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(545), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55204] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(557), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55254] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(503), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55304] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(563), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55354] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(595), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55404] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(569), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55454] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(585), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55504] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(1630), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(619), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55558] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(387), 27,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55608] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1634), 1,
      anon_sym_LPAREN,
    ACTIONS(1636), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(660), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55662] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1638), 1,
      anon_sym_LPAREN,
    ACTIONS(1640), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(682), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55716] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(642), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55765] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(619), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55816] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(646), 26,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55865] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(666), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55914] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(670), 26,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [55963] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(674), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56012] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(688), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56061] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(656), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56110] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(650), 26,
      anon_sym_RBRACE,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56159] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(619), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56210] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(619), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56261] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 11,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56327] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 14,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56389] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1662), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1660), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56437] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(852), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56485] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1666), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1664), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56533] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(836), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56581] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(816), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(814), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56649] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1672), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1670), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56697] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1676), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1674), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56745] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(786), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56793] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(782), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56841] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(778), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [56889] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1680), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1678), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56937] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1684), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1682), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [56985] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(752), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(750), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57053] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(742), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57101] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(738), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57149] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(700), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57197] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(708), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57245] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1688), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1686), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [57293] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(734), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57341] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1692), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1690), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [57389] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 7,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(734), 19,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57445] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 7,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(734), 18,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57503] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 12,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57567] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 9,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(734), 21,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57619] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(734), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57667] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1696), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1694), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [57715] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1700), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1698), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [57763] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(856), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57811] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1704), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1702), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57879] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(826), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(824), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [57947] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(1708), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(1706), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58015] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(830), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(828), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58083] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(704), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58131] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(840), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58179] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(848), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58227] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(798), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58275] = 14,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1646), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1654), 1,
      anon_sym_PIPE_GT,
    ACTIONS(1668), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1642), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(1644), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(1648), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(812), 3,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(1650), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(1652), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(1658), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(810), 10,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58343] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(794), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58391] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(818), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58439] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(844), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58487] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(802), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58535] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 11,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      sym__decimal,
      sym__discard_name,
      sym__name,
    ACTIONS(790), 25,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [58583] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1712), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1710), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58631] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1716), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1714), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58679] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1720), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1718), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58727] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1724), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1722), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58775] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1728), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1726), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58823] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1732), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1730), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58871] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1734), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58919] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1740), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1738), 28,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [58967] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1746), 1,
      anon_sym_SLASH,
    STATE(684), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1742), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1744), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59018] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1752), 1,
      anon_sym_SLASH,
    STATE(683), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1748), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1750), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59069] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1746), 1,
      anon_sym_SLASH,
    STATE(683), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1755), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1757), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59120] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1748), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_DOT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1750), 20,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_SLASH,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59166] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1759), 1,
      anon_sym_LPAREN,
    STATE(746), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(381), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59216] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1765), 1,
      anon_sym_LPAREN,
    STATE(747), 1,
      sym_type_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1761), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1763), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59266] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1771), 1,
      anon_sym_LPAREN,
    STATE(786), 1,
      sym_type_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1767), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1769), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59315] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(557), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59360] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(563), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59405] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1773), 15,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1775), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59450] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1781), 1,
      anon_sym_DOT,
    ACTIONS(1783), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1777), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1779), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59499] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(521), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59544] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(525), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59589] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(585), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59634] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(1785), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59683] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1793), 1,
      anon_sym_LPAREN,
    STATE(780), 1,
      sym__attribute_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1789), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1791), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59732] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1795), 15,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1797), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [59777] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(545), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59822] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(599), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59867] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(387), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59912] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(503), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [59957] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(579), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60002] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(595), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60047] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(1799), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60096] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1802), 1,
      anon_sym_LPAREN,
    STATE(794), 1,
      sym_constant_record_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1624), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1628), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60145] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1804), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1806), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60189] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1808), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1810), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60233] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(607), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1812), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1814), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60279] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1816), 1,
      anon_sym_RBRACE,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1901), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [60363] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1840), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1857), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [60447] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1842), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1844), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60491] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1850), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1846), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1848), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60537] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(688), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60581] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1852), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1854), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60625] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1856), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DASH_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1858), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60669] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(646), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60713] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1864), 1,
      anon_sym_LBRACE,
    ACTIONS(1866), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1860), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1862), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [60761] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1868), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1844), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [60845] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1870), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1882), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [60929] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(666), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [60973] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1872), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1874), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61017] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1876), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1878), 19,
      anon_sym_if,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61061] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(642), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61105] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1884), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1880), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1882), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61151] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1886), 1,
      anon_sym_LPAREN,
    ACTIONS(1888), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(682), 22,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61199] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1890), 1,
      anon_sym_LPAREN,
    ACTIONS(1892), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(660), 22,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61247] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1894), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DASH_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1896), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61291] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1898), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1943), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [61375] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(670), 24,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61419] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1900), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DASH_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1902), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61463] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1904), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1906), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61507] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1908), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1944), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [61591] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1914), 1,
      anon_sym_LBRACE,
    ACTIONS(1916), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1910), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1912), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61639] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1918), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_DASH_GT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1920), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61683] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1926), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1922), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1924), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61729] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(674), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61773] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1932), 1,
      anon_sym_LBRACE,
    ACTIONS(1934), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1928), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1930), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [61821] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(656), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61865] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1936), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [61911] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(1938), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    STATE(1960), 1,
      sym_case_clauses,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(823), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [61995] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1940), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1942), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62039] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1944), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62085] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62131] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1950), 1,
      anon_sym_LBRACE,
    ACTIONS(1952), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1946), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1948), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62179] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(650), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62223] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1954), 14,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_EQ,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1956), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62267] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1660), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1662), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62310] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(852), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62353] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1958), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1960), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62396] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(790), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62439] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1962), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1964), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62482] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1966), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1968), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62525] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1970), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1972), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62568] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1974), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1976), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62611] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1978), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1980), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62654] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1982), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1984), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62697] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1986), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1988), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62740] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(798), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62783] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1990), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1992), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62826] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1994), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1996), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [62869] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1998), 1,
      anon_sym_RPAREN,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1547), 1,
      sym_record_pattern_argument,
    STATE(1615), 1,
      sym__pattern,
    STATE(1847), 1,
      sym_pattern_spread,
    STATE(1897), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [62952] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(794), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [62995] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2006), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2008), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63038] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2010), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2012), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63081] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2014), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2016), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63124] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1734), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1736), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63167] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1726), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1728), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63210] = 23,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2018), 1,
      anon_sym_RBRACE,
    ACTIONS(2020), 1,
      anon_sym_POUND,
    ACTIONS(2023), 1,
      anon_sym_LBRACK,
    ACTIONS(2026), 1,
      anon_sym_LT_LT,
    ACTIONS(2029), 1,
      anon_sym_DASH,
    ACTIONS(2032), 1,
      anon_sym_DQUOTE,
    ACTIONS(2035), 1,
      sym_float,
    ACTIONS(2041), 1,
      sym__decimal,
    ACTIONS(2044), 1,
      sym__discard_name,
    ACTIONS(2047), 1,
      sym__name,
    ACTIONS(2050), 1,
      sym__upname,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(769), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(2038), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [63291] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1718), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1720), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63334] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2053), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2055), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63377] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1670), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1672), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63420] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(818), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [63463] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(844), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [63506] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1698), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1700), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63549] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(802), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [63592] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2057), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2059), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63635] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(848), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [63678] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2061), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2063), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63721] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2065), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2067), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63764] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2073), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2069), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2071), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63809] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(840), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [63852] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2075), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2077), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63895] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2083), 1,
      anon_sym_size,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2079), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1562), 5,
      sym__expression_bit_string_segment_option,
      sym__expression_bit_string_named_segment_option,
      sym__expression_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [63952] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2091), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2093), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [63995] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2095), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2097), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64038] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1946), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1948), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64081] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2099), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2101), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64124] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2103), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2105), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64167] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2107), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2109), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64210] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2111), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2113), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64253] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2115), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2117), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64296] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2119), 1,
      anon_sym_RPAREN,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1404), 1,
      sym_record_pattern_argument,
    STATE(1615), 1,
      sym__pattern,
    STATE(1894), 1,
      sym_pattern_spread,
    STATE(1897), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [64379] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1686), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1688), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64422] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(734), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [64465] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1682), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1684), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64508] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1678), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1680), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64551] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2121), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2123), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64594] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2127), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2125), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1633), 5,
      sym__pattern_bit_string_segment_option,
      sym__pattern_bit_string_named_segment_option,
      sym__pattern_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [64651] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1710), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1712), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64694] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2129), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2131), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64737] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2133), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2135), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64780] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2137), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2139), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64823] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1694), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1696), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64866] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2141), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2143), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64909] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2145), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2147), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64952] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1690), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1692), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [64995] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2149), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2151), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65038] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2153), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2155), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65081] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2159), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2157), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1542), 5,
      sym__constant_bit_string_segment_option,
      sym__constant_bit_string_named_segment_option,
      sym__constant_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [65138] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1674), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1676), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65181] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2161), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2163), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65224] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2127), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2165), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1633), 5,
      sym__pattern_bit_string_segment_option,
      sym__pattern_bit_string_named_segment_option,
      sym__pattern_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [65281] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2167), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2169), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65324] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2171), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2173), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65367] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2175), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2177), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65410] = 24,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2002), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2179), 1,
      anon_sym_RPAREN,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1547), 1,
      sym_record_pattern_argument,
    STATE(1615), 1,
      sym__pattern,
    STATE(1897), 1,
      sym_label,
    STATE(1962), 1,
      sym_pattern_spread,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [65493] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(734), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [65536] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(708), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [65579] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2181), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2183), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65622] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2185), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2187), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65665] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2159), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2189), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1542), 5,
      sym__constant_bit_string_segment_option,
      sym__constant_bit_string_named_segment_option,
      sym__constant_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [65722] = 23,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2191), 1,
      anon_sym_RBRACE,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1433), 1,
      sym_case_clause_pattern,
    STATE(1524), 1,
      sym_case_clause_patterns,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(769), 2,
      sym_case_clause,
      aux_sym_case_clauses_repeat1,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [65803] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(700), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [65846] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1664), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1666), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [65889] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(738), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [65932] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(742), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [65975] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2193), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2195), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66018] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(778), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [66061] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2197), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2199), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66104] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(782), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [66147] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(786), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [66190] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2201), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2203), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66233] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2205), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2207), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66276] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(836), 23,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [66319] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2083), 1,
      anon_sym_size,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2209), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1562), 5,
      sym__expression_bit_string_segment_option,
      sym__expression_bit_string_named_segment_option,
      sym__expression_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [66376] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2211), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2213), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66419] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1714), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1716), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66462] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2215), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2217), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66505] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1722), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1724), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66548] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1738), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1740), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66591] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1730), 13,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(1732), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66634] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2223), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2219), 12,
      ts_builtin_sym_end,
      anon_sym_AT,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
    ACTIONS(2221), 18,
      anon_sym_if,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym_visibility_modifier,
      sym_opacity_modifier,
      sym__decimal,
      sym__name,
  [66679] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2127), 1,
      anon_sym_size,
    STATE(1665), 1,
      sym_pattern_bit_string_segment_options,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1429), 5,
      sym__pattern_bit_string_segment_option,
      sym__pattern_bit_string_named_segment_option,
      sym__pattern_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [66735] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2233), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1105), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [66807] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2235), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1113), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [66879] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2237), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1108), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [66951] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2249), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1126), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67023] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(750), 7,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [67083] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2271), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1131), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67155] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2273), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1109), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67227] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(581), 9,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_PIPE,
    ACTIONS(579), 21,
      anon_sym_if,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [67269] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(828), 7,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [67329] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2275), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1110), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67401] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2277), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1119), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67473] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2279), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1124), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67545] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2281), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1125), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67617] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2283), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1127), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67689] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2285), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1128), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67761] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(718), 1,
      anon_sym_BANG,
    ACTIONS(2239), 1,
      anon_sym_LBRACE,
    ACTIONS(2241), 1,
      anon_sym_POUND,
    ACTIONS(2243), 1,
      anon_sym_LBRACK,
    ACTIONS(2245), 1,
      anon_sym_LT_LT,
    ACTIONS(2247), 1,
      anon_sym_DASH,
    ACTIONS(2287), 1,
      sym_float,
    STATE(1098), 1,
      sym_identifier,
    STATE(1116), 1,
      sym__case_clause_guard_binary_expression,
    STATE(1117), 1,
      sym__case_clause_tuple_access,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1046), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1129), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [67833] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(810), 7,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [67893] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 16,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_DOT_DOT,
  [67943] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(814), 7,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [68003] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 15,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_DOT_DOT,
  [68055] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2289), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1114), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [68127] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(241), 1,
      anon_sym_DQUOTE,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(249), 1,
      sym__name,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(1122), 1,
      anon_sym_BANG,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(2225), 1,
      anon_sym_LBRACE,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2291), 1,
      sym_float,
    STATE(1102), 1,
      sym_identifier,
    STATE(1136), 1,
      sym__case_clause_tuple_access,
    STATE(1137), 1,
      sym__case_clause_guard_binary_expression,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1115), 11,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym__case_clause_guard_expression,
      sym__case_clause_guard_unit,
      sym_boolean_negation,
      sym_string,
      sym_integer,
  [68199] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2083), 1,
      anon_sym_size,
    ACTIONS(2089), 1,
      anon_sym_unit,
    STATE(1735), 1,
      sym_expression_bit_string_segment_options,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1474), 5,
      sym__expression_bit_string_segment_option,
      sym__expression_bit_string_named_segment_option,
      sym__expression_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [68255] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2159), 1,
      anon_sym_size,
    STATE(1751), 1,
      sym_constant_bit_string_segment_options,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1443), 5,
      sym__constant_bit_string_segment_option,
      sym__constant_bit_string_named_segment_option,
      sym__constant_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [68311] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(824), 7,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [68371] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(509), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 9,
      anon_sym_DOT,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(607), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_as,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [68415] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 6,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
    ACTIONS(734), 18,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_DOT_DOT,
  [68461] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 11,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_DOT_DOT,
  [68515] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 9,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_DOT_DOT,
  [68571] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 8,
      anon_sym_RBRACE,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_DOT_DOT,
  [68629] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2083), 1,
      anon_sym_size,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1562), 5,
      sym__expression_bit_string_segment_option,
      sym__expression_bit_string_named_segment_option,
      sym__expression_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [68682] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2293), 1,
      anon_sym_RBRACK,
    ACTIONS(2295), 1,
      anon_sym_DOT_DOT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1396), 1,
      sym__pattern,
    STATE(1938), 1,
      sym_list_pattern_tail,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [68759] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2127), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1633), 5,
      sym__pattern_bit_string_segment_option,
      sym__pattern_bit_string_named_segment_option,
      sym__pattern_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [68812] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(2089), 1,
      anon_sym_unit,
    ACTIONS(2159), 1,
      anon_sym_size,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2087), 3,
      anon_sym_utf8,
      anon_sym_utf16,
      anon_sym_utf32,
    STATE(1542), 5,
      sym__constant_bit_string_segment_option,
      sym__constant_bit_string_named_segment_option,
      sym__constant_bit_string_segment_option_size,
      sym_integer,
      sym__bit_string_segment_option,
    ACTIONS(2085), 14,
      anon_sym_binary,
      anon_sym_bytes,
      anon_sym_int,
      anon_sym_float,
      anon_sym_bit_string,
      anon_sym_bits,
      anon_sym_utf8_codepoint,
      anon_sym_utf16_codepoint,
      anon_sym_utf32_codepoint,
      anon_sym_signed,
      anon_sym_unsigned,
      anon_sym_big,
      anon_sym_little,
      anon_sym_native,
  [68865] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2295), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2297), 1,
      anon_sym_RBRACK,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    STATE(1833), 1,
      sym_list_pattern_tail,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [68942] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2299), 1,
      anon_sym_if,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2301), 2,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69017] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2303), 1,
      anon_sym_if,
    ACTIONS(2305), 1,
      anon_sym_DASH_GT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1554), 1,
      sym_case_clause_pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69094] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2307), 1,
      anon_sym_if,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2309), 2,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69169] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1544), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    ACTIONS(2311), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69242] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2295), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2313), 1,
      anon_sym_RBRACK,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    STATE(1932), 1,
      sym_list_pattern_tail,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69319] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2315), 1,
      anon_sym_LT_DASH,
    ACTIONS(2317), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1454), 1,
      sym__pattern,
    STATE(1552), 1,
      sym_use_assignment,
    STATE(1835), 1,
      sym_use_assignments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1415), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69396] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2319), 1,
      anon_sym_if,
    ACTIONS(2321), 1,
      anon_sym_DASH_GT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1554), 1,
      sym_case_clause_pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69473] = 22,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2317), 1,
      sym_float,
    ACTIONS(2323), 1,
      anon_sym_LT_DASH,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1454), 1,
      sym__pattern,
    STATE(1552), 1,
      sym_use_assignment,
    STATE(1875), 1,
      sym_use_assignments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1415), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69550] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2317), 1,
      sym_float,
    ACTIONS(2325), 1,
      anon_sym_LT_DASH,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1454), 1,
      sym__pattern,
    STATE(1690), 1,
      sym_use_assignment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1415), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69624] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1547), 1,
      sym_record_pattern_argument,
    STATE(1615), 1,
      sym__pattern,
    STATE(1897), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69698] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2327), 1,
      anon_sym_GT_GT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1582), 1,
      sym__pattern,
    STATE(1723), 1,
      sym_pattern_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69772] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2317), 1,
      sym_float,
    ACTIONS(2329), 1,
      anon_sym_LT_DASH,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1454), 1,
      sym__pattern,
    STATE(1690), 1,
      sym_use_assignment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1415), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69846] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2331), 1,
      anon_sym_GT_GT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1582), 1,
      sym__pattern,
    STATE(1645), 1,
      sym_pattern_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69920] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2333), 1,
      anon_sym_GT_GT,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1582), 1,
      sym__pattern,
    STATE(1723), 1,
      sym_pattern_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [69994] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2335), 1,
      anon_sym_assert,
    STATE(1103), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1553), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70068] = 21,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2337), 1,
      anon_sym_assert,
    STATE(771), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1545), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70142] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1582), 1,
      sym__pattern,
    STATE(1723), 1,
      sym_pattern_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70213] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1053), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1593), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70284] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2339), 1,
      anon_sym_COMMA,
    ACTIONS(2341), 1,
      anon_sym_RBRACK,
    ACTIONS(2343), 1,
      anon_sym_DOT_DOT,
    STATE(1422), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [70347] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2345), 1,
      anon_sym_COMMA,
    ACTIONS(2347), 1,
      anon_sym_RBRACK,
    ACTIONS(2349), 1,
      anon_sym_DOT_DOT,
    STATE(1430), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [70410] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2351), 1,
      anon_sym_RPAREN,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1595), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70481] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1410), 1,
      sym__pattern,
    STATE(1554), 1,
      sym_case_clause_pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70552] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(667), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1627), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70623] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(352), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1570), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70694] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(725), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1634), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70765] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1088), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1656), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70836] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2353), 1,
      anon_sym_COMMA,
    ACTIONS(2355), 1,
      anon_sym_RBRACK,
    ACTIONS(2357), 1,
      anon_sym_DOT_DOT,
    STATE(1468), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [70899] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(527), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1656), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [70970] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2359), 1,
      anon_sym_RPAREN,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71041] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2361), 1,
      anon_sym_LPAREN,
    STATE(994), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(383), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(381), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71084] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2363), 1,
      anon_sym_COMMA,
    ACTIONS(2365), 1,
      anon_sym_RBRACK,
    ACTIONS(2367), 1,
      anon_sym_DOT_DOT,
    STATE(1431), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71147] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2369), 1,
      anon_sym_COMMA,
    ACTIONS(2371), 1,
      anon_sym_RBRACK,
    ACTIONS(2373), 1,
      anon_sym_DOT_DOT,
    STATE(1459), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71210] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(158), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1634), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71281] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2317), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1454), 1,
      sym__pattern,
    STATE(1690), 1,
      sym_use_assignment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1415), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71352] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    ACTIONS(2375), 1,
      anon_sym_RPAREN,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71423] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2377), 1,
      anon_sym_COMMA,
    ACTIONS(2379), 1,
      anon_sym_RBRACK,
    ACTIONS(2381), 1,
      anon_sym_DOT_DOT,
    STATE(1420), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71486] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(759), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1570), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71557] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(759), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1548), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71628] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(2383), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT,
  [71685] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1886), 1,
      anon_sym_LPAREN,
    ACTIONS(2385), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(682), 17,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71728] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1890), 1,
      anon_sym_LPAREN,
    ACTIONS(2387), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(660), 17,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71771] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(352), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1553), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71842] = 20,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(188), 1,
      sym__assignment,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1545), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [71913] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2389), 1,
      anon_sym_COMMA,
    ACTIONS(2391), 1,
      anon_sym_RBRACK,
    ACTIONS(2393), 1,
      anon_sym_DOT_DOT,
    STATE(1448), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [71976] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2395), 1,
      anon_sym_RPAREN,
    ACTIONS(2397), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1609), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [72042] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2399), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1515), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [72108] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2401), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1594), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [72174] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(511), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(509), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72212] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(613), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(611), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72250] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2403), 1,
      anon_sym_LBRACE,
    ACTIONS(2405), 1,
      anon_sym_COMMA,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    STATE(1625), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72310] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(609), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(607), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72348] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2427), 1,
      anon_sym_COMMA,
    ACTIONS(2429), 1,
      anon_sym_RPAREN,
    STATE(1583), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72408] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(605), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(603), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72446] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2431), 1,
      anon_sym_COMMA,
    ACTIONS(2433), 1,
      anon_sym_RPAREN,
    STATE(1559), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72506] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2435), 1,
      anon_sym_LPAREN,
    ACTIONS(2437), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(662), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(660), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72548] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2439), 1,
      anon_sym_RPAREN,
    ACTIONS(2441), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1803), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [72614] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2443), 1,
      anon_sym_LPAREN,
    ACTIONS(2445), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(684), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(682), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72656] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(2447), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72698] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(625), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72740] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2451), 1,
      anon_sym_COMMA,
    ACTIONS(2453), 1,
      anon_sym_RPAREN,
    STATE(1622), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72800] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    ACTIONS(2455), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1803), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [72866] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(601), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(599), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72904] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2457), 1,
      anon_sym_COMMA,
    ACTIONS(2459), 1,
      anon_sym_RPAREN,
    STATE(1640), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [72964] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2461), 1,
      anon_sym_COMMA,
    ACTIONS(2463), 1,
      anon_sym_RPAREN,
    STATE(1650), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73024] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(597), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(595), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73062] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2465), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73128] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(523), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(521), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73166] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(505), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(503), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73204] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(527), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(525), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73242] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(571), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(569), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73280] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2004), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1309), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [73348] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2467), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73414] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(547), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(545), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73452] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(559), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(557), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73490] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2469), 1,
      anon_sym_COMMA,
    ACTIONS(2471), 1,
      anon_sym_RPAREN,
    STATE(1589), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73550] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2473), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73616] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(565), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(563), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73654] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(587), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(585), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73692] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2475), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73758] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(389), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(387), 18,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [73796] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2477), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73862] = 19,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1828), 1,
      sym_float,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    STATE(1351), 1,
      sym__pattern,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1386), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [73930] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    ACTIONS(2479), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [73996] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2481), 1,
      anon_sym_COMMA,
    ACTIONS(2483), 1,
      anon_sym_RPAREN,
    STATE(1487), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74056] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    ACTIONS(2485), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1803), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74122] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    ACTIONS(2487), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    STATE(1803), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74188] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2489), 1,
      anon_sym_GT_GT,
    ACTIONS(2491), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1498), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74251] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(654), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74290] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2493), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74353] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2495), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1246), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [74418] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(644), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(642), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74455] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2497), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74518] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(648), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(646), 17,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74555] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1803), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74618] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(668), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(666), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74655] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(828), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74710] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(824), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74765] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(672), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(670), 17,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74802] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2501), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2499), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1749), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [74863] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(734), 3,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74916] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(734), 4,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [74967] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 6,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [75016] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 10,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [75063] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(750), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75118] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(814), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75173] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(676), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(674), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75210] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1824), 1,
      anon_sym_DASH,
    ACTIONS(1826), 1,
      anon_sym_DQUOTE,
    ACTIONS(1832), 1,
      sym__decimal,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(1836), 1,
      sym__name,
    ACTIONS(1838), 1,
      sym__upname,
    ACTIONS(2495), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1304), 1,
      sym_identifier,
    STATE(1342), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(1830), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1246), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [75275] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2503), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1599), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75338] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2505), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75401] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2507), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75464] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(690), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(688), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75501] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(658), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(656), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75538] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2509), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1641), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75601] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2511), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75664] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(652), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(650), 17,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75701] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2513), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75756] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2397), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1800), 1,
      sym_constant_record_argument,
    STATE(1881), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75819] = 18,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(1818), 1,
      anon_sym_POUND,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_LT_LT,
    ACTIONS(1834), 1,
      sym__discard_name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2495), 1,
      sym_float,
    STATE(1256), 1,
      sym__pattern_binary_expression,
    STATE(1290), 1,
      sym_identifier,
    STATE(1316), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1180), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1246), 7,
      sym__pattern_expression,
      sym_record_pattern,
      sym_tuple_pattern,
      sym__pattern_bit_string,
      sym_list_pattern,
      sym_integer,
      sym_discard,
  [75884] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(623), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [75923] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    ACTIONS(2515), 1,
      anon_sym_GT_GT,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [75986] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2517), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76041] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1630), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [76104] = 17,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(724), 1,
      sym__name,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2441), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1483), 1,
      sym_attribute_value,
    STATE(1912), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1717), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [76167] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2383), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76222] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2411), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2413), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2421), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(810), 2,
      anon_sym_LBRACE,
      anon_sym_COMMA,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2415), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2417), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2419), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76277] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2519), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76332] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(640), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(621), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(619), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76371] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 6,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
    ACTIONS(734), 13,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
  [76412] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2407), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2409), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2423), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2425), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 11,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
  [76457] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(850), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(848), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76493] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(858), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(856), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76529] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(792), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(790), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76565] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2521), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76619] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2523), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76673] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(814), 1,
      anon_sym_DASH_GT,
    ACTIONS(2529), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76727] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2545), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76781] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2547), 1,
      anon_sym_RBRACK,
    ACTIONS(2549), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1598), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [76841] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2551), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76895] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2553), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [76949] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2555), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77003] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2557), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77057] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2559), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77111] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(838), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(836), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77147] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2561), 1,
      anon_sym_RBRACK,
    ACTIONS(2563), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1635), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [77207] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2565), 1,
      anon_sym_RPAREN,
    ACTIONS(2567), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1607), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [77267] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2569), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77321] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2571), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77375] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2573), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77429] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2575), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77483] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2577), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77537] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(750), 1,
      anon_sym_DASH_GT,
    ACTIONS(2529), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77591] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2491), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    STATE(1748), 1,
      sym_constant_bit_string_segment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1638), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [77651] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2579), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77705] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2581), 1,
      anon_sym_RPAREN,
    ACTIONS(2583), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [77765] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2585), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77819] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2587), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77873] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2589), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77927] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2591), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [77981] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2593), 1,
      anon_sym_RPAREN,
    ACTIONS(2595), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1511), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78041] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2597), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78101] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2599), 1,
      anon_sym_RBRACK,
    ACTIONS(2601), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1497), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78161] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2603), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78221] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(788), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(786), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78257] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2605), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78317] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2607), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78377] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(854), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(852), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78413] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2609), 1,
      anon_sym_LPAREN,
    STATE(1123), 1,
      sym_constant_record_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1628), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1624), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [78453] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(804), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(802), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78489] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(846), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(844), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78525] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(821), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(818), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78561] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(796), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(794), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78597] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2611), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78657] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(810), 1,
      anon_sym_DASH_GT,
    ACTIONS(2529), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78711] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(800), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(798), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78747] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2613), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78807] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2615), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78861] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2617), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [78921] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(784), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(782), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [78957] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2619), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79011] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2621), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79065] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2623), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79119] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2625), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79173] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2627), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [79233] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2629), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79287] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(842), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(840), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79323] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(828), 1,
      anon_sym_DASH_GT,
    ACTIONS(2529), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79377] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2631), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [79437] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(824), 1,
      anon_sym_DASH_GT,
    ACTIONS(2529), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79491] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(706), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(704), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79527] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(734), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79563] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(736), 6,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
    ACTIONS(734), 12,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
  [79603] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2531), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(734), 2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79655] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2633), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79709] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2635), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79763] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2533), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(734), 3,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79813] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2535), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2537), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 5,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [79861] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2539), 1,
      anon_sym_PIPE_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 9,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [79907] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2525), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2527), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2541), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(736), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2543), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
    ACTIONS(734), 10,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
  [79951] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(734), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [79987] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2637), 1,
      anon_sym_RPAREN,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80047] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2639), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80101] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2641), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80155] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(710), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(708), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80191] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(702), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(700), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80227] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(740), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(738), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80263] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(744), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(742), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80299] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2643), 1,
      anon_sym_RPAREN,
    ACTIONS(2645), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1604), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80359] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2255), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2257), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2265), 1,
      anon_sym_PIPE_GT,
    ACTIONS(2647), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2251), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2253), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2259), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2267), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2261), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2263), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2269), 4,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80413] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2649), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1882), 11,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym__decimal,
      sym__name,
    ACTIONS(1880), 12,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [80451] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(780), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(778), 16,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PIPE_GT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
      anon_sym_LT_GT,
  [80487] = 16,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    ACTIONS(2651), 1,
      anon_sym_RBRACK,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80547] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1996), 11,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym__decimal,
      sym__name,
    ACTIONS(1994), 12,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [80582] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2583), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1531), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80639] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2653), 1,
      sym_float,
    STATE(1299), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1315), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80696] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2655), 1,
      sym_float,
    STATE(1299), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1333), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80753] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(2657), 1,
      anon_sym_POUND,
    ACTIONS(2659), 1,
      anon_sym_LBRACK,
    ACTIONS(2661), 1,
      anon_sym_LT_LT,
    ACTIONS(2663), 1,
      anon_sym_DASH,
    ACTIONS(2665), 1,
      sym_float,
    STATE(713), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(706), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(785), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80810] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(2657), 1,
      anon_sym_POUND,
    ACTIONS(2659), 1,
      anon_sym_LBRACK,
    ACTIONS(2661), 1,
      anon_sym_LT_LT,
    ACTIONS(2663), 1,
      anon_sym_DASH,
    ACTIONS(2667), 1,
      sym_float,
    STATE(713), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(706), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(821), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80867] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2669), 1,
      sym_float,
    STATE(1299), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1337), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [80924] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2674), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2676), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2671), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [80961] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(2657), 1,
      anon_sym_POUND,
    ACTIONS(2659), 1,
      anon_sym_LBRACK,
    ACTIONS(2661), 1,
      anon_sym_LT_LT,
    ACTIONS(2663), 1,
      anon_sym_DASH,
    ACTIONS(2679), 1,
      sym_float,
    STATE(713), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(706), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(758), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [81018] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(2657), 1,
      anon_sym_POUND,
    ACTIONS(2659), 1,
      anon_sym_LBRACK,
    ACTIONS(2661), 1,
      anon_sym_LT_LT,
    ACTIONS(2663), 1,
      anon_sym_DASH,
    ACTIONS(2681), 1,
      sym_float,
    STATE(713), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(706), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(828), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [81075] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2683), 1,
      sym_float,
    STATE(1380), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1757), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [81132] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2685), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2676), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2671), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81169] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2055), 11,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_todo,
      anon_sym_panic,
      anon_sym_echo,
      anon_sym_case,
      anon_sym_let,
      anon_sym_assert,
      anon_sym_use,
      sym__decimal,
      sym__name,
    ACTIONS(2053), 12,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LBRACK,
      anon_sym_LT_LT,
      anon_sym_BANG,
      anon_sym_DQUOTE,
      sym_float,
      sym__hex,
      sym__octal,
      sym__binary,
      sym__upname,
  [81204] = 15,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2000), 1,
      anon_sym_DASH,
    ACTIONS(2227), 1,
      anon_sym_POUND,
    ACTIONS(2229), 1,
      anon_sym_LBRACK,
    ACTIONS(2231), 1,
      anon_sym_LT_LT,
    ACTIONS(2687), 1,
      sym_float,
    STATE(1299), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(604), 2,
      sym_constructor_name,
      sym_remote_constructor_name,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1336), 8,
      sym__constant_value,
      sym_constant_tuple,
      sym_constant_list,
      sym__constant_bit_string,
      sym_constant_record,
      sym_constant_field_access,
      sym_string,
      sym_integer,
  [81261] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2693), 1,
      anon_sym_DASH_GT,
    ACTIONS(2695), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2697), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2691), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2699), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2705), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2701), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2703), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [81311] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1680), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1678), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81345] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2709), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2711), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81379] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2713), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2715), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81413] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2713), 6,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
    ACTIONS(2715), 11,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
  [81451] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2697), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2691), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2699), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2705), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2715), 2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2701), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2703), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [81499] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(581), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(579), 21,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [81533] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2121), 22,
      anon_sym_if,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [81565] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2691), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2699), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2705), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2715), 3,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(2701), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2703), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [81611] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2691), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2705), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2701), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2703), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2715), 5,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [81655] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2689), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2691), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2705), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2707), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2713), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2715), 9,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [81697] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2719), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2717), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81731] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2723), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2721), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81765] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2727), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2725), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81799] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2713), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2715), 9,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [81841] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1662), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1660), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81875] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2727), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2725), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81909] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1684), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1682), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81943] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1688), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1686), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [81977] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2737), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2739), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
    ACTIONS(2715), 5,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
  [82021] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2741), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2715), 3,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2737), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2739), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [82067] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2743), 1,
      anon_sym_RBRACE,
    ACTIONS(2745), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2747), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2741), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2737), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2739), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [82117] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2747), 1,
      anon_sym_AMP_AMP,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2715), 2,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2741), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2737), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2739), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [82165] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2713), 6,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
    ACTIONS(2715), 11,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
  [82203] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2713), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2715), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82237] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2709), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2711), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82271] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2745), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2747), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2749), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2729), 2,
      anon_sym_SLASH,
      anon_sym_STAR,
    ACTIONS(2731), 2,
      anon_sym_DASH,
      anon_sym_PLUS,
    ACTIONS(2733), 2,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
    ACTIONS(2741), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2735), 3,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
    ACTIONS(2737), 4,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
    ACTIONS(2739), 4,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
  [82321] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1700), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1698), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82355] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1672), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1670), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82389] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1720), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1718), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82423] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1728), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1726), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82457] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2723), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2721), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82491] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2719), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(2717), 14,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82525] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1736), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1734), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82559] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1712), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1710), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82593] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1696), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1694), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82627] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1692), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1690), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82661] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1666), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1664), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82695] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1716), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1714), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82729] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1676), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1674), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82763] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1724), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1722), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82797] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1740), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1738), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82831] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1732), 8,
      anon_sym_SLASH,
      anon_sym_DASH,
      anon_sym_LT,
      anon_sym_LT_EQ,
      anon_sym_GT,
      anon_sym_GT_EQ,
      anon_sym_PLUS,
      anon_sym_STAR,
    ACTIONS(1730), 14,
      anon_sym_RBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_AMP_AMP,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_DOT,
      anon_sym_LT_EQ_DOT,
      anon_sym_GT_DOT,
      anon_sym_GT_EQ_DOT,
      anon_sym_PLUS_DOT,
      anon_sym_DASH_DOT,
      anon_sym_STAR_DOT,
      anon_sym_SLASH_DOT,
      anon_sym_PERCENT,
  [82865] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2751), 1,
      anon_sym_RBRACE,
    ACTIONS(2753), 1,
      anon_sym_AT,
    ACTIONS(2756), 1,
      anon_sym_import,
    ACTIONS(2759), 1,
      anon_sym_type,
    ACTIONS(2762), 1,
      anon_sym_const,
    ACTIONS(2765), 1,
      anon_sym_fn,
    ACTIONS(2768), 1,
      anon_sym_external,
    ACTIONS(2771), 1,
      sym_visibility_modifier,
    ACTIONS(2774), 1,
      sym_opacity_modifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1148), 10,
      sym__module_statement,
      sym_attribute,
      sym_import,
      sym_constant,
      sym_external_type,
      sym_external_function,
      sym_function,
      sym_type_definition,
      sym_type_alias,
      aux_sym_target_group_repeat1,
  [82912] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2777), 1,
      anon_sym_RBRACE,
    ACTIONS(2779), 1,
      anon_sym_AT,
    ACTIONS(2781), 1,
      anon_sym_import,
    ACTIONS(2783), 1,
      anon_sym_type,
    ACTIONS(2785), 1,
      anon_sym_const,
    ACTIONS(2787), 1,
      anon_sym_fn,
    ACTIONS(2789), 1,
      anon_sym_external,
    ACTIONS(2791), 1,
      sym_visibility_modifier,
    ACTIONS(2793), 1,
      sym_opacity_modifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1150), 10,
      sym__module_statement,
      sym_attribute,
      sym_import,
      sym_constant,
      sym_external_type,
      sym_external_function,
      sym_function,
      sym_type_definition,
      sym_type_alias,
      aux_sym_target_group_repeat1,
  [82959] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2779), 1,
      anon_sym_AT,
    ACTIONS(2781), 1,
      anon_sym_import,
    ACTIONS(2783), 1,
      anon_sym_type,
    ACTIONS(2785), 1,
      anon_sym_const,
    ACTIONS(2787), 1,
      anon_sym_fn,
    ACTIONS(2789), 1,
      anon_sym_external,
    ACTIONS(2791), 1,
      sym_visibility_modifier,
    ACTIONS(2793), 1,
      sym_opacity_modifier,
    ACTIONS(2795), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1148), 10,
      sym__module_statement,
      sym_attribute,
      sym_import,
      sym_constant,
      sym_external_type,
      sym_external_function,
      sym_function,
      sym_type_definition,
      sym_type_alias,
      aux_sym_target_group_repeat1,
  [83006] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2797), 1,
      anon_sym_RPAREN,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2805), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1581), 1,
      sym_data_constructor_argument,
    STATE(1826), 1,
      sym_identifier,
    STATE(1858), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1780), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83053] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2805), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2809), 1,
      anon_sym_RPAREN,
    STATE(1673), 1,
      sym_data_constructor_argument,
    STATE(1826), 1,
      sym_identifier,
    STATE(1858), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1780), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83100] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2805), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2811), 1,
      anon_sym_RPAREN,
    STATE(1673), 1,
      sym_data_constructor_argument,
    STATE(1826), 1,
      sym_identifier,
    STATE(1858), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1780), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83147] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2805), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1673), 1,
      sym_data_constructor_argument,
    STATE(1826), 1,
      sym_identifier,
    STATE(1858), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1780), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83191] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2813), 1,
      anon_sym_RPAREN,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1747), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83235] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2817), 1,
      anon_sym_RPAREN,
    STATE(1747), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83279] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2819), 1,
      anon_sym_RPAREN,
    STATE(1711), 1,
      sym_identifier,
    STATE(1720), 1,
      sym_external_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1710), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83323] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2821), 1,
      anon_sym_RPAREN,
    STATE(1747), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83367] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2823), 1,
      anon_sym_RPAREN,
    STATE(1573), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83411] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2825), 1,
      anon_sym_RPAREN,
    STATE(1747), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83455] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2827), 1,
      anon_sym_LPAREN,
    STATE(1205), 1,
      sym_type_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1767), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [83485] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(607), 2,
      anon_sym_COLON,
      anon_sym_DOT,
    ACTIONS(1812), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [83513] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1773), 16,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [83539] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2829), 1,
      anon_sym_RPAREN,
    STATE(1711), 1,
      sym_identifier,
    STATE(1720), 1,
      sym_external_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1710), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83583] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2831), 1,
      anon_sym_RPAREN,
    STATE(1518), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83627] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2833), 1,
      anon_sym_RPAREN,
    STATE(1636), 1,
      sym_external_function_parameter,
    STATE(1711), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1710), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83671] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1795), 15,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [83696] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2835), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83737] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2837), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1517), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83778] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2839), 1,
      anon_sym_RPAREN,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1728), 1,
      sym_constant_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1807), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [83821] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2845), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83862] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1711), 1,
      sym_identifier,
    STATE(1720), 1,
      sym_external_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1710), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83903] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2847), 1,
      anon_sym_RPAREN,
    STATE(1728), 1,
      sym_constant_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1807), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [83946] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2849), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [83987] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1747), 1,
      sym_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1791), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84028] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2851), 1,
      anon_sym_RPAREN,
    STATE(1563), 1,
      sym_constant_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1807), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84071] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2853), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84112] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2855), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1577), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84153] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2857), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84194] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2861), 1,
      anon_sym_LPAREN,
    STATE(1262), 1,
      sym_record_pattern_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2859), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [84223] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2863), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1575), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84264] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    ACTIONS(2865), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84305] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1955), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84343] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2867), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1566), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84383] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2197), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [84407] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(816), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84445] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1782), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84483] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1941), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84521] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2171), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [84545] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1937), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84583] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2879), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1523), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84623] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2881), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1766), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84663] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1477), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84701] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1728), 1,
      sym_constant_type_argument,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1807), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84741] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1825), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84779] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2115), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [84803] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1827), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84841] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1970), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [84865] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1982), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [84889] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1311), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [84927] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2883), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1766), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [84967] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1335), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85005] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1225), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85043] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2885), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1766), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [85083] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2095), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85107] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(752), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85145] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(781), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85183] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1933), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85221] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(837), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85259] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(760), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85297] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1738), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85335] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(843), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85373] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1731), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85411] = 11,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    ACTIONS(2887), 1,
      anon_sym_RPAREN,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1766), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [85451] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2215), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85475] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1939), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85513] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2869), 1,
      anon_sym_POUND,
    ACTIONS(2871), 1,
      anon_sym_fn,
    ACTIONS(2873), 1,
      sym__discard_name,
    ACTIONS(2875), 1,
      sym__name,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(688), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(777), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85551] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1958), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85575] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1292), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85613] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1907), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85651] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2057), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85675] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1221), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85713] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1325), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85751] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1889), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85789] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1962), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85813] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1966), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85837] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1295), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85875] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1923), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85913] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2799), 1,
      anon_sym_POUND,
    ACTIONS(2801), 1,
      anon_sym_fn,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2815), 1,
      sym__name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1161), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1929), 6,
      sym__type,
      sym_type_hole,
      sym_tuple_type,
      sym_function_type,
      sym_type,
      sym_type_var,
  [85951] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2167), 14,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      anon_sym_LT_DASH,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [85975] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1766), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [86012] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2889), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86035] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2891), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86058] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2893), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86081] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2895), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86104] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2897), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86127] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2899), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86150] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2895), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86173] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2901), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86196] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2903), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86219] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2905), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86242] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2907), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86265] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2909), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86288] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2911), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86311] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1493), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [86348] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2913), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86371] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2915), 1,
      anon_sym_LPAREN,
    STATE(1274), 1,
      sym_type_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1761), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86398] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2917), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86421] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2919), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86444] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2921), 1,
      anon_sym_SLASH,
    STATE(1261), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1755), 11,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86471] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2923), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86494] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1820), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [86531] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2925), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86554] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2925), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86577] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2927), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86600] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2929), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86623] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2931), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86646] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2921), 1,
      anon_sym_SLASH,
    STATE(1250), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1742), 11,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86673] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2933), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86696] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2913), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86719] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2935), 1,
      anon_sym_SLASH,
    STATE(1261), 1,
      aux_sym_module_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1748), 11,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86746] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2938), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86769] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2803), 1,
      sym__discard_name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(2841), 1,
      anon_sym_POUND,
    ACTIONS(2843), 1,
      anon_sym_fn,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1419), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
    STATE(1565), 5,
      sym__constant_type,
      sym_constant_tuple_type,
      sym_constant_function_type,
      sym_constant_type,
      sym_type_hole,
  [86806] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2940), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86829] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2942), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86852] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2944), 13,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86875] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1750), 1,
      anon_sym_SLASH,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1748), 11,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86899] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1904), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86920] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2948), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1946), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86945] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1918), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [86966] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2950), 11,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DASH_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [86987] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2952), 1,
      anon_sym_RPAREN,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1811), 1,
      sym_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87028] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1940), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87049] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1954), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87070] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    ACTIONS(2958), 1,
      anon_sym_RPAREN,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1811), 1,
      sym_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87111] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2960), 1,
      anon_sym_LPAREN,
    STATE(1313), 1,
      sym__attribute_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1789), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87136] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1808), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87157] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    ACTIONS(2962), 1,
      anon_sym_RPAREN,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1811), 1,
      sym_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87198] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    ACTIONS(2964), 1,
      anon_sym_RPAREN,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1811), 1,
      sym_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87239] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1856), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87260] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2966), 1,
      anon_sym_DOT,
    ACTIONS(2968), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1777), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87285] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2970), 1,
      anon_sym_LBRACE,
    ACTIONS(2972), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1910), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87310] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1894), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87331] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2974), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1928), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87356] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2978), 1,
      anon_sym_LBRACE,
    ACTIONS(2980), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1860), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87381] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1842), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_EQ,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87402] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    ACTIONS(2982), 1,
      anon_sym_RPAREN,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1491), 1,
      sym_function_parameter,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87443] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1900), 11,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_DASH_GT,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87464] = 13,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    ACTIONS(2984), 1,
      anon_sym_RPAREN,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1488), 1,
      sym_function_parameter,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87505] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2988), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2986), 9,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [87527] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1852), 10,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87547] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2990), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2219), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87569] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2992), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1922), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87591] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1872), 10,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87611] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2994), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2069), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87633] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2996), 1,
      anon_sym_DASH,
    ACTIONS(2998), 1,
      sym__name,
    STATE(593), 1,
      sym_constructor_name,
    STATE(1118), 1,
      sym_integer,
    STATE(1135), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [87667] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1876), 10,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87687] = 12,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(2956), 1,
      sym__name,
    STATE(1441), 1,
      sym_label,
    STATE(1442), 1,
      sym__discard_param,
    STATE(1447), 1,
      sym__name_param,
    STATE(1449), 1,
      sym__labeled_name_param,
    STATE(1470), 1,
      sym__labeled_discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1811), 1,
      sym_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [87725] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3000), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1846), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87747] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(251), 1,
      sym__upname,
    ACTIONS(3002), 1,
      anon_sym_DASH,
    ACTIONS(3004), 1,
      sym__name,
    STATE(678), 1,
      sym_label,
    STATE(949), 1,
      sym_constructor_name,
    STATE(1121), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [87781] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1804), 10,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_as,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87801] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2061), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87820] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2161), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87839] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3006), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2986), 8,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [87860] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2006), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87879] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(607), 9,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DOT,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [87898] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1946), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87917] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(611), 9,
      anon_sym_if,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [87936] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3010), 1,
      anon_sym_as,
    ACTIONS(3012), 1,
      anon_sym_LT_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3008), 7,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_DOT_DOT,
  [87959] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2149), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87978] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1990), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [87997] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2145), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88016] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2065), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88035] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2201), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88054] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2091), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88073] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3014), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2986), 8,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
      anon_sym_LT_GT,
      anon_sym_DOT_DOT,
  [88094] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2153), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88113] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2141), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88132] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2014), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88151] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2137), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88170] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(569), 9,
      anon_sym_if,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [88189] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1974), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88208] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2075), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88227] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2129), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88246] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2211), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88265] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2099), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88284] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2205), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88303] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1978), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88322] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    ACTIONS(3016), 1,
      sym__name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
    STATE(1966), 3,
      sym__pattern_bit_string_segment_argument,
      sym_integer,
      sym_identifier,
  [88349] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2103), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88368] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2111), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88387] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2010), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88406] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2185), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88425] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2181), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88444] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2175), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88463] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2193), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88482] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1986), 9,
      anon_sym_RBRACE,
      anon_sym_AT,
      anon_sym_import,
      anon_sym_type,
      anon_sym_const,
      anon_sym_fn,
      anon_sym_external,
      sym_visibility_modifier,
      sym_opacity_modifier,
  [88501] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(3018), 1,
      anon_sym_RPAREN,
    STATE(1444), 1,
      sym__name_param,
    STATE(1455), 1,
      sym__discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1756), 1,
      sym_anonymous_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88533] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(3020), 1,
      anon_sym_RPAREN,
    STATE(1444), 1,
      sym__name_param,
    STATE(1455), 1,
      sym__discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1756), 1,
      sym_anonymous_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88565] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(704), 8,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [88583] = 10,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(3022), 1,
      anon_sym_RPAREN,
    STATE(1444), 1,
      sym__name_param,
    STATE(1455), 1,
      sym__discard_param,
    STATE(1610), 1,
      sym_anonymous_function_parameter,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88615] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3024), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2986), 7,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [88635] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(856), 8,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [88653] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(603), 8,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_as,
      anon_sym_DASH_GT,
      anon_sym_LT_GT,
      anon_sym_PIPE,
      anon_sym_LT_DASH,
  [88671] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3026), 1,
      anon_sym_RBRACE,
    ACTIONS(3028), 1,
      anon_sym_type,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    STATE(1732), 1,
      sym_unqualified_import,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88700] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    STATE(1444), 1,
      sym__name_param,
    STATE(1455), 1,
      sym__discard_param,
    STATE(1613), 1,
      sym_identifier,
    STATE(1619), 1,
      sym_discard,
    STATE(1756), 1,
      sym_anonymous_function_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88729] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    ACTIONS(3030), 1,
      anon_sym_RBRACE,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    STATE(1732), 1,
      sym_unqualified_import,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88758] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    ACTIONS(3032), 1,
      anon_sym_RBRACE,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    STATE(1732), 1,
      sym_unqualified_import,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88787] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    ACTIONS(3034), 1,
      anon_sym_RBRACE,
    STATE(1501), 1,
      sym_unqualified_import,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88816] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    ACTIONS(3036), 1,
      anon_sym_RBRACE,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    STATE(1732), 1,
      sym_unqualified_import,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88845] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3038), 7,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_DASH_GT,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
  [88862] = 9,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    ACTIONS(3040), 1,
      anon_sym_RBRACE,
    STATE(1659), 1,
      sym_unqualified_import,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [88891] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3042), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3046), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1361), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [88913] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    STATE(1852), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [88935] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1331), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [88959] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3048), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [88981] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3052), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3054), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1385), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89003] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(97), 1,
      sym__decimal,
    ACTIONS(2996), 1,
      anon_sym_DASH,
    STATE(693), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(95), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89025] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2877), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(791), 1,
      sym_type_name,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(687), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89049] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3056), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3058), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1362), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89071] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3060), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89093] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3062), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89115] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(3064), 1,
      anon_sym_DASH,
    STATE(242), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89137] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1793), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89161] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3066), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3068), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1356), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89183] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1182), 1,
      sym__decimal,
    ACTIONS(3070), 1,
      anon_sym_DASH,
    STATE(451), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1180), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89205] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(59), 1,
      sym__decimal,
    ACTIONS(3072), 1,
      anon_sym_DASH,
    STATE(86), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(57), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89227] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1707), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89251] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3074), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89273] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1282), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89297] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1307), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89321] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1725), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89345] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3076), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3078), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1369), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89367] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3080), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3082), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1379), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89389] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(247), 1,
      sym__decimal,
    ACTIONS(3002), 1,
      anon_sym_DASH,
    STATE(946), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(245), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89411] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1682), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89435] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3084), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89457] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3086), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3088), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1382), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89479] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3090), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89501] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3092), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1846), 5,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_GT_GT,
  [89519] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1018), 1,
      sym__decimal,
    ACTIONS(3094), 1,
      anon_sym_DASH,
    STATE(605), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1016), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89541] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3096), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89563] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2877), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(734), 1,
      sym_type_name,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(687), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89587] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2877), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(787), 1,
      sym_type_name,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(687), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89611] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3050), 1,
      sym_quoted_content,
    ACTIONS(3098), 1,
      anon_sym_DQUOTE2,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89633] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3100), 1,
      anon_sym_as,
    ACTIONS(3102), 1,
      anon_sym_LT_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3008), 4,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
  [89653] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3104), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3106), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3044), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1377), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89675] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3108), 1,
      anon_sym_DQUOTE2,
    ACTIONS(3113), 1,
      sym_quoted_content,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3110), 2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
    STATE(1388), 2,
      sym_escape_sequence,
      aux_sym_string_repeat1,
  [89697] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    STATE(1818), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89719] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2877), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(745), 1,
      sym_type_name,
    STATE(1896), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(687), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89743] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(427), 1,
      sym__decimal,
    ACTIONS(2081), 1,
      anon_sym_DASH,
    STATE(242), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(425), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89765] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1102), 1,
      sym__decimal,
    ACTIONS(3116), 1,
      anon_sym_DASH,
    STATE(30), 1,
      sym_integer,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1100), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [89787] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1269), 1,
      sym_type_name,
    STATE(1826), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1247), 2,
      sym_type_identifier,
      sym_remote_type_identifier,
  [89811] = 8,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2807), 1,
      sym__upname,
    ACTIONS(3028), 1,
      anon_sym_type,
    STATE(1661), 1,
      sym_identifier,
    STATE(1662), 1,
      sym_type_identifier,
    STATE(1732), 1,
      sym_unqualified_import,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [89837] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(509), 1,
      sym__discard_name,
    ACTIONS(511), 1,
      sym__name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(607), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [89856] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2295), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3118), 1,
      anon_sym_COMMA,
    ACTIONS(3120), 1,
      anon_sym_RBRACK,
    STATE(1400), 1,
      aux_sym_case_clause_pattern_repeat1,
    STATE(1860), 1,
      sym_list_pattern_tail,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [89879] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3122), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2383), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT,
  [89898] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3125), 1,
      anon_sym_COMMA,
    STATE(1398), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3128), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_DOT_DOT,
  [89917] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(3130), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1884), 2,
      sym_identifier,
      sym_discard,
  [89938] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2295), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3132), 1,
      anon_sym_COMMA,
    ACTIONS(3134), 1,
      anon_sym_RBRACK,
    STATE(1398), 1,
      aux_sym_case_clause_pattern_repeat1,
    STATE(1930), 1,
      sym_list_pattern_tail,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [89961] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1522), 1,
      anon_sym_LPAREN,
    ACTIONS(1888), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(682), 3,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_GT_GT,
  [89980] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1853), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90001] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1908), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90022] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2002), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3136), 1,
      anon_sym_COMMA,
    ACTIONS(3138), 1,
      anon_sym_RPAREN,
    STATE(1412), 1,
      aux_sym_record_pattern_arguments_repeat1,
    STATE(1911), 1,
      sym_pattern_spread,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90045] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1910), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90066] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3142), 1,
      anon_sym_COMMA,
    STATE(1408), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3140), 3,
      anon_sym_if,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
  [90085] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3144), 1,
      anon_sym_RBRACE,
    ACTIONS(3146), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1407), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90106] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3149), 1,
      anon_sym_COMMA,
    STATE(1408), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3128), 3,
      anon_sym_if,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
  [90125] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(3152), 1,
      anon_sym_RBRACE,
    STATE(1464), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1407), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90146] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3156), 1,
      anon_sym_COMMA,
    STATE(1406), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3154), 3,
      anon_sym_if,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
  [90165] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1518), 1,
      anon_sym_LPAREN,
    ACTIONS(1892), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(660), 3,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_GT_GT,
  [90184] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1998), 1,
      anon_sym_RPAREN,
    ACTIONS(2002), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3158), 1,
      anon_sym_COMMA,
    STATE(1438), 1,
      aux_sym_record_pattern_arguments_repeat1,
    STATE(1847), 1,
      sym_pattern_spread,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90207] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1920), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90228] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1956), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90249] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3100), 1,
      anon_sym_as,
    ACTIONS(3160), 1,
      anon_sym_LT_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3008), 3,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_LT_DASH,
  [90268] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3162), 1,
      anon_sym_type,
    ACTIONS(3164), 1,
      anon_sym_const,
    ACTIONS(3166), 1,
      anon_sym_fn,
    ACTIONS(3168), 1,
      anon_sym_external,
    ACTIONS(3170), 1,
      sym_opacity_modifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90291] = 7,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3172), 1,
      anon_sym_type,
    ACTIONS(3174), 1,
      anon_sym_const,
    ACTIONS(3176), 1,
      anon_sym_fn,
    ACTIONS(3178), 1,
      anon_sym_external,
    ACTIONS(3180), 1,
      sym_opacity_modifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90314] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(1464), 1,
      sym_constructor_name,
    STATE(1888), 1,
      sym_data_constructors,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1409), 2,
      sym_data_constructor,
      aux_sym_data_constructors_repeat1,
  [90335] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3182), 1,
      anon_sym_LPAREN,
    STATE(1652), 1,
      sym_constant_type_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3184), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [90354] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(940), 1,
      anon_sym_RBRACK,
    ACTIONS(942), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3186), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90374] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1303), 2,
      sym_identifier,
      sym_discard,
  [90392] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(988), 1,
      anon_sym_RBRACK,
    ACTIONS(990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3188), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90412] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3192), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3190), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [90428] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3194), 1,
      anon_sym_DASH,
    STATE(1465), 1,
      aux_sym_expression_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2079), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [90446] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3198), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3196), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [90462] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3202), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3200), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [90478] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    STATE(1959), 1,
      sym_record_update_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90498] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    STATE(1893), 1,
      sym_record_update_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90518] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3208), 1,
      anon_sym_DASH,
    STATE(1451), 1,
      aux_sym_pattern_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3206), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [90536] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(930), 1,
      anon_sym_RBRACK,
    ACTIONS(932), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3210), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90556] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(936), 1,
      anon_sym_RBRACK,
    ACTIONS(938), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3212), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90576] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3216), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3214), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [90592] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3220), 1,
      anon_sym_PIPE,
    STATE(1460), 1,
      aux_sym_case_clause_patterns_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3218), 2,
      anon_sym_if,
      anon_sym_DASH_GT,
  [90610] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3224), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3222), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [90626] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    ACTIONS(3204), 1,
      sym__name,
    STATE(234), 1,
      sym_constructor_name,
    STATE(678), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90646] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3226), 1,
      anon_sym_DASH,
    STATE(1467), 1,
      aux_sym_constant_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2189), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [90664] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3228), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3231), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [90682] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3233), 1,
      anon_sym_COMMA,
    STATE(1438), 1,
      aux_sym_record_pattern_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3236), 2,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
  [90700] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    STATE(1940), 1,
      sym_record_update_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90720] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1854), 1,
      sym_record_update_arguments,
    STATE(1865), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90740] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(429), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    STATE(1647), 1,
      sym_identifier,
    STATE(1658), 1,
      sym_discard,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90760] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1701), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3238), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [90778] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3244), 1,
      anon_sym_DASH,
    STATE(1436), 1,
      aux_sym_constant_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3242), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [90796] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1716), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3246), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [90814] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    STATE(1890), 1,
      sym_record_update_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90834] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(99), 1,
      sym__name,
    ACTIONS(2954), 1,
      sym__discard_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(1322), 2,
      sym_identifier,
      sym_discard,
  [90852] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1705), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3238), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [90870] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(956), 1,
      anon_sym_RBRACK,
    ACTIONS(958), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3248), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90890] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1703), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [90908] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(63), 1,
      sym__upname,
    ACTIONS(3252), 1,
      sym__name,
    STATE(103), 1,
      sym_constructor_name,
    STATE(768), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90928] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3254), 1,
      anon_sym_DASH,
    STATE(1475), 1,
      aux_sym_pattern_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2165), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [90946] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1838), 1,
      sym_record_update_arguments,
    STATE(1865), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [90966] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(3256), 1,
      sym__discard_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(755), 2,
      sym_identifier,
      sym_discard,
  [90984] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1692), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3258), 2,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
  [91002] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1724), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3246), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [91020] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3262), 1,
      anon_sym_PIPE,
    STATE(1456), 1,
      aux_sym_case_clause_patterns_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3260), 2,
      anon_sym_if,
      anon_sym_DASH_GT,
  [91038] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    ACTIONS(3265), 1,
      anon_sym_RPAREN,
    STATE(1809), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91058] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1539), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    STATE(1906), 1,
      sym_record_update_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91078] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(944), 1,
      anon_sym_RBRACK,
    ACTIONS(946), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3267), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91098] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3269), 1,
      anon_sym_PIPE,
    STATE(1456), 1,
      aux_sym_case_clause_patterns_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2321), 2,
      anon_sym_if,
      anon_sym_DASH_GT,
  [91116] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3273), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3271), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [91132] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3277), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3275), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [91148] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3279), 4,
      sym_quoted_content,
      anon_sym_DQUOTE2,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
  [91162] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3283), 1,
      anon_sym_LPAREN,
    STATE(1681), 1,
      sym_data_constructor_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3281), 2,
      anon_sym_RBRACE,
      sym__upname,
  [91180] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3287), 1,
      anon_sym_DASH,
    STATE(1465), 1,
      aux_sym_expression_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3285), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [91198] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    ACTIONS(3290), 1,
      anon_sym_RPAREN,
    STATE(1809), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91218] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3294), 1,
      anon_sym_DASH,
    STATE(1467), 1,
      aux_sym_constant_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3292), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [91236] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(980), 1,
      anon_sym_RBRACK,
    ACTIONS(982), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91256] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3301), 1,
      sym__decimal,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3299), 3,
      sym__hex,
      sym__octal,
      sym__binary,
  [91272] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    STATE(1698), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3250), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [91290] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1754), 1,
      sym_anonymous_function_parameters,
    STATE(1762), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91310] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(61), 1,
      sym__name,
    ACTIONS(3256), 1,
      sym__discard_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    STATE(812), 2,
      sym_identifier,
      sym_discard,
  [91328] = 6,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(101), 1,
      sym__upname,
    ACTIONS(2998), 1,
      sym__name,
    STATE(593), 1,
      sym_constructor_name,
    STATE(678), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91348] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3307), 1,
      anon_sym_DASH,
    STATE(1424), 1,
      aux_sym_expression_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3305), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [91366] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3311), 1,
      anon_sym_DASH,
    STATE(1475), 1,
      aux_sym_pattern_bit_string_segment_options_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3309), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [91384] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(509), 1,
      anon_sym_COLON,
    ACTIONS(607), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(1812), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [91402] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3314), 4,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_LT_DASH,
  [91416] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2487), 1,
      anon_sym_RPAREN,
    ACTIONS(3316), 1,
      anon_sym_COMMA,
    STATE(1600), 1,
      aux_sym__attribute_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91433] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(766), 1,
      sym_external_function_body,
    STATE(1678), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91450] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3318), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [91463] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3320), 1,
      anon_sym_COMMA,
    ACTIONS(3322), 1,
      anon_sym_RPAREN,
    STATE(1496), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91480] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2821), 1,
      anon_sym_RPAREN,
    ACTIONS(3324), 1,
      anon_sym_COMMA,
    STATE(1556), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91497] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3326), 1,
      anon_sym_COMMA,
    ACTIONS(3328), 1,
      anon_sym_RPAREN,
    STATE(1478), 1,
      aux_sym__attribute_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91514] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2835), 1,
      anon_sym_RPAREN,
    ACTIONS(3330), 1,
      anon_sym_COMMA,
    STATE(1541), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91531] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3332), 1,
      anon_sym_LPAREN,
    ACTIONS(3334), 1,
      anon_sym_DASH_GT,
    STATE(1931), 1,
      sym_function_parameter_types,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91548] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3336), 1,
      anon_sym_COMMA,
    ACTIONS(3338), 1,
      anon_sym_RPAREN,
    STATE(1504), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91565] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1036), 1,
      anon_sym_RPAREN,
    ACTIONS(3340), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91582] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3342), 1,
      anon_sym_COMMA,
    ACTIONS(3344), 1,
      anon_sym_RPAREN,
    STATE(1664), 1,
      aux_sym_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91599] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2849), 1,
      anon_sym_RPAREN,
    ACTIONS(3346), 1,
      anon_sym_COMMA,
    STATE(1541), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91616] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3018), 1,
      anon_sym_RPAREN,
    ACTIONS(3348), 1,
      anon_sym_COMMA,
    STATE(1655), 1,
      aux_sym_anonymous_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91633] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3350), 1,
      anon_sym_COMMA,
    ACTIONS(3352), 1,
      anon_sym_RPAREN,
    STATE(1506), 1,
      aux_sym_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91650] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2887), 1,
      anon_sym_RPAREN,
    ACTIONS(3354), 1,
      anon_sym_COMMA,
    STATE(1616), 1,
      aux_sym_constant_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91667] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3356), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [91680] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3358), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [91693] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2839), 1,
      anon_sym_RPAREN,
    ACTIONS(3360), 1,
      anon_sym_COMMA,
    STATE(1637), 1,
      aux_sym_constant_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91710] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(880), 1,
      anon_sym_RPAREN,
    ACTIONS(3362), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91727] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3364), 1,
      anon_sym_COMMA,
    ACTIONS(3366), 1,
      anon_sym_RBRACK,
    STATE(1512), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91744] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3368), 1,
      anon_sym_COMMA,
    ACTIONS(3370), 1,
      anon_sym_GT_GT,
    STATE(1514), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91761] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3372), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [91774] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3374), 1,
      anon_sym_COMMA,
    ACTIONS(3377), 1,
      anon_sym_RPAREN,
    STATE(1500), 1,
      aux_sym_data_constructor_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91791] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3379), 1,
      anon_sym_RBRACE,
    ACTIONS(3381), 1,
      anon_sym_COMMA,
    STATE(1516), 1,
      aux_sym_unqualified_imports_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91808] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3383), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [91821] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3385), 1,
      anon_sym_RPAREN,
    ACTIONS(3387), 1,
      sym__name,
    STATE(1794), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91838] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3385), 1,
      anon_sym_RPAREN,
    ACTIONS(3389), 1,
      anon_sym_COMMA,
    STATE(1571), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91855] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2809), 1,
      anon_sym_RPAREN,
    ACTIONS(3391), 1,
      anon_sym_COMMA,
    STATE(1500), 1,
      aux_sym_data_constructor_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91872] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2958), 1,
      anon_sym_RPAREN,
    ACTIONS(3393), 1,
      anon_sym_COMMA,
    STATE(1561), 1,
      aux_sym_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91889] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3395), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [91902] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3397), 1,
      anon_sym_RBRACE,
    ACTIONS(3399), 1,
      anon_sym_COMMA,
    STATE(1508), 1,
      aux_sym_unqualified_imports_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91919] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3402), 1,
      anon_sym_COMMA,
    ACTIONS(3405), 1,
      anon_sym_GT_GT,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91936] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3407), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [91949] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3409), 1,
      anon_sym_COMMA,
    ACTIONS(3411), 1,
      anon_sym_RPAREN,
    STATE(1525), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91966] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2597), 1,
      anon_sym_RBRACK,
    ACTIONS(3413), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [91983] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3415), 1,
      anon_sym_COMMA,
    ACTIONS(3418), 1,
      anon_sym_RPAREN,
    STATE(1513), 1,
      aux_sym_external_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92000] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2493), 1,
      anon_sym_GT_GT,
    ACTIONS(3420), 1,
      anon_sym_COMMA,
    STATE(1521), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92017] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3422), 1,
      anon_sym_COMMA,
    ACTIONS(3424), 1,
      anon_sym_RPAREN,
    STATE(1529), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92034] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3026), 1,
      anon_sym_RBRACE,
    ACTIONS(3426), 1,
      anon_sym_COMMA,
    STATE(1508), 1,
      aux_sym_unqualified_imports_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92051] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3428), 1,
      anon_sym_COMMA,
    ACTIONS(3430), 1,
      anon_sym_RPAREN,
    STATE(1532), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92068] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3432), 1,
      anon_sym_COMMA,
    ACTIONS(3434), 1,
      anon_sym_RPAREN,
    STATE(1534), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92085] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    ACTIONS(3436), 1,
      anon_sym_RPAREN,
    STATE(1794), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92102] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2473), 1,
      anon_sym_RPAREN,
    ACTIONS(3438), 1,
      anon_sym_COMMA,
    STATE(1569), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92119] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3440), 1,
      anon_sym_COMMA,
    ACTIONS(3443), 1,
      anon_sym_GT_GT,
    STATE(1521), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92136] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2383), 1,
      anon_sym_LBRACE,
    ACTIONS(3445), 1,
      anon_sym_COMMA,
    STATE(1522), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92153] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3448), 1,
      anon_sym_COMMA,
    ACTIONS(3450), 1,
      anon_sym_RPAREN,
    STATE(1492), 1,
      aux_sym_constant_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92170] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3452), 1,
      anon_sym_if,
    ACTIONS(3454), 1,
      anon_sym_DASH_GT,
    STATE(1935), 1,
      sym_case_clause_guard,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92187] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2605), 1,
      anon_sym_RPAREN,
    ACTIONS(3456), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92204] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3458), 1,
      anon_sym_COLON,
    ACTIONS(3460), 1,
      anon_sym_EQ,
    STATE(1942), 1,
      sym__constant_type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92221] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3458), 1,
      anon_sym_COLON,
    ACTIONS(3462), 1,
      anon_sym_EQ,
    STATE(1877), 1,
      sym__constant_type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92238] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3466), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3464), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [92253] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2477), 1,
      anon_sym_RPAREN,
    ACTIONS(3468), 1,
      anon_sym_COMMA,
    STATE(1569), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92270] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3470), 1,
      anon_sym_COMMA,
    ACTIONS(3472), 1,
      anon_sym_GT_GT,
    STATE(1603), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92287] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3231), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [92300] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2857), 1,
      anon_sym_RPAREN,
    ACTIONS(3474), 1,
      anon_sym_COMMA,
    STATE(1541), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92317] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2617), 1,
      anon_sym_RPAREN,
    ACTIONS(3476), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92334] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2825), 1,
      anon_sym_RPAREN,
    ACTIONS(3478), 1,
      anon_sym_COMMA,
    STATE(1556), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92351] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3480), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [92364] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1324), 1,
      sym_external_function_body,
    STATE(1777), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92381] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(1809), 1,
      sym_record_update_argument,
    STATE(1865), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92398] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3482), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [92411] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3484), 1,
      anon_sym_COMMA,
    ACTIONS(3486), 1,
      anon_sym_RPAREN,
    STATE(1602), 1,
      aux_sym_record_update_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92428] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1274), 1,
      anon_sym_GT_GT,
    ACTIONS(3488), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92445] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3490), 1,
      anon_sym_COMMA,
    ACTIONS(3493), 1,
      anon_sym_RPAREN,
    STATE(1541), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92462] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3292), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [92475] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    ACTIONS(3495), 1,
      anon_sym_RPAREN,
    STATE(1486), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92492] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3497), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
  [92505] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3499), 1,
      anon_sym_EQ,
    STATE(1945), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92522] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1319), 1,
      sym_external_function_body,
    STATE(1777), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92539] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3236), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
  [92552] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3501), 1,
      anon_sym_EQ,
    STATE(1839), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92569] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3503), 1,
      anon_sym_COMMA,
    ACTIONS(3506), 1,
      anon_sym_RPAREN,
    STATE(1549), 1,
      aux_sym_record_update_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92586] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2881), 1,
      anon_sym_RPAREN,
    ACTIONS(3508), 1,
      anon_sym_COMMA,
    STATE(1616), 1,
      aux_sym_constant_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92603] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    STATE(1898), 1,
      sym_target,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3510), 2,
      anon_sym_erlang,
      anon_sym_javascript,
  [92618] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3512), 1,
      anon_sym_COMMA,
    ACTIONS(3514), 1,
      anon_sym_LT_DASH,
    STATE(1639), 1,
      aux_sym_use_assignments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92635] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3516), 1,
      anon_sym_EQ,
    STATE(1887), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92652] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3260), 3,
      anon_sym_if,
      anon_sym_DASH_GT,
      anon_sym_PIPE,
  [92665] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(801), 1,
      sym_external_function_body,
    STATE(1678), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92682] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3518), 1,
      anon_sym_COMMA,
    ACTIONS(3521), 1,
      anon_sym_RPAREN,
    STATE(1556), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92699] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2829), 1,
      anon_sym_RPAREN,
    ACTIONS(3523), 1,
      anon_sym_COMMA,
    STATE(1513), 1,
      aux_sym_external_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92716] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3525), 1,
      anon_sym_COMMA,
    ACTIONS(3527), 1,
      anon_sym_GT_GT,
    STATE(1540), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92733] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1048), 1,
      anon_sym_RPAREN,
    ACTIONS(3529), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92750] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3458), 1,
      anon_sym_COLON,
    ACTIONS(3531), 1,
      anon_sym_EQ,
    STATE(1861), 1,
      sym__constant_type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92767] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3533), 1,
      anon_sym_COMMA,
    ACTIONS(3536), 1,
      anon_sym_RPAREN,
    STATE(1561), 1,
      aux_sym_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92784] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3285), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [92797] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3538), 1,
      anon_sym_COMMA,
    ACTIONS(3540), 1,
      anon_sym_RPAREN,
    STATE(1495), 1,
      aux_sym_constant_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92814] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3542), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [92827] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3544), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [92840] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3546), 1,
      anon_sym_COMMA,
    ACTIONS(3548), 1,
      anon_sym_RPAREN,
    STATE(1550), 1,
      aux_sym_constant_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92857] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3550), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [92870] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3552), 1,
      anon_sym_COMMA,
    ACTIONS(3554), 1,
      anon_sym_GT_GT,
    STATE(1579), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92887] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3556), 1,
      anon_sym_COMMA,
    ACTIONS(3559), 1,
      anon_sym_RPAREN,
    STATE(1569), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92904] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3561), 1,
      anon_sym_EQ,
    STATE(1869), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92921] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3563), 1,
      anon_sym_COMMA,
    ACTIONS(3566), 1,
      anon_sym_RPAREN,
    STATE(1571), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92938] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    ACTIONS(3568), 1,
      anon_sym_RPAREN,
    STATE(1794), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92955] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3570), 1,
      anon_sym_COMMA,
    ACTIONS(3572), 1,
      anon_sym_RPAREN,
    STATE(1482), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92972] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3574), 1,
      anon_sym_LPAREN,
    ACTIONS(3576), 1,
      anon_sym_DASH_GT,
    STATE(1954), 1,
      sym_constant_function_parameter_types,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [92989] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3578), 1,
      anon_sym_COMMA,
    ACTIONS(3580), 1,
      anon_sym_RPAREN,
    STATE(1484), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93006] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3582), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [93019] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3584), 1,
      anon_sym_COMMA,
    ACTIONS(3586), 1,
      anon_sym_RPAREN,
    STATE(1489), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93036] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3588), 1,
      anon_sym_COMMA,
    ACTIONS(3590), 1,
      anon_sym_RPAREN,
    STATE(1644), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93053] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1316), 1,
      anon_sym_GT_GT,
    ACTIONS(3592), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93070] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(900), 1,
      anon_sym_RPAREN,
    ACTIONS(3594), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93087] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3596), 1,
      anon_sym_COMMA,
    ACTIONS(3598), 1,
      anon_sym_RPAREN,
    STATE(1505), 1,
      aux_sym_data_constructor_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93104] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3602), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3600), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [93119] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1116), 1,
      anon_sym_RPAREN,
    ACTIONS(3604), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93136] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3606), 1,
      anon_sym_COMMA,
    ACTIONS(3608), 1,
      anon_sym_RPAREN,
    STATE(1580), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93153] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3610), 1,
      anon_sym_COMMA,
    ACTIONS(3612), 1,
      anon_sym_RPAREN,
    STATE(1596), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93170] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    ACTIONS(3614), 1,
      anon_sym_RPAREN,
    STATE(1578), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93187] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1467), 1,
      anon_sym_GT_GT,
    ACTIONS(3616), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93204] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3618), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [93217] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1206), 1,
      anon_sym_RPAREN,
    ACTIONS(3620), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93234] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3030), 1,
      anon_sym_RBRACE,
    ACTIONS(3622), 1,
      anon_sym_COMMA,
    STATE(1508), 1,
      aux_sym_unqualified_imports_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93251] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3626), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3624), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [93266] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3332), 1,
      anon_sym_LPAREN,
    ACTIONS(3628), 1,
      anon_sym_DASH_GT,
    STATE(1958), 1,
      sym_function_parameter_types,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93283] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3630), 1,
      anon_sym_EQ,
    STATE(1870), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93300] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3632), 1,
      anon_sym_COMMA,
    ACTIONS(3634), 1,
      anon_sym_RPAREN,
    STATE(1520), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93317] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3636), 1,
      anon_sym_COMMA,
    ACTIONS(3638), 1,
      anon_sym_RPAREN,
    STATE(1629), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93334] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(922), 1,
      anon_sym_RPAREN,
    ACTIONS(3640), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93351] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3642), 1,
      anon_sym_COMMA,
    ACTIONS(3644), 1,
      anon_sym_GT_GT,
    STATE(1587), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93368] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3648), 1,
      anon_sym_RBRACK,
    STATE(1606), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93385] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3650), 1,
      anon_sym_COMMA,
    ACTIONS(3652), 1,
      anon_sym_GT_GT,
    STATE(1608), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93402] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3654), 1,
      anon_sym_COMMA,
    ACTIONS(3657), 1,
      anon_sym_RPAREN,
    STATE(1600), 1,
      aux_sym__attribute_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93419] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2515), 1,
      anon_sym_GT_GT,
    ACTIONS(3659), 1,
      anon_sym_COMMA,
    STATE(1521), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93436] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3290), 1,
      anon_sym_RPAREN,
    ACTIONS(3661), 1,
      anon_sym_COMMA,
    STATE(1549), 1,
      aux_sym_record_update_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93453] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1447), 1,
      anon_sym_GT_GT,
    ACTIONS(3663), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93470] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3665), 1,
      anon_sym_COMMA,
    ACTIONS(3667), 1,
      anon_sym_RPAREN,
    STATE(1614), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93487] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2603), 1,
      anon_sym_RBRACK,
    ACTIONS(3669), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93504] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2627), 1,
      anon_sym_RBRACK,
    ACTIONS(3671), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93521] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3673), 1,
      anon_sym_COMMA,
    ACTIONS(3675), 1,
      anon_sym_RPAREN,
    STATE(1533), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93538] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2505), 1,
      anon_sym_GT_GT,
    ACTIONS(3677), 1,
      anon_sym_COMMA,
    STATE(1521), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93555] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3679), 1,
      anon_sym_COMMA,
    ACTIONS(3681), 1,
      anon_sym_RPAREN,
    STATE(1618), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93572] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3683), 1,
      anon_sym_COMMA,
    ACTIONS(3685), 1,
      anon_sym_RPAREN,
    STATE(1490), 1,
      aux_sym_anonymous_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93589] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2327), 1,
      anon_sym_GT_GT,
    ACTIONS(3687), 1,
      anon_sym_COMMA,
    STATE(1626), 1,
      aux_sym__pattern_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93606] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3689), 1,
      anon_sym_COMMA,
    ACTIONS(3692), 1,
      anon_sym_RPAREN,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93623] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3694), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [93636] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2613), 1,
      anon_sym_RPAREN,
    ACTIONS(3696), 1,
      anon_sym_COMMA,
    STATE(1437), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93653] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3698), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
  [93666] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3700), 1,
      anon_sym_COMMA,
    ACTIONS(3703), 1,
      anon_sym_RPAREN,
    STATE(1616), 1,
      aux_sym_constant_tuple_type_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93683] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(910), 1,
      anon_sym_RPAREN,
    ACTIONS(3705), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93700] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2467), 1,
      anon_sym_RPAREN,
    ACTIONS(3707), 1,
      anon_sym_COMMA,
    STATE(1569), 1,
      aux_sym_constant_record_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93717] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3709), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [93730] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3711), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [93743] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3713), 1,
      anon_sym_COMMA,
    ACTIONS(3716), 1,
      anon_sym_LT_DASH,
    STATE(1621), 1,
      aux_sym_use_assignments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93760] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1136), 1,
      anon_sym_RPAREN,
    ACTIONS(3718), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93777] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3720), 1,
      anon_sym_COMMA,
    ACTIONS(3722), 1,
      anon_sym_RPAREN,
    STATE(1617), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93794] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1485), 1,
      anon_sym_GT_GT,
    ACTIONS(3724), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93811] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3726), 1,
      anon_sym_LBRACE,
    ACTIONS(3728), 1,
      anon_sym_COMMA,
    STATE(1522), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93828] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3730), 1,
      anon_sym_COMMA,
    ACTIONS(3733), 1,
      anon_sym_GT_GT,
    STATE(1626), 1,
      aux_sym__pattern_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93845] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3735), 1,
      anon_sym_EQ,
    STATE(1914), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93862] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3737), 1,
      anon_sym_COMMA,
    ACTIONS(3739), 1,
      anon_sym_GT_GT,
    STATE(1624), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93879] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3741), 1,
      anon_sym_COMMA,
    ACTIONS(3743), 1,
      anon_sym_RPAREN,
    STATE(1398), 1,
      aux_sym_case_clause_pattern_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93896] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3745), 1,
      anon_sym_COMMA,
    ACTIONS(3747), 1,
      anon_sym_RPAREN,
    STATE(1649), 1,
      aux_sym__attribute_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93913] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(914), 1,
      anon_sym_RPAREN,
    ACTIONS(3749), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93930] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3751), 1,
      anon_sym_COMMA,
    ACTIONS(3753), 1,
      anon_sym_GT_GT,
    STATE(1643), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93947] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3309), 3,
      anon_sym_COMMA,
      anon_sym_GT_GT,
      anon_sym_DASH,
  [93960] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3755), 1,
      anon_sym_EQ,
    STATE(1963), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93977] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3757), 1,
      anon_sym_COMMA,
    ACTIONS(3759), 1,
      anon_sym_RBRACK,
    STATE(1605), 1,
      aux_sym_constant_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [93994] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3761), 1,
      anon_sym_COMMA,
    ACTIONS(3763), 1,
      anon_sym_RPAREN,
    STATE(1557), 1,
      aux_sym_external_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94011] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3765), 1,
      anon_sym_COMMA,
    ACTIONS(3768), 1,
      anon_sym_RPAREN,
    STATE(1637), 1,
      aux_sym_constant_type_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94028] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3772), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3770), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [94043] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2325), 1,
      anon_sym_LT_DASH,
    ACTIONS(3774), 1,
      anon_sym_COMMA,
    STATE(1621), 1,
      aux_sym_use_assignments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94060] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1148), 1,
      anon_sym_RPAREN,
    ACTIONS(3776), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94077] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3778), 1,
      anon_sym_COMMA,
    ACTIONS(3780), 1,
      anon_sym_GT_GT,
    STATE(1601), 1,
      aux_sym__constant_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94094] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    ACTIONS(3782), 1,
      anon_sym_RPAREN,
    STATE(1794), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94111] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1377), 1,
      anon_sym_GT_GT,
    ACTIONS(3784), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94128] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3782), 1,
      anon_sym_RPAREN,
    ACTIONS(3786), 1,
      anon_sym_COMMA,
    STATE(1571), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94145] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3788), 1,
      anon_sym_COMMA,
    ACTIONS(3790), 1,
      anon_sym_GT_GT,
    STATE(1611), 1,
      aux_sym__pattern_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94162] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1510), 1,
      anon_sym_GT_GT,
    ACTIONS(3792), 1,
      anon_sym_COMMA,
    STATE(1509), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94179] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3794), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [94192] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3796), 1,
      anon_sym_COMMA,
    ACTIONS(3798), 1,
      anon_sym_RPAREN,
    STATE(1657), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94209] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2455), 1,
      anon_sym_RPAREN,
    ACTIONS(3800), 1,
      anon_sym_COMMA,
    STATE(1600), 1,
      aux_sym__attribute_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94226] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1194), 1,
      anon_sym_RPAREN,
    ACTIONS(3802), 1,
      anon_sym_COMMA,
    STATE(1397), 1,
      aux_sym_tuple_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94243] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(884), 1,
      anon_sym_RPAREN,
    ACTIONS(3804), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94260] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3806), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ,
  [94273] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3808), 1,
      anon_sym_COMMA,
    ACTIONS(3810), 1,
      anon_sym_GT_GT,
    STATE(1646), 1,
      aux_sym__expression_bit_string_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94290] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3812), 1,
      anon_sym_COMMA,
    ACTIONS(3814), 1,
      anon_sym_RPAREN,
    STATE(1631), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94307] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3816), 1,
      anon_sym_COMMA,
    ACTIONS(3819), 1,
      anon_sym_RPAREN,
    STATE(1655), 1,
      aux_sym_anonymous_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94324] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3240), 1,
      anon_sym_COLON,
    ACTIONS(3821), 1,
      anon_sym_EQ,
    STATE(1948), 1,
      sym__type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94341] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(898), 1,
      anon_sym_RPAREN,
    ACTIONS(3823), 1,
      anon_sym_COMMA,
    STATE(1612), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94358] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3825), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [94371] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3827), 1,
      anon_sym_RBRACE,
    ACTIONS(3829), 1,
      anon_sym_COMMA,
    STATE(1590), 1,
      aux_sym_unqualified_imports_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94388] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3831), 1,
      anon_sym_COMMA,
    ACTIONS(3833), 1,
      anon_sym_RPAREN,
    STATE(1651), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94405] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3837), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3835), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [94420] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3839), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3835), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [94435] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3458), 1,
      anon_sym_COLON,
    ACTIONS(3841), 1,
      anon_sym_EQ,
    STATE(1816), 1,
      sym__constant_type_annotation,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94452] = 5,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2964), 1,
      anon_sym_RPAREN,
    ACTIONS(3843), 1,
      anon_sym_COMMA,
    STATE(1561), 1,
      aux_sym_function_parameters_repeat1,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94469] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3845), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [94481] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3847), 1,
      anon_sym_LPAREN,
    STATE(113), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94495] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1669), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94509] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1670), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94523] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3849), 1,
      anon_sym_LPAREN,
    STATE(1947), 1,
      sym_external_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94537] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3849), 1,
      anon_sym_LPAREN,
    STATE(1946), 1,
      sym_external_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94551] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1814), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94565] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3851), 1,
      sym__name,
    STATE(697), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94579] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3377), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94591] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1702), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94605] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2513), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94617] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3853), 1,
      sym__name,
    STATE(692), 1,
      sym_module,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94631] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1591), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94645] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(53), 1,
      anon_sym_DQUOTE,
    STATE(764), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94659] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3855), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94671] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3387), 1,
      sym__name,
    STATE(1794), 1,
      sym_type_parameter,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94685] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3857), 2,
      anon_sym_RBRACE,
      sym__upname,
  [94697] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1950), 1,
      anon_sym_LBRACE,
    ACTIONS(1952), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94711] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1709), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94725] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1733), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94739] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1714), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94753] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3859), 1,
      anon_sym_LPAREN,
    STATE(477), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94767] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1527), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94781] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3861), 1,
      anon_sym_LPAREN,
    STATE(738), 1,
      sym_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94795] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1186), 1,
      sym__upname,
    STATE(354), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94809] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3716), 2,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
  [94821] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3863), 1,
      sym__name,
    STATE(449), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94835] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3865), 2,
      anon_sym_COMMA,
      anon_sym_LT_DASH,
  [94847] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3867), 2,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
  [94859] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(101), 1,
      sym__upname,
    STATE(593), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94873] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1718), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94887] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1754), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94901] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3869), 2,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
  [94913] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3871), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94925] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3873), 1,
      sym__name,
    STATE(1260), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94939] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1726), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94953] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3875), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94965] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3877), 1,
      anon_sym_LBRACE,
    ACTIONS(3879), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [94979] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3871), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [94991] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1271), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95005] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3875), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95017] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3881), 1,
      anon_sym_type,
    ACTIONS(3883), 1,
      anon_sym_fn,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95031] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3885), 1,
      anon_sym_LBRACE,
    ACTIONS(3887), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95045] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1022), 1,
      sym__upname,
    STATE(613), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95059] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3889), 1,
      anon_sym_LBRACE,
    ACTIONS(3891), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95073] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3893), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95085] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3895), 1,
      anon_sym_COLON,
    ACTIONS(3897), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95099] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3899), 2,
      anon_sym_RBRACE,
      sym__upname,
  [95111] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3901), 1,
      anon_sym_LPAREN,
    STATE(601), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95125] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3903), 1,
      anon_sym_LBRACE,
    ACTIONS(3905), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95139] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3907), 1,
      sym__name,
    STATE(23), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95153] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3909), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95165] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3911), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95177] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3913), 1,
      anon_sym_LBRACE,
    ACTIONS(3915), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95191] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3917), 1,
      anon_sym_LPAREN,
    STATE(60), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95205] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3418), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95217] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3919), 1,
      sym__name,
    STATE(610), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95231] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3921), 1,
      anon_sym_LPAREN,
    STATE(1284), 1,
      sym_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95245] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3733), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [95257] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3909), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95269] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3923), 1,
      anon_sym_LBRACE,
    ACTIONS(3925), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95283] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3927), 1,
      anon_sym_LBRACE,
    ACTIONS(3929), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95297] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3931), 2,
      anon_sym_RBRACE,
      sym__upname,
  [95309] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3768), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95321] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3933), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [95333] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3405), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [95345] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3935), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95357] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3397), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [95369] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3849), 1,
      anon_sym_LPAREN,
    STATE(1904), 1,
      sym_external_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95383] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3849), 1,
      anon_sym_LPAREN,
    STATE(1823), 1,
      sym_external_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95397] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3937), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [95409] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1740), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95423] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3939), 1,
      anon_sym_LBRACE,
    ACTIONS(3941), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95437] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3493), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95449] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3692), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95461] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3943), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [95473] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3303), 1,
      anon_sym_LPAREN,
    STATE(1737), 1,
      sym_anonymous_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95487] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1873), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95501] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1874), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95515] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3945), 1,
      anon_sym_COMMA,
    ACTIONS(3947), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95529] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3949), 1,
      sym__name,
    STATE(1260), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95543] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(2517), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95555] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3521), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95567] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3443), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [95579] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3951), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95591] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3953), 2,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
  [95603] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3955), 2,
      anon_sym_COMMA,
      anon_sym_GT_GT,
  [95615] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3957), 2,
      anon_sym_RBRACE,
      sym__upname,
  [95627] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1106), 1,
      sym__upname,
    STATE(49), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95641] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3959), 1,
      anon_sym_LBRACE,
    ACTIONS(3961), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95655] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3963), 1,
      anon_sym_LPAREN,
    STATE(941), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95669] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3819), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95681] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3965), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95693] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3967), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95705] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(431), 1,
      sym__upname,
    STATE(234), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95719] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1918), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95733] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1921), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95747] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3861), 1,
      anon_sym_LPAREN,
    STATE(718), 1,
      sym_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95761] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3004), 1,
      sym__name,
    STATE(947), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95775] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3949), 1,
      sym__name,
    STATE(1271), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95789] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1734), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95803] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3703), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [95815] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1729), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95829] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1729), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95843] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1916), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95857] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(251), 1,
      sym__upname,
    STATE(949), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95871] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1922), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95885] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1951), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95899] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3873), 1,
      sym__name,
    STATE(1276), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95913] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1952), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95927] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3969), 1,
      sym__name,
    STATE(1281), 1,
      sym_module,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95941] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2877), 1,
      sym__upname,
    STATE(698), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95955] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(91), 1,
      anon_sym_DQUOTE,
    STATE(1305), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95969] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3971), 2,
      anon_sym_RBRACE,
      sym__upname,
  [95981] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1964), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [95995] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3973), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96007] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1965), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96021] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3975), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96033] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3977), 1,
      anon_sym_LBRACE,
    STATE(736), 1,
      sym_unqualified_imports,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96047] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3979), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96059] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3981), 1,
      anon_sym_LBRACE,
    STATE(1293), 1,
      sym_unqualified_imports,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96073] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1851), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96087] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1848), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96101] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1838), 1,
      sym__upname,
    STATE(1321), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96115] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3204), 1,
      sym__name,
    STATE(249), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96129] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3983), 1,
      anon_sym_type,
    ACTIONS(3985), 1,
      anon_sym_fn,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96143] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3987), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96155] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3989), 1,
      anon_sym_type,
    ACTIONS(3991), 1,
      anon_sym_fn,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96169] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2946), 1,
      anon_sym_LBRACE,
    ACTIONS(2948), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96183] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3566), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96195] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(63), 1,
      sym__upname,
    STATE(103), 1,
      sym_constructor_name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96209] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1560), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96223] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1872), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96237] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(421), 1,
      anon_sym_DQUOTE,
    STATE(1876), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96251] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2998), 1,
      sym__name,
    STATE(702), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96265] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3559), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96277] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1526), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96291] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1722), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96305] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3657), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96317] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3252), 1,
      sym__name,
    STATE(94), 1,
      sym_label,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96331] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3993), 1,
      anon_sym_LPAREN,
    STATE(260), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96345] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2807), 1,
      sym__upname,
    STATE(1167), 1,
      sym_type_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96359] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3995), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96371] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3997), 1,
      anon_sym_LPAREN,
    STATE(700), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96385] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3506), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96397] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3999), 2,
      anon_sym_LBRACE,
      anon_sym_DASH_GT,
  [96409] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
    ACTIONS(3536), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [96421] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4001), 1,
      anon_sym_type,
    ACTIONS(4003), 1,
      anon_sym_fn,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96435] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1688), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96449] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3921), 1,
      anon_sym_LPAREN,
    STATE(1285), 1,
      sym_function_parameters,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96463] = 4,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3016), 1,
      sym__name,
    STATE(1663), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96477] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4005), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96488] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4007), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96499] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4009), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96510] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4011), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96521] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4013), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96532] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4015), 1,
      anon_sym_type,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96543] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4017), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96554] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4019), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96565] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4021), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96576] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4023), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96587] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3897), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96598] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4025), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96609] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4027), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96620] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4029), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96631] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4031), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96642] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4033), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96653] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4035), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96664] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4037), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96675] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4039), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96686] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4041), 1,
      anon_sym_LT_DASH,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96697] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4043), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96708] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4045), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96719] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4047), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96730] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4049), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96741] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4051), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96752] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4053), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96763] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4055), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96774] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4057), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96785] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4059), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96796] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4061), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96807] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4063), 1,
      sym__name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96818] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(2179), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96829] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4065), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96840] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4067), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96851] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4069), 1,
      sym__name,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96862] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4071), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96873] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4073), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96884] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4075), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96895] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4077), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96906] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4079), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96917] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4081), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96928] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4083), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96939] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4085), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96950] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4087), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96961] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4089), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96972] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4091), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96983] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4093), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [96994] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4095), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97005] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4097), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97016] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4099), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97027] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4101), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97038] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4103), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97049] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4105), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97060] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4107), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97071] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4109), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97082] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4111), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97093] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4113), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97104] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4115), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97115] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4117), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97126] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4119), 1,
      anon_sym_LT_DASH,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97137] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4121), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97148] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4123), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97159] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4125), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97170] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4127), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97181] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4129), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97192] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4131), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97203] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4133), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97214] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4135), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97225] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4137), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97236] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4139), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97247] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4141), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97258] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4143), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97269] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4145), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97280] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4147), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97291] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4149), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97302] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4151), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97313] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4153), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97324] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4155), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97335] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(3138), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97346] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4157), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97357] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4159), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97368] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4161), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97379] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4163), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97390] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4165), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97401] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4167), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97412] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4169), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97423] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4171), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97434] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4173), 1,
      anon_sym_type,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97445] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4175), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97456] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4177), 1,
      anon_sym_type,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97467] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4179), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97478] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4181), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97489] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4183), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97500] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4185), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97511] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4187), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97522] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(1998), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97533] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4189), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97544] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4191), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97555] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4193), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97566] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4195), 1,
      ts_builtin_sym_end,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97577] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4197), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97588] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4199), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97599] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4201), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97610] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4203), 1,
      anon_sym_type,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97621] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4205), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97632] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4207), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97643] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4209), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97654] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4211), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97665] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4213), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97676] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4215), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97687] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4217), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97698] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4219), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97709] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4221), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97720] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4223), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97731] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4225), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97742] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4227), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97753] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4229), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97764] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4231), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97775] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4233), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97786] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4235), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97797] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4237), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97808] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4239), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97819] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4241), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97830] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4243), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97841] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4245), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97852] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4247), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97863] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4249), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97874] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4251), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97885] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4253), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97896] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4255), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97907] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4257), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97918] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4259), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97929] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4261), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97940] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4263), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97951] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4265), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97962] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4267), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97973] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4269), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97984] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4271), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [97995] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4273), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98006] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4275), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98017] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4277), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98028] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4279), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98039] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4281), 1,
      anon_sym_DASH_GT,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98050] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4283), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98061] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4285), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98072] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4287), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98083] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4289), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98094] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4291), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98105] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4293), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98116] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4295), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
  [98127] = 3,
    ACTIONS(3), 1,
      sym_module_comment,
    ACTIONS(4297), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_statement_comment,
      sym_comment,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(14)] = 0,
  [SMALL_STATE(15)] = 73,
  [SMALL_STATE(16)] = 141,
  [SMALL_STATE(17)] = 271,
  [SMALL_STATE(18)] = 401,
  [SMALL_STATE(19)] = 531,
  [SMALL_STATE(20)] = 661,
  [SMALL_STATE(21)] = 791,
  [SMALL_STATE(22)] = 921,
  [SMALL_STATE(23)] = 1051,
  [SMALL_STATE(24)] = 1119,
  [SMALL_STATE(25)] = 1249,
  [SMALL_STATE(26)] = 1317,
  [SMALL_STATE(27)] = 1447,
  [SMALL_STATE(28)] = 1577,
  [SMALL_STATE(29)] = 1707,
  [SMALL_STATE(30)] = 1837,
  [SMALL_STATE(31)] = 1905,
  [SMALL_STATE(32)] = 1973,
  [SMALL_STATE(33)] = 2103,
  [SMALL_STATE(34)] = 2233,
  [SMALL_STATE(35)] = 2363,
  [SMALL_STATE(36)] = 2493,
  [SMALL_STATE(37)] = 2623,
  [SMALL_STATE(38)] = 2753,
  [SMALL_STATE(39)] = 2883,
  [SMALL_STATE(40)] = 3013,
  [SMALL_STATE(41)] = 3081,
  [SMALL_STATE(42)] = 3211,
  [SMALL_STATE(43)] = 3341,
  [SMALL_STATE(44)] = 3471,
  [SMALL_STATE(45)] = 3543,
  [SMALL_STATE(46)] = 3611,
  [SMALL_STATE(47)] = 3741,
  [SMALL_STATE(48)] = 3809,
  [SMALL_STATE(49)] = 3939,
  [SMALL_STATE(50)] = 4007,
  [SMALL_STATE(51)] = 4137,
  [SMALL_STATE(52)] = 4267,
  [SMALL_STATE(53)] = 4397,
  [SMALL_STATE(54)] = 4465,
  [SMALL_STATE(55)] = 4595,
  [SMALL_STATE(56)] = 4663,
  [SMALL_STATE(57)] = 4793,
  [SMALL_STATE(58)] = 4923,
  [SMALL_STATE(59)] = 5053,
  [SMALL_STATE(60)] = 5121,
  [SMALL_STATE(61)] = 5189,
  [SMALL_STATE(62)] = 5257,
  [SMALL_STATE(63)] = 5325,
  [SMALL_STATE(64)] = 5393,
  [SMALL_STATE(65)] = 5523,
  [SMALL_STATE(66)] = 5653,
  [SMALL_STATE(67)] = 5725,
  [SMALL_STATE(68)] = 5855,
  [SMALL_STATE(69)] = 5985,
  [SMALL_STATE(70)] = 6115,
  [SMALL_STATE(71)] = 6187,
  [SMALL_STATE(72)] = 6317,
  [SMALL_STATE(73)] = 6384,
  [SMALL_STATE(74)] = 6511,
  [SMALL_STATE(75)] = 6578,
  [SMALL_STATE(76)] = 6705,
  [SMALL_STATE(77)] = 6832,
  [SMALL_STATE(78)] = 6899,
  [SMALL_STATE(79)] = 7026,
  [SMALL_STATE(80)] = 7153,
  [SMALL_STATE(81)] = 7222,
  [SMALL_STATE(82)] = 7349,
  [SMALL_STATE(83)] = 7476,
  [SMALL_STATE(84)] = 7603,
  [SMALL_STATE(85)] = 7730,
  [SMALL_STATE(86)] = 7799,
  [SMALL_STATE(87)] = 7866,
  [SMALL_STATE(88)] = 7933,
  [SMALL_STATE(89)] = 8000,
  [SMALL_STATE(90)] = 8067,
  [SMALL_STATE(91)] = 8134,
  [SMALL_STATE(92)] = 8261,
  [SMALL_STATE(93)] = 8330,
  [SMALL_STATE(94)] = 8397,
  [SMALL_STATE(95)] = 8464,
  [SMALL_STATE(96)] = 8533,
  [SMALL_STATE(97)] = 8660,
  [SMALL_STATE(98)] = 8727,
  [SMALL_STATE(99)] = 8794,
  [SMALL_STATE(100)] = 8921,
  [SMALL_STATE(101)] = 8988,
  [SMALL_STATE(102)] = 9115,
  [SMALL_STATE(103)] = 9242,
  [SMALL_STATE(104)] = 9309,
  [SMALL_STATE(105)] = 9376,
  [SMALL_STATE(106)] = 9443,
  [SMALL_STATE(107)] = 9570,
  [SMALL_STATE(108)] = 9697,
  [SMALL_STATE(109)] = 9764,
  [SMALL_STATE(110)] = 9835,
  [SMALL_STATE(111)] = 9904,
  [SMALL_STATE(112)] = 10031,
  [SMALL_STATE(113)] = 10158,
  [SMALL_STATE(114)] = 10225,
  [SMALL_STATE(115)] = 10296,
  [SMALL_STATE(116)] = 10423,
  [SMALL_STATE(117)] = 10490,
  [SMALL_STATE(118)] = 10617,
  [SMALL_STATE(119)] = 10684,
  [SMALL_STATE(120)] = 10751,
  [SMALL_STATE(121)] = 10818,
  [SMALL_STATE(122)] = 10945,
  [SMALL_STATE(123)] = 11012,
  [SMALL_STATE(124)] = 11079,
  [SMALL_STATE(125)] = 11150,
  [SMALL_STATE(126)] = 11221,
  [SMALL_STATE(127)] = 11287,
  [SMALL_STATE(128)] = 11353,
  [SMALL_STATE(129)] = 11419,
  [SMALL_STATE(130)] = 11485,
  [SMALL_STATE(131)] = 11551,
  [SMALL_STATE(132)] = 11681,
  [SMALL_STATE(133)] = 11811,
  [SMALL_STATE(134)] = 11941,
  [SMALL_STATE(135)] = 12007,
  [SMALL_STATE(136)] = 12073,
  [SMALL_STATE(137)] = 12139,
  [SMALL_STATE(138)] = 12207,
  [SMALL_STATE(139)] = 12273,
  [SMALL_STATE(140)] = 12339,
  [SMALL_STATE(141)] = 12405,
  [SMALL_STATE(142)] = 12535,
  [SMALL_STATE(143)] = 12603,
  [SMALL_STATE(144)] = 12671,
  [SMALL_STATE(145)] = 12757,
  [SMALL_STATE(146)] = 12823,
  [SMALL_STATE(147)] = 12889,
  [SMALL_STATE(148)] = 12963,
  [SMALL_STATE(149)] = 13093,
  [SMALL_STATE(150)] = 13169,
  [SMALL_STATE(151)] = 13249,
  [SMALL_STATE(152)] = 13331,
  [SMALL_STATE(153)] = 13397,
  [SMALL_STATE(154)] = 13463,
  [SMALL_STATE(155)] = 13529,
  [SMALL_STATE(156)] = 13595,
  [SMALL_STATE(157)] = 13661,
  [SMALL_STATE(158)] = 13727,
  [SMALL_STATE(159)] = 13793,
  [SMALL_STATE(160)] = 13859,
  [SMALL_STATE(161)] = 13925,
  [SMALL_STATE(162)] = 14055,
  [SMALL_STATE(163)] = 14139,
  [SMALL_STATE(164)] = 14225,
  [SMALL_STATE(165)] = 14295,
  [SMALL_STATE(166)] = 14361,
  [SMALL_STATE(167)] = 14447,
  [SMALL_STATE(168)] = 14513,
  [SMALL_STATE(169)] = 14599,
  [SMALL_STATE(170)] = 14685,
  [SMALL_STATE(171)] = 14815,
  [SMALL_STATE(172)] = 14881,
  [SMALL_STATE(173)] = 14947,
  [SMALL_STATE(174)] = 15013,
  [SMALL_STATE(175)] = 15079,
  [SMALL_STATE(176)] = 15145,
  [SMALL_STATE(177)] = 15211,
  [SMALL_STATE(178)] = 15294,
  [SMALL_STATE(179)] = 15359,
  [SMALL_STATE(180)] = 15444,
  [SMALL_STATE(181)] = 15571,
  [SMALL_STATE(182)] = 15656,
  [SMALL_STATE(183)] = 15783,
  [SMALL_STATE(184)] = 15848,
  [SMALL_STATE(185)] = 15975,
  [SMALL_STATE(186)] = 16102,
  [SMALL_STATE(187)] = 16167,
  [SMALL_STATE(188)] = 16232,
  [SMALL_STATE(189)] = 16297,
  [SMALL_STATE(190)] = 16382,
  [SMALL_STATE(191)] = 16509,
  [SMALL_STATE(192)] = 16636,
  [SMALL_STATE(193)] = 16763,
  [SMALL_STATE(194)] = 16828,
  [SMALL_STATE(195)] = 16955,
  [SMALL_STATE(196)] = 17020,
  [SMALL_STATE(197)] = 17089,
  [SMALL_STATE(198)] = 17216,
  [SMALL_STATE(199)] = 17297,
  [SMALL_STATE(200)] = 17376,
  [SMALL_STATE(201)] = 17451,
  [SMALL_STATE(202)] = 17536,
  [SMALL_STATE(203)] = 17609,
  [SMALL_STATE(204)] = 17674,
  [SMALL_STATE(205)] = 17739,
  [SMALL_STATE(206)] = 17804,
  [SMALL_STATE(207)] = 17869,
  [SMALL_STATE(208)] = 17934,
  [SMALL_STATE(209)] = 17999,
  [SMALL_STATE(210)] = 18126,
  [SMALL_STATE(211)] = 18191,
  [SMALL_STATE(212)] = 18318,
  [SMALL_STATE(213)] = 18445,
  [SMALL_STATE(214)] = 18530,
  [SMALL_STATE(215)] = 18657,
  [SMALL_STATE(216)] = 18742,
  [SMALL_STATE(217)] = 18869,
  [SMALL_STATE(218)] = 18934,
  [SMALL_STATE(219)] = 18999,
  [SMALL_STATE(220)] = 19064,
  [SMALL_STATE(221)] = 19191,
  [SMALL_STATE(222)] = 19276,
  [SMALL_STATE(223)] = 19361,
  [SMALL_STATE(224)] = 19488,
  [SMALL_STATE(225)] = 19553,
  [SMALL_STATE(226)] = 19618,
  [SMALL_STATE(227)] = 19683,
  [SMALL_STATE(228)] = 19810,
  [SMALL_STATE(229)] = 19937,
  [SMALL_STATE(230)] = 20064,
  [SMALL_STATE(231)] = 20149,
  [SMALL_STATE(232)] = 20276,
  [SMALL_STATE(233)] = 20341,
  [SMALL_STATE(234)] = 20468,
  [SMALL_STATE(235)] = 20532,
  [SMALL_STATE(236)] = 20596,
  [SMALL_STATE(237)] = 20660,
  [SMALL_STATE(238)] = 20784,
  [SMALL_STATE(239)] = 20906,
  [SMALL_STATE(240)] = 20970,
  [SMALL_STATE(241)] = 21037,
  [SMALL_STATE(242)] = 21155,
  [SMALL_STATE(243)] = 21217,
  [SMALL_STATE(244)] = 21335,
  [SMALL_STATE(245)] = 21397,
  [SMALL_STATE(246)] = 21515,
  [SMALL_STATE(247)] = 21633,
  [SMALL_STATE(248)] = 21695,
  [SMALL_STATE(249)] = 21757,
  [SMALL_STATE(250)] = 21819,
  [SMALL_STATE(251)] = 21881,
  [SMALL_STATE(252)] = 21999,
  [SMALL_STATE(253)] = 22117,
  [SMALL_STATE(254)] = 22235,
  [SMALL_STATE(255)] = 22353,
  [SMALL_STATE(256)] = 22471,
  [SMALL_STATE(257)] = 22587,
  [SMALL_STATE(258)] = 22649,
  [SMALL_STATE(259)] = 22767,
  [SMALL_STATE(260)] = 22829,
  [SMALL_STATE(261)] = 22891,
  [SMALL_STATE(262)] = 22953,
  [SMALL_STATE(263)] = 23015,
  [SMALL_STATE(264)] = 23081,
  [SMALL_STATE(265)] = 23199,
  [SMALL_STATE(266)] = 23317,
  [SMALL_STATE(267)] = 23383,
  [SMALL_STATE(268)] = 23445,
  [SMALL_STATE(269)] = 23563,
  [SMALL_STATE(270)] = 23681,
  [SMALL_STATE(271)] = 23796,
  [SMALL_STATE(272)] = 23911,
  [SMALL_STATE(273)] = 23974,
  [SMALL_STATE(274)] = 24037,
  [SMALL_STATE(275)] = 24100,
  [SMALL_STATE(276)] = 24215,
  [SMALL_STATE(277)] = 24330,
  [SMALL_STATE(278)] = 24445,
  [SMALL_STATE(279)] = 24560,
  [SMALL_STATE(280)] = 24675,
  [SMALL_STATE(281)] = 24790,
  [SMALL_STATE(282)] = 24905,
  [SMALL_STATE(283)] = 25020,
  [SMALL_STATE(284)] = 25081,
  [SMALL_STATE(285)] = 25196,
  [SMALL_STATE(286)] = 25257,
  [SMALL_STATE(287)] = 25372,
  [SMALL_STATE(288)] = 25487,
  [SMALL_STATE(289)] = 25602,
  [SMALL_STATE(290)] = 25717,
  [SMALL_STATE(291)] = 25832,
  [SMALL_STATE(292)] = 25893,
  [SMALL_STATE(293)] = 26008,
  [SMALL_STATE(294)] = 26123,
  [SMALL_STATE(295)] = 26184,
  [SMALL_STATE(296)] = 26245,
  [SMALL_STATE(297)] = 26306,
  [SMALL_STATE(298)] = 26367,
  [SMALL_STATE(299)] = 26428,
  [SMALL_STATE(300)] = 26543,
  [SMALL_STATE(301)] = 26658,
  [SMALL_STATE(302)] = 26773,
  [SMALL_STATE(303)] = 26888,
  [SMALL_STATE(304)] = 27003,
  [SMALL_STATE(305)] = 27118,
  [SMALL_STATE(306)] = 27233,
  [SMALL_STATE(307)] = 27348,
  [SMALL_STATE(308)] = 27463,
  [SMALL_STATE(309)] = 27578,
  [SMALL_STATE(310)] = 27693,
  [SMALL_STATE(311)] = 27808,
  [SMALL_STATE(312)] = 27923,
  [SMALL_STATE(313)] = 28038,
  [SMALL_STATE(314)] = 28153,
  [SMALL_STATE(315)] = 28268,
  [SMALL_STATE(316)] = 28333,
  [SMALL_STATE(317)] = 28448,
  [SMALL_STATE(318)] = 28563,
  [SMALL_STATE(319)] = 28678,
  [SMALL_STATE(320)] = 28793,
  [SMALL_STATE(321)] = 28908,
  [SMALL_STATE(322)] = 29023,
  [SMALL_STATE(323)] = 29138,
  [SMALL_STATE(324)] = 29253,
  [SMALL_STATE(325)] = 29368,
  [SMALL_STATE(326)] = 29480,
  [SMALL_STATE(327)] = 29592,
  [SMALL_STATE(328)] = 29704,
  [SMALL_STATE(329)] = 29816,
  [SMALL_STATE(330)] = 29928,
  [SMALL_STATE(331)] = 30040,
  [SMALL_STATE(332)] = 30152,
  [SMALL_STATE(333)] = 30264,
  [SMALL_STATE(334)] = 30376,
  [SMALL_STATE(335)] = 30488,
  [SMALL_STATE(336)] = 30604,
  [SMALL_STATE(337)] = 30716,
  [SMALL_STATE(338)] = 30828,
  [SMALL_STATE(339)] = 30944,
  [SMALL_STATE(340)] = 31060,
  [SMALL_STATE(341)] = 31172,
  [SMALL_STATE(342)] = 31288,
  [SMALL_STATE(343)] = 31348,
  [SMALL_STATE(344)] = 31460,
  [SMALL_STATE(345)] = 31520,
  [SMALL_STATE(346)] = 31632,
  [SMALL_STATE(347)] = 31692,
  [SMALL_STATE(348)] = 31752,
  [SMALL_STATE(349)] = 31864,
  [SMALL_STATE(350)] = 31976,
  [SMALL_STATE(351)] = 32088,
  [SMALL_STATE(352)] = 32200,
  [SMALL_STATE(353)] = 32260,
  [SMALL_STATE(354)] = 32372,
  [SMALL_STATE(355)] = 32432,
  [SMALL_STATE(356)] = 32492,
  [SMALL_STATE(357)] = 32604,
  [SMALL_STATE(358)] = 32716,
  [SMALL_STATE(359)] = 32828,
  [SMALL_STATE(360)] = 32888,
  [SMALL_STATE(361)] = 33004,
  [SMALL_STATE(362)] = 33116,
  [SMALL_STATE(363)] = 33176,
  [SMALL_STATE(364)] = 33288,
  [SMALL_STATE(365)] = 33400,
  [SMALL_STATE(366)] = 33512,
  [SMALL_STATE(367)] = 33624,
  [SMALL_STATE(368)] = 33736,
  [SMALL_STATE(369)] = 33848,
  [SMALL_STATE(370)] = 33960,
  [SMALL_STATE(371)] = 34072,
  [SMALL_STATE(372)] = 34184,
  [SMALL_STATE(373)] = 34300,
  [SMALL_STATE(374)] = 34412,
  [SMALL_STATE(375)] = 34524,
  [SMALL_STATE(376)] = 34636,
  [SMALL_STATE(377)] = 34696,
  [SMALL_STATE(378)] = 34756,
  [SMALL_STATE(379)] = 34868,
  [SMALL_STATE(380)] = 34980,
  [SMALL_STATE(381)] = 35092,
  [SMALL_STATE(382)] = 35204,
  [SMALL_STATE(383)] = 35316,
  [SMALL_STATE(384)] = 35432,
  [SMALL_STATE(385)] = 35544,
  [SMALL_STATE(386)] = 35656,
  [SMALL_STATE(387)] = 35768,
  [SMALL_STATE(388)] = 35880,
  [SMALL_STATE(389)] = 35992,
  [SMALL_STATE(390)] = 36104,
  [SMALL_STATE(391)] = 36216,
  [SMALL_STATE(392)] = 36328,
  [SMALL_STATE(393)] = 36440,
  [SMALL_STATE(394)] = 36552,
  [SMALL_STATE(395)] = 36664,
  [SMALL_STATE(396)] = 36776,
  [SMALL_STATE(397)] = 36892,
  [SMALL_STATE(398)] = 37004,
  [SMALL_STATE(399)] = 37116,
  [SMALL_STATE(400)] = 37228,
  [SMALL_STATE(401)] = 37340,
  [SMALL_STATE(402)] = 37452,
  [SMALL_STATE(403)] = 37564,
  [SMALL_STATE(404)] = 37676,
  [SMALL_STATE(405)] = 37788,
  [SMALL_STATE(406)] = 37900,
  [SMALL_STATE(407)] = 38012,
  [SMALL_STATE(408)] = 38124,
  [SMALL_STATE(409)] = 38236,
  [SMALL_STATE(410)] = 38348,
  [SMALL_STATE(411)] = 38460,
  [SMALL_STATE(412)] = 38572,
  [SMALL_STATE(413)] = 38684,
  [SMALL_STATE(414)] = 38744,
  [SMALL_STATE(415)] = 38804,
  [SMALL_STATE(416)] = 38920,
  [SMALL_STATE(417)] = 38980,
  [SMALL_STATE(418)] = 39092,
  [SMALL_STATE(419)] = 39204,
  [SMALL_STATE(420)] = 39316,
  [SMALL_STATE(421)] = 39428,
  [SMALL_STATE(422)] = 39540,
  [SMALL_STATE(423)] = 39652,
  [SMALL_STATE(424)] = 39764,
  [SMALL_STATE(425)] = 39876,
  [SMALL_STATE(426)] = 39936,
  [SMALL_STATE(427)] = 40048,
  [SMALL_STATE(428)] = 40160,
  [SMALL_STATE(429)] = 40220,
  [SMALL_STATE(430)] = 40332,
  [SMALL_STATE(431)] = 40392,
  [SMALL_STATE(432)] = 40504,
  [SMALL_STATE(433)] = 40616,
  [SMALL_STATE(434)] = 40728,
  [SMALL_STATE(435)] = 40840,
  [SMALL_STATE(436)] = 40952,
  [SMALL_STATE(437)] = 41012,
  [SMALL_STATE(438)] = 41124,
  [SMALL_STATE(439)] = 41236,
  [SMALL_STATE(440)] = 41296,
  [SMALL_STATE(441)] = 41408,
  [SMALL_STATE(442)] = 41520,
  [SMALL_STATE(443)] = 41632,
  [SMALL_STATE(444)] = 41744,
  [SMALL_STATE(445)] = 41804,
  [SMALL_STATE(446)] = 41916,
  [SMALL_STATE(447)] = 42028,
  [SMALL_STATE(448)] = 42140,
  [SMALL_STATE(449)] = 42200,
  [SMALL_STATE(450)] = 42260,
  [SMALL_STATE(451)] = 42320,
  [SMALL_STATE(452)] = 42380,
  [SMALL_STATE(453)] = 42492,
  [SMALL_STATE(454)] = 42552,
  [SMALL_STATE(455)] = 42668,
  [SMALL_STATE(456)] = 42728,
  [SMALL_STATE(457)] = 42844,
  [SMALL_STATE(458)] = 42956,
  [SMALL_STATE(459)] = 43068,
  [SMALL_STATE(460)] = 43180,
  [SMALL_STATE(461)] = 43292,
  [SMALL_STATE(462)] = 43404,
  [SMALL_STATE(463)] = 43520,
  [SMALL_STATE(464)] = 43632,
  [SMALL_STATE(465)] = 43748,
  [SMALL_STATE(466)] = 43860,
  [SMALL_STATE(467)] = 43920,
  [SMALL_STATE(468)] = 44036,
  [SMALL_STATE(469)] = 44152,
  [SMALL_STATE(470)] = 44264,
  [SMALL_STATE(471)] = 44376,
  [SMALL_STATE(472)] = 44488,
  [SMALL_STATE(473)] = 44604,
  [SMALL_STATE(474)] = 44716,
  [SMALL_STATE(475)] = 44828,
  [SMALL_STATE(476)] = 44888,
  [SMALL_STATE(477)] = 45004,
  [SMALL_STATE(478)] = 45064,
  [SMALL_STATE(479)] = 45176,
  [SMALL_STATE(480)] = 45288,
  [SMALL_STATE(481)] = 45400,
  [SMALL_STATE(482)] = 45512,
  [SMALL_STATE(483)] = 45624,
  [SMALL_STATE(484)] = 45736,
  [SMALL_STATE(485)] = 45848,
  [SMALL_STATE(486)] = 45960,
  [SMALL_STATE(487)] = 46072,
  [SMALL_STATE(488)] = 46188,
  [SMALL_STATE(489)] = 46300,
  [SMALL_STATE(490)] = 46412,
  [SMALL_STATE(491)] = 46472,
  [SMALL_STATE(492)] = 46588,
  [SMALL_STATE(493)] = 46700,
  [SMALL_STATE(494)] = 46812,
  [SMALL_STATE(495)] = 46872,
  [SMALL_STATE(496)] = 46988,
  [SMALL_STATE(497)] = 47100,
  [SMALL_STATE(498)] = 47164,
  [SMALL_STATE(499)] = 47228,
  [SMALL_STATE(500)] = 47340,
  [SMALL_STATE(501)] = 47400,
  [SMALL_STATE(502)] = 47512,
  [SMALL_STATE(503)] = 47572,
  [SMALL_STATE(504)] = 47684,
  [SMALL_STATE(505)] = 47800,
  [SMALL_STATE(506)] = 47912,
  [SMALL_STATE(507)] = 47971,
  [SMALL_STATE(508)] = 48032,
  [SMALL_STATE(509)] = 48093,
  [SMALL_STATE(510)] = 48154,
  [SMALL_STATE(511)] = 48215,
  [SMALL_STATE(512)] = 48276,
  [SMALL_STATE(513)] = 48339,
  [SMALL_STATE(514)] = 48398,
  [SMALL_STATE(515)] = 48461,
  [SMALL_STATE(516)] = 48574,
  [SMALL_STATE(517)] = 48633,
  [SMALL_STATE(518)] = 48692,
  [SMALL_STATE(519)] = 48751,
  [SMALL_STATE(520)] = 48810,
  [SMALL_STATE(521)] = 48869,
  [SMALL_STATE(522)] = 48928,
  [SMALL_STATE(523)] = 48987,
  [SMALL_STATE(524)] = 49045,
  [SMALL_STATE(525)] = 49123,
  [SMALL_STATE(526)] = 49201,
  [SMALL_STATE(527)] = 49311,
  [SMALL_STATE(528)] = 49369,
  [SMALL_STATE(529)] = 49427,
  [SMALL_STATE(530)] = 49485,
  [SMALL_STATE(531)] = 49595,
  [SMALL_STATE(532)] = 49705,
  [SMALL_STATE(533)] = 49763,
  [SMALL_STATE(534)] = 49821,
  [SMALL_STATE(535)] = 49931,
  [SMALL_STATE(536)] = 50009,
  [SMALL_STATE(537)] = 50087,
  [SMALL_STATE(538)] = 50197,
  [SMALL_STATE(539)] = 50307,
  [SMALL_STATE(540)] = 50417,
  [SMALL_STATE(541)] = 50527,
  [SMALL_STATE(542)] = 50589,
  [SMALL_STATE(543)] = 50665,
  [SMALL_STATE(544)] = 50775,
  [SMALL_STATE(545)] = 50849,
  [SMALL_STATE(546)] = 50921,
  [SMALL_STATE(547)] = 50989,
  [SMALL_STATE(548)] = 51099,
  [SMALL_STATE(549)] = 51165,
  [SMALL_STATE(550)] = 51275,
  [SMALL_STATE(551)] = 51333,
  [SMALL_STATE(552)] = 51391,
  [SMALL_STATE(553)] = 51449,
  [SMALL_STATE(554)] = 51559,
  [SMALL_STATE(555)] = 51617,
  [SMALL_STATE(556)] = 51675,
  [SMALL_STATE(557)] = 51753,
  [SMALL_STATE(558)] = 51811,
  [SMALL_STATE(559)] = 51869,
  [SMALL_STATE(560)] = 51927,
  [SMALL_STATE(561)] = 51985,
  [SMALL_STATE(562)] = 52043,
  [SMALL_STATE(563)] = 52101,
  [SMALL_STATE(564)] = 52211,
  [SMALL_STATE(565)] = 52269,
  [SMALL_STATE(566)] = 52379,
  [SMALL_STATE(567)] = 52437,
  [SMALL_STATE(568)] = 52547,
  [SMALL_STATE(569)] = 52657,
  [SMALL_STATE(570)] = 52715,
  [SMALL_STATE(571)] = 52825,
  [SMALL_STATE(572)] = 52883,
  [SMALL_STATE(573)] = 52993,
  [SMALL_STATE(574)] = 53060,
  [SMALL_STATE(575)] = 53121,
  [SMALL_STATE(576)] = 53178,
  [SMALL_STATE(577)] = 53255,
  [SMALL_STATE(578)] = 53330,
  [SMALL_STATE(579)] = 53407,
  [SMALL_STATE(580)] = 53484,
  [SMALL_STATE(581)] = 53561,
  [SMALL_STATE(582)] = 53634,
  [SMALL_STATE(583)] = 53711,
  [SMALL_STATE(584)] = 53782,
  [SMALL_STATE(585)] = 53859,
  [SMALL_STATE(586)] = 53936,
  [SMALL_STATE(587)] = 54001,
  [SMALL_STATE(588)] = 54078,
  [SMALL_STATE(589)] = 54135,
  [SMALL_STATE(590)] = 54212,
  [SMALL_STATE(591)] = 54267,
  [SMALL_STATE(592)] = 54322,
  [SMALL_STATE(593)] = 54377,
  [SMALL_STATE(594)] = 54432,
  [SMALL_STATE(595)] = 54485,
  [SMALL_STATE(596)] = 54538,
  [SMALL_STATE(597)] = 54591,
  [SMALL_STATE(598)] = 54646,
  [SMALL_STATE(599)] = 54700,
  [SMALL_STATE(600)] = 54750,
  [SMALL_STATE(601)] = 54800,
  [SMALL_STATE(602)] = 54850,
  [SMALL_STATE(603)] = 54900,
  [SMALL_STATE(604)] = 54950,
  [SMALL_STATE(605)] = 55004,
  [SMALL_STATE(606)] = 55054,
  [SMALL_STATE(607)] = 55104,
  [SMALL_STATE(608)] = 55154,
  [SMALL_STATE(609)] = 55204,
  [SMALL_STATE(610)] = 55254,
  [SMALL_STATE(611)] = 55304,
  [SMALL_STATE(612)] = 55354,
  [SMALL_STATE(613)] = 55404,
  [SMALL_STATE(614)] = 55454,
  [SMALL_STATE(615)] = 55504,
  [SMALL_STATE(616)] = 55558,
  [SMALL_STATE(617)] = 55608,
  [SMALL_STATE(618)] = 55662,
  [SMALL_STATE(619)] = 55716,
  [SMALL_STATE(620)] = 55765,
  [SMALL_STATE(621)] = 55816,
  [SMALL_STATE(622)] = 55865,
  [SMALL_STATE(623)] = 55914,
  [SMALL_STATE(624)] = 55963,
  [SMALL_STATE(625)] = 56012,
  [SMALL_STATE(626)] = 56061,
  [SMALL_STATE(627)] = 56110,
  [SMALL_STATE(628)] = 56159,
  [SMALL_STATE(629)] = 56210,
  [SMALL_STATE(630)] = 56261,
  [SMALL_STATE(631)] = 56327,
  [SMALL_STATE(632)] = 56389,
  [SMALL_STATE(633)] = 56437,
  [SMALL_STATE(634)] = 56485,
  [SMALL_STATE(635)] = 56533,
  [SMALL_STATE(636)] = 56581,
  [SMALL_STATE(637)] = 56649,
  [SMALL_STATE(638)] = 56697,
  [SMALL_STATE(639)] = 56745,
  [SMALL_STATE(640)] = 56793,
  [SMALL_STATE(641)] = 56841,
  [SMALL_STATE(642)] = 56889,
  [SMALL_STATE(643)] = 56937,
  [SMALL_STATE(644)] = 56985,
  [SMALL_STATE(645)] = 57053,
  [SMALL_STATE(646)] = 57101,
  [SMALL_STATE(647)] = 57149,
  [SMALL_STATE(648)] = 57197,
  [SMALL_STATE(649)] = 57245,
  [SMALL_STATE(650)] = 57293,
  [SMALL_STATE(651)] = 57341,
  [SMALL_STATE(652)] = 57389,
  [SMALL_STATE(653)] = 57445,
  [SMALL_STATE(654)] = 57503,
  [SMALL_STATE(655)] = 57567,
  [SMALL_STATE(656)] = 57619,
  [SMALL_STATE(657)] = 57667,
  [SMALL_STATE(658)] = 57715,
  [SMALL_STATE(659)] = 57763,
  [SMALL_STATE(660)] = 57811,
  [SMALL_STATE(661)] = 57879,
  [SMALL_STATE(662)] = 57947,
  [SMALL_STATE(663)] = 58015,
  [SMALL_STATE(664)] = 58083,
  [SMALL_STATE(665)] = 58131,
  [SMALL_STATE(666)] = 58179,
  [SMALL_STATE(667)] = 58227,
  [SMALL_STATE(668)] = 58275,
  [SMALL_STATE(669)] = 58343,
  [SMALL_STATE(670)] = 58391,
  [SMALL_STATE(671)] = 58439,
  [SMALL_STATE(672)] = 58487,
  [SMALL_STATE(673)] = 58535,
  [SMALL_STATE(674)] = 58583,
  [SMALL_STATE(675)] = 58631,
  [SMALL_STATE(676)] = 58679,
  [SMALL_STATE(677)] = 58727,
  [SMALL_STATE(678)] = 58775,
  [SMALL_STATE(679)] = 58823,
  [SMALL_STATE(680)] = 58871,
  [SMALL_STATE(681)] = 58919,
  [SMALL_STATE(682)] = 58967,
  [SMALL_STATE(683)] = 59018,
  [SMALL_STATE(684)] = 59069,
  [SMALL_STATE(685)] = 59120,
  [SMALL_STATE(686)] = 59166,
  [SMALL_STATE(687)] = 59216,
  [SMALL_STATE(688)] = 59266,
  [SMALL_STATE(689)] = 59315,
  [SMALL_STATE(690)] = 59360,
  [SMALL_STATE(691)] = 59405,
  [SMALL_STATE(692)] = 59450,
  [SMALL_STATE(693)] = 59499,
  [SMALL_STATE(694)] = 59544,
  [SMALL_STATE(695)] = 59589,
  [SMALL_STATE(696)] = 59634,
  [SMALL_STATE(697)] = 59683,
  [SMALL_STATE(698)] = 59732,
  [SMALL_STATE(699)] = 59777,
  [SMALL_STATE(700)] = 59822,
  [SMALL_STATE(701)] = 59867,
  [SMALL_STATE(702)] = 59912,
  [SMALL_STATE(703)] = 59957,
  [SMALL_STATE(704)] = 60002,
  [SMALL_STATE(705)] = 60047,
  [SMALL_STATE(706)] = 60096,
  [SMALL_STATE(707)] = 60145,
  [SMALL_STATE(708)] = 60189,
  [SMALL_STATE(709)] = 60233,
  [SMALL_STATE(710)] = 60279,
  [SMALL_STATE(711)] = 60363,
  [SMALL_STATE(712)] = 60447,
  [SMALL_STATE(713)] = 60491,
  [SMALL_STATE(714)] = 60537,
  [SMALL_STATE(715)] = 60581,
  [SMALL_STATE(716)] = 60625,
  [SMALL_STATE(717)] = 60669,
  [SMALL_STATE(718)] = 60713,
  [SMALL_STATE(719)] = 60761,
  [SMALL_STATE(720)] = 60845,
  [SMALL_STATE(721)] = 60929,
  [SMALL_STATE(722)] = 60973,
  [SMALL_STATE(723)] = 61017,
  [SMALL_STATE(724)] = 61061,
  [SMALL_STATE(725)] = 61105,
  [SMALL_STATE(726)] = 61151,
  [SMALL_STATE(727)] = 61199,
  [SMALL_STATE(728)] = 61247,
  [SMALL_STATE(729)] = 61291,
  [SMALL_STATE(730)] = 61375,
  [SMALL_STATE(731)] = 61419,
  [SMALL_STATE(732)] = 61463,
  [SMALL_STATE(733)] = 61507,
  [SMALL_STATE(734)] = 61591,
  [SMALL_STATE(735)] = 61639,
  [SMALL_STATE(736)] = 61683,
  [SMALL_STATE(737)] = 61729,
  [SMALL_STATE(738)] = 61773,
  [SMALL_STATE(739)] = 61821,
  [SMALL_STATE(740)] = 61865,
  [SMALL_STATE(741)] = 61911,
  [SMALL_STATE(742)] = 61995,
  [SMALL_STATE(743)] = 62039,
  [SMALL_STATE(744)] = 62085,
  [SMALL_STATE(745)] = 62131,
  [SMALL_STATE(746)] = 62179,
  [SMALL_STATE(747)] = 62223,
  [SMALL_STATE(748)] = 62267,
  [SMALL_STATE(749)] = 62310,
  [SMALL_STATE(750)] = 62353,
  [SMALL_STATE(751)] = 62396,
  [SMALL_STATE(752)] = 62439,
  [SMALL_STATE(753)] = 62482,
  [SMALL_STATE(754)] = 62525,
  [SMALL_STATE(755)] = 62568,
  [SMALL_STATE(756)] = 62611,
  [SMALL_STATE(757)] = 62654,
  [SMALL_STATE(758)] = 62697,
  [SMALL_STATE(759)] = 62740,
  [SMALL_STATE(760)] = 62783,
  [SMALL_STATE(761)] = 62826,
  [SMALL_STATE(762)] = 62869,
  [SMALL_STATE(763)] = 62952,
  [SMALL_STATE(764)] = 62995,
  [SMALL_STATE(765)] = 63038,
  [SMALL_STATE(766)] = 63081,
  [SMALL_STATE(767)] = 63124,
  [SMALL_STATE(768)] = 63167,
  [SMALL_STATE(769)] = 63210,
  [SMALL_STATE(770)] = 63291,
  [SMALL_STATE(771)] = 63334,
  [SMALL_STATE(772)] = 63377,
  [SMALL_STATE(773)] = 63420,
  [SMALL_STATE(774)] = 63463,
  [SMALL_STATE(775)] = 63506,
  [SMALL_STATE(776)] = 63549,
  [SMALL_STATE(777)] = 63592,
  [SMALL_STATE(778)] = 63635,
  [SMALL_STATE(779)] = 63678,
  [SMALL_STATE(780)] = 63721,
  [SMALL_STATE(781)] = 63764,
  [SMALL_STATE(782)] = 63809,
  [SMALL_STATE(783)] = 63852,
  [SMALL_STATE(784)] = 63895,
  [SMALL_STATE(785)] = 63952,
  [SMALL_STATE(786)] = 63995,
  [SMALL_STATE(787)] = 64038,
  [SMALL_STATE(788)] = 64081,
  [SMALL_STATE(789)] = 64124,
  [SMALL_STATE(790)] = 64167,
  [SMALL_STATE(791)] = 64210,
  [SMALL_STATE(792)] = 64253,
  [SMALL_STATE(793)] = 64296,
  [SMALL_STATE(794)] = 64379,
  [SMALL_STATE(795)] = 64422,
  [SMALL_STATE(796)] = 64465,
  [SMALL_STATE(797)] = 64508,
  [SMALL_STATE(798)] = 64551,
  [SMALL_STATE(799)] = 64594,
  [SMALL_STATE(800)] = 64651,
  [SMALL_STATE(801)] = 64694,
  [SMALL_STATE(802)] = 64737,
  [SMALL_STATE(803)] = 64780,
  [SMALL_STATE(804)] = 64823,
  [SMALL_STATE(805)] = 64866,
  [SMALL_STATE(806)] = 64909,
  [SMALL_STATE(807)] = 64952,
  [SMALL_STATE(808)] = 64995,
  [SMALL_STATE(809)] = 65038,
  [SMALL_STATE(810)] = 65081,
  [SMALL_STATE(811)] = 65138,
  [SMALL_STATE(812)] = 65181,
  [SMALL_STATE(813)] = 65224,
  [SMALL_STATE(814)] = 65281,
  [SMALL_STATE(815)] = 65324,
  [SMALL_STATE(816)] = 65367,
  [SMALL_STATE(817)] = 65410,
  [SMALL_STATE(818)] = 65493,
  [SMALL_STATE(819)] = 65536,
  [SMALL_STATE(820)] = 65579,
  [SMALL_STATE(821)] = 65622,
  [SMALL_STATE(822)] = 65665,
  [SMALL_STATE(823)] = 65722,
  [SMALL_STATE(824)] = 65803,
  [SMALL_STATE(825)] = 65846,
  [SMALL_STATE(826)] = 65889,
  [SMALL_STATE(827)] = 65932,
  [SMALL_STATE(828)] = 65975,
  [SMALL_STATE(829)] = 66018,
  [SMALL_STATE(830)] = 66061,
  [SMALL_STATE(831)] = 66104,
  [SMALL_STATE(832)] = 66147,
  [SMALL_STATE(833)] = 66190,
  [SMALL_STATE(834)] = 66233,
  [SMALL_STATE(835)] = 66276,
  [SMALL_STATE(836)] = 66319,
  [SMALL_STATE(837)] = 66376,
  [SMALL_STATE(838)] = 66419,
  [SMALL_STATE(839)] = 66462,
  [SMALL_STATE(840)] = 66505,
  [SMALL_STATE(841)] = 66548,
  [SMALL_STATE(842)] = 66591,
  [SMALL_STATE(843)] = 66634,
  [SMALL_STATE(844)] = 66679,
  [SMALL_STATE(845)] = 66735,
  [SMALL_STATE(846)] = 66807,
  [SMALL_STATE(847)] = 66879,
  [SMALL_STATE(848)] = 66951,
  [SMALL_STATE(849)] = 67023,
  [SMALL_STATE(850)] = 67083,
  [SMALL_STATE(851)] = 67155,
  [SMALL_STATE(852)] = 67227,
  [SMALL_STATE(853)] = 67269,
  [SMALL_STATE(854)] = 67329,
  [SMALL_STATE(855)] = 67401,
  [SMALL_STATE(856)] = 67473,
  [SMALL_STATE(857)] = 67545,
  [SMALL_STATE(858)] = 67617,
  [SMALL_STATE(859)] = 67689,
  [SMALL_STATE(860)] = 67761,
  [SMALL_STATE(861)] = 67833,
  [SMALL_STATE(862)] = 67893,
  [SMALL_STATE(863)] = 67943,
  [SMALL_STATE(864)] = 68003,
  [SMALL_STATE(865)] = 68055,
  [SMALL_STATE(866)] = 68127,
  [SMALL_STATE(867)] = 68199,
  [SMALL_STATE(868)] = 68255,
  [SMALL_STATE(869)] = 68311,
  [SMALL_STATE(870)] = 68371,
  [SMALL_STATE(871)] = 68415,
  [SMALL_STATE(872)] = 68461,
  [SMALL_STATE(873)] = 68515,
  [SMALL_STATE(874)] = 68571,
  [SMALL_STATE(875)] = 68629,
  [SMALL_STATE(876)] = 68682,
  [SMALL_STATE(877)] = 68759,
  [SMALL_STATE(878)] = 68812,
  [SMALL_STATE(879)] = 68865,
  [SMALL_STATE(880)] = 68942,
  [SMALL_STATE(881)] = 69017,
  [SMALL_STATE(882)] = 69094,
  [SMALL_STATE(883)] = 69169,
  [SMALL_STATE(884)] = 69242,
  [SMALL_STATE(885)] = 69319,
  [SMALL_STATE(886)] = 69396,
  [SMALL_STATE(887)] = 69473,
  [SMALL_STATE(888)] = 69550,
  [SMALL_STATE(889)] = 69624,
  [SMALL_STATE(890)] = 69698,
  [SMALL_STATE(891)] = 69772,
  [SMALL_STATE(892)] = 69846,
  [SMALL_STATE(893)] = 69920,
  [SMALL_STATE(894)] = 69994,
  [SMALL_STATE(895)] = 70068,
  [SMALL_STATE(896)] = 70142,
  [SMALL_STATE(897)] = 70213,
  [SMALL_STATE(898)] = 70284,
  [SMALL_STATE(899)] = 70347,
  [SMALL_STATE(900)] = 70410,
  [SMALL_STATE(901)] = 70481,
  [SMALL_STATE(902)] = 70552,
  [SMALL_STATE(903)] = 70623,
  [SMALL_STATE(904)] = 70694,
  [SMALL_STATE(905)] = 70765,
  [SMALL_STATE(906)] = 70836,
  [SMALL_STATE(907)] = 70899,
  [SMALL_STATE(908)] = 70970,
  [SMALL_STATE(909)] = 71041,
  [SMALL_STATE(910)] = 71084,
  [SMALL_STATE(911)] = 71147,
  [SMALL_STATE(912)] = 71210,
  [SMALL_STATE(913)] = 71281,
  [SMALL_STATE(914)] = 71352,
  [SMALL_STATE(915)] = 71423,
  [SMALL_STATE(916)] = 71486,
  [SMALL_STATE(917)] = 71557,
  [SMALL_STATE(918)] = 71628,
  [SMALL_STATE(919)] = 71685,
  [SMALL_STATE(920)] = 71728,
  [SMALL_STATE(921)] = 71771,
  [SMALL_STATE(922)] = 71842,
  [SMALL_STATE(923)] = 71913,
  [SMALL_STATE(924)] = 71976,
  [SMALL_STATE(925)] = 72042,
  [SMALL_STATE(926)] = 72108,
  [SMALL_STATE(927)] = 72174,
  [SMALL_STATE(928)] = 72212,
  [SMALL_STATE(929)] = 72250,
  [SMALL_STATE(930)] = 72310,
  [SMALL_STATE(931)] = 72348,
  [SMALL_STATE(932)] = 72408,
  [SMALL_STATE(933)] = 72446,
  [SMALL_STATE(934)] = 72506,
  [SMALL_STATE(935)] = 72548,
  [SMALL_STATE(936)] = 72614,
  [SMALL_STATE(937)] = 72656,
  [SMALL_STATE(938)] = 72698,
  [SMALL_STATE(939)] = 72740,
  [SMALL_STATE(940)] = 72800,
  [SMALL_STATE(941)] = 72866,
  [SMALL_STATE(942)] = 72904,
  [SMALL_STATE(943)] = 72964,
  [SMALL_STATE(944)] = 73024,
  [SMALL_STATE(945)] = 73062,
  [SMALL_STATE(946)] = 73128,
  [SMALL_STATE(947)] = 73166,
  [SMALL_STATE(948)] = 73204,
  [SMALL_STATE(949)] = 73242,
  [SMALL_STATE(950)] = 73280,
  [SMALL_STATE(951)] = 73348,
  [SMALL_STATE(952)] = 73414,
  [SMALL_STATE(953)] = 73452,
  [SMALL_STATE(954)] = 73490,
  [SMALL_STATE(955)] = 73550,
  [SMALL_STATE(956)] = 73616,
  [SMALL_STATE(957)] = 73654,
  [SMALL_STATE(958)] = 73692,
  [SMALL_STATE(959)] = 73758,
  [SMALL_STATE(960)] = 73796,
  [SMALL_STATE(961)] = 73862,
  [SMALL_STATE(962)] = 73930,
  [SMALL_STATE(963)] = 73996,
  [SMALL_STATE(964)] = 74056,
  [SMALL_STATE(965)] = 74122,
  [SMALL_STATE(966)] = 74188,
  [SMALL_STATE(967)] = 74251,
  [SMALL_STATE(968)] = 74290,
  [SMALL_STATE(969)] = 74353,
  [SMALL_STATE(970)] = 74418,
  [SMALL_STATE(971)] = 74455,
  [SMALL_STATE(972)] = 74518,
  [SMALL_STATE(973)] = 74555,
  [SMALL_STATE(974)] = 74618,
  [SMALL_STATE(975)] = 74655,
  [SMALL_STATE(976)] = 74710,
  [SMALL_STATE(977)] = 74765,
  [SMALL_STATE(978)] = 74802,
  [SMALL_STATE(979)] = 74863,
  [SMALL_STATE(980)] = 74916,
  [SMALL_STATE(981)] = 74967,
  [SMALL_STATE(982)] = 75016,
  [SMALL_STATE(983)] = 75063,
  [SMALL_STATE(984)] = 75118,
  [SMALL_STATE(985)] = 75173,
  [SMALL_STATE(986)] = 75210,
  [SMALL_STATE(987)] = 75275,
  [SMALL_STATE(988)] = 75338,
  [SMALL_STATE(989)] = 75401,
  [SMALL_STATE(990)] = 75464,
  [SMALL_STATE(991)] = 75501,
  [SMALL_STATE(992)] = 75538,
  [SMALL_STATE(993)] = 75601,
  [SMALL_STATE(994)] = 75664,
  [SMALL_STATE(995)] = 75701,
  [SMALL_STATE(996)] = 75756,
  [SMALL_STATE(997)] = 75819,
  [SMALL_STATE(998)] = 75884,
  [SMALL_STATE(999)] = 75923,
  [SMALL_STATE(1000)] = 75986,
  [SMALL_STATE(1001)] = 76041,
  [SMALL_STATE(1002)] = 76104,
  [SMALL_STATE(1003)] = 76167,
  [SMALL_STATE(1004)] = 76222,
  [SMALL_STATE(1005)] = 76277,
  [SMALL_STATE(1006)] = 76332,
  [SMALL_STATE(1007)] = 76371,
  [SMALL_STATE(1008)] = 76412,
  [SMALL_STATE(1009)] = 76457,
  [SMALL_STATE(1010)] = 76493,
  [SMALL_STATE(1011)] = 76529,
  [SMALL_STATE(1012)] = 76565,
  [SMALL_STATE(1013)] = 76619,
  [SMALL_STATE(1014)] = 76673,
  [SMALL_STATE(1015)] = 76727,
  [SMALL_STATE(1016)] = 76781,
  [SMALL_STATE(1017)] = 76841,
  [SMALL_STATE(1018)] = 76895,
  [SMALL_STATE(1019)] = 76949,
  [SMALL_STATE(1020)] = 77003,
  [SMALL_STATE(1021)] = 77057,
  [SMALL_STATE(1022)] = 77111,
  [SMALL_STATE(1023)] = 77147,
  [SMALL_STATE(1024)] = 77207,
  [SMALL_STATE(1025)] = 77267,
  [SMALL_STATE(1026)] = 77321,
  [SMALL_STATE(1027)] = 77375,
  [SMALL_STATE(1028)] = 77429,
  [SMALL_STATE(1029)] = 77483,
  [SMALL_STATE(1030)] = 77537,
  [SMALL_STATE(1031)] = 77591,
  [SMALL_STATE(1032)] = 77651,
  [SMALL_STATE(1033)] = 77705,
  [SMALL_STATE(1034)] = 77765,
  [SMALL_STATE(1035)] = 77819,
  [SMALL_STATE(1036)] = 77873,
  [SMALL_STATE(1037)] = 77927,
  [SMALL_STATE(1038)] = 77981,
  [SMALL_STATE(1039)] = 78041,
  [SMALL_STATE(1040)] = 78101,
  [SMALL_STATE(1041)] = 78161,
  [SMALL_STATE(1042)] = 78221,
  [SMALL_STATE(1043)] = 78257,
  [SMALL_STATE(1044)] = 78317,
  [SMALL_STATE(1045)] = 78377,
  [SMALL_STATE(1046)] = 78413,
  [SMALL_STATE(1047)] = 78453,
  [SMALL_STATE(1048)] = 78489,
  [SMALL_STATE(1049)] = 78525,
  [SMALL_STATE(1050)] = 78561,
  [SMALL_STATE(1051)] = 78597,
  [SMALL_STATE(1052)] = 78657,
  [SMALL_STATE(1053)] = 78711,
  [SMALL_STATE(1054)] = 78747,
  [SMALL_STATE(1055)] = 78807,
  [SMALL_STATE(1056)] = 78861,
  [SMALL_STATE(1057)] = 78921,
  [SMALL_STATE(1058)] = 78957,
  [SMALL_STATE(1059)] = 79011,
  [SMALL_STATE(1060)] = 79065,
  [SMALL_STATE(1061)] = 79119,
  [SMALL_STATE(1062)] = 79173,
  [SMALL_STATE(1063)] = 79233,
  [SMALL_STATE(1064)] = 79287,
  [SMALL_STATE(1065)] = 79323,
  [SMALL_STATE(1066)] = 79377,
  [SMALL_STATE(1067)] = 79437,
  [SMALL_STATE(1068)] = 79491,
  [SMALL_STATE(1069)] = 79527,
  [SMALL_STATE(1070)] = 79563,
  [SMALL_STATE(1071)] = 79603,
  [SMALL_STATE(1072)] = 79655,
  [SMALL_STATE(1073)] = 79709,
  [SMALL_STATE(1074)] = 79763,
  [SMALL_STATE(1075)] = 79813,
  [SMALL_STATE(1076)] = 79861,
  [SMALL_STATE(1077)] = 79907,
  [SMALL_STATE(1078)] = 79951,
  [SMALL_STATE(1079)] = 79987,
  [SMALL_STATE(1080)] = 80047,
  [SMALL_STATE(1081)] = 80101,
  [SMALL_STATE(1082)] = 80155,
  [SMALL_STATE(1083)] = 80191,
  [SMALL_STATE(1084)] = 80227,
  [SMALL_STATE(1085)] = 80263,
  [SMALL_STATE(1086)] = 80299,
  [SMALL_STATE(1087)] = 80359,
  [SMALL_STATE(1088)] = 80413,
  [SMALL_STATE(1089)] = 80451,
  [SMALL_STATE(1090)] = 80487,
  [SMALL_STATE(1091)] = 80547,
  [SMALL_STATE(1092)] = 80582,
  [SMALL_STATE(1093)] = 80639,
  [SMALL_STATE(1094)] = 80696,
  [SMALL_STATE(1095)] = 80753,
  [SMALL_STATE(1096)] = 80810,
  [SMALL_STATE(1097)] = 80867,
  [SMALL_STATE(1098)] = 80924,
  [SMALL_STATE(1099)] = 80961,
  [SMALL_STATE(1100)] = 81018,
  [SMALL_STATE(1101)] = 81075,
  [SMALL_STATE(1102)] = 81132,
  [SMALL_STATE(1103)] = 81169,
  [SMALL_STATE(1104)] = 81204,
  [SMALL_STATE(1105)] = 81261,
  [SMALL_STATE(1106)] = 81311,
  [SMALL_STATE(1107)] = 81345,
  [SMALL_STATE(1108)] = 81379,
  [SMALL_STATE(1109)] = 81413,
  [SMALL_STATE(1110)] = 81451,
  [SMALL_STATE(1111)] = 81499,
  [SMALL_STATE(1112)] = 81533,
  [SMALL_STATE(1113)] = 81565,
  [SMALL_STATE(1114)] = 81611,
  [SMALL_STATE(1115)] = 81655,
  [SMALL_STATE(1116)] = 81697,
  [SMALL_STATE(1117)] = 81731,
  [SMALL_STATE(1118)] = 81765,
  [SMALL_STATE(1119)] = 81799,
  [SMALL_STATE(1120)] = 81841,
  [SMALL_STATE(1121)] = 81875,
  [SMALL_STATE(1122)] = 81909,
  [SMALL_STATE(1123)] = 81943,
  [SMALL_STATE(1124)] = 81977,
  [SMALL_STATE(1125)] = 82021,
  [SMALL_STATE(1126)] = 82067,
  [SMALL_STATE(1127)] = 82117,
  [SMALL_STATE(1128)] = 82165,
  [SMALL_STATE(1129)] = 82203,
  [SMALL_STATE(1130)] = 82237,
  [SMALL_STATE(1131)] = 82271,
  [SMALL_STATE(1132)] = 82321,
  [SMALL_STATE(1133)] = 82355,
  [SMALL_STATE(1134)] = 82389,
  [SMALL_STATE(1135)] = 82423,
  [SMALL_STATE(1136)] = 82457,
  [SMALL_STATE(1137)] = 82491,
  [SMALL_STATE(1138)] = 82525,
  [SMALL_STATE(1139)] = 82559,
  [SMALL_STATE(1140)] = 82593,
  [SMALL_STATE(1141)] = 82627,
  [SMALL_STATE(1142)] = 82661,
  [SMALL_STATE(1143)] = 82695,
  [SMALL_STATE(1144)] = 82729,
  [SMALL_STATE(1145)] = 82763,
  [SMALL_STATE(1146)] = 82797,
  [SMALL_STATE(1147)] = 82831,
  [SMALL_STATE(1148)] = 82865,
  [SMALL_STATE(1149)] = 82912,
  [SMALL_STATE(1150)] = 82959,
  [SMALL_STATE(1151)] = 83006,
  [SMALL_STATE(1152)] = 83053,
  [SMALL_STATE(1153)] = 83100,
  [SMALL_STATE(1154)] = 83147,
  [SMALL_STATE(1155)] = 83191,
  [SMALL_STATE(1156)] = 83235,
  [SMALL_STATE(1157)] = 83279,
  [SMALL_STATE(1158)] = 83323,
  [SMALL_STATE(1159)] = 83367,
  [SMALL_STATE(1160)] = 83411,
  [SMALL_STATE(1161)] = 83455,
  [SMALL_STATE(1162)] = 83485,
  [SMALL_STATE(1163)] = 83513,
  [SMALL_STATE(1164)] = 83539,
  [SMALL_STATE(1165)] = 83583,
  [SMALL_STATE(1166)] = 83627,
  [SMALL_STATE(1167)] = 83671,
  [SMALL_STATE(1168)] = 83696,
  [SMALL_STATE(1169)] = 83737,
  [SMALL_STATE(1170)] = 83778,
  [SMALL_STATE(1171)] = 83821,
  [SMALL_STATE(1172)] = 83862,
  [SMALL_STATE(1173)] = 83903,
  [SMALL_STATE(1174)] = 83946,
  [SMALL_STATE(1175)] = 83987,
  [SMALL_STATE(1176)] = 84028,
  [SMALL_STATE(1177)] = 84071,
  [SMALL_STATE(1178)] = 84112,
  [SMALL_STATE(1179)] = 84153,
  [SMALL_STATE(1180)] = 84194,
  [SMALL_STATE(1181)] = 84223,
  [SMALL_STATE(1182)] = 84264,
  [SMALL_STATE(1183)] = 84305,
  [SMALL_STATE(1184)] = 84343,
  [SMALL_STATE(1185)] = 84383,
  [SMALL_STATE(1186)] = 84407,
  [SMALL_STATE(1187)] = 84445,
  [SMALL_STATE(1188)] = 84483,
  [SMALL_STATE(1189)] = 84521,
  [SMALL_STATE(1190)] = 84545,
  [SMALL_STATE(1191)] = 84583,
  [SMALL_STATE(1192)] = 84623,
  [SMALL_STATE(1193)] = 84663,
  [SMALL_STATE(1194)] = 84701,
  [SMALL_STATE(1195)] = 84741,
  [SMALL_STATE(1196)] = 84779,
  [SMALL_STATE(1197)] = 84803,
  [SMALL_STATE(1198)] = 84841,
  [SMALL_STATE(1199)] = 84865,
  [SMALL_STATE(1200)] = 84889,
  [SMALL_STATE(1201)] = 84927,
  [SMALL_STATE(1202)] = 84967,
  [SMALL_STATE(1203)] = 85005,
  [SMALL_STATE(1204)] = 85043,
  [SMALL_STATE(1205)] = 85083,
  [SMALL_STATE(1206)] = 85107,
  [SMALL_STATE(1207)] = 85145,
  [SMALL_STATE(1208)] = 85183,
  [SMALL_STATE(1209)] = 85221,
  [SMALL_STATE(1210)] = 85259,
  [SMALL_STATE(1211)] = 85297,
  [SMALL_STATE(1212)] = 85335,
  [SMALL_STATE(1213)] = 85373,
  [SMALL_STATE(1214)] = 85411,
  [SMALL_STATE(1215)] = 85451,
  [SMALL_STATE(1216)] = 85475,
  [SMALL_STATE(1217)] = 85513,
  [SMALL_STATE(1218)] = 85551,
  [SMALL_STATE(1219)] = 85575,
  [SMALL_STATE(1220)] = 85613,
  [SMALL_STATE(1221)] = 85651,
  [SMALL_STATE(1222)] = 85675,
  [SMALL_STATE(1223)] = 85713,
  [SMALL_STATE(1224)] = 85751,
  [SMALL_STATE(1225)] = 85789,
  [SMALL_STATE(1226)] = 85813,
  [SMALL_STATE(1227)] = 85837,
  [SMALL_STATE(1228)] = 85875,
  [SMALL_STATE(1229)] = 85913,
  [SMALL_STATE(1230)] = 85951,
  [SMALL_STATE(1231)] = 85975,
  [SMALL_STATE(1232)] = 86012,
  [SMALL_STATE(1233)] = 86035,
  [SMALL_STATE(1234)] = 86058,
  [SMALL_STATE(1235)] = 86081,
  [SMALL_STATE(1236)] = 86104,
  [SMALL_STATE(1237)] = 86127,
  [SMALL_STATE(1238)] = 86150,
  [SMALL_STATE(1239)] = 86173,
  [SMALL_STATE(1240)] = 86196,
  [SMALL_STATE(1241)] = 86219,
  [SMALL_STATE(1242)] = 86242,
  [SMALL_STATE(1243)] = 86265,
  [SMALL_STATE(1244)] = 86288,
  [SMALL_STATE(1245)] = 86311,
  [SMALL_STATE(1246)] = 86348,
  [SMALL_STATE(1247)] = 86371,
  [SMALL_STATE(1248)] = 86398,
  [SMALL_STATE(1249)] = 86421,
  [SMALL_STATE(1250)] = 86444,
  [SMALL_STATE(1251)] = 86471,
  [SMALL_STATE(1252)] = 86494,
  [SMALL_STATE(1253)] = 86531,
  [SMALL_STATE(1254)] = 86554,
  [SMALL_STATE(1255)] = 86577,
  [SMALL_STATE(1256)] = 86600,
  [SMALL_STATE(1257)] = 86623,
  [SMALL_STATE(1258)] = 86646,
  [SMALL_STATE(1259)] = 86673,
  [SMALL_STATE(1260)] = 86696,
  [SMALL_STATE(1261)] = 86719,
  [SMALL_STATE(1262)] = 86746,
  [SMALL_STATE(1263)] = 86769,
  [SMALL_STATE(1264)] = 86806,
  [SMALL_STATE(1265)] = 86829,
  [SMALL_STATE(1266)] = 86852,
  [SMALL_STATE(1267)] = 86875,
  [SMALL_STATE(1268)] = 86899,
  [SMALL_STATE(1269)] = 86920,
  [SMALL_STATE(1270)] = 86945,
  [SMALL_STATE(1271)] = 86966,
  [SMALL_STATE(1272)] = 86987,
  [SMALL_STATE(1273)] = 87028,
  [SMALL_STATE(1274)] = 87049,
  [SMALL_STATE(1275)] = 87070,
  [SMALL_STATE(1276)] = 87111,
  [SMALL_STATE(1277)] = 87136,
  [SMALL_STATE(1278)] = 87157,
  [SMALL_STATE(1279)] = 87198,
  [SMALL_STATE(1280)] = 87239,
  [SMALL_STATE(1281)] = 87260,
  [SMALL_STATE(1282)] = 87285,
  [SMALL_STATE(1283)] = 87310,
  [SMALL_STATE(1284)] = 87331,
  [SMALL_STATE(1285)] = 87356,
  [SMALL_STATE(1286)] = 87381,
  [SMALL_STATE(1287)] = 87402,
  [SMALL_STATE(1288)] = 87443,
  [SMALL_STATE(1289)] = 87464,
  [SMALL_STATE(1290)] = 87505,
  [SMALL_STATE(1291)] = 87527,
  [SMALL_STATE(1292)] = 87547,
  [SMALL_STATE(1293)] = 87569,
  [SMALL_STATE(1294)] = 87591,
  [SMALL_STATE(1295)] = 87611,
  [SMALL_STATE(1296)] = 87633,
  [SMALL_STATE(1297)] = 87667,
  [SMALL_STATE(1298)] = 87687,
  [SMALL_STATE(1299)] = 87725,
  [SMALL_STATE(1300)] = 87747,
  [SMALL_STATE(1301)] = 87781,
  [SMALL_STATE(1302)] = 87801,
  [SMALL_STATE(1303)] = 87820,
  [SMALL_STATE(1304)] = 87839,
  [SMALL_STATE(1305)] = 87860,
  [SMALL_STATE(1306)] = 87879,
  [SMALL_STATE(1307)] = 87898,
  [SMALL_STATE(1308)] = 87917,
  [SMALL_STATE(1309)] = 87936,
  [SMALL_STATE(1310)] = 87959,
  [SMALL_STATE(1311)] = 87978,
  [SMALL_STATE(1312)] = 87997,
  [SMALL_STATE(1313)] = 88016,
  [SMALL_STATE(1314)] = 88035,
  [SMALL_STATE(1315)] = 88054,
  [SMALL_STATE(1316)] = 88073,
  [SMALL_STATE(1317)] = 88094,
  [SMALL_STATE(1318)] = 88113,
  [SMALL_STATE(1319)] = 88132,
  [SMALL_STATE(1320)] = 88151,
  [SMALL_STATE(1321)] = 88170,
  [SMALL_STATE(1322)] = 88189,
  [SMALL_STATE(1323)] = 88208,
  [SMALL_STATE(1324)] = 88227,
  [SMALL_STATE(1325)] = 88246,
  [SMALL_STATE(1326)] = 88265,
  [SMALL_STATE(1327)] = 88284,
  [SMALL_STATE(1328)] = 88303,
  [SMALL_STATE(1329)] = 88322,
  [SMALL_STATE(1330)] = 88349,
  [SMALL_STATE(1331)] = 88368,
  [SMALL_STATE(1332)] = 88387,
  [SMALL_STATE(1333)] = 88406,
  [SMALL_STATE(1334)] = 88425,
  [SMALL_STATE(1335)] = 88444,
  [SMALL_STATE(1336)] = 88463,
  [SMALL_STATE(1337)] = 88482,
  [SMALL_STATE(1338)] = 88501,
  [SMALL_STATE(1339)] = 88533,
  [SMALL_STATE(1340)] = 88565,
  [SMALL_STATE(1341)] = 88583,
  [SMALL_STATE(1342)] = 88615,
  [SMALL_STATE(1343)] = 88635,
  [SMALL_STATE(1344)] = 88653,
  [SMALL_STATE(1345)] = 88671,
  [SMALL_STATE(1346)] = 88700,
  [SMALL_STATE(1347)] = 88729,
  [SMALL_STATE(1348)] = 88758,
  [SMALL_STATE(1349)] = 88787,
  [SMALL_STATE(1350)] = 88816,
  [SMALL_STATE(1351)] = 88845,
  [SMALL_STATE(1352)] = 88862,
  [SMALL_STATE(1353)] = 88891,
  [SMALL_STATE(1354)] = 88913,
  [SMALL_STATE(1355)] = 88935,
  [SMALL_STATE(1356)] = 88959,
  [SMALL_STATE(1357)] = 88981,
  [SMALL_STATE(1358)] = 89003,
  [SMALL_STATE(1359)] = 89025,
  [SMALL_STATE(1360)] = 89049,
  [SMALL_STATE(1361)] = 89071,
  [SMALL_STATE(1362)] = 89093,
  [SMALL_STATE(1363)] = 89115,
  [SMALL_STATE(1364)] = 89137,
  [SMALL_STATE(1365)] = 89161,
  [SMALL_STATE(1366)] = 89183,
  [SMALL_STATE(1367)] = 89205,
  [SMALL_STATE(1368)] = 89227,
  [SMALL_STATE(1369)] = 89251,
  [SMALL_STATE(1370)] = 89273,
  [SMALL_STATE(1371)] = 89297,
  [SMALL_STATE(1372)] = 89321,
  [SMALL_STATE(1373)] = 89345,
  [SMALL_STATE(1374)] = 89367,
  [SMALL_STATE(1375)] = 89389,
  [SMALL_STATE(1376)] = 89411,
  [SMALL_STATE(1377)] = 89435,
  [SMALL_STATE(1378)] = 89457,
  [SMALL_STATE(1379)] = 89479,
  [SMALL_STATE(1380)] = 89501,
  [SMALL_STATE(1381)] = 89519,
  [SMALL_STATE(1382)] = 89541,
  [SMALL_STATE(1383)] = 89563,
  [SMALL_STATE(1384)] = 89587,
  [SMALL_STATE(1385)] = 89611,
  [SMALL_STATE(1386)] = 89633,
  [SMALL_STATE(1387)] = 89653,
  [SMALL_STATE(1388)] = 89675,
  [SMALL_STATE(1389)] = 89697,
  [SMALL_STATE(1390)] = 89719,
  [SMALL_STATE(1391)] = 89743,
  [SMALL_STATE(1392)] = 89765,
  [SMALL_STATE(1393)] = 89787,
  [SMALL_STATE(1394)] = 89811,
  [SMALL_STATE(1395)] = 89837,
  [SMALL_STATE(1396)] = 89856,
  [SMALL_STATE(1397)] = 89879,
  [SMALL_STATE(1398)] = 89898,
  [SMALL_STATE(1399)] = 89917,
  [SMALL_STATE(1400)] = 89938,
  [SMALL_STATE(1401)] = 89961,
  [SMALL_STATE(1402)] = 89980,
  [SMALL_STATE(1403)] = 90001,
  [SMALL_STATE(1404)] = 90022,
  [SMALL_STATE(1405)] = 90045,
  [SMALL_STATE(1406)] = 90066,
  [SMALL_STATE(1407)] = 90085,
  [SMALL_STATE(1408)] = 90106,
  [SMALL_STATE(1409)] = 90125,
  [SMALL_STATE(1410)] = 90146,
  [SMALL_STATE(1411)] = 90165,
  [SMALL_STATE(1412)] = 90184,
  [SMALL_STATE(1413)] = 90207,
  [SMALL_STATE(1414)] = 90228,
  [SMALL_STATE(1415)] = 90249,
  [SMALL_STATE(1416)] = 90268,
  [SMALL_STATE(1417)] = 90291,
  [SMALL_STATE(1418)] = 90314,
  [SMALL_STATE(1419)] = 90335,
  [SMALL_STATE(1420)] = 90354,
  [SMALL_STATE(1421)] = 90374,
  [SMALL_STATE(1422)] = 90392,
  [SMALL_STATE(1423)] = 90412,
  [SMALL_STATE(1424)] = 90428,
  [SMALL_STATE(1425)] = 90446,
  [SMALL_STATE(1426)] = 90462,
  [SMALL_STATE(1427)] = 90478,
  [SMALL_STATE(1428)] = 90498,
  [SMALL_STATE(1429)] = 90518,
  [SMALL_STATE(1430)] = 90536,
  [SMALL_STATE(1431)] = 90556,
  [SMALL_STATE(1432)] = 90576,
  [SMALL_STATE(1433)] = 90592,
  [SMALL_STATE(1434)] = 90610,
  [SMALL_STATE(1435)] = 90626,
  [SMALL_STATE(1436)] = 90646,
  [SMALL_STATE(1437)] = 90664,
  [SMALL_STATE(1438)] = 90682,
  [SMALL_STATE(1439)] = 90700,
  [SMALL_STATE(1440)] = 90720,
  [SMALL_STATE(1441)] = 90740,
  [SMALL_STATE(1442)] = 90760,
  [SMALL_STATE(1443)] = 90778,
  [SMALL_STATE(1444)] = 90796,
  [SMALL_STATE(1445)] = 90814,
  [SMALL_STATE(1446)] = 90834,
  [SMALL_STATE(1447)] = 90852,
  [SMALL_STATE(1448)] = 90870,
  [SMALL_STATE(1449)] = 90890,
  [SMALL_STATE(1450)] = 90908,
  [SMALL_STATE(1451)] = 90928,
  [SMALL_STATE(1452)] = 90946,
  [SMALL_STATE(1453)] = 90966,
  [SMALL_STATE(1454)] = 90984,
  [SMALL_STATE(1455)] = 91002,
  [SMALL_STATE(1456)] = 91020,
  [SMALL_STATE(1457)] = 91038,
  [SMALL_STATE(1458)] = 91058,
  [SMALL_STATE(1459)] = 91078,
  [SMALL_STATE(1460)] = 91098,
  [SMALL_STATE(1461)] = 91116,
  [SMALL_STATE(1462)] = 91132,
  [SMALL_STATE(1463)] = 91148,
  [SMALL_STATE(1464)] = 91162,
  [SMALL_STATE(1465)] = 91180,
  [SMALL_STATE(1466)] = 91198,
  [SMALL_STATE(1467)] = 91218,
  [SMALL_STATE(1468)] = 91236,
  [SMALL_STATE(1469)] = 91256,
  [SMALL_STATE(1470)] = 91272,
  [SMALL_STATE(1471)] = 91290,
  [SMALL_STATE(1472)] = 91310,
  [SMALL_STATE(1473)] = 91328,
  [SMALL_STATE(1474)] = 91348,
  [SMALL_STATE(1475)] = 91366,
  [SMALL_STATE(1476)] = 91384,
  [SMALL_STATE(1477)] = 91402,
  [SMALL_STATE(1478)] = 91416,
  [SMALL_STATE(1479)] = 91433,
  [SMALL_STATE(1480)] = 91450,
  [SMALL_STATE(1481)] = 91463,
  [SMALL_STATE(1482)] = 91480,
  [SMALL_STATE(1483)] = 91497,
  [SMALL_STATE(1484)] = 91514,
  [SMALL_STATE(1485)] = 91531,
  [SMALL_STATE(1486)] = 91548,
  [SMALL_STATE(1487)] = 91565,
  [SMALL_STATE(1488)] = 91582,
  [SMALL_STATE(1489)] = 91599,
  [SMALL_STATE(1490)] = 91616,
  [SMALL_STATE(1491)] = 91633,
  [SMALL_STATE(1492)] = 91650,
  [SMALL_STATE(1493)] = 91667,
  [SMALL_STATE(1494)] = 91680,
  [SMALL_STATE(1495)] = 91693,
  [SMALL_STATE(1496)] = 91710,
  [SMALL_STATE(1497)] = 91727,
  [SMALL_STATE(1498)] = 91744,
  [SMALL_STATE(1499)] = 91761,
  [SMALL_STATE(1500)] = 91774,
  [SMALL_STATE(1501)] = 91791,
  [SMALL_STATE(1502)] = 91808,
  [SMALL_STATE(1503)] = 91821,
  [SMALL_STATE(1504)] = 91838,
  [SMALL_STATE(1505)] = 91855,
  [SMALL_STATE(1506)] = 91872,
  [SMALL_STATE(1507)] = 91889,
  [SMALL_STATE(1508)] = 91902,
  [SMALL_STATE(1509)] = 91919,
  [SMALL_STATE(1510)] = 91936,
  [SMALL_STATE(1511)] = 91949,
  [SMALL_STATE(1512)] = 91966,
  [SMALL_STATE(1513)] = 91983,
  [SMALL_STATE(1514)] = 92000,
  [SMALL_STATE(1515)] = 92017,
  [SMALL_STATE(1516)] = 92034,
  [SMALL_STATE(1517)] = 92051,
  [SMALL_STATE(1518)] = 92068,
  [SMALL_STATE(1519)] = 92085,
  [SMALL_STATE(1520)] = 92102,
  [SMALL_STATE(1521)] = 92119,
  [SMALL_STATE(1522)] = 92136,
  [SMALL_STATE(1523)] = 92153,
  [SMALL_STATE(1524)] = 92170,
  [SMALL_STATE(1525)] = 92187,
  [SMALL_STATE(1526)] = 92204,
  [SMALL_STATE(1527)] = 92221,
  [SMALL_STATE(1528)] = 92238,
  [SMALL_STATE(1529)] = 92253,
  [SMALL_STATE(1530)] = 92270,
  [SMALL_STATE(1531)] = 92287,
  [SMALL_STATE(1532)] = 92300,
  [SMALL_STATE(1533)] = 92317,
  [SMALL_STATE(1534)] = 92334,
  [SMALL_STATE(1535)] = 92351,
  [SMALL_STATE(1536)] = 92364,
  [SMALL_STATE(1537)] = 92381,
  [SMALL_STATE(1538)] = 92398,
  [SMALL_STATE(1539)] = 92411,
  [SMALL_STATE(1540)] = 92428,
  [SMALL_STATE(1541)] = 92445,
  [SMALL_STATE(1542)] = 92462,
  [SMALL_STATE(1543)] = 92475,
  [SMALL_STATE(1544)] = 92492,
  [SMALL_STATE(1545)] = 92505,
  [SMALL_STATE(1546)] = 92522,
  [SMALL_STATE(1547)] = 92539,
  [SMALL_STATE(1548)] = 92552,
  [SMALL_STATE(1549)] = 92569,
  [SMALL_STATE(1550)] = 92586,
  [SMALL_STATE(1551)] = 92603,
  [SMALL_STATE(1552)] = 92618,
  [SMALL_STATE(1553)] = 92635,
  [SMALL_STATE(1554)] = 92652,
  [SMALL_STATE(1555)] = 92665,
  [SMALL_STATE(1556)] = 92682,
  [SMALL_STATE(1557)] = 92699,
  [SMALL_STATE(1558)] = 92716,
  [SMALL_STATE(1559)] = 92733,
  [SMALL_STATE(1560)] = 92750,
  [SMALL_STATE(1561)] = 92767,
  [SMALL_STATE(1562)] = 92784,
  [SMALL_STATE(1563)] = 92797,
  [SMALL_STATE(1564)] = 92814,
  [SMALL_STATE(1565)] = 92827,
  [SMALL_STATE(1566)] = 92840,
  [SMALL_STATE(1567)] = 92857,
  [SMALL_STATE(1568)] = 92870,
  [SMALL_STATE(1569)] = 92887,
  [SMALL_STATE(1570)] = 92904,
  [SMALL_STATE(1571)] = 92921,
  [SMALL_STATE(1572)] = 92938,
  [SMALL_STATE(1573)] = 92955,
  [SMALL_STATE(1574)] = 92972,
  [SMALL_STATE(1575)] = 92989,
  [SMALL_STATE(1576)] = 93006,
  [SMALL_STATE(1577)] = 93019,
  [SMALL_STATE(1578)] = 93036,
  [SMALL_STATE(1579)] = 93053,
  [SMALL_STATE(1580)] = 93070,
  [SMALL_STATE(1581)] = 93087,
  [SMALL_STATE(1582)] = 93104,
  [SMALL_STATE(1583)] = 93119,
  [SMALL_STATE(1584)] = 93136,
  [SMALL_STATE(1585)] = 93153,
  [SMALL_STATE(1586)] = 93170,
  [SMALL_STATE(1587)] = 93187,
  [SMALL_STATE(1588)] = 93204,
  [SMALL_STATE(1589)] = 93217,
  [SMALL_STATE(1590)] = 93234,
  [SMALL_STATE(1591)] = 93251,
  [SMALL_STATE(1592)] = 93266,
  [SMALL_STATE(1593)] = 93283,
  [SMALL_STATE(1594)] = 93300,
  [SMALL_STATE(1595)] = 93317,
  [SMALL_STATE(1596)] = 93334,
  [SMALL_STATE(1597)] = 93351,
  [SMALL_STATE(1598)] = 93368,
  [SMALL_STATE(1599)] = 93385,
  [SMALL_STATE(1600)] = 93402,
  [SMALL_STATE(1601)] = 93419,
  [SMALL_STATE(1602)] = 93436,
  [SMALL_STATE(1603)] = 93453,
  [SMALL_STATE(1604)] = 93470,
  [SMALL_STATE(1605)] = 93487,
  [SMALL_STATE(1606)] = 93504,
  [SMALL_STATE(1607)] = 93521,
  [SMALL_STATE(1608)] = 93538,
  [SMALL_STATE(1609)] = 93555,
  [SMALL_STATE(1610)] = 93572,
  [SMALL_STATE(1611)] = 93589,
  [SMALL_STATE(1612)] = 93606,
  [SMALL_STATE(1613)] = 93623,
  [SMALL_STATE(1614)] = 93636,
  [SMALL_STATE(1615)] = 93653,
  [SMALL_STATE(1616)] = 93666,
  [SMALL_STATE(1617)] = 93683,
  [SMALL_STATE(1618)] = 93700,
  [SMALL_STATE(1619)] = 93717,
  [SMALL_STATE(1620)] = 93730,
  [SMALL_STATE(1621)] = 93743,
  [SMALL_STATE(1622)] = 93760,
  [SMALL_STATE(1623)] = 93777,
  [SMALL_STATE(1624)] = 93794,
  [SMALL_STATE(1625)] = 93811,
  [SMALL_STATE(1626)] = 93828,
  [SMALL_STATE(1627)] = 93845,
  [SMALL_STATE(1628)] = 93862,
  [SMALL_STATE(1629)] = 93879,
  [SMALL_STATE(1630)] = 93896,
  [SMALL_STATE(1631)] = 93913,
  [SMALL_STATE(1632)] = 93930,
  [SMALL_STATE(1633)] = 93947,
  [SMALL_STATE(1634)] = 93960,
  [SMALL_STATE(1635)] = 93977,
  [SMALL_STATE(1636)] = 93994,
  [SMALL_STATE(1637)] = 94011,
  [SMALL_STATE(1638)] = 94028,
  [SMALL_STATE(1639)] = 94043,
  [SMALL_STATE(1640)] = 94060,
  [SMALL_STATE(1641)] = 94077,
  [SMALL_STATE(1642)] = 94094,
  [SMALL_STATE(1643)] = 94111,
  [SMALL_STATE(1644)] = 94128,
  [SMALL_STATE(1645)] = 94145,
  [SMALL_STATE(1646)] = 94162,
  [SMALL_STATE(1647)] = 94179,
  [SMALL_STATE(1648)] = 94192,
  [SMALL_STATE(1649)] = 94209,
  [SMALL_STATE(1650)] = 94226,
  [SMALL_STATE(1651)] = 94243,
  [SMALL_STATE(1652)] = 94260,
  [SMALL_STATE(1653)] = 94273,
  [SMALL_STATE(1654)] = 94290,
  [SMALL_STATE(1655)] = 94307,
  [SMALL_STATE(1656)] = 94324,
  [SMALL_STATE(1657)] = 94341,
  [SMALL_STATE(1658)] = 94358,
  [SMALL_STATE(1659)] = 94371,
  [SMALL_STATE(1660)] = 94388,
  [SMALL_STATE(1661)] = 94405,
  [SMALL_STATE(1662)] = 94420,
  [SMALL_STATE(1663)] = 94435,
  [SMALL_STATE(1664)] = 94452,
  [SMALL_STATE(1665)] = 94469,
  [SMALL_STATE(1666)] = 94481,
  [SMALL_STATE(1667)] = 94495,
  [SMALL_STATE(1668)] = 94509,
  [SMALL_STATE(1669)] = 94523,
  [SMALL_STATE(1670)] = 94537,
  [SMALL_STATE(1671)] = 94551,
  [SMALL_STATE(1672)] = 94565,
  [SMALL_STATE(1673)] = 94579,
  [SMALL_STATE(1674)] = 94591,
  [SMALL_STATE(1675)] = 94605,
  [SMALL_STATE(1676)] = 94617,
  [SMALL_STATE(1677)] = 94631,
  [SMALL_STATE(1678)] = 94645,
  [SMALL_STATE(1679)] = 94659,
  [SMALL_STATE(1680)] = 94671,
  [SMALL_STATE(1681)] = 94685,
  [SMALL_STATE(1682)] = 94697,
  [SMALL_STATE(1683)] = 94711,
  [SMALL_STATE(1684)] = 94725,
  [SMALL_STATE(1685)] = 94739,
  [SMALL_STATE(1686)] = 94753,
  [SMALL_STATE(1687)] = 94767,
  [SMALL_STATE(1688)] = 94781,
  [SMALL_STATE(1689)] = 94795,
  [SMALL_STATE(1690)] = 94809,
  [SMALL_STATE(1691)] = 94821,
  [SMALL_STATE(1692)] = 94835,
  [SMALL_STATE(1693)] = 94847,
  [SMALL_STATE(1694)] = 94859,
  [SMALL_STATE(1695)] = 94873,
  [SMALL_STATE(1696)] = 94887,
  [SMALL_STATE(1697)] = 94901,
  [SMALL_STATE(1698)] = 94913,
  [SMALL_STATE(1699)] = 94925,
  [SMALL_STATE(1700)] = 94939,
  [SMALL_STATE(1701)] = 94953,
  [SMALL_STATE(1702)] = 94965,
  [SMALL_STATE(1703)] = 94979,
  [SMALL_STATE(1704)] = 94991,
  [SMALL_STATE(1705)] = 95005,
  [SMALL_STATE(1706)] = 95017,
  [SMALL_STATE(1707)] = 95031,
  [SMALL_STATE(1708)] = 95045,
  [SMALL_STATE(1709)] = 95059,
  [SMALL_STATE(1710)] = 95073,
  [SMALL_STATE(1711)] = 95085,
  [SMALL_STATE(1712)] = 95099,
  [SMALL_STATE(1713)] = 95111,
  [SMALL_STATE(1714)] = 95125,
  [SMALL_STATE(1715)] = 95139,
  [SMALL_STATE(1716)] = 95153,
  [SMALL_STATE(1717)] = 95165,
  [SMALL_STATE(1718)] = 95177,
  [SMALL_STATE(1719)] = 95191,
  [SMALL_STATE(1720)] = 95205,
  [SMALL_STATE(1721)] = 95217,
  [SMALL_STATE(1722)] = 95231,
  [SMALL_STATE(1723)] = 95245,
  [SMALL_STATE(1724)] = 95257,
  [SMALL_STATE(1725)] = 95269,
  [SMALL_STATE(1726)] = 95283,
  [SMALL_STATE(1727)] = 95297,
  [SMALL_STATE(1728)] = 95309,
  [SMALL_STATE(1729)] = 95321,
  [SMALL_STATE(1730)] = 95333,
  [SMALL_STATE(1731)] = 95345,
  [SMALL_STATE(1732)] = 95357,
  [SMALL_STATE(1733)] = 95369,
  [SMALL_STATE(1734)] = 95383,
  [SMALL_STATE(1735)] = 95397,
  [SMALL_STATE(1736)] = 95409,
  [SMALL_STATE(1737)] = 95423,
  [SMALL_STATE(1738)] = 95437,
  [SMALL_STATE(1739)] = 95449,
  [SMALL_STATE(1740)] = 95461,
  [SMALL_STATE(1741)] = 95473,
  [SMALL_STATE(1742)] = 95487,
  [SMALL_STATE(1743)] = 95501,
  [SMALL_STATE(1744)] = 95515,
  [SMALL_STATE(1745)] = 95529,
  [SMALL_STATE(1746)] = 95543,
  [SMALL_STATE(1747)] = 95555,
  [SMALL_STATE(1748)] = 95567,
  [SMALL_STATE(1749)] = 95579,
  [SMALL_STATE(1750)] = 95591,
  [SMALL_STATE(1751)] = 95603,
  [SMALL_STATE(1752)] = 95615,
  [SMALL_STATE(1753)] = 95627,
  [SMALL_STATE(1754)] = 95641,
  [SMALL_STATE(1755)] = 95655,
  [SMALL_STATE(1756)] = 95669,
  [SMALL_STATE(1757)] = 95681,
  [SMALL_STATE(1758)] = 95693,
  [SMALL_STATE(1759)] = 95705,
  [SMALL_STATE(1760)] = 95719,
  [SMALL_STATE(1761)] = 95733,
  [SMALL_STATE(1762)] = 95747,
  [SMALL_STATE(1763)] = 95761,
  [SMALL_STATE(1764)] = 95775,
  [SMALL_STATE(1765)] = 95789,
  [SMALL_STATE(1766)] = 95803,
  [SMALL_STATE(1767)] = 95815,
  [SMALL_STATE(1768)] = 95829,
  [SMALL_STATE(1769)] = 95843,
  [SMALL_STATE(1770)] = 95857,
  [SMALL_STATE(1771)] = 95871,
  [SMALL_STATE(1772)] = 95885,
  [SMALL_STATE(1773)] = 95899,
  [SMALL_STATE(1774)] = 95913,
  [SMALL_STATE(1775)] = 95927,
  [SMALL_STATE(1776)] = 95941,
  [SMALL_STATE(1777)] = 95955,
  [SMALL_STATE(1778)] = 95969,
  [SMALL_STATE(1779)] = 95981,
  [SMALL_STATE(1780)] = 95995,
  [SMALL_STATE(1781)] = 96007,
  [SMALL_STATE(1782)] = 96021,
  [SMALL_STATE(1783)] = 96033,
  [SMALL_STATE(1784)] = 96047,
  [SMALL_STATE(1785)] = 96059,
  [SMALL_STATE(1786)] = 96073,
  [SMALL_STATE(1787)] = 96087,
  [SMALL_STATE(1788)] = 96101,
  [SMALL_STATE(1789)] = 96115,
  [SMALL_STATE(1790)] = 96129,
  [SMALL_STATE(1791)] = 96143,
  [SMALL_STATE(1792)] = 96155,
  [SMALL_STATE(1793)] = 96169,
  [SMALL_STATE(1794)] = 96183,
  [SMALL_STATE(1795)] = 96195,
  [SMALL_STATE(1796)] = 96209,
  [SMALL_STATE(1797)] = 96223,
  [SMALL_STATE(1798)] = 96237,
  [SMALL_STATE(1799)] = 96251,
  [SMALL_STATE(1800)] = 96265,
  [SMALL_STATE(1801)] = 96277,
  [SMALL_STATE(1802)] = 96291,
  [SMALL_STATE(1803)] = 96305,
  [SMALL_STATE(1804)] = 96317,
  [SMALL_STATE(1805)] = 96331,
  [SMALL_STATE(1806)] = 96345,
  [SMALL_STATE(1807)] = 96359,
  [SMALL_STATE(1808)] = 96371,
  [SMALL_STATE(1809)] = 96385,
  [SMALL_STATE(1810)] = 96397,
  [SMALL_STATE(1811)] = 96409,
  [SMALL_STATE(1812)] = 96421,
  [SMALL_STATE(1813)] = 96435,
  [SMALL_STATE(1814)] = 96449,
  [SMALL_STATE(1815)] = 96463,
  [SMALL_STATE(1816)] = 96477,
  [SMALL_STATE(1817)] = 96488,
  [SMALL_STATE(1818)] = 96499,
  [SMALL_STATE(1819)] = 96510,
  [SMALL_STATE(1820)] = 96521,
  [SMALL_STATE(1821)] = 96532,
  [SMALL_STATE(1822)] = 96543,
  [SMALL_STATE(1823)] = 96554,
  [SMALL_STATE(1824)] = 96565,
  [SMALL_STATE(1825)] = 96576,
  [SMALL_STATE(1826)] = 96587,
  [SMALL_STATE(1827)] = 96598,
  [SMALL_STATE(1828)] = 96609,
  [SMALL_STATE(1829)] = 96620,
  [SMALL_STATE(1830)] = 96631,
  [SMALL_STATE(1831)] = 96642,
  [SMALL_STATE(1832)] = 96653,
  [SMALL_STATE(1833)] = 96664,
  [SMALL_STATE(1834)] = 96675,
  [SMALL_STATE(1835)] = 96686,
  [SMALL_STATE(1836)] = 96697,
  [SMALL_STATE(1837)] = 96708,
  [SMALL_STATE(1838)] = 96719,
  [SMALL_STATE(1839)] = 96730,
  [SMALL_STATE(1840)] = 96741,
  [SMALL_STATE(1841)] = 96752,
  [SMALL_STATE(1842)] = 96763,
  [SMALL_STATE(1843)] = 96774,
  [SMALL_STATE(1844)] = 96785,
  [SMALL_STATE(1845)] = 96796,
  [SMALL_STATE(1846)] = 96807,
  [SMALL_STATE(1847)] = 96818,
  [SMALL_STATE(1848)] = 96829,
  [SMALL_STATE(1849)] = 96840,
  [SMALL_STATE(1850)] = 96851,
  [SMALL_STATE(1851)] = 96862,
  [SMALL_STATE(1852)] = 96873,
  [SMALL_STATE(1853)] = 96884,
  [SMALL_STATE(1854)] = 96895,
  [SMALL_STATE(1855)] = 96906,
  [SMALL_STATE(1856)] = 96917,
  [SMALL_STATE(1857)] = 96928,
  [SMALL_STATE(1858)] = 96939,
  [SMALL_STATE(1859)] = 96950,
  [SMALL_STATE(1860)] = 96961,
  [SMALL_STATE(1861)] = 96972,
  [SMALL_STATE(1862)] = 96983,
  [SMALL_STATE(1863)] = 96994,
  [SMALL_STATE(1864)] = 97005,
  [SMALL_STATE(1865)] = 97016,
  [SMALL_STATE(1866)] = 97027,
  [SMALL_STATE(1867)] = 97038,
  [SMALL_STATE(1868)] = 97049,
  [SMALL_STATE(1869)] = 97060,
  [SMALL_STATE(1870)] = 97071,
  [SMALL_STATE(1871)] = 97082,
  [SMALL_STATE(1872)] = 97093,
  [SMALL_STATE(1873)] = 97104,
  [SMALL_STATE(1874)] = 97115,
  [SMALL_STATE(1875)] = 97126,
  [SMALL_STATE(1876)] = 97137,
  [SMALL_STATE(1877)] = 97148,
  [SMALL_STATE(1878)] = 97159,
  [SMALL_STATE(1879)] = 97170,
  [SMALL_STATE(1880)] = 97181,
  [SMALL_STATE(1881)] = 97192,
  [SMALL_STATE(1882)] = 97203,
  [SMALL_STATE(1883)] = 97214,
  [SMALL_STATE(1884)] = 97225,
  [SMALL_STATE(1885)] = 97236,
  [SMALL_STATE(1886)] = 97247,
  [SMALL_STATE(1887)] = 97258,
  [SMALL_STATE(1888)] = 97269,
  [SMALL_STATE(1889)] = 97280,
  [SMALL_STATE(1890)] = 97291,
  [SMALL_STATE(1891)] = 97302,
  [SMALL_STATE(1892)] = 97313,
  [SMALL_STATE(1893)] = 97324,
  [SMALL_STATE(1894)] = 97335,
  [SMALL_STATE(1895)] = 97346,
  [SMALL_STATE(1896)] = 97357,
  [SMALL_STATE(1897)] = 97368,
  [SMALL_STATE(1898)] = 97379,
  [SMALL_STATE(1899)] = 97390,
  [SMALL_STATE(1900)] = 97401,
  [SMALL_STATE(1901)] = 97412,
  [SMALL_STATE(1902)] = 97423,
  [SMALL_STATE(1903)] = 97434,
  [SMALL_STATE(1904)] = 97445,
  [SMALL_STATE(1905)] = 97456,
  [SMALL_STATE(1906)] = 97467,
  [SMALL_STATE(1907)] = 97478,
  [SMALL_STATE(1908)] = 97489,
  [SMALL_STATE(1909)] = 97500,
  [SMALL_STATE(1910)] = 97511,
  [SMALL_STATE(1911)] = 97522,
  [SMALL_STATE(1912)] = 97533,
  [SMALL_STATE(1913)] = 97544,
  [SMALL_STATE(1914)] = 97555,
  [SMALL_STATE(1915)] = 97566,
  [SMALL_STATE(1916)] = 97577,
  [SMALL_STATE(1917)] = 97588,
  [SMALL_STATE(1918)] = 97599,
  [SMALL_STATE(1919)] = 97610,
  [SMALL_STATE(1920)] = 97621,
  [SMALL_STATE(1921)] = 97632,
  [SMALL_STATE(1922)] = 97643,
  [SMALL_STATE(1923)] = 97654,
  [SMALL_STATE(1924)] = 97665,
  [SMALL_STATE(1925)] = 97676,
  [SMALL_STATE(1926)] = 97687,
  [SMALL_STATE(1927)] = 97698,
  [SMALL_STATE(1928)] = 97709,
  [SMALL_STATE(1929)] = 97720,
  [SMALL_STATE(1930)] = 97731,
  [SMALL_STATE(1931)] = 97742,
  [SMALL_STATE(1932)] = 97753,
  [SMALL_STATE(1933)] = 97764,
  [SMALL_STATE(1934)] = 97775,
  [SMALL_STATE(1935)] = 97786,
  [SMALL_STATE(1936)] = 97797,
  [SMALL_STATE(1937)] = 97808,
  [SMALL_STATE(1938)] = 97819,
  [SMALL_STATE(1939)] = 97830,
  [SMALL_STATE(1940)] = 97841,
  [SMALL_STATE(1941)] = 97852,
  [SMALL_STATE(1942)] = 97863,
  [SMALL_STATE(1943)] = 97874,
  [SMALL_STATE(1944)] = 97885,
  [SMALL_STATE(1945)] = 97896,
  [SMALL_STATE(1946)] = 97907,
  [SMALL_STATE(1947)] = 97918,
  [SMALL_STATE(1948)] = 97929,
  [SMALL_STATE(1949)] = 97940,
  [SMALL_STATE(1950)] = 97951,
  [SMALL_STATE(1951)] = 97962,
  [SMALL_STATE(1952)] = 97973,
  [SMALL_STATE(1953)] = 97984,
  [SMALL_STATE(1954)] = 97995,
  [SMALL_STATE(1955)] = 98006,
  [SMALL_STATE(1956)] = 98017,
  [SMALL_STATE(1957)] = 98028,
  [SMALL_STATE(1958)] = 98039,
  [SMALL_STATE(1959)] = 98050,
  [SMALL_STATE(1960)] = 98061,
  [SMALL_STATE(1961)] = 98072,
  [SMALL_STATE(1962)] = 98083,
  [SMALL_STATE(1963)] = 98094,
  [SMALL_STATE(1964)] = 98105,
  [SMALL_STATE(1965)] = 98116,
  [SMALL_STATE(1966)] = 98127,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1551),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1672),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1676),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1383),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1687),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(543),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1471),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1706),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(125),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(124),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(374),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(895),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(922),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1919),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(118),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [65] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pipeline_echo, 1),
  [67] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pipeline_echo, 1),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1700),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(290),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(916),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(592),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2),
  [105] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1551),
  [108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(117),
  [111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1672),
  [114] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1676),
  [117] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1383),
  [120] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1687),
  [123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1950),
  [126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(276),
  [129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(341),
  [132] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(543),
  [135] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1471),
  [138] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1706),
  [141] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(125),
  [144] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(124),
  [147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(374),
  [150] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(286),
  [153] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(895),
  [156] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(922),
  [159] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(887),
  [162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(534),
  [165] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1416),
  [168] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1919),
  [171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1373),
  [174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(230),
  [177] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(120),
  [180] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(120),
  [183] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(118),
  [186] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(116),
  [189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1),
  [191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1683),
  [195] = {.entry = {.count = 1, .reusable = false}}, SHIFT(618),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(617),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(419),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(919),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(920),
  [211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(530),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1891),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1674),
  [229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(936),
  [231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(934),
  [233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(447),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(897),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(547),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(932),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(930),
  [251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2),
  [255] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2),
  [257] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(117),
  [260] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1950),
  [263] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(276),
  [266] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(341),
  [269] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(543),
  [272] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1696),
  [275] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(125),
  [278] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(124),
  [281] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(374),
  [284] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(286),
  [287] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(895),
  [290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(922),
  [293] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(887),
  [296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(534),
  [299] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1373),
  [302] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(230),
  [305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(120),
  [308] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(120),
  [311] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(118),
  [314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(116),
  [317] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1),
  [319] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1),
  [321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(117),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(1950),
  [327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(276),
  [330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(341),
  [333] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(543),
  [336] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(1696),
  [339] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(125),
  [342] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(124),
  [345] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(374),
  [348] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(286),
  [351] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(895),
  [354] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(922),
  [357] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(887),
  [360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(534),
  [363] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(1373),
  [366] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(230),
  [369] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(120),
  [372] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(120),
  [375] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(118),
  [378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1), SHIFT(116),
  [381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record, 1, .production_id = 1),
  [383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record, 1, .production_id = 1),
  [385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 5),
  [389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 5),
  [391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1822),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [401] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [403] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1741),
  [405] = {.entry = {.count = 1, .reusable = false}}, SHIFT(514),
  [407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(512),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(894),
  [415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(921),
  [417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(885),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(235),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [439] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(111),
  [442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1822),
  [445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(292),
  [448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(467),
  [451] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(570),
  [454] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1741),
  [457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(514),
  [460] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(512),
  [463] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(328),
  [466] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(293),
  [469] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(894),
  [472] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(921),
  [475] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(885),
  [478] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(563),
  [481] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(1357),
  [484] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(582),
  [487] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(235),
  [490] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(235),
  [493] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(128),
  [496] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__statement_seq, 2), SHIFT_REPEAT(236),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_access, 3, .production_id = 21),
  [505] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_access, 3, .production_id = 21),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label, 1),
  [511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_label, 1),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_access, 3, .production_id = 20),
  [523] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_access, 3, .production_id = 20),
  [525] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2),
  [527] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 4, .production_id = 31),
  [547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 4, .production_id = 31),
  [549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3),
  [559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3),
  [561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case, 5, .production_id = 46),
  [565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case, 5, .production_id = 46),
  [567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_remote_constructor_name, 3, .production_id = 22),
  [571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_remote_constructor_name, 3, .production_id = 22),
  [573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer, 2),
  [581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer, 2),
  [583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4),
  [587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4),
  [589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [595] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3),
  [597] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3),
  [599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 2, .production_id = 8),
  [601] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 2, .production_id = 8),
  [603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer, 1),
  [605] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer, 1),
  [607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier, 1),
  [609] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_identifier, 1),
  [611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_name, 1),
  [613] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_name, 1),
  [615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_unit, 1),
  [621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_unit, 1),
  [623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_function_expression, 1),
  [625] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(765),
  [634] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1753),
  [638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1),
  [642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update, 7, .production_id = 72),
  [644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_update, 7, .production_id = 72),
  [646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 7, .production_id = 68),
  [648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 7, .production_id = 68),
  [650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record, 2, .production_id = 9),
  [652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record, 2, .production_id = 9),
  [654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_record_expression, 1),
  [656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 3),
  [658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 3),
  [660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_panic, 1),
  [662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_panic, 1),
  [664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 6),
  [668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 6),
  [670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function, 5, .production_id = 41),
  [672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_anonymous_function, 5, .production_id = 41),
  [674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 5),
  [676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 5),
  [678] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1795),
  [682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_todo, 1),
  [684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_todo, 1),
  [686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 4),
  [690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 4),
  [692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1771),
  [694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(384),
  [696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1769),
  [698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(380),
  [700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_bit_string, 4),
  [702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_bit_string, 4),
  [704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3),
  [706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3),
  [708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 4),
  [710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 4),
  [712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [714] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(995),
  [722] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1679),
  [724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(870),
  [726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, .production_id = 19),
  [736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, .production_id = 19),
  [738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_todo, 4, .production_id = 14),
  [740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_todo, 4, .production_id = 14),
  [742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_panic, 4, .production_id = 14),
  [744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_panic, 4, .production_id = 14),
  [746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__assignment, 3, .production_id = 34),
  [752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__assignment, 3, .production_id = 34),
  [754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(409),
  [756] = {.entry = {.count = 1, .reusable = false}}, SHIFT(410),
  [758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [764] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 5, .production_id = 40),
  [780] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 5, .production_id = 40),
  [782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 5),
  [784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 5),
  [786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_bit_string, 5),
  [788] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_bit_string, 5),
  [790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_negation, 2),
  [792] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_negation, 2),
  [794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer_negation, 2),
  [796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_integer_negation, 2),
  [798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert, 2, .production_id = 5),
  [800] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert, 2, .production_id = 5),
  [802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 2),
  [804] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 2),
  [806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_echo, 2),
  [812] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_echo, 2),
  [814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__assignment, 4, .production_id = 52),
  [816] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__assignment, 4, .production_id = 52),
  [818] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_integer, 1), REDUCE(sym_integer, 2),
  [821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_integer, 1), REDUCE(sym_integer, 2),
  [824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_panic, 3, .production_id = 14),
  [826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_panic, 3, .production_id = 14),
  [828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_todo, 3, .production_id = 14),
  [830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_todo, 3, .production_id = 14),
  [832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 6, .production_id = 57),
  [838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 6, .production_id = 57),
  [840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_bit_string, 3),
  [842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_bit_string, 3),
  [844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_bit_string, 2),
  [846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_bit_string, 2),
  [848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 3),
  [850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 3),
  [852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list, 7, .production_id = 67),
  [854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list, 7, .production_id = 67),
  [856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2),
  [858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2),
  [860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(433),
  [862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(435),
  [864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(445),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [886] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use, 3, .production_id = 17),
  [888] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use, 3, .production_id = 17),
  [890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [902] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_assert, 5, .production_id = 51),
  [904] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_assert, 5, .production_id = 51),
  [906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use, 4, .production_id = 36),
  [908] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use, 4, .production_id = 36),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 1),
  [918] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 1),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 2, .production_id = 38),
  [926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1082),
  [932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update_argument, 2, .production_id = 38),
  [970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1005),
  [972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [976] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1759),
  [980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [1000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [1002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [1004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [1006] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [1008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(9),
  [1010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [1012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [1014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [1016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(607),
  [1018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [1020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(606),
  [1022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [1024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(531),
  [1026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(10),
  [1028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [1030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [1032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [1034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [1036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [1038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [1042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [1044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [1046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [1048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [1050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [1052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(963),
  [1054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [1056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [1058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [1060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [1062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [1064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [1066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(862),
  [1068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [1070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [1072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [1074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [1076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [1080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [1082] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1695),
  [1084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [1086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [1088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [1092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(912),
  [1094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [1096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [1098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [1100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [1102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(61),
  [1104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [1106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [1108] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1696),
  [1110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [1112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [1114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [1116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [1118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(549),
  [1120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(11),
  [1122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [1124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [1126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [1128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [1130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [1132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [1134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [1136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [1138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [1140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [1144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [1150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [1152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [1154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [1156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [1158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [1160] = {.entry = {.count = 1, .reusable = false}}, SHIFT(526),
  [1162] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1685),
  [1164] = {.entry = {.count = 1, .reusable = false}}, SHIFT(509),
  [1166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(508),
  [1168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [1170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(288),
  [1172] = {.entry = {.count = 1, .reusable = false}}, SHIFT(907),
  [1174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [1176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1374),
  [1178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [1180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [1182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(416),
  [1184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [1186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [1188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [1190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [1192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [1194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [1196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [1198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [1200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [1202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [1204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [1206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [1208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [1210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [1212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [1214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [1216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [1218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [1220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [1222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [1224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [1226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [1228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [1230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [1232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [1234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1401),
  [1236] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [1238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(903),
  [1240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [1242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [1244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [1246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [1248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(832),
  [1250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [1252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [1254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [1256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [1258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [1260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [1262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [1264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [1266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [1268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [1270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [1272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [1274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [1276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [1278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [1280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [1282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [1284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [1286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [1288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [1290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [1292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [1294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [1296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [1298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [1300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [1302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [1304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [1306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [1308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [1310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [1312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [1314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [1316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [1318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [1322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [1324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [1326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [1328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [1330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [1332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [1334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [1336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [1338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [1340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [1342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [1344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [1346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [1348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [1350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [1352] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_case_subjects, 3), SHIFT(84),
  [1355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [1357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [1359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [1361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [1363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [1365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [1367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [1369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [1371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [1373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [1375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [1377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [1379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [1381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [1383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [1385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [1387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [1389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [1391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [1393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [1395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [1397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [1399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [1401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [1403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [1407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [1409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [1411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [1413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [1415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [1417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [1419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [1421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [1423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [1427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [1429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [1431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [1433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1076),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [1437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [1439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [1443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [1445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [1449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [1451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1048),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [1457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [1463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [1465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [1467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [1469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [1471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [1473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [1475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [1477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [1479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [1481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [1483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [1485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [1487] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_case_subjects, 2), SHIFT(84),
  [1490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [1492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [1494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [1496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [1498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [1500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [1502] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1689),
  [1506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [1508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [1510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [1512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [1514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1772),
  [1516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1774),
  [1518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1798),
  [1520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [1522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [1524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(349),
  [1526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(486),
  [1528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(485),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [1532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [1534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(479),
  [1538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [1540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [1542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [1544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(486),
  [1546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [1548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [1550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [1552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [1554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [1556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [1558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(773),
  [1560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [1562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [1564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [1566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [1568] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [1570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [1572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [1574] = {.entry = {.count = 1, .reusable = false}}, SHIFT(670),
  [1576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [1578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [1580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [1582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [1584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [1586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [1588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [1590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1049),
  [1592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [1594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [1596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [1598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [1600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(346),
  [1602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [1604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [1606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [1608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [1610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [1612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [1614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [1616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [1618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [1620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [1622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [1624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record, 1, .production_id = 1),
  [1626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [1628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record, 1, .production_id = 1),
  [1630] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1708),
  [1634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [1636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [1638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [1640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(401),
  [1642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [1644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(392),
  [1646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [1648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [1650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [1652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [1654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [1656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [1658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [1660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_arguments, 5),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record_arguments, 5),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple, 6),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_tuple, 6),
  [1668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [1670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_list, 3),
  [1672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_list, 3),
  [1674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_arguments, 3),
  [1676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record_arguments, 3),
  [1678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_list, 2),
  [1680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_list, 2),
  [1682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_bit_string, 2),
  [1684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant_bit_string, 2),
  [1686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record, 2, .production_id = 9),
  [1688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record, 2, .production_id = 9),
  [1690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_bit_string, 4),
  [1692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant_bit_string, 4),
  [1694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_list, 4),
  [1696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_list, 4),
  [1698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple, 3),
  [1700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_tuple, 3),
  [1702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause, 4, .production_id = 70),
  [1704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause, 4, .production_id = 70),
  [1706] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause, 3, .production_id = 60),
  [1708] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause, 3, .production_id = 60),
  [1710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple, 4),
  [1712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_tuple, 4),
  [1714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_arguments, 4),
  [1716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record_arguments, 4),
  [1718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_bit_string, 3),
  [1720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant_bit_string, 3),
  [1722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_bit_string, 5),
  [1724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant_bit_string, 5),
  [1726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_field_access, 3, .production_id = 21),
  [1728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_field_access, 3, .production_id = 21),
  [1730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple, 5),
  [1732] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_tuple, 5),
  [1734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_arguments, 2),
  [1736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_record_arguments, 2),
  [1738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_list, 5),
  [1740] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant_list, 5),
  [1742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 1),
  [1744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 1),
  [1746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1850),
  [1748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_repeat1, 2),
  [1750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2),
  [1752] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2), SHIFT_REPEAT(1850),
  [1755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 2),
  [1757] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 2),
  [1759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [1761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 1, .production_id = 1),
  [1763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_name, 1, .production_id = 1),
  [1765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [1767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 1, .production_id = 1),
  [1769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type, 1, .production_id = 1),
  [1771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [1773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_identifier, 1),
  [1775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_identifier, 1),
  [1777] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 2, .production_id = 3),
  [1779] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 2, .production_id = 3),
  [1781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [1783] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1472),
  [1785] = {.entry = {.count = 3, .reusable = false}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1694),
  [1789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, .production_id = 2),
  [1791] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 2, .production_id = 2),
  [1793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1001),
  [1795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_remote_type_identifier, 3, .production_id = 22),
  [1797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_remote_type_identifier, 3, .production_id = 22),
  [1799] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1),
  [1802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [1804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_imports, 5),
  [1806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unqualified_imports, 5),
  [1808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 2),
  [1810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 2),
  [1812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_var, 1),
  [1814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_var, 1),
  [1816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [1818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [1820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [1822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [1824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [1826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [1828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1386),
  [1830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [1832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1344),
  [1834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1112),
  [1836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1306),
  [1838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [1840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [1842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 3),
  [1844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 3),
  [1846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_value, 1),
  [1848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant_value, 1),
  [1850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [1852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_imports, 3),
  [1854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unqualified_imports, 3),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 4),
  [1858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_parameters, 4),
  [1860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 3, .production_id = 13),
  [1862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 3, .production_id = 13),
  [1864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [1866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [1868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [1870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [1872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_imports, 2),
  [1874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unqualified_imports, 2),
  [1876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_imports, 4),
  [1878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unqualified_imports, 4),
  [1880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_assert, 3, .production_id = 16),
  [1882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_assert, 3, .production_id = 16),
  [1884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [1886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [1888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [1890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [1892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [1894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 3),
  [1896] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_parameters, 3),
  [1898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [1900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 5),
  [1902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_parameters, 5),
  [1904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 5),
  [1906] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 5),
  [1908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [1910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_type, 2),
  [1912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_type, 2),
  [1914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1418),
  [1916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1186),
  [1918] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 2),
  [1920] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_parameters, 2),
  [1922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 4, .production_id = 24),
  [1924] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 4, .production_id = 24),
  [1926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1453),
  [1928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 4, .production_id = 37),
  [1930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 4, .production_id = 37),
  [1932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [1934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [1936] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__maybe_record_expression, 1),
  [1938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [1940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 4),
  [1942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 4),
  [1944] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__maybe_tuple_expression, 1),
  [1946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_type, 3),
  [1948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_type, 3),
  [1950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [1952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1210),
  [1954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_name, 2, .production_id = 11),
  [1956] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_name, 2, .production_id = 11),
  [1958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 2),
  [1960] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 2),
  [1962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, .production_id = 56),
  [1964] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, .production_id = 56),
  [1966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 3),
  [1968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 3),
  [1970] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 6),
  [1972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 6),
  [1974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 6, .production_id = 55),
  [1976] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 6, .production_id = 55),
  [1978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_arguments, 4),
  [1980] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_arguments, 4),
  [1982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 5),
  [1984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 5),
  [1986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant, 5, .production_id = 53),
  [1988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant, 5, .production_id = 53),
  [1990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_alias, 5),
  [1992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_alias, 5),
  [1994] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__statement_seq, 1),
  [1996] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__statement_seq, 1),
  [1998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [2000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [2002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1744),
  [2004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [2006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_body, 2),
  [2008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_function_body, 2),
  [2010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 9, .production_id = 76),
  [2012] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 9, .production_id = 76),
  [2014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function, 9, .production_id = 77),
  [2016] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_function, 9, .production_id = 77),
  [2018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2),
  [2020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1864),
  [2023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(876),
  [2026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(892),
  [2029] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1462),
  [2032] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1378),
  [2035] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1386),
  [2038] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1344),
  [2041] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1344),
  [2044] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1112),
  [2047] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1306),
  [2050] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clauses_repeat1, 2), SHIFT_REPEAT(1308),
  [2053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let, 2, .production_id = 5),
  [2055] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let, 2, .production_id = 5),
  [2057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 4, .production_id = 66),
  [2059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 4, .production_id = 66),
  [2061] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 8, .production_id = 64),
  [2063] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 8, .production_id = 64),
  [2065] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, .production_id = 10),
  [2067] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 3, .production_id = 10),
  [2069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 5, .production_id = 44),
  [2071] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 5, .production_id = 44),
  [2073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [2075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 5, .production_id = 13),
  [2077] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 5, .production_id = 13),
  [2079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_bit_string_segment_options, 2),
  [2081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [2083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [2085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1535),
  [2089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [2091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant, 5, .production_id = 39),
  [2093] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant, 5, .production_id = 39),
  [2095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 2, .production_id = 9),
  [2097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type, 2, .production_id = 9),
  [2099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5),
  [2101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5),
  [2103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_arguments, 3),
  [2105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_arguments, 3),
  [2107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_target_group, 5, .production_id = 23),
  [2109] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_target_group, 5, .production_id = 23),
  [2111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_type, 4),
  [2113] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_type, 4),
  [2115] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 4),
  [2117] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 4),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [2121] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_discard, 1),
  [2123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_discard, 1),
  [2125] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_bit_string_segment_options, 3),
  [2127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [2129] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function, 8, .production_id = 75),
  [2131] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_function, 8, .production_id = 75),
  [2133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_target_group, 4, .production_id = 23),
  [2135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_target_group, 4, .production_id = 23),
  [2137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 7, .production_id = 44),
  [2139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 7, .production_id = 44),
  [2141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 8, .production_id = 74),
  [2143] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 8, .production_id = 74),
  [2145] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 7, .production_id = 71),
  [2147] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 7, .production_id = 71),
  [2149] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 7),
  [2151] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 7),
  [2153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 6, .production_id = 58),
  [2155] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 6, .production_id = 58),
  [2157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_bit_string_segment_options, 3),
  [2159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [2161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 4, .production_id = 25),
  [2163] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 4, .production_id = 25),
  [2165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_bit_string_segment_options, 2),
  [2167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 3),
  [2169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 3),
  [2171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_hole, 1),
  [2173] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_hole, 1),
  [2175] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_alias, 4),
  [2177] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_alias, 4),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [2181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 6),
  [2183] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 6),
  [2185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant, 4, .production_id = 27),
  [2187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant, 4, .production_id = 27),
  [2189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_bit_string_segment_options, 2),
  [2191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clauses, 1),
  [2193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant, 6, .production_id = 63),
  [2195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constant, 6, .production_id = 63),
  [2197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 4),
  [2199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 4),
  [2201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_arguments, 5),
  [2203] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_arguments, 5),
  [2205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 6, .production_id = 37),
  [2207] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 6, .production_id = 37),
  [2209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_bit_string_segment_options, 3),
  [2211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_alias, 6),
  [2213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_alias, 6),
  [2215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 5),
  [2217] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 5),
  [2219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function, 6, .production_id = 64),
  [2221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function, 6, .production_id = 64),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1840),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [2231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [2235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [2237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [2245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [2247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [2249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1126),
  [2251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [2253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [2255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [2257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [2259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [2263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [2265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [2269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [2271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [2275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [2279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [2281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [2283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [2285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [2287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [2289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [2293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [2297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1253),
  [2299] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause_pattern, 2, .production_id = 7),
  [2301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_pattern, 2, .production_id = 7),
  [2303] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause_patterns, 3),
  [2305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_patterns, 3),
  [2307] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause_pattern, 3, .production_id = 47),
  [2309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_pattern, 3, .production_id = 47),
  [2311] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_argument, 2, .production_id = 38),
  [2313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [2317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [2319] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_case_clause_patterns, 2),
  [2321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_patterns, 2),
  [2323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [2325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_assignments, 2),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [2329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_assignments, 3),
  [2331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [2333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [2335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(905),
  [2337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(904),
  [2339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [2341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [2343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [2345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [2347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [2349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [2351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [2353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [2355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [2357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [2359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [2361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [2363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [2365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [2367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [2369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [2371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [2373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [2375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [2377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [2379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [2381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [2383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tuple_repeat1, 2),
  [2385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [2387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [2389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [2391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [2393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [2395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1138),
  [2397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [2399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [2403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_subjects, 1),
  [2405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [2407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(482),
  [2409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(480),
  [2411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [2413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [2415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [2417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(408),
  [2419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [2421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [2423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [2425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [2427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [2429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [2431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [2433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [2435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [2437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [2439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [2441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1717),
  [2443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [2445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [2447] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym__maybe_tuple_expression, 1), REDUCE(sym__maybe_record_expression, 1), SHIFT(1770),
  [2451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [2453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [2455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [2457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [2459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [2463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [2465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [2467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [2469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [2471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [2473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [2475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [2477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [2479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [2481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [2483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [2485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [2487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [2489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(796),
  [2491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [2493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [2495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1246),
  [2497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [2499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_argument, 2, .production_id = 38),
  [2501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [2503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [2505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [2507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [2509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [2511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [2513] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 1, .production_id = 4),
  [2515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [2517] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_argument, 3, .production_id = 54),
  [2519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update_argument, 3, .production_id = 54),
  [2521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(557),
  [2523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [2525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [2527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(493),
  [2529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [2531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [2533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [2535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(457),
  [2537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [2539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [2541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [2543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [2545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [2547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [2549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [2551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [2553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [2555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [2557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [2559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [2561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [2563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [2565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [2567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [2569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [2571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [2573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [2575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [2577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [2579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [2581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [2583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [2585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [2587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [2589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [2591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [2593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [2595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1511),
  [2597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [2599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(797),
  [2601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [2603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [2605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [2607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [2609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [2611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [2613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [2615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [2617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [2619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [2621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [2623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [2625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [2627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1140),
  [2629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [2631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [2633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [2635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [2637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [2639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [2641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [2643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [2647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(829),
  [2649] = {.entry = {.count = 1, .reusable = false}}, SHIFT(473),
  [2651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [2653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [2655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1333),
  [2657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [2659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [2661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [2663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [2665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [2667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [2669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1337),
  [2671] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__constant_value, 1), REDUCE(sym__case_clause_guard_unit, 1),
  [2674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [2676] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__constant_value, 1), REDUCE(sym__case_clause_guard_unit, 1),
  [2679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [2681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [2683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [2685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [2687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [2689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(847),
  [2691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [2693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_guard, 2),
  [2695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [2697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(846),
  [2699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [2701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(866),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [2705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [2709] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__case_clause_guard_unit, 3),
  [2711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__case_clause_guard_unit, 3),
  [2713] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__case_clause_guard_binary_expression, 3, .production_id = 19),
  [2715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__case_clause_guard_binary_expression, 3, .production_id = 19),
  [2717] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__case_clause_guard_expression, 1, .production_id = 6),
  [2719] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__case_clause_guard_expression, 1, .production_id = 6),
  [2721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__case_clause_guard_unit, 1, .production_id = 59),
  [2723] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__case_clause_guard_unit, 1, .production_id = 59),
  [2725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__case_clause_tuple_access, 3, .production_id = 20),
  [2727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__case_clause_tuple_access, 3, .production_id = 20),
  [2729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(860),
  [2731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(859),
  [2733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(860),
  [2737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(855),
  [2739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [2743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [2745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [2747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(857),
  [2749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [2751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2),
  [2753] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1773),
  [2756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1775),
  [2759] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1370),
  [2762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1796),
  [2765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1671),
  [2768] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1792),
  [2771] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1417),
  [2774] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_target_group_repeat1, 2), SHIFT_REPEAT(1903),
  [2777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1775),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1671),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1792),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1903),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1778),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1831),
  [2801] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1485),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [2805] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1476),
  [2807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [2809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [2811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [2813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [2815] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1162),
  [2817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [2819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1900),
  [2821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [2823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [2825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [2827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [2829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [2831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [2833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [2835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1837),
  [2837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [2839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1834),
  [2843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1574),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [2847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1510),
  [2849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [2851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [2853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1926),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [2859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 1, .production_id = 1),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [2863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1843),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [2867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1567),
  [2869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [2871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1592),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [2875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(709),
  [2877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [2879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [2881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [2883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [2885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [2887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [2889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_arguments, 5),
  [2891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 4, .production_id = 49),
  [2893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 2),
  [2895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 4, .production_id = 32),
  [2897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_bit_string, 4),
  [2899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 6, .production_id = 49),
  [2901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_arguments, 6),
  [2903] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 4, .production_id = 48),
  [2905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 5, .production_id = 48),
  [2907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 5, .production_id = 32),
  [2909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_bit_string, 3),
  [2911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 5, .production_id = 61),
  [2913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_binary_expression, 3, .production_id = 19),
  [2915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [2917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_arguments, 3),
  [2919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 6, .production_id = 61),
  [2921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1846),
  [2923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_bit_string, 2),
  [2925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 5, .production_id = 49),
  [2927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3),
  [2929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_expression, 1, .production_id = 6),
  [2931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 3),
  [2933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3, .production_id = 32),
  [2935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_module_repeat1, 2), SHIFT_REPEAT(1846),
  [2938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 2, .production_id = 9),
  [2940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_arguments, 4),
  [2942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_bit_string, 5),
  [2944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_arguments, 2),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1413),
  [2948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [2950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 3, .production_id = 33),
  [2952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1288),
  [2954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [2956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1395),
  [2958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [2960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [2964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [2966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [2968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [2972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [2978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [2984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [2986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_expression, 1),
  [2988] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1694),
  [2990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [2992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1446),
  [2994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [2996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [3000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [3002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [3004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [3008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [3012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(997),
  [3014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [3016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1693),
  [3020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [3022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [3024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [3026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [3028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(723),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [3034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [3036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_clause_pattern_repeat1, 2, .production_id = 32),
  [3040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [3042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [3044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [3060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [3062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1379),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1469),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [3108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2),
  [3110] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(1463),
  [3113] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2), SHIFT_REPEAT(1388),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1434),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [3122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_repeat1, 2), SHIFT_REPEAT(386),
  [3125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clause_pattern_repeat1, 2, .production_id = 47), SHIFT_REPEAT(950),
  [3128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_clause_pattern_repeat1, 2, .production_id = 47),
  [3130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern_tail, 1),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [3134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [3136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [3138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [3140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_pattern, 2, .production_id = 47),
  [3142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [3144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_data_constructors_repeat1, 2),
  [3146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_data_constructors_repeat1, 2), SHIFT_REPEAT(236),
  [3149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clause_pattern_repeat1, 2, .production_id = 47), SHIFT_REPEAT(961),
  [3152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructors, 1),
  [3154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_pattern, 1, .production_id = 7),
  [3156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [3158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [3160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [3164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [3166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1813),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1812),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1821),
  [3172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [3174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [3176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [3178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1790),
  [3180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1905),
  [3182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [3184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type, 1, .production_id = 1),
  [3186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [3188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [3192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [3196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [3198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(703),
  [3200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [3202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1111),
  [3204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [3206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_bit_string_segment_options, 1),
  [3208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [3210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [3212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [3214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [3216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [3218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_clause_patterns, 1),
  [3220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [3222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [3224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(53),
  [3226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [3228] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_tuple_repeat1, 2), SHIFT_REPEAT(1092),
  [3231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_tuple_repeat1, 2),
  [3233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_pattern_arguments_repeat1, 2), SHIFT_REPEAT(889),
  [3236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_pattern_arguments_repeat1, 2),
  [3238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 1, .production_id = 12),
  [3240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [3242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_bit_string_segment_options, 1),
  [3244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [3246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameter, 1, .production_id = 12),
  [3248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [3250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 1, .production_id = 30),
  [3252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [3254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [3256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [3258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_assignment, 1, .production_id = 7),
  [3260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_clause_patterns_repeat1, 2),
  [3262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_clause_patterns_repeat1, 2), SHIFT_REPEAT(901),
  [3265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update_arguments, 3),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [3269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [3271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [3273] = {.entry = {.count = 1, .reusable = false}}, SHIFT(466),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [3277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(852),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_escape_sequence, 1),
  [3281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor, 1, .production_id = 1),
  [3283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [3285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_bit_string_segment_options_repeat1, 2),
  [3287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_bit_string_segment_options_repeat1, 2), SHIFT_REPEAT(875),
  [3290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update_arguments, 2),
  [3292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_bit_string_segment_options_repeat1, 2),
  [3294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_bit_string_segment_options_repeat1, 2), SHIFT_REPEAT(878),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [3301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [3305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_bit_string_segment_options, 1),
  [3307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [3309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_pattern_bit_string_segment_options_repeat1, 2),
  [3311] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_pattern_bit_string_segment_options_repeat1, 2), SHIFT_REPEAT(877),
  [3314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_annotation, 2, .production_id = 26),
  [3316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [3318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple_type, 4),
  [3320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [3322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [3324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [3326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [3328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [3330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [3332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1503),
  [3338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [3340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [3342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [3344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [3346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [3348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [3350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [3352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1283),
  [3354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [3356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_type, 4, .production_id = 66),
  [3358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type_arguments, 3),
  [3360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [3362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [3364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1039),
  [3366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [3368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [3370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [3372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_bit_string_segment_option_size, 4),
  [3374] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_data_constructor_arguments_repeat1, 2), SHIFT_REPEAT(1154),
  [3377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_data_constructor_arguments_repeat1, 2),
  [3379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1291),
  [3381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [3383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__bit_string_segment_option, 4),
  [3385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [3387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [3389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [3391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [3393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [3395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple_type, 6),
  [3397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_unqualified_imports_repeat1, 2),
  [3399] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_unqualified_imports_repeat1, 2), SHIFT_REPEAT(1394),
  [3402] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__expression_bit_string_repeat1, 2), SHIFT_REPEAT(515),
  [3405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__expression_bit_string_repeat1, 2),
  [3407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type_arguments, 5),
  [3409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [3415] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_function_parameters_repeat1, 2), SHIFT_REPEAT(1172),
  [3418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_function_parameters_repeat1, 2),
  [3420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [3422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [3424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [3426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [3428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [3430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [3432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [3436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1268),
  [3438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [3440] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constant_bit_string_repeat1, 2), SHIFT_REPEAT(1031),
  [3443] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__constant_bit_string_repeat1, 2),
  [3445] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_repeat1, 2), SHIFT_REPEAT(429),
  [3448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1214),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1934),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(845),
  [3454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [3456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [3458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1097),
  [3462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [3464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_bit_string_segment, 1, .production_id = 4),
  [3466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [3468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [3470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [3474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [3478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__bit_string_segment_option, 1),
  [3482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_bit_string_segment_option_size, 4),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [3486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_update_arguments, 1),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [3490] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_type_repeat1, 2), SHIFT_REPEAT(1211),
  [3493] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tuple_type_repeat1, 2),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [3497] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_argument, 3, .production_id = 62),
  [3499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [3501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [3503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_update_arguments_repeat1, 2), SHIFT_REPEAT(1537),
  [3506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_update_arguments_repeat1, 2),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_assignments, 1),
  [3516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [3518] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2), SHIFT_REPEAT(1175),
  [3521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2),
  [3523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1157),
  [3525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [3527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [3529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [3531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [3533] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_parameters_repeat1, 2), SHIFT_REPEAT(1298),
  [3536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_parameters_repeat1, 2),
  [3538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [3540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [3542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type_arguments, 2),
  [3544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_type, 3, .production_id = 56),
  [3546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [3550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple_type, 3),
  [3552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [3554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [3556] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_record_arguments_repeat1, 2), SHIFT_REPEAT(996),
  [3559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_record_arguments_repeat1, 2),
  [3561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [3563] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2), SHIFT_REPEAT(1680),
  [3566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2),
  [3568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [3570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [3576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [3578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1949),
  [3582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern_bit_string_segment_option_size, 4),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [3588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [3590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [3594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [3596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1727),
  [3600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_bit_string_segment, 1, .production_id = 15),
  [3602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [3604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [3606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [3608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [3618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_tuple_type, 5),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [3622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [3624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_import, 2, .production_id = 2),
  [3626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [3630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [3642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [3654] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__attribute_arguments_repeat1, 2), SHIFT_REPEAT(973),
  [3657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__attribute_arguments_repeat1, 2),
  [3659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [3661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [3663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [3665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1054),
  [3667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [3669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [3671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [3673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [3675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [3677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [3679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [3681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [3683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [3685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [3687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [3689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2), SHIFT_REPEAT(237),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2),
  [3694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name_param, 1, .production_id = 1),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [3698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern_argument, 1, .production_id = 35),
  [3700] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_tuple_type_repeat1, 2), SHIFT_REPEAT(1231),
  [3703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_tuple_type_repeat1, 2),
  [3705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [3707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [3709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__discard_param, 1, .production_id = 1),
  [3711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type_arguments, 4),
  [3713] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_assignments_repeat1, 2), SHIFT_REPEAT(913),
  [3716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_use_assignments_repeat1, 2),
  [3718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [3726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_subjects, 2),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [3730] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__pattern_bit_string_repeat1, 2), SHIFT_REPEAT(896),
  [3733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__pattern_bit_string_repeat1, 2),
  [3735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [3737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [3739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [3741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [3743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [3745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [3747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [3749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [3751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [3753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [3755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [3757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [3759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [3761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [3763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1824),
  [3765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_type_arguments_repeat1, 2), SHIFT_REPEAT(1194),
  [3768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_type_arguments_repeat1, 2),
  [3770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_bit_string_segment, 1, .production_id = 4),
  [3772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__labeled_name_param, 2, .production_id = 43),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [3798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [3804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [3806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type, 2, .production_id = 9),
  [3808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [3812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [3814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [3816] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_anonymous_function_parameters_repeat1, 2), SHIFT_REPEAT(1346),
  [3819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_anonymous_function_parameters_repeat1, 2),
  [3821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [3823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [3825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__labeled_discard_param, 2, .production_id = 43),
  [3827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [3829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1347),
  [3831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [3833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [3835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_import, 1, .production_id = 1),
  [3837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [3839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1768),
  [3841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [3843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [3845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_bit_string_segment, 3, .production_id = 50),
  [3847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [3849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1166),
  [3851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [3853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [3855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hole, 1),
  [3857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor, 2, .production_id = 9),
  [3859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [3861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [3863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [3865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_assignment, 2, .production_id = 18),
  [3867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameters, 4),
  [3869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameters, 3),
  [3871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 2, .production_id = 42),
  [3873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(590),
  [3875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter, 2, .production_id = 29),
  [3877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [3879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [3881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [3883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [3885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [3887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1223),
  [3889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [3891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [3893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameter, 1, .production_id = 45),
  [3895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [3897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [3899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_arguments, 5),
  [3901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [3903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [3905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [3907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [3909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameter, 2, .production_id = 29),
  [3911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_value, 1),
  [3913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [3915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [3917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [3919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [3921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [3923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1402),
  [3925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1209),
  [3927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [3929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [3931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_arguments, 3),
  [3933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_import, 3, .production_id = 65),
  [3935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_argument, 3, .production_id = 54),
  [3937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_bit_string_segment, 3, .production_id = 28),
  [3939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [3941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [3943] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unqualified_import, 4, .production_id = 73),
  [3945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [3947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_spread, 1),
  [3949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [3951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_argument, 3, .production_id = 54),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameters, 2),
  [3955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_bit_string_segment, 3, .production_id = 28),
  [3957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_arguments, 4),
  [3959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [3961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [3963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [3965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_value, 3, .production_id = 54),
  [3967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_record_argument, 1, .production_id = 4),
  [3969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [3971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_arguments, 2),
  [3973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_data_constructor_argument, 1, .production_id = 4),
  [3975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameter, 3, .production_id = 69),
  [3977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [3979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter, 1),
  [3981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [3983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [3985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_argument, 1),
  [3989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1668),
  [3993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [3995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_type_argument, 1),
  [3997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [3999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_anonymous_function_parameters, 5),
  [4001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [4003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [4005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [4009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [4011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [4013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant_type_annotation, 2, .production_id = 26),
  [4015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [4017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [4019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [4021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameters, 3),
  [4023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [4025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [4027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [4029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [4031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [4033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1169),
  [4035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_parameter_types, 2),
  [4037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [4039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [4041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [4043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [4045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_types, 4),
  [4047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [4049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [4051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [4053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [4055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [4057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_types, 2),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [4063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [4065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [4067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [4069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [4071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [4073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [4075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [4077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [4079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [4081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [4083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [4085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [4087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [4089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1235),
  [4091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [4093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_parameter_types, 4),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [4097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [4107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [4109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [4111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [4115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [4117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [4121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [4123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1095),
  [4125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1721),
  [4131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [4133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [4137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern_tail, 2),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [4141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [4153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [4155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [4157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [4159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [4161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(883),
  [4163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [4165] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_target, 1),
  [4167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameters, 5),
  [4169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [4171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1804),
  [4173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [4175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [4177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [4179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [4181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [4183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [4185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [4187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [4189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [4191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [4193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [4195] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [4197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [4199] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_parameter_types, 5),
  [4201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [4203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [4205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [4207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [4209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [4211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [4213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern_spread, 2),
  [4215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [4217] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_types, 5),
  [4219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameters, 4),
  [4221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [4223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [4233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_function_parameter_types, 3),
  [4235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [4237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [4239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [4241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [4243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [4245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [4247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [4251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [4253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [4255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [4257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [4259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [4261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [4263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameter_types, 3),
  [4265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [4267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [4269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [4271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_function_parameters, 2),
  [4273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [4275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [4279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [4283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [4285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [4287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [4289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [4291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(471),
  [4293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [4295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [4297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_gleam_external_scanner_create(void);
void tree_sitter_gleam_external_scanner_destroy(void *);
bool tree_sitter_gleam_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_gleam_external_scanner_serialize(void *, char *);
void tree_sitter_gleam_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef _WIN32
#define extern __declspec(dllexport)
#endif

extern const TSLanguage *tree_sitter_gleam(void) {
  static const TSLanguage language = {
    .version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_gleam_external_scanner_create,
      tree_sitter_gleam_external_scanner_destroy,
      tree_sitter_gleam_external_scanner_scan,
      tree_sitter_gleam_external_scanner_serialize,
      tree_sitter_gleam_external_scanner_deserialize,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
